/*
Company :		ZEFYR Technologies
File :			BMPfiles.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	15/04/2003

Description : BMP load/save support

Features : 8bit greyscale non compressed BMP files		
    
Functions : 

	getBMPsize: get the size of a BMP file (necessary for buffer allocation )
    loadBMP:	load a BMP file into an array of unsigned char
    saveBMP:	save a BMP file to disk from an array of unsigned char
	saveBmp32:  save a BMP file to disk from an array of unsigned long
	saveBmp32S: save a BMP file to disk from an array of signed long
	saveBmpR:   save a BMP file to disk from an array of float
	saveBmpC:   save a BMP file to disk from an array of complexe numbers (two floats)
	saveBmpS:	save a BMP file to disk from an array of char
	saveBmpS16:	save a BMP file to disk from an array of short

Necessary package : 
	errorCodes.h
	customTypes.h

Comments :none

*/

#ifndef __ZEFYR_BMP_FILES_H
#define __ZEFYR_BMP_FILES_H

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Returns the size of a BMP image.
 * This function calculates the size we need to allocate to store  the BMP file in a buffer
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName	Name of the BMP file
 * @param[out]		p_size		Size in bytes of the BMP image data
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 */
INT32 getBmpSize( char *p_fileName, INT32 *p_size );

/**
 * Loads a BMP file in a memory buffer.
 * This function loads a BMP file and stores its image data in a buffer
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName		Name of the BMP file
 * @param[out]		p_buffer		Buffer in which the image will be stored (must be allocated by the caller)
 * @param[in]		p_bufferSize	Size of buffer p_buffer
 * @param[out]		p_width			Width of the image (in pixels)
 * @param[out]		p_height		Height of the image (in pixels)
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				getBmpSize
 */
INT32 loadBmp( char *p_fileName, UINT8 *p_buffer, INT32 p_bufferSize, INT32 *p_width, INT32 *p_height );

/**
 * Save a BMP file from a memory buffer
 * This function saves a BMP file from data passed from a buffer containing UINT8 data.
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName		Name of the BMP file
 * @param[in]		p_buffer		Buffer containing the image data
 * @param[in]		p_width			Width of the image (in pixels)
 * @param[in]		p_height		Height of the image (in pixels)
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				UINT8, saveBmp32, saveBmp32S, saveBmpR, saveBmpC, saveBmpS, saveBmpS16
 */
INT32 saveBmp( char *p_fileName, UINT8 *p_buffer, INT32 p_width, INT32 p_height );

/**
 * Save a BMP file from a memory buffer
 * This function saves a BMP file from data passed from a buffer containing UINT32 data.
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName		Name of the BMP file
 * @param[in]		p_buffer		Buffer containing the image data
 * @param[in]		p_width			Width of the image (in pixels)
 * @param[in]		p_height		Height of the image (in pixels)
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				UINT32, saveBmp, saveBmp32S, saveBmpR, saveBmpC, saveBmpS, saveBmpS16
 */
INT32 saveBmp32( char *p_fileName, UINT32 *p_buffer, INT32 p_width, INT32 p_height );

/**
 * Save a BMP file from a memory buffer
 * This function saves a BMP file from data passed from a buffer containing INT32 data.
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName		Name of the BMP file
 * @param[in]		p_buffer		Buffer containing the image data
 * @param[in]		p_width			Width of the image (in pixels)
 * @param[in]		p_height		Height of the image (in pixels)
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				INT32, saveBmp, saveBmp32, saveBmpR, saveBmpC, saveBmpS, saveBmpS16
 */
INT32 saveBmp32S( char *p_fileName, INT32 *p_buffer, INT32 p_width, INT32 p_height );

/**
 * Save a BMP file from a memory buffer
 * This function saves a BMP file from data passed from a buffer containing REAL32 data.
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName		Name of the BMP file
 * @param[in]		p_buffer		Buffer containing the image data
 * @param[in]		p_width			Width of the image (in pixels)
 * @param[in]		p_height		Height of the image (in pixels)
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				REAL32, saveBmp, saveBmp32, saveBmp32S, saveBmpC, saveBmpS, saveBmpS16
 */
INT32 saveBmpR( char *p_fileName, REAL32 *p_buffer, INT32 p_width, INT32 p_height );

/**
 * Save a BMP file from a memory buffer
 * This function saves a BMP file from data passed from a buffer containing COMPLEXE data.
 * COMPLEXE data is defined as two consecutive REAL32 value : one for the real part, one for the imaginary one.
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName		Name of the BMP file
 * @param[in]		p_buffer		Buffer containing the image data
 * @param[in]		p_width			Width of the image (in pixels)
 * @param[in]		p_height		Height of the image (in pixels)
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				REAL32, saveBmp, saveBmp32, saveBmp32S, saveBmpR, saveBmpS, saveBmpS16
 */
INT32 saveBmpC( char *p_fileName, REAL32 *p_buffer, INT32 p_width, INT32 p_height );

/**
 * Save a BMP file from a memory buffer
 * This function saves a BMP file from data passed from a buffer containing multiple consecutive INT8 data representing different image plans.
 * Only one plan is saved
 * This function is only used for internal debugging
 *
 * @param[in]		p_fileName		Name of the BMP file
 * @param[in]		p_buffer		Buffer containing the image data
 * @param[in]		p_width			Width of the image (in pixels)
 * @param[in]		p_height		Height of the image (in pixels)
 * @param[in]		p_nbPlans		Number of plans in the image
 * @param[in]		p_plan			Defines which plan will be saved to disk
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				INT8, saveBmp, saveBmp32, saveBmp32S, saveBmpC, saveBmpR, saveBmpS16
 */
INT32 saveBmpS( char *p_fileName, INT8 *p_buffer, INT32 p_width, INT32 p_height, UINT32 p_nbPlans, UINT32 p_plan );

/**
 * Save a BMP file from a memory buffer
 * This function saves a BMP file from data passed from a buffer containing multiple consecutive INT16 data representing different image plans.
 * Only one plan is saved
 * This function is only used for internal debugging
 * 
 * @param[in]		p_fileName		Name of the BMP file
 * @param[in]		p_buffer		Buffer containing the image data
 * @param[in]		p_width			Width of the image (in pixels)
 * @param[in]		p_height		Height of the image (in pixels)
 * @param[in]		p_nbPlans		Number of plans in the image
 * @param[in]		p_plan			Defines which plan will be saved to disk
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				INT16, saveBmp, saveBmp32, saveBmp32S, saveBmpC, saveBmpR, saveBmpS
 */
INT32 saveBmpS16( char *p_fileName, INT16 *p_buffer, INT32 p_width, INT32 p_height, UINT32 p_nbPlans, UINT32 p_plan );

#ifdef __cplusplus
}
#endif

#endif //__ZEFYR_BMP_FILES_H
