/*
Company :		ZEFYR Technologies
File :			coresDeltas.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : core and deltas position determination

Features : routines to process cores position
    
Functions : 

	findCoresDeltas	: main function that allow to find cores and deltas in an image
	getCoreDeltasXY	: retreives the coordinates of the cores ans deltas from a coreMap processed with poincare function	
	processOriMap	: calculate orientation map from a gradient image
	scaleOriMap		: scale an orientation map
	findCoreOrientation	: processes the orientation of the main core of the image
	getVerticalAxis : Searches the position of the vertical axis of symmetry.
	getHorizontalAxis : Searches the position of the horizontal axis the "core" of an arch image.

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	filters.h
	zMaths.h
	coresDeltas.h
	geometry.h

Comments : none

*/


#ifndef __ZEFYR_CORES_DELTAS_H
#define __ZEFYR_CORES_DELTAS_H

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Main function of this file : Looks for core and deltas in the fingerprint image and store it into the template structure.
 * Every data needed by this function is in p_data
 * 
 * @param[in,out]	p_data	Structure containing the input image, the template structure, and temporary working buffers
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 *
 * @see				BioEngineData
 */
INT32 findCoresDeltas( BioEngineData *p_data );

/**
 * Finds the position of the core and deltas and stores it into the template structure.
 * 
 * @param[in]		p_src		Source image.containing the poincar indexes of the original image
 * @param[in]		p_quality	Quality map : bad pixels won't be processed
 * @param[in]		p_width		Width of p_width
 * @param[in]		p_height	Height of p_height
 * @param[in,out]	p_template	Template structure : will be filled with de position of the core and deltas (if the function finds them).
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 * 
 * @see				poincare
 */
INT32 getCoreDeltasXY( INT8 *p_src, UINT8 *p_quality, UINT32 p_width, UINT32 p_height, Template *p_template );

/**
 * Processes the orientation map of the fingerprint image.
 * The orientation map is processed from the gradient of the source image.
 * 
 * @param[in]		p_gradient	Gradient of the source image packed in a INT16 buffer. Even indexes = X gradient. Odd indexes = Y gradient.
 * @param[out]		p_ori		Orientation	map generated by the function. p_ori will be smaller than p_gradient (default is 4X smaller)
 * @param[in]		p_quality	Quality map : bad pixels won't be processed	
 * @param[in]		p_width		Width of p_gradient
 * @param[in]		p_height	Height of p_gradient
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 * 
 * @see				sobelGradient
 */
INT32 processOriMap( INT16 *p_gradient, UINT8 *p_ori, UINT8 *p_quality, UINT32 p_width, UINT32 p_height );

/**
 * Scales an orientation map by an integer factor.
 * 
 * @param[in]		p_src		Source orientation map
 * @param[out]		p_dst		Destination scaled orientation map
 * @param[in]		p_qual		Quality map : bad pixels won't be processed	
 * @param[in]		p_width		Width of p_src
 * @param[in]		p_height	Height of p_src
 * @param[in]		p_scale		Image scaling factor. p_dst will be of size (p_width*p_scale,p_height*p_scale)
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 */
INT32 scaleOriMap( UINT8 *p_src, UINT8 *p_dst, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 p_scale );

/**
 * Processes the orientation of the main fingerprint core.
 * 
 * @param[in]		p_ori			Orientation map of the fingerprint image
 * @param[in]		p_width			Width of p_curvature
 * @param[in]		p_height		Height of p_curvature
 * @param[in,out]	p_template		Template structure tha t will be filled with the center (main core...) orientation.
 * 
 * @return			ZEF_OK if everything was ok, an error code otherwise.
 * 
 * @see				processOriMap
 */
INT32 findCoreOrientation( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template );

/**	Searches the position of the vertical axis of symmetry.
 * 
 *	@param[in]		p_ori		Orientation map	
 *	@param[in]		p_qual		Quality map (used as a mask to avoid processing unusefull pixels)
 *	@param[in]		p_width		Width of the orientation map
 *	@param[in]		p_height	Height of the orientation map	
 *	@param[out]		p_symAxis	x Position of the vertical axis of symmetriy	
 * 
 *	@return			ZEF_OK if everything was ok, an error code otherwise.			
*/
INT32 getVerticalAxis( UINT8 *p_ori, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 *p_symAxis );

/**	Searches the position of the horizontal axis the "core" of an arch image.
 *  This point is placed along the vertical axis of symmetry at the point with max curvature.
 * 
 *	@param[in]		p_ori			Orientation map	
 *	@param[in]		p_qual			Quality map (used as a mask to avoid processing unusefull pixels)
 *	@param[in]		p_width			Width of the orientation map
 *	@param[in]		p_height		Height of the orientation map	
 *	@param[in]		p_yAxis			x Position of the vertical axis of symmetriy
 *	@param[out]		p_symAxis		y Position of the "core" point (which is not really a core since arches don't have cores)
 *	@param[in]		p_filterSize	Widht of the vertical (centerd on p_yAxis) analysed to find the core point.	
 * 
 *	@return			ZEF_OK if everything was ok, an error code otherwise.			
*/
INT32 getHorizontalAxis( UINT8 *p_ori, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 p_yAxis, UINT32 *p_symAxis, UINT32 p_filterSize );

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_CORES_DELTAS_H
