/*
Company :		ZEFYR Technologies
File :			customTypes.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	05/05/2003

Description : Custom types defined for SDK use

Features : defines some custom types		
    
Functions : 

  	swapUInt32 : swap long data type from/to little indian (intel) from/to big indian (motorola) 
	swapUInt16: swap short data type from/to little indian (intel) from/to big indian (motorola) 

Structures : 

	Coord16	: used for coordinates representation
	Rect16 : defines a rectangular area
	Minutia : minutia coordinates, orientation and quality
	MinutiaNeighborhood : infrmations used for minutiae post-processing
	Template : template data generated by the algorithm : minutiae, cores, etc...
	TpsData : data stored when thin-plate-spline are used to estimate and correct deformation

Necessary package : none

	staticParameters.h

Comments :none

*/

#ifndef __ZEFYR_CUSTOM_TYPES_H
#define __ZEFYR_CUSTOM_TYPES_H

#include "staticParameters.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __GNUC__
#if defined(__i686__) && !defined(_M_IX86)
	#define _M_IX86 600
#elif defined(__i586__) && !defined(_M_IX86)
	#define _M_IX86 500
#elif defined(__i486__) && !defined(_M_IX86)
	#define _M_IX86 400
#elif defined(__i386__) && !defined(_M_IX86)
	#define _M_IX86 300
#endif
#endif // __GNUC__

/** Our definition of truth... ;-) */
#define TRUE				1

/** Our definition of FALSE */
#define FALSE				0

#ifndef _BASETSD_H_

/** Internal type for a 32 bits signed integer */
typedef	signed int			INT32;

/** Internal type for a 32 bits unsigned integer */
typedef	unsigned int		UINT32;

/** Internal type for a 16 bits signed integer */
typedef	signed short		INT16;

/** Internal type for a 16 bits unsigned integer */
typedef unsigned short		UINT16;

/** Internal type for a 8 bits signed integer */
typedef signed char			INT8;

/** Internal type for a 8 bits unsigned integer */
typedef unsigned char		UINT8;

#endif // _BASETSD_H_

/** Internal type for a 32 floating point value */
typedef float				REAL32;

/** Internal type for a boolean */
typedef signed char			BOOL8;

/** Internal type for a fixed point 16 bits value */
typedef	signed short		FIXED16;

/** Internal type for a fixed point 32 bits value */
typedef	signed int			FIXED32;

/** Define a point with signed 16 bits X and Y coordinates */
typedef struct _Coord16
{
	INT16 m_x; /**< X coordinate */
	INT16 m_y; /**< Y coordinate */

} Coord16, *PCoord16;

/** Defines a rectangle */
typedef struct _Rect16
{
	INT16 m_left;		/**< Top-left corner X coordinate */
	INT16 m_right;		/**< Bottom-Right corner X coordinate */
	INT16 m_top;		/**< Top-left corner XYcoordinate */
	INT16 m_bottom;		/**< Bottom-Right corner Y coordinate */

} Rect16, *PRect16;

/** Defines a minutia : position, quality and orientation */
typedef struct _Minutia
{
	Coord16	m_coord;	/**< Position of the minutia */
	UINT8	m_qual;		/**< Quality of the minutia */
	UINT8	m_ori;		/**< Orientation of the minutia */

} Minutia, *PMinutia;

/** Defines the neighborhood of a minutiae during its processing  */
typedef struct _MinutiaNeighborhood
{
	BOOL8	m_typeMultiple;						/**< if TRUE, minutia is a biffurcation, else it's ridge ending */
	Coord16	m_coord;							/**< Position of the minutia */
	INT16	m_nbMinutiae;						/**< number of minutiae really kept around the minutiae */
	INT16	m_maxNbMinutiae;					/**< number of neighboring minutiae */
	INT16	m_neighbors[MAX_MINUTIAE_NEIGHBOR]; /**< indexes of neighboring minutiae */
	UINT8	m_qual;								/**< Quality of the minutia */
	INT16	m_ori;								/**< 360 minutia orientation */
} MinutiaNeighborhood, *PMinutiaNeighborhood;

/** Defines THE template structure */
typedef struct _Template
{
	// header
	UINT32		m_size;			/**< Size of the template */
	UINT32		m_version;		/**< Version of the template */
	// core and delta part		
	UINT8		m_nbCores;		/**< Number of cores */
	UINT8		m_nbDeltas;		/**< Number of deltas */
	Coord16		*m_cores;		/**< Array of cores */
	Coord16		*m_deltas;		/**< Array of deltas */
	UINT8		m_oriCore;		/**< Orientation of the main core */
	// Minutiae part
	Rect16		m_boundBox;		/**< Bounding rectangle for all minutiae*/
	UINT16		m_nbMinutiae;	/**< Number of minutiae stored in the template*/
	Minutia		*m_minutiae;	/**< Minutiae array*/
	// Other Infos
	UINT8		m_sign[16];		/**< core signature : used to speed up identification */
	UINT16		m_meanFreq;		/**< mean fingerprint ridges spacing frequency */
	UINT8		m_blocSize;		/**< Bloc size for local informations (bloc orientation and quality */
	UINT8		m_localW;		/**< Width of local infos arrays */
	UINT8		m_localH;		/**< Height of local infos arrays */
	INT16		m_localShiftX;	/**< X shift between center of the arrays and first core */
	INT16		m_localShiftY;	/**< Y shift between center of the arrays and first core */ 
	UINT8		*m_ori;			/**< local infos : bloc orientation */
	UINT8		*m_qual;		/**< local infos : bloc quality */

} Template,*PTemplate;

#ifndef ZEF_FIXED_POINT
/** Data structure used for thin plate spline non-linear deformation correction.
 *  Floating point version */
	typedef struct _TpsData
	{
		REAL32	**m_matrixA;
		REAL32	*m_vectorBx;
		REAL32	*m_vectorBy;
		REAL32	*m_srcX;
		REAL32	*m_srcY;
		INT32	*m_index;
		INT32	m_nbPts;
		UINT32	m_error;
		INT32	m_cptPts;
	} TpsData,*PTpsData;
#else
/** Data structure used for thin plate spline non-linear deformation correction.
 *  Fixed point version */
	typedef struct _TpsData
	{
		INT32	**m_matrixA;
		INT32	*m_vectorBx;
		INT32	*m_vectorBy;
		INT32	*m_srcX;
		INT32	*m_srcY;
		INT32	*m_index;
		INT32	m_nbPts;
		UINT32	m_error;
		INT32	m_cptPts;
	} TpsData,*PTpsData;
#endif // ZEF_FIXED_POINT

// functions

/** Used to swap UINT32 data for indiannes independance.
 *	@param[in]		p_data source data
 *  @returns		p_data with all bytes swapped
 */
UINT32 swapUInt32(UINT32 p_data);

/** Used to swap UINT16 data for indiannes independance.
 *	@param[in]		p_data source data
 *  @returns		p_data with all bytes swapped
 */
UINT16 swapUInt16(UINT16 p_data);

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_CUSTOM_TYPES_H
