/*
Company :		ZEFYR Technologies
File :			dpBio.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	19/06/2003

Description : digiPass BioEngine functions

Features : bioEngine...
    
Functions : 

  	dpInit :	initialize everything
	dpUnInit :	free everything
	dpGetMaxTemplateSize : return the max size a template can be (with the actual parameters)		
	dpGetTemplateSize	:	return the size of a given template
	dpGenerateTemplate : generate a template from an image
	dpGetTemplateQuality : return the quality of a given template
	dpSetSecurity	: set the global security threshold
	dpGetSecurity	: get the global security threshold
	dpIdentifyTemplate : compared a template to many others
	dpSetDbMaxSearch	: set maximum seach depth during identification
	dpVerifyTemplate : Function comparing two templates and returning if they match
	dpVerifyAndRefineTemplate : Matches two templates and if the matching score is good makes a composite template
	dpLoadParameters : load the custom bioEngine parameters


Necessary package : 

  none

Comments : none

*/

#ifndef __ZEFYR_DP_BIO_H
#define __ZEFYR_DP_BIO_H


// ******** BioEgine Warnings ********
#define WARNING_OFFSET			2

/** No match was found during authentication/identification */
#define ZEF_NO_MATCH			WARNING_OFFSET

/** Too many minutiae where found during template generation. 
 *  Some of them where not stored in the template */
#define	ZEF_TOO_MANY_MINUTIAE	WARNING_OFFSET+1

/** A core was found, but too close from the border of the image and so was discarded. */
#define ZEF_CORE_CLOSE_TO_BORDER	WARNING_OFFSET+2

// ******** BioEngine Common Errors ********

/** Everything worked, no error */
#define ZEF_OK						1

/** NULL value */
#define ZEF_NULL					(void*)0;

/** A NULL pointer was passed to the function */
#define	ZEF_NULL_POINTER			-1

/** The size of the buffer passed as argument is to small */
#define ZEF_BUFFER_TO_SMALL			-2

/** Memory allocation problem. 
 *  Not enough memory available */
#define ZEF_OUT_OF_MEMORY			-3

/** Error during math function processing */
#define	ZEF_MATHS_ERROR				-4

/** Fingerprint template checksum verification failed.
 *  The template can not be loaded */
#define	ZEF_CHECKSUM_FAILED			-5

/** The version of the template that was loaded is incompatible with the current version of the bioEngine. 
 *  The template can not be loaded. */
#define	ZEF_BAD_VERSION				-6

/** The security threshold passed as argument was not in the range 0-MAX_SCORE (usually MAX_SCORE = 1000) */
#define ZEF_BAD_SECU_THRESHOLD		-7

/** The max identification search depth parameter was not in the range 0-100 */
#define	ZEF_BAD_MAX_SEARCH			-8

/** DpInit has not been called */
#define	ZEF_ENGINE_NOT_INITIALIZED	-9

/** The size passed as argument was zero or an invalid value */
#define	ZEF_BAD_SIZE				-10

/** (Internal error) Image was not cropped to a size multiple of bloc size */
#define ZEF_NOT_BLOC_MULTIPLE		-11

/** The parameters passed as argument to dpSetSpeed were not valid. */
#define ZEF_BAD_SPEED_PARAM			-12

/** The identification process has been stopped by a call to dpStopIdentification */
#define ZEF_STOPPED					-13

// ******** ProcessPreview flags ********

/** A finger has been detected */
#define ZEF_FLAG_FINGER_DETECTED	1

/** Quality is good enough. The user can remove his finger from the sensor. */
#define ZEF_FLAG_REMOVE_FINGER		2

/** Quality is good. This image could be used for fingerprint processing. */
#define ZEF_FLAG_BEST_IMG			4

/** The user must move his finger up */
#define ZEF_FLAG_MOVE_UP			8

/** The user must move his finger down */
#define ZEF_FLAG_MOVE_DOWN			16

/** The user must move his finger left */
#define ZEF_FLAG_MOVE_LEFT			32

/** The user must move his finger right */
#define ZEF_FLAG_MOVE_RIGHT			64

/** The user must press his finger stronger on the sensor.
 *  The image is to light or too small */
#define ZEF_FLAG_PRESS_STRONGER		128
			

#if !defined(WINAPI) && defined(WIN32)
	#define WINAPI __stdcall
#elif !defined(WINAPI)
	#define WINAPI 
#endif

#ifdef __cplusplus
	extern "C" {
#endif

// ******** BioEgine Functions ********

/** Initializes the library (internal memory buffer, pre-compiled data, etc.)
 *  Must be called before any other function. 
 *  If you call a function without a prior call to dpInit the function call will fail.
 *
 *  @return		ZEF_OK if initialization was ok. An error code listed in the error codes table otherwise.
 */
int				WINAPI	dpInit( void );

/** Free all system resources allocated by the library for its internal processing.
 *  Must be called at the end of the calling program. 
 *  Any dpBio function call made after calling dpUnInit will fail.
 * 
 *  @return	ZEF_OK if uninitialization was ok. An error code listed in the error codes table otherwise.	
 */
int				WINAPI	dpUnInit( void );

/** Returns a null terminating string containing the version number of the bioEngine. 
 *  The string is a static string, it should not be deallocated.
 * 
 *  @return		A pointer to a null terminating string containing the version number.
 */
char *			WINAPI  dpGetEngineVersion( void );

/** Returns the maximum size a template can be. 
 *  This size can vary if you change the algorithm parameters using loadParameters. 
 *  So if you have to load custom parameters, always call getMaxTemplateSize after the parameters loading.
 * 
 *	@return		If no error occurred, max size of a template in bytes. An error code listed in the error codes table otherwise.
 */
int				WINAPI	dpGetMaxTemplateSize( void );

/**	Returns the size in bytes of the template passed as argument
 *
 *	@param[in]	p_templateBuffer	A pointer to the buffer containing the template
 *
 *	@return		If no error occurred, size in bytes of the template. An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpGetTemplateSize( unsigned char *p_templateBuffer );

/** Generates a fingerprint template from a fingerprint image. 
 *  The image must be 8 bit grayscale. 
 *
 *	@param[in]		p_imgData			A pointer to the 8 bit greyscale image of a fingerprint. 
 *										Data must be stored line by line in the buffer with first line being the to of the image. 
 *										No line padding.
 *	@param[in]		p_width				Width of the image in pixels
 *	@param[in]		p_height			Height of the image in pixels
 *	@param[out]		p_templateBuffer	A pointer to the buffer that will receive the fingerprint template.
 *										This buffer must be allocated by the caller with enough space to contain one template. @see dpGetMaxTemplateSize
 *
 *	@return		ZEF_OK if template generation was ok. An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpGenerateTemplate( unsigned char *p_imgData, unsigned int p_width, unsigned int p_height, unsigned char *p_templateBuffer ); 

/** Compares two fingerprint templates and indicates if they match or not. 
 *  Returns also the matching score.
 *  
 *	@param[in]		p_templateBuffer1	A pointer to the first fingerprint template to be compared
 *	@param[in]		p_templateBuffer2	A pointer to the second fingerprint template to be compared
 *	@param[out]		p_score				Pointer to the variable that will receive the matching score if the two templates match. 
 *										Is filled with 0 otherwise.
 * 
 *	@return			ZEF_OK if the two template are from the same finger, ZEF_NO_MATCH if they don't match. 
 *					An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpVerifyTemplate( unsigned char *p_templateBuffer1, unsigned char *p_templateBuffer2, unsigned int *p_score );

/** Compares two fingerprint templates and indicates if they match or not. 
 *  Returns also the matching score. If the matching score is above a certain threshold, 
 *  the first template will be updated with information coming from the second template.
 *  It will generate a composite template containing more information of better quality.
 *
 *	@param[in]		p_templateBuffer1	A pointer to the first fingerprint template to be compared
 *	@param[in]		p_templateBuffer2	A pointer to the second fingerprint template to be compared
 *	@param[out]		p_score				Pointer to the variable that will receive the matching score if the two templates match. 
 *										Is filled with 0 otherwise.
 *	@param[out]		p_refined			Pointer to a boolean that will receive TRUE if the first template has been updated, FALSE otherwise.
 *
 *	@return			ZEF_OK if the two template are from the same finger, ZEF_NO_MATCH if they don't match. 
 *					An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpVerifyAndRefineTemplate( unsigned char *p_templateBuffer1, unsigned char *p_templateBuffer2, unsigned int *p_score, char *p_refined );

/** Return the quality of the fingerprint template passed as argument.
 *  Quality ranges from 0 to 100.
 *
 *	@param[in]		p_templateBuffer	A pointer to the fingerprint template whose quality is to be evaluated
 *
 *	@return			If no error occurred, a quality value between 0 and 100. 
 *					An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpGetTemplateQuality( unsigned char *p_templateBuffer );

/** Load custom bioEngine parameters. 
 *  Usually you won't have to use this function. 
 *  But in certain circumstances (for example : to tune the bioEngine to a specific sensor), 
 *  ZEFYR Technologies could provide you a file to be loaded with this function.
 *
 *	@param[in]		p_buffer			A pointer to the buffer containing the parameters loaded from a file
 *	@param[in]		p_size				size of the parameters data buffer
 *
 *	@return			ZEF_OK if dpLoadParameters executed ok. An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpLoadParameters( unsigned char *p_buffer, unsigned int p_size );

/** Set the security level (FAR) to a predefined threshold.
 *  If the threshold is very high then security will be very high but there will be more rejection.
 * 
 *	@param[in]		p_securityLevel		Value of the security level. 
 *										This value ranges from 0 to MAX_SCORE (default=1000).
 *										Typical values are :
 *											- FAR = 10-0 security level = 0
 *											- FAR = 10-1 security level = 25
 *											- FAR = 10-2 security level = 50
 *											- FAR = 10-3 security level = 75
 *											- FAR = 10-4 security level = 100
 *											- FAR = 10-5 security level = 125
 *											- FAR = 10-6 security level = 150	\n\n
 *										WARNING: these typical values can change if some custom
 *										bioEngine parameters are loaded with dpLoadParameters.	
 *	
 *	@return			ZEF_OK if dpSetSecurity executed ok. An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpSetSecurity( unsigned int p_securityLevel );

/** Retrieves the security level that has been set before by dpGetSecurity
 *
 *	@param[out]		p_securityLevel		Pointer to a variable that will receive the security level.
 *										This value will range from 0 to MAX_SCORE (default=1000).
 *
 *	@return			ZEF_OK if dpGetSecurity executed ok. An error code listed in the error codes table otherwise.
 *
 *	@see dpSetSecurity
*/
int				WINAPI	dpGetSecurity( unsigned int *p_securityLevel );

/**	Check if there is a match between a fingerprint template and a set of other fingerprint templates. 
 *	If the algorithm finds a match, it returns the index of the template that match and the matching score.
 * 
 *	@param[in]		p_templatesArray	An array containing pointers to buffers where fingerprint templates are stored.
 *										This array sis the database in which the candidate template will be searched.
 *	@param[in]		p_nbTemplates		Contains the number of elements in p_templatesArray.
 *	@param[in]		p_templateBuffer	Pointer to a buffer containing the candidate template.
 *	@param[out]		p_score				Will receive the matching score if a match is found.
 *	@param[out]		p_templatePos		If a match is found, contains the index of the matching template.
 *										Otherwise it contains the number of elements in templatesArray
 *										(the value of p_nbTemplates) or (in case of error) the index of the
 *										template that caused the error.
 * 
 *	@return			ZEF_OK if a match was found, if a match was not found the function return ZEF_NO_MATCH. 
 *					An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpIdentifyTemplate( unsigned char **p_templatesArray, unsigned int p_nbTemplates, unsigned char *p_templateBuffer, 
											unsigned int *p_score, unsigned int *p_templatePos );
/** This function allows you to stop an identification if you consider it is taking to much time. 
 *	Its main goal is to allow to cancel identification processes.
 *
 *	@param[in]		p_templateBuffer	Pointer to the buffer that was passed as parameter to dpIdentifyTemplate. 
 *										This buffer will identify which thread must be ended in case of multiple identification running in different threads.
 *
 *	@return			ZEF_OK if there was no error. An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpStopIdentification( unsigned char *p_templateBuffer );

/** Set the maximum percentages of the template array database that have to be searched before returning that a match was not found.
 *  This function can be useful to speed up a little bit searches with big databases, but you must know that decreasing the search depth can increase the FRR.
 *
 *	@param[in]		p_maxSearch			Value ranging from 0 to 100 telling the max percentage of the database to be searched
 *
 *	@return			ZEF_OK if there was no error. An error code listed in the error codes table otherwise.
*/ 
int				WINAPI	dpSetDbMaxSearch( unsigned int p_maxSearch );

/** Sets the speed of the algorithm. By default the algorithm is set to normal speed which gives the best performance at a reasonable speed. 
 *  If you really need faster responses, you can accelerate template generation and/or template matching by setting speed to 1 or 2.
 *  This will accelerate the algorithm up to 3 times, but will increase the FRR a little.
 * 
 *	@param[in]		p_type				Can be one of the following values :
 *											- 0 set speed for both template generation and template matching.
 *											- 1 set speed for template generation.
 *											- 2 set speed for template matching.
 *	@param[in]		p_speed				Defines the speed of the algorithm :
 *											- 0 normal speed : the algorithm have it's best FAR/FRR performances.
 *											- 1 medium speed : the algorithm is faster (up to 1.5X normal speed), but FAR/FRR performances are a little lower.
 *											- 2 high speed : the algorithm is really fast (up to 3X normal speed), but FAR/FRR performances are lower.
 *	
 *	@return			ZEF_OK if there was no error. An error code listed in the error codes table otherwise.
*/
int				WINAPI	dpSetSpeed( unsigned int p_type, int p_speed );

/** Generates a preview image of a fingerprint image. 
 *  This preview doesn't show the details of the fingerprint and thus preserves the users privacy.
 *  This function also generates flags that can be used to help the users fingerprint positioning.
 *
 *	@param[in]		p_imgData			Buffer containing the user's fingerprint image.		
 *	@param[in]		p_width				Witdh of p_imgData.
 *	@param[in]		p_height			Height of p_imgData.
 *	@param[out]		p_preview			buffer that will contain the preview image after the call of the function. 
 *										Must be allocated withe the same size as p_imgData.	
 *	@param[out]		p_flags				Can be a combination of the following flags :
 *											- ZEF_FLAG_FINGER_DETECTED : a finger has been detected in imgData.
 *											- ZEF_FLAG_REMOVE_FINGER : imgData's quality is good enough. The user can remove his finger from the sensor.
 *											- ZEF_FLAG_BEST_IMG : imgData quality is good. This image could be used for fingerprint processing.
 *											- ZEF_FLAG_MOVE_UP : The user must move his finger up
 *											- ZEF_FLAG_MOVE_DOWN : The user must move his finger down
 *											- ZEF_FLAG_MOVE_LEFT : The user must move his finger left
 *											- ZEF_FLAG_MOVE_RIGHT : The user must move his finger right
 *											- ZEF_FLAG_PRESS_STRONGER : : The user must press his finger stronger on the sensor.
 *
 *	@return			ZEF_OK if there was no error. An error code listed in the error codes table otherwise.
*/
int				WINAPI  dpProcessPreview( unsigned char *p_imgData, unsigned int p_width, 
										 unsigned int p_height, unsigned char *p_preview, unsigned int *p_flags );


#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_DP_BIO_H