/*
Company :		ZEFYR Technologies
File :			enhancements.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : fingerprint enhancements

Features : differents fingerprint specific enhancements
    
Functions : 

	enhance : main enhancement routine
	fillBlocArray :	cuts the source image into overlapping blocs and store those blocs in a array
	fftBlocArray :	compute the FFT of all blocs in a bloc array
	polarSpectrum : compute the polar spectrum of each bloc of a bloc array
	freqSpectrum :	compute the main frequency of each spectrum bloc of an array of blocs
	powBlocFFT :	enhance some blocs of an array of bloc by multipl
	applyGaborFilter : filter each bloc with N oriented gabor filters tuned the the frequency processed in freqSpectrum function
	gaborFusion : use the N filterd image to find the best orientation map for the image
	findRegions : find regions of connected pixels with the same orientation
	fillRegions : replace regions that are to different from their neighboors
	findPixels :  find pixels that are too different from their neighboors
	imageFusion : reconstruct the final image based on the orientation map
	qualityThreshold : threshold the quality map to define a mask
	maskImage : removes used part of the image, defined by a mask.

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	FFT2D.h
	zMaths.h
	filters.h
	greyLevelTrans.h
	geometry.h
	quicksort.h

Comments : none

*/

#ifndef __ZEFYR_ENHANCEMENTS_H
#define __ZEFYR_ENHANCEMENTS_H

#ifndef ZEF_FIXED_POINT

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Initializes all the data used for fingerprint image enhancement.
 *	Initialization = internal buffers allocation + precompiled tables generation
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise
*/
INT32 initEnhance( void );


/**	Uninitializes and frees all data initialized by initEnhance
 *
 * @return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise
*/
INT32 unInitEnhance( void );


/**	The enhance function
 * 
 *	@param[in]		p_data	Structure containing the image image that will be enhanced. 
 *					Contains also pointers for temporaty buffers that will be allocated and used by "enhance" function	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 enhance( BioEngineData *p_data );


/**	"Cuts" the image into overlapping blocs.
 *	These bloc will then be used by most of the enhancement functions of this file (mostly for filering)
 *	By default we use 64x64 blocs
 * 
 *	@param[in]		p_src			Source image		
 *	@param[out]		p_blocArray	2D Array of buffer that will receive all the image blocs. 
 *									All buffers are allocated by the calling function.	
 *	@param[in]		p_width			Width of the source image
 *	@param[in]		p_height		Height of the source image
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise					
*/
INT32 fillBlocArray( UINT8 *p_src, REAL32 **p_blocArray, UINT32 p_width, UINT32 p_height );


/**	Performs an inplace Fast Fourier Transform on each bloc of the bloc arrays passed as parameter
 * 
 *	@param[in,out]	p_blocArray		2D Arrays of image blocs (size = 64x64 by default)	
 *	@param[in]		p_nbBlocsX		Number blocs in the X dimention of p_blocArray array
 *	@param[in]		p_nbBlocsY		Number blocs in the Y dimention of p_blocArray array
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 fftBlocArray( REAL32 **p_blocArray, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY );


/**	Processes the power spectrum of each bloc.
 *  Power spectrums are processed in polar coordinates in order to ease further frequency estimation process
 * 
 *	@param[in]		p_blocArray		2D Arrays of FFT of image blocs (size = 64x64 by default)			
 *	@param[in]		p_nbBlocsX		Number blocs in the X dimention of p_blocArray array
 *	@param[in]		p_nbBlocsY		Number blocs in the Y dimention of p_blocArray array
 *	@param[out]		p_spectrum		2D Array of polar spectrums of each bloc of p_blocArray
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 polarSpectrum( REAL32 **p_blocArray, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32	**p_spectrum );


/**	Estimated the main frequency of each bloc passed as parameter from its polar spectrum.
 *  Also outputs a qualiy map, representing the confidence level of the frequency estimation of each bloc.
 * 
 *	@param[in]		p_spectrum		2D Array of polar spectrums of each bloc.		
 *	@param[in]		p_nbBlocsX		Number blocs in the X dimention of p_spectrum array
 *	@param[in]		p_nbBlocsY		Number blocs in the Y dimention of p_spectrum array
 *	@param[out]		p_freq			Buffer where the frequency estimation of each bloc will  be stored (size = p_nbBlocsX x p_nbBlocsY)
 *	@param[out]		p_qual			Buffer where the quality of the frequency estimation of each bloc will  be stored (size = p_nbBlocsX x p_nbBlocsY)
 *	@param[out]		p_meanFreq		Mean frequency of all the processd blocs
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 freqSpectrum( REAL32 **p_spectrum, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_freq, REAL32 *p_qual, REAL32 *p_meanFreq );


/**	Boosts frequencies with high magnitude and attenuated frequencies with low amplitude
 *  This processing is only applied to blocs with low frequency estimation quality (because it
 *  tends to "erase" minutiae on good quality blocs).
 * 
 *	@param[in,out]	p_spectrum		2D Array of polar spectrums of each bloc. Some of its blocs will be "boosted" if
 *									their quality (stored in p_qual) is too low.
 *	@param[in]		p_nbBlocsX		Number blocs in the X dimention of p_spectrum array
 *	@param[in]		p_nbBlocsY		Number blocs in the Y dimention of p_spectrum array
 *	@param[in]		p_qual			Buffer containing the quality of the frequency estimation of each bloc (size = p_nbBlocsX x p_nbBlocsY)
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 powBlocFFT( REAL32 **p_spectrum, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_qual );


/**	Convoluates each bloc passed as parameter with a set of N  oriented gabor filters.
 *	Usually N=8 => mean an orientation stop of 22.5.
 *  Each Gabor filter has a frequency parameter adapter to the bloc it will be convoluted with.
 * 
 *	@param[in]		p_blocArray		2D Array of FFT of image blocs (size = 64x64 by default)
 *	@param[out]		p_images		Buffer which will be filled with  an interlaced "multiplan" image.
 *									Each plan represents the input image filtered by one of the N directed Gabor filter.
 *	@param[in]		p_nbBlocsX		Number blocs in the X dimention of p_spectrum array
 *	@param[in]		p_nbBlocsY		Number blocs in the Y dimention of p_spectrum array
 *	@param[in]		p_freq			Buffer containing the frequency estimation of each bloc (size = p_nbBlocsX x p_nbBlocsY)
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 applyGaborFilter(  REAL32 **p_blocArray, INT8 *p_images, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_freq );


/**	Builds one image from the N oriented gabor filtered images passed as input.
 *  The goal of this function is to guess the best orientation from the N filters responses.
 *	It also build a oriention map and quality map for the input image. 
 *  These maps wil be very usefull to determine minutiae orientation and quality.
 * 
 *	@param[in]		p_images		Buffer containing an interlaced "multiplan" image. Each plan represents the input image filtered by one of the N directed Gabor filter.	
 *	@param[in]		p_width			Width of p_images
 *	@param[in]		p_height		Height of p_images
 *	@param[out]		p_quality		Quality map derived from the filtered image magnitude
 *	@param[out]		p_orient		Orientation map  reprensting fingerprint lines flow orientation.
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 gaborFusion( INT8 *p_images, UINT32 p_width, UINT32 p_height, UINT8 *p_quality, UINT8 *p_orient );


/**	Detects if some regions of the input image need orientation reprocessing
 *  Regions that need re-processing are regions whose orientaion is too much different 
 *  from the one of its neighboring pixels and whose size is relatively small
 * 
 *	@param[in]		p_orient		Orientation map of the source image	
 *	@param[out]		p_regions		Region that need re-processing
 *	@param[in]		p_width			Width of p_orient
 *	@param[in]		p_height		Height of p_orient
 *	@param[in]		p_threshold		Threshold under which orientaions will need re-processing
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 findRegions( UINT8 *p_orient, UINT32 *p_regions, UINT32 p_width, UINT32 p_height, REAL32 p_threshold  );


/**	Re-Process some of the image orientations
 * 
 *	@param[in,out]	p_orient		Orientation map of the source image		
 *	@param[in,out]	p_quality		Quality map of the source image
 *	@param[in]		p_intensities	Intensity of each filterd image (needed to know which gabor filtered image has the higher magnitude)
 *	@param[in,out]	p_regions		Regions (pixels) that need to be reprocesed
 *	@param[in]		p_width			Width of p_orient
 *	@param[in]		p_height		Height of p_orient
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 fillRegions( UINT8 *p_orient, UINT8 *p_quality, UINT32 *p_intensities, UINT32 *p_regions, UINT32 p_width, UINT32 p_height );


/**	Detects if some pixels of the input image need orientation reprocessing
 *  Pixels that need re-processing are pixels whose orientaion is too much different 
 *  from the one of its neighboring pixels
 * 
 *	@param[in]		p_img			Orientation map of the source image	
 *	@param[out]		p_pixels		Pixels that need re-processing	
 *	@param[in]		p_width			Width of p_img
 *	@param[in]		p_height		Height of p_img
 *	@param[in]		p_threshold		Threshold under which orientaions will need re-processing
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 findPixels( UINT8 *p_img, UINT32 *p_pixels, UINT32 p_width, UINT32 p_height, REAL32 p_threshold );


/**	Mixes pixels from the N Gabor filterd image and from the orientation map found by gaborFusion function
 * 
 *	@param[in]		p_images		Buffer containing an interlaced "multiplan" image. 
 *									Each plan represents the input image filtered by one of the N directed Gabor filter.		
 *	@param[in]		p_orient		Orientation map found by gaborFusion function
 *	@param[out]		p_dst			Destination "enhanced" image.
 *	@param[in]		p_width			Width of p_images
 *	@param[in]		p_height		Height of p_images
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 imageFusion( INT8 *p_images, UINT8 *p_orient, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );


/**	Transforms the quality map so that pixel with quality below a certain threashold a a quality of zero.
 * 
 *	@param[in,out]	p_quality		Quality map of the "enhanced" image
 *	@param[in]		p_width			Width of p_quality
 *	@param[in]		p_height		Height of p_quality
 *	@param[in]		p_threshold		Thresholf under which quality must be set to zero
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise	
*/
INT32 qualityThreshold( UINT8 *p_quality, UINT32 p_width, UINT32 p_height, UINT32 p_threshold );

/**	Set to 255 parts of the image that are declared as unusefull int the mask image.
 *	Unusefull pixels have zero value in the mask image
 * 
 *	@param			p_src		Source image		
 *	@param			p_dst		Destination (masked) image
 *	@param			p_width		Width of source image
 *	@param			p_height	Height of source image	
 *	@param			p_mask		Mask for image (size = p_width/FUSION_BLOC_SIZE x p_height/FUSION_BLOC_SIZE )
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 maskImage( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT8 *p_mask );

#ifdef __cplusplus
}
#endif

#endif // ZEF_FIXED_POINT

#endif // __ZEFYR_ENHANCEMENTS_H
