/**
 * Company :		ZEFYR Technologies
 * File :			filters.h
 * Author :		Matthieu Perreira Da Silva 
 * Fisrt release :	09/05/2003
 * 
 * Description : filters like Gabor, median, erosion, dilation, etc.
 * 
 * Features : recursive and non recursive filtering		
 *     
 * Functions : 
 * 
 * 	medianFind : helper function that find the median element of an array of n elements
 * 	medianFilter : recursive and non recursive median filter
 * 	medianFind32 : helper function that find the median element of an array of n elements (for long)
 * 	medianFilter32 : recursive and non recursive median filter (for long)
 * 	binaryMedianFilter : recursive and non recursive median filter (binary)
 * 	erosionFilter32 : morphological filter : grayscale erosion (for long) 
 * 	generateGaborFFT : generates a gabor filter in the frequency domain
 * 	freqConvolution : double convolution in frequancy domain (used with real filters)
 * 	poincare : process poincare index of pixels of an orientation map
 * 	sobelGradient : process X and Y gradient with sobel filters
 * 	curvature : process curvature of an orientation map
 * 
 * Necessary package : 
 * 
 * 	customTypes.h
 * 	staticParameters.h
 * 	errorCodes.h
 * 	zMaths.h
 * 	limits.h
 * 
 * Comments : none
 * 
 */

#ifndef __ZEFYR_FILTERS_H
#define __ZEFYR_FILTERS_H

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Finds the median value of an array of UINT8.
 * This findMedian routine is based on the algorithm described in
 * "Numerical recipes in C", Second Edition.
 * It returns the median value of an array of n elements.
 * Elements passed in array p_arr are moved during the processing.
 * You need to take it into account when using this function.
 * 
 * @param[in,out]	p_arr	Array from which the median value has to be found
 * @param[in]		p_n		Number of elements in array p_arr
 * 
 * @return			The median value of the array p_arr
 *
 * @see				medianFilter, medianFind32, medianFilter32
 */
UINT8 medianFind(UINT8 p_arr[], UINT32 p_n);

/**
 * Median filter.
 * Replaces the value of each pixel in the image by the median value of its neigbours.
 * If you pass the same image as source and destination you will have a recursif filter.
 * You will have a normal filter otherwise
 * 
 * @param[in]		p_src			Source image
 * @param[out]		p_dst			Destination image
 * @param[in]		p_width			Horizontal size of source and destination images
 * @param[in]		p_height		Vertical size of ource and destination images
 * @param[in]		p_filterSize	Size of the structuring element (size of the filter)
 * 
 * @return			Error code
 *
 * @see				medianFind, medianFind32, medianFilter32
 */
INT32 medianFilter( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );

/**
 * Finds the median value of an array of UINT8.
 * Same as medianFind, but for UINT32 data type.
 * @see		medianFilter, medianFind, medianFilter32
 */
UINT32 medianFind32(UINT32 p_arr[], UINT32 p_n);

/**
 * Median filter.
 * Replaces the value of each pixel in the image by the median value of its neigbours.
 * Same as medianFind, but for UINT32 data type.
 * @see		medianFilter, medianFind32, medianFilter32
 */
INT32 medianFilter32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );

/**
 * Binary median filter.
 * Replaces the value of each pixel in the image by the median value of its neigbours.
 * Is only valid for images with 0 and 255 values.
 * If you pass the same image as source and destination you will have a recursif filter.
 * You will have a normal filter otherwise.
 * 
 * @param[in]		p_src			Source image
 * @param[out]		p_dst			Destination image
 * @param[in]		p_width			Horizontal size of source and destination images
 * @param[in]		p_height		Vertical size of ource and destination images
 * @param[in]		p_filterSize	Size of the structuring element
 * 
 * @return			Error code
 *
 * @see				medianFilter, medianFilter32
 */
INT32 binaryMedianFilter( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );

/**
 * Erosion filter.
 * Closes holes in black filed regions. Can be used as a preprocession to skeltonization.
 * If you pass the same image as source and destination you will have a recursif filter.
 * You will have a normal filter otherwise.
 * 
 * @param[in]		p_src			Source image
 * @param[out]		p_dst			Destination image
 * @param[in]		p_width			Horizontal size of source and destination images
 * @param[in]		p_height		Vertical size of ource and destination images
 * @param[in]		p_filterSize	Size of the structuring element
 * 
 * @return			Error code
 */
INT32 erosionFilter32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );

#ifndef ZEF_FIXED_POINT
/**
 * Gabor filter genereration in Fourier space
 * A Gabor filter is a frequency and orientation selective filter.
 * It is usually used to detect repetitive oriented pattern in textures.
 * 
 * @param[out]		p_filter		Destination filter
 * @param[in]		p_size			Size of the filer
 * @param[in]		p_freq			Frequency to which the filter will be sensitive
 * @param[in]		p_angle			Orientation to which the filter will be sensitive
 * 
 * @return			Error code
 */
INT32 generateGaborFFT( REAL32 *p_filter, REAL32 p_size, REAL32 p_freq, REAL32 p_angle );

/**
 * Convolution in Fourier space
 * Time/freq space convolution with a real filter (non imaginary part)
 * In fact its just some multiplications since we are in fourier space.
 * It allows to perform fast convolution of real data and real filters.
 * 
 * @param[in]		p_src			Fourier transform of the source image
 * @param[out]		p_dst			result of th convolution in th Fourier space.
 * @param[in]		p_filter		Fourier transform of the filter 
 * 
 * @return			Error code
 */
INT32 freqConvolution( REAL32 *p_src, REAL32 *p_dst, REAL32 *p_filter );
#endif // ZEF_FIXED_POINT

/**
 * Processes Poincare index for each pixel of an image
 * Replaces the value of each pixel in the image by the poincare index of that pixel
 * This index is based on neigouring orientation of the pixel
 * 
 * @param[in]		p_src			Source image (orientation map)
 * @param[out]		p_dst			Destination image
 * @param[in]		p_width			Horizontal size of source and destination images
 * @param[in]		p_height		Vertical size of ource and destination images
 * @param[in]		p_filterSize	Size of the neigbouring area.
 * 
 * @return			Error code
 */
INT32 poincare( UINT8 *p_src, INT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );

/**
 * Processes the gradient of the image using a sobel filter
 * Generates two gradient maps (vertical and horizontal) from the source image
 * 
 * @param[in]		p_src			Source image
 * @param[out]		p_gradient		Destination map (the image is twice as large as src, size for each pixel we store dx and dy
 * @param[in]		p_width			Horizontal size of source image
 * @param[in]		p_height		Vertical size of source image
 * 
 * @return			Error code
 */
INT32 sobelGradient( UINT8 *p_src, INT16 *p_gradient, UINT32 p_width, UINT32 p_height );

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_FILTERS_H 
