/*
Company :		ZEFYR Technologies
File :			FixedPpoint.c
Author :		Matthieu Perreira Da Silva
Fisrt release :	24/03/2004

Description : FixedPoint MACROS

Functions : 

    cdft: Complex Discrete Fourier Transform
    rdft: Real Discrete Fourier Transform

MACROS : 

    void cdft(int p_size, int p_sign, float *p_data, int *p_inPlace, float *p_trigo);
	void rdft(int p_size, int p_sign, float *p_data, int *p_inPlace, float *p_trigo);

Comments : 

Necessary package : none

*/

#include "customTypes.h"

#ifndef __ZEFYR_FIXED_POINT_H
#define __ZEFYR_FIXED_POINT_H

/** Q7_8 format = 1 bits for integer part + 14 bits for the fractional part */
#define Q1_14	14

/** Q7_8 format = 7 bits for integer part + 8 bits for the fractional part */
#define	Q7_8	8

/** Q6_9 format = 6 bits for integer part + 9 bits for the fractional part */
#define	Q6_9	9

/** Q4_11 format = 4 bits for integer part + 11 bits for the fractional part */
#define Q4_11	11

/** Q15_16 format = 15 bits for integer part + 16 bits for the fractional part */
#define	Q15_16  16

/** 1.0f in Q1_14 representation */
#define CONST_1_0_Q1_14 16384

/** 0.5f in Q1_14 representation */
#define CONST_0_5_Q1_14 8192 


// ******** The basic operations performed on two numbers a and b of fixed ********
// ******** point q format returning the answer in q format				   ********

/** Fixed point add (a and b with same q representation) */
#define FADD(a,b) ((a)+(b))

/** Fixed point substract (a and b with same q representation)*/
#define FSUB(a,b) ((a)-(b))

/** Fixed point multiply (a and b with same q representation)*/
#define FMUL(a,b,q) (((a)*(b)+(INT16)(1<<(q-1)))>>(q))

/** Fixed point divide (a and b with same q representation)*/
#define FDIV(a,b,q) (((a)<<(q))/(b))


// ******** The basic operations where a is of fixed point q format and b is an integer ********

/** Adds an integer to a fixed point number */
#define FADDI(a,b,q) ((a)+((b)<<(q)))

/** Substracts an integer from a fixed point number */
#define FSUBI(a,b,q) ((a)-((b)<<(q)))

/** multipies a fixed point number by an integer*/
#define FMULI(a,b) ((a)*(b))

/** divides a fixed point number by an integer*/
#define FDIVI(a,b) ((a)/(b))

/** convert a from q1 format to q2 format */
#define FCONV(a, q1, q2) (((q2)>(q1)) ? (a)<<((q2)-(q1)) : (a)>>((q1)-(q2)))


// ******** the general operation between a in q1 format and b in q2 format ********
// ******** returning the result in q3 format								********

/** Fixed point add (a and b with different q representation) */
#define FADDG(a,b,q1,q2,q3) (FCONV(a,q1,q3)+FCONV(b,q2,q3))

/** Fixed point substract (a and b with different q representation)*/
#define FSUBG(a,b,q1,q2,q3) (FCONV(a,q1,q3)-FCONV(b,q2,q3))

/** Fixed point multiply (a and b with different q representation)*/
#define FMULG(a,b,q1,q2,q3) FCONV((a)*(b), (q1)+(q2), q3)

/** Fixed point divide (a and b with different q representation)*/
#define FDIVG(a,b,q1,q2,q3) (FCONV(a, q1, (q2)+(q3))/(b))


// ******** convert to and from floating point ********

/** Convert a floating point number to a 32 bits fixed point number */
#define FTOFIX32(d,q) ((FIXED32)( (d)*(REAL32)(1L<<(q))))

/** Convert a floating point number to a 16 bits fixed point number */
#define FTOFIX16(d,q) ((FIXED16)( (d)*(REAL32)(1<<(q))))

/** Convert an integer number to a 32 bits fixed point number */
#define ITOFIX32(d,q) ((FIXED32)( (d)<<(q) ))

/** Convert an integer number to a 16 bits fixed point number */
#define ITOFIX16(d,q) ((FIXED16)( (d)<<(q) ))

/** Convert fixed point number to a floating point number */
#define TOFLT(a,q) ( (REAL32)(a) / (REAL32)(1L<<(q)) )

/** Convert fixed point number to an integer number */
#define TOINT(a,q) ( (INT32)(a) / (1L<<(q)) )

#endif //__ZEFYR_FIXED_POINT_H
