/*
Company :		ZEFYR Technologies
File :			geometry.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	07/05/2003

Description : Geometric trnasformations functions

Features : different transformation functions : scaling, rotations, etc.		
    
Functions : 

	scale			: scale an image by a real factor (nearest neighbour )
	integerScale	: scale an image by an integer factor (bilinear interpolation)
	integerScale32	: scale an image by an integer factor (bilinear interpolation)
	integerScaleS	: scale an image by an integer factor (bilinear interpolation)
	scaleOri		: scale an orientation map
	checkBlocMultiple : crops an image to a bloc size multiple 

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h

Comments : none

*/

#ifndef __ZEFYR_GEOMETRY_H
#define __ZEFYR_GEOMETRY_H

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Image scaleing by an integer factor 
 *  scale p_src by p_scale and stores the result in p_dst
 *  p_dst must be allocated by the caller with enough space to hold an image
 *  of p_width*p_height*p_scale*p_scale bytes
 *  p_src and p_dst can be the same pointer (in place processing )
 *
 *	@param[in]		p_src		Source image	
 *	@param[out]		p_dst		Destination (scaled) image
 *	@param[in]		p_width		Width of the source image
 *	@param[in]		p_height	Height of the source image	
 *	@param[in]		p_scale		Integer scaling factor
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 scale( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale );

/** Check if the size of the image is a multiple of SEGM_BLOC_SIZE
 *  if not it crops the image so that its size is a multiple of SEGM_BLOC_SIZE.
 *  No memory allocation is done since the dest image size will be smaller or equal to  the original size
 * 
 *	@param[in,out]	p_src		Source image [in], cropped image [out]
 *	@param[in,out]	p_width		Width of the source image [in], Width of the cropped image [out]
 *	@param[in,out]	p_height	Height of the source image [in] Height of the cropped image [out]		
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 checkBlocMultiple( UINT8 *p_src, UINT32 *p_width, UINT32 *p_height );

/** Image scaleing by an integer factor 
 *  scale p_src by p_scale and stores the result in p_dst
 *  p_dst must be allocated by the caller with enough space to hold an image
 *  of p_width*p_height*p_scale*p_scale bytes
 *  p_src and p_dst cannot be the same pointer (no in place processing )
 *
 *	@param[in]		p_src		Source image	
 *	@param[out]		p_dst		Destination (scaled) image
 *	@param[in]		p_width		Width of the source image
 *	@param[in]		p_height	Height of the source image	
 *	@param[in]		p_scale		Integer scaling factor
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 integerScale( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale );

/** a version of integerScale for UINT32 data arrays 
 * 
 *	@see integerScale
*/
INT32 integerScale32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale );

/** a version of integerScale for INT8 data arrays 
 * 
 *	@see integerScale
*/
INT32 integerScaleS( INT8 *p_src, INT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale );

/** Scaling adapted to orientation maps
 *	
 *	@param[in]		p_src		Source image	
 *	@param[out]		p_dst		Destination (scaled) image
 *	@param[in]		p_width		Width of the source image
 *	@param[in]		p_height	Height of the source image	
 *	@param[in]		p_factor	Integer scaling factor	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 scaleOri( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_factor );

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_GEOMETRY_H 
