/*
Company :		ZEFYR Technologies
File :			greyLevelTrans.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	06/05/2003

Description : Image normalization functions, K-means, etc.

Features : normalization functions		
    
Functions : 

	meanStdDev : evaluation the mean and standard deviation of an image
	localMeanStdDev : evaluation the mean and standard deviation of blocs of an image
	normalize : set the mean and standard deviation of an image to the given values
	normalizeLocally : set the mean and standard deviation of pixel neiborhoods to the given values
	kMeans : classify the pixels of an image using a k-means algortihms (useb for segmentation)
	thresholdImage : binarize an image, based on the threshold level

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	zMaths.h
	geometry.h
	BMPfiles.h
	filters.h

Comments : none

*/

#ifndef __ZEFYR_NORMALIZATION_H
#define __ZEFYR_NORMALIZATION_H

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**	Evaluate the mean and standard deviation of an image.
 *  The input image is left unchanged
 * 
 *	@param[in]		p_src			Source image
 *	@param[in]		p_width			Width of the source image
 *	@param[in]		p_height		Height of the source image
 *	@param[out]		p_mean			Mean of source image
 *	@param[out]		p_stdDev		Standard deviation of source image
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 meanStdDev( UINT8 *p_src, UINT32 p_width, UINT32 p_height, INT32 *p_mean, INT32 *p_stdDev );

/**	Normalizes image p_src into p_dst with p_mean mean and p_stdDev standard deviation.
 *	if p_mean = -1 p_dst mean is left unchanged (is left to p_src mean) \n
 *	if p_stdDev = -1 p_dst standard deviation is left unchanged (is left to p_src standard deviation) \n
 *  p_dst can be the same as p_src (in place processing )
 * 
 *	@param[in]		p_src		Source image
 *	@param[out]		p_dst		Destination (normalized) image
 *	@param[in]		p_width		Width of source image
 *	@param[in]		p_height	Height of source image
 *	@param[in]		p_mean		Target mean for normalized image (p_dst)
 *	@param[in]		p_stdDev	Target standard deviation for normalized image (p_dst)	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 normalize( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_mean, INT32 p_stdDev );

/**	Evaluates the local mean and standard deviation on SEGM_BLOC_SIZExSEGM_BLOC_SIZE blocs of an image.
 *	p_mean and p_stdDev arrays must be allocated before calling this function. \n
 *  Their max size is (p_width/SEGM_BLOC_SIZE)*((p_height/SEGM_BLOC_SIZE) \n 
 *	p_width and p_height must be multiples of SEGM_BLOC_SIZE \n
 *  p_src is left unchanged
 * 
 *	@param[in]		p_src		Source image
 *	@param[in]		p_width		Width of source image
 *	@param[in]		p_height	Height of source image
 *	@param[out]		p_mean		Image representing the mean of each SEGM_BLOC_SIZExSEGM_BLOC_SIZE bloc of the source image
 *	@param[out]		p_stdDev	Image representing the standard deviation of each SEGM_BLOC_SIZExSEGM_BLOC_SIZE bloc of the source image	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 localMeanStdDev( UINT8 *p_src, UINT32 p_width, UINT32 p_height, UINT32 *p_mean, UINT32 *p_stdDev );

/** normalizes image p_src into p_dst with p_mean mean and p_stdDev standard deviation.
 *  Normalization is done pixel by pixel on a local neiborhood. 
 *  The size of the neiborhood is defined by SEGM_BLOC_SIZE static parameter. \n
 *  if p_mean = -1 p_dst mean is left unchanged (is left to p_src mean) \n
 *  if p_stdDev = -1 p_dst standard deviation is left unchanged (is left to p_src standard deviation) \n
 *  p_dst can be the same as p_src (in place processing )
 * 
 *	@param[in]		p_src		Source image	
 *	@param[out]		p_dst		Destination (Locally normalized) image
 *	@param[in]		p_width		Width of source image
 *	@param[in]		p_height	Heigt of source image
 *	@param[in]		p_mean		Target mean for normalized image (p_dst)
 *	@param[in]		p_stdDev	Target standard deviation for normalized image (p_dst)		
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 normalizeLocally( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_mean, INT32 p_stdDev );

/** k-means on the image LUT.
 *  If you choose 4 classes you'll get an image with 4 grey level (0,1,2,3) \n
 *  p_src and p_dst can be the same pointer
 *
 *	@param[in]		p_src			Source image
 *	@param[out]		p_dst			Destination (kMeans) image
 *	@param[in]		p_width			Width of the source image
 *	@param[in]		p_height		Height of the source image
 *	@param[in]		p_nbClasses		Number of classes for the image LUT kMean classification.
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 kMeans( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_nbClasses );


/** Image threshold (Binarization).
 *  If pixel value > threshold dst image is 255 else dst image is 0 
 * 
 *	@param[in]		p_src			Source image
 *	@param[out]		p_dst			Destination image
 *	@param[in]		p_width			Width of source image
 *	@param[in]		p_height		Height of source image
 *	@param[in]		p_threshold		Threshold under which p_dst pixels will be set to 0.
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 thresholdImage( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_threshold );

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_NORMALIZATION_H
