/*
Company :		ZEFYR Technologies
File :			localParameters.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	11/06/2003

Description : local parameters determination

Features : fill template with orientation and quality infos	
    
Functions : 

	processLocalParameters : main function used to process local parameters
	setLocalParamSize : precalculate some info used to process local parameters (arrays size, shifts, etc.)
	fillLocalParams : processes local parameters (bloc orientation, bloc quality, etc)
	generateIndex : Genrate a local infos index based on the orientations around the main core

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	zMaths.h

Comments : none

*/

#ifndef __ZEFYR_LOCAL_PARAMETERS_H
#define __ZEFYR_LOCAL_PARAMETERS_H

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**	Main local parameters function
 *  Fills the template structure with local parameter values processed from quality and orientation maps
 * 
 *	@param[in,out]		p_data		BioEngine data structure	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise						
*/
INT32 processLocalParameters( BioEngineData *p_data );	

/**	Analyses the quality map in order to determine local parameters sizes
 *  Processed parameters are : 
 *  - local info tables width and height
 *  - local info bloc size
 *  - local info shifts relative to the main fingerprint core
 * 
 *	@param[in]		p_quality		Quality map (from which the different size parameters are processed
 *	@param[in]		p_width			Width of the quality map
 *	@param[in]		p_height		Height of the quality map
 *	@param[in,out]	p_template		Template corresponding to our source image	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 setLocalParamSize( UINT8 *p_quality, UINT32 p_width, UINT32 p_height, Template *p_template );

/**	Fills the orientation and quality arrays of the template structure
 *  These arrays are filled with blocwise orientation and quality
 * 
 *	@param[in]		p_qual			Quality map
 *	@param[in]		p_ori		    Orientation map
 *	@param			p_width			Width of quality and orientation maps
 *	@param			p_height		Height of quality and orientation maps
 *	@param			p_template		Template corresponding to our source image. (will be filled with local ori and qual).
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 fillLocalParams( UINT8  *p_qual, UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template );

/**	Generate a core index that will be used to speed up identifications
 *  This feature is not very mature, but its a simple trick that helps identifcation.
 *  Core index = local orientation differences around the main fingerprint core.
 *  Together with main core orientation, it provide an translation and rotation invariant index.
 * 
 *	@param			p_ori			Orientation map
 *	@param			p_width			Width of  orientation map
 *	@param			p_height		Height of orientation map
 *	@param			p_template		Template corresponding to our source image. (will be filled with core index).
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 generateIndex( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template );

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_LOCAL_PARAMETERS_H
