/*
Company :		ZEFYR Technologies
File :			matcher.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	02/07/2003

Description : Minutiae matching algorithm
Features : Fast matching !!!
    
Functions : 

	matchTemplate : main routine for template matching
	fastVerifyTemplate : same as verifyTemplate, but using a fast core based registration algorithm
	registerMinutiae : Main registering function, align to minutiae sets
	register1stPass : first step of minutiae alignment, finds a approximate alignment
	register2ndPass : second step of minutiae alignment, refines the fisrt estimation
	applyRegistration : modifies a template using  the registration informations (rotation, translation) 
	findMatches : searches the corresponding minutiae between two aligned templates
	estimateDeformations : uses a thin plate spline model to estimate the local deformation of the fingerprints
	applyDeformations : deforms a template using the model calculated with estimateDeformations
	matchOri : calculate a matching score based on bloc orientations
	computeScore : Main function to calculate the matching score of two aligned templates 
	findRotation : Fast registration algorithm used in fastVerifyTemplate
	scoreIndex : processes a distance score used for linear classification during identification
	sortTemplates : sort an array of template according to their distance score

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	zMaths.h
	tools.h
	tpsWarper.h
	quicksort.h

Comments : none

*/

#ifndef __ZEFYR_MATCHER_H
#define __ZEFYR_MATCHER_H

#ifndef ZEF_FIXED_POINT

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif


/**	Mid level template matching function.
 *  Given two templates as input, this function will compare them and determine if they match.
 *  It return the matching score and a 2D array of matching minutiae.
 * 
 *	@param[in]		p_template1			First template to compare (reference template)
 *	@param[in]		p_template2			Second template to compare (candidate template)
 *	@param[out]		p_matches			2D array of size [p_nbMatches,2] containing pairs of matching minutiae.
 *	@param[out]		p_nbMatches			Number of matching minutiae found
 *	@param[out]		p_score				Matching score (usually matching 0 <= score <= 1000
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise
 * 
 *  @see			verifyTemplate
*/
INT32	matchTemplate( Template *p_template1, Template *p_template2, INT32 **p_matches, 
						INT32 *p_nbMatches, UINT32 *p_score );

/**	High level template matching function.
 *  Given two packed templates buffers as input, this function will compare them and determine if they match
 *  It returns the matching score computed during the comparison. \n
 *  This function is identical to verifyTemplate except it performs registration using the findRotation 
 *	function (a fast core based registration function).
 * 
 *	@param[in]		p_templateBuffer1	First packed template to compare (reference template)	
 *	@param[in]		p_templateBuffer2	Second packed template to compare (candidate template)	
 *	@param[out]		p_score				Matching score (usually matching 0 <= score <= 1000
 * 
 *	@return			ZEF_OK if the two templates match, ZEF_NO_MATCH if everything was ok but the templates do not match.
 *					An error code listed in "errorCodes.h" otherwise
 *
 *	@see			verifyTemplate
*/
INT32	fastVerifyTemplate( UINT8 *p_templateBuffer1, UINT8 *p_templateBuffer2, UINT32 *p_score );

/**	Mid level template registration (alignment) function.
 *  Given to templates as input, this function aligns the two templates and outputs the translation and rotation
 *  needed to align p_templ2 with p_templ1.
 *	Performs a two stages registration process using functions register1stPass and register2ndPass
 * 
 *	@param[in]		p_templ1		Reference template	
 *	@param[in]		p_templ2		Candidate template
 *	@param[out]		p_dx			X translation between  p_templ1 and p_templ2
 *	@param[out]		p_dy			Y translation between  p_templ1 and p_templ2
 *	@param[out]		p_dTheta		Rotation p_templ1 and p_templ2
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise	
 *
 *	@see			register1stPass, register2ndPass
*/
INT32	registerMinutiae( Template *p_templ1, Template *p_templ2, INT32 *p_dx, INT32 *p_dy, INT32 *p_dTheta );

/**	Low level template registration (alignment) function.
 *  Given to templates as input, this function aligns roughly the two templates 
 *  and outputs the translation and rotation.
 * 
 *	@param[in]		p_templ1		Reference template	
 *	@param[in]		p_templ2		Candidate template
 *	@param[out]		p_dx			X translation between  p_templ1 and p_templ2
 *	@param[out]		p_dy			Y translation between  p_templ1 and p_templ2
 *	@param[out]		p_dTheta		Rotation p_templ1 and p_templ2		
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise
*/
INT32	register1stPass( Template *p_templ1, Template *p_templ2, INT32 *p_dx, INT32 *p_dy, INT32 *p_dTheta );

/**	Low level template registration (alignment) function.
 *  Given to templates as input and their approximated relative translation and rotations, 
 *	this function aligns precisely the two templates and outputs the corrected translation and rotation.
 * 
 *	@param[in]		p_templ1		Reference template	
 *	@param[in]		p_templ2		Candidate template
 *	@param[in,out]	p_dx			X translation between  p_templ1 and p_templ2
 *	@param[in,out]	p_dy			Y translation between  p_templ1 and p_templ2
 *	@param[in,out]	p_dTheta		Rotation p_templ1 and p_templ2		
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32	register2ndPass( Template *p_templ1, Template *p_templ2, INT32 *p_dx, INT32 *p_dy, INT32 *p_dTheta );

/**	Rotates and translates a template
 *  Given a template and some rotation and translation parameter, this function translate and rotates all the data
 *  of the input template (minutiae, local informations, core orientation, etc.)
 * 
 *	@param[in,out]	p_templ			Template to be rotated and translated
 *	@param[in]		p_dx			X translation to be applied
 *	@param[in]		p_dy			Y translation to be applied
 *	@param[in]		p_dTheta		Rotation
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32	applyRegistration( Template *p_templ, INT32 p_dx, INT32 p_dy, INT32 p_dTheta );

/**	Given two aligned templates, this function will find all matching minutiae and return them in an array.
 * 
 *	@param[in]		p_templ1		Reference template	
 *	@param[in]		p_templ2		Candidate template
 *	@param[out]		p_matches		2D array of size [p_nbMatches,2] containing the matching minutiae pairs
 *	@param[out]		p_nbMatches		number of matching minutiae pairs in p_matches
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32	findMatches( Template *p_templ1, Template *p_templ2, INT32 **p_matches, INT32 *p_nbMatches );

/**	Estimated the plastic deformations between two templates using a Thin Plate Spline algorithme
 * 
 *	@param[in]		p_templ1		Reference template	
 *	@param[in]		p_templ2		Candidate template
 *	@param[in]		p_matches		2D array of size [p_nbMatches,2] containing the matching minutiae pairs
 *	@param[in]		p_cptMatches	number of matching minutiae pairs in p_matches	
 *	@param[in]		p_alpha			Level of "elasticity" of the TPS model. p_alpha = 0.0 mean very elastic model. p_alpha = 1.0 means very little elasticity
 *	@param[out]		p_data			TPS deformation parameters (will be used by applyDeformations to modify a template)
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
 *
 *	@see			estimateDeformations
*/
INT32   estimateDeformations( Template *p_templ1, Template *p_templ2, INT32 **p_matches, UINT32 p_cptMatches, REAL32 p_alpha, TpsData *p_data );

/**	Deforms a template according to the TPS model data processed by findMatches
 * 
 *	@param[in,out]	p_templ			Template that will be "deformed"
 *	@param[out]		p_diffs			Value representing the "amount" of deformation that was applied to p_templ.
									The higher this value, the more p_templ has been deformed.
 *	@param[in]		p_data			TPS deformation parameters (computed by estimateDeformations)
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise	
 *
 *	@see			findMatches
*/
INT32	applyDeformations( Template *p_templ, UINT32 *p_diffs, TpsData *p_data );

/**	Compares the local orientation data stored in each template
 * 
 *	@param[in]		p_templ1		
 *	@param[in]		p_templ2		
 * 
 *	@return			Maching score of the orientation fields matching. The lower the value, the best the matching.			
*/
UINT32  matchOri( Template *p_templ1, Template *p_templ2 );

/**	Computes the matching score between two aligned templates
 *  The score is base not only on the number of matching minutiae, but alos on their relative positionning error,
 *  their quality, and other global and local parameters (global frequency, local orientations, etc).
 * 
 *	@param[in]		p_templ1		Reference template	
 *	@param[in]		p_templ2		Candidate template
 *	@param[in]		p_matches		2D array of size [p_nbMatches,2] containing the matching minutiae pairs
 *	@param[in]		p_nbMatches		number of matching minutiae pairs in p_matches
 *	@param[out]		p_score			Matching score
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32   computeScore( Template *p_templ1, Template *p_templ2, INT32 **p_matches, INT32 p_nbMatches, UINT32 *p_score );


/**	Fast (core based) low level template alignment function.
 *  This functions aligns two template provided their core position are accurate 
 *	(it only searched for a rotation parameter, so if a translation is needed it won't find it!)
 * 
 *	@param[in]		p_templ1		Reference template		
 *	@param[in]		p_templ2		Candidate template
 *	@param[out]		p_dTheta		Rotation between p_templ1 and p_templ2
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32	findRotation( Template *p_templ1, Template *p_templ2, INT32 *p_dTheta );

/**	Processes the core index score beetween to tempates
 *  This core index score is then used by sortTemplates in order to speed up identification process
 * 
 *	@param[in]		p_templ1		Reference template
 *	@param[in]		p_templ2		Candidate template
 * 
 *	@return			The core index score. The lower it is, the best the match is.
 *
 *  @see			sortTemplates
*/
UINT32	scoreIndex( Template *p_templ1, Template *p_templ2 );

/**	Sort templates according to their "core index" score.
 * 
 *	@param[in]		p_templateBuffer		Refernece template	
 *	@param[in]		p_templateBufferArray	Array of candidate templates	
 *	@param[in]		p_nbTemplates			Number of templates in p_templateBufferArray
 *	@param[out]		p_coreIndexes			Array containing the position of the templates in p_templateBufferArray, 
 *											sorted by "core index" score (for templates that have a core)
 *	@param[out]		p_coreDist				Array containing the "core index" scores (for templates that have a core)
 *	@param[out]		p_noCoreIndexe			Array containing the position of the templates in p_templateBufferArray,
 *											sorted by "global distance" score (for templates that no not have a core)
 *	@param[out]		p_noCoreGlobalDist		Array containing the "global distance" scores (for templates that do not have a core)
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise
 *
 *	@see			scoreIndex
*/
INT32	sortTemplates( UINT8 *p_templateBuffer, UINT8 **p_templateBufferArray, UINT32 p_nbTemplates,
					 UINT32 *p_coreIndexes, UINT32 *p_coreDist, UINT32 *p_noCoreIndexe, UINT32 *p_noCoreGlobalDist  );

#ifdef __cplusplus
}
#endif

#endif // ZEF_FIXED_POINT

#endif //__ZEFYR_MATCHER_H
