/*
Company :		ZEFYR Technologies
File :			minutiae.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	13/06/2003

Description : Minutiae pre-processing, finding and post-processing

Features : differents fingerprint specific enhancements
    
Functions :

	extractMinutiae : main function used for minutiae extraction
	detectMinutiae : find minutiae in a skeletonize binary image
	followLine : internal function used to find minutiae neighboors
	writeMinutiae : write minutiae into a template
	cleanMinutiae : remove fake minutiae
	checkMinutia : internal function used when cleaning minutiae, check if a minutiae have to be erased
	reconnectSkeleton : reconnect some minutiae that are in front of each other and with the same orientation

Necessary package : 

	customTypes.h
	staticParameters.h
	dynamicParameters.h
	errorCodes.h
	skeleton.h
	zMaths.h

Comments : none

*/

#ifndef __ZEFYR_MINUTIAE_H
#define __ZEFYR_MINUTIAE_H

#include "customTypes.h"
#include "dynamicParameters.h"

#ifdef __cplusplus
extern "C" {
#endif

/**	High level function that extract minutiae from an enhanced fingerprint image 
 *	and store them in a template structure.
 * 
 *	@param[in,out]	p_data		BioEngine data structure. Contains Fingerprint image, 
 *								orientation map, quality map, template, etc.	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 extractMinutiae( BioEngineData *p_data );

/**	Mid level function that extracts minutiae points from a skeletonized fingerprint image.
 * 
 *	@param[in]		p_img			Skeletonized source fingerprint image
 *	@param[in]		p_qual			Quality map (used as a mask to avoid processing unwanted pixels)
 *	@param[in]		p_width			Width of p_img
 *	@param[in]		p_height		Height of p_img
 *	@param[out]		p_minutiae		Array of MinutiaNeighborhood structures (allocated by the calling function)
 *	@param[out]		p_nbMinutiae	Number of minutiae found by the function	
 *	@param[in]		p_searchSize	Searching size used to explore minutiae neighborhood
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 detectMinutiae( UINT8 *p_img, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 *p_nbMinutiae, UINT32 p_searchSize );

/**	Low level (recursive) function used by detectMinutiae to explore a minutia neighborhood according to one direction.
 * 
 *	@param[in,out]	p_img			Skeletonized source fingerprint image
 *	@param[in]		p_width			Width of p_img
 *	@param[in]		p_height		Height of p_img
 *	@param[in]		p_x				X coodinate of the minutia whose neighborhood have to be analysed
 *	@param[in]		p_y				Y coodinate of the minutia whose neighborhood have to be analysed
 *	@param[out]		p_retX			X Coordinates of the neighboring minutia found (if one is found)
 *	@param[out]		p_retY			Y Coordinates of the neighboring minutia found (if one is found)
 *	@param[in]		p_searchSize	Searching size used to explore minutiae neighborhood (recusion limit)	
 *	@param[in]		p_oldDir		Direction in which the function is searching
 * 
 *	@return			The type of minutia found close to the analysed minutiae (1 = ridge ending, 3 = bifurcation)			
*/
UINT32 followLine( UINT8 *p_img, UINT32 p_width, UINT32 p_height, UINT32 p_x, UINT32 p_y, UINT32 *p_retX, UINT32 *p_retY, UINT32 p_searchSize, UINT32 p_oldDir );

/**	Processes the orientation of each minutiae found by detectMinutiae
 *  Orientation is determined using the orientation map AND the skeleton informations. 
 *  This allow to have minutiae oriention over 360 (and not 180 if we ahd only used orientation map)
 * 
 *	@param[in]		p_img			Skeletonized source fingerprint image
 *	@param[in]		p_ori			FIngerprint orientation map processed during the enhancement phase
 *	@param[in]		p_width			Width of p_img
 *	@param[in]		p_height		Height of p_img
 *	@param[in,out]	p_minutiae		Array of MinutiaNeighborhood structures
 *	@param[in]		p_nbMinutiae	Number of minutiae in p_minutiae array
 *	@param			p_searchSize	Searching size used to explore minutiae neighborhood	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 detectMinutiaeOri( UINT8 *p_img, UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 p_nbMinutiae, UINT32 p_searchSize );

/**	Write the minutiae in the template structure
 * 
 *	@param[in]		p_qual			Quality map
 *	@param[in]		p_width			Width of source image (quality map is smaller)
 *	@param[in]		p_height		Height of source image (quality map is smaller)
 *	@param[in]		p_minutiae		Array of MinutiaNeighborhood structures 
 *									(contains all the needed minutiae information)
 *	@param[in]		p_nbMinutiae	Number of minutiae in p_minutiae array	
 *	@param[in,out]	p_template		Template structure that will be filled with minutiae information
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 writeMinutiae( UINT8 *p_qual, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 p_nbMinutiae, Template *p_template );

/**	Removes false minutiae (Mid level function).
 *  Minutiae removal is done using rules based on the "common" minutiae topology.
 * 
 *	@param[in]		p_ori			Orientation map
 *	@param[in]		p_width			Width of orientation map	
 *	@param[in]		p_height		Height of orientation map
 *	@param[in,out]	p_minutiae		Array of MinutiaNeighborhood structures 
 *	@param[in,out]	p_nbMinutiae	Number of minutiae in p_minutiae array		
 *	@param[in]		p_size			Depth of analyse for minutiae cleaning
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 cleanMinutiae( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 *p_nbMinutiae, UINT32 p_size );

/**	Removes minutia (Low vele function)
 *  Recursive function called by cleanMinutiae.
 * 
 *	@param[in]		p_ori			Orientation map
 *	@param[in]		p_width			Width of orientation map
 *	@param[in]		p_height		Height of orientation map
 *	@param[in,out]	p_minutiae		Array of MinutiaNeighborhood structures 
 *	@param[in]		p_index			Index (in p_minutiae) of the minutia currently beeing analysed
 *	@param[in]		p_size			Depth of analyse for minutiae cleaning
 *	@param[in,out]	oldMinutiae		Array of minutiae that have already been analysed
 *	@param[in,out]	oldCount		Number of calls already done to checkMinutia (to avoid infinite loops)
 * 
 *	@return			True if we have kept the minutia, FALSE otherwise.			
*/
BOOL8 checkMinutia( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 p_index, UINT32 p_size, UINT32 *oldMinutiae, UINT32 *oldCount );

/**	Reconnects some of the broken lines of the skeleton
 *  If two minutiae (ridge ending) have the same direction, are in front of each other, and close enough
 *  we remove them and reconnect the skeleton.
 * 
 *	@param[in]		p_ori			Orientation map
 *	@param[in]		p_width			Width of orientation map
 *	@param[in]		p_height		Height of orientation map	
 *	@param[in,out]	p_minutiae		Array of MinutiaNeighborhood structures 
 *	@param[in,out]	p_nbMinutiae	Number of minutiae in p_minutiae array		
 *	@param[in]		p_reconSize		Size under which skeleton parts will be reconnected
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 reconnectSkeleton( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 *p_nbMinutiae, UINT32 p_reconSize );

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_MINUTIAE_H
