/*
Company :		ZEFYR Technologies
File :			preProcessings.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : fingerprint pre-Processings

Features : differents fingerprint specific pre-processings		
    
Functions : 

	preProcess : main function applying the preprocessings
	firstMask : quickly evaluate a first mask to crop the image and only process real fingrprint data (less background)
	binarizeMask : binarize the mask calculated with firstMask
	cleanMask : keep only the biggest 8-connected region of a mask
	calculateCropSize : calculate the size of the image once pre-processed
	cropFromMask : crop the image to reduce its size and adapt it the the processings taht will follow

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	greyLevelTrans.h
	filters.h
	geometry.h

Comments : none

*/

#ifndef __ZEFYR_PRE_PROCESSINGS_H
#define __ZEFYR_PRE_PROCESSINGS_H

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/** crop structure, for internal use */
typedef struct _CropStruct
{
	UINT32	m_xMin,m_xMax,m_yMin,m_yMax;
	UINT32	m_newWidth, m_newHeight;

} CropStruct,*PCropStruct;

/**	High level pre-processing function.
 *  Normalizes the image and prepares it for the enhancement process
 * 
 *	@param[in,out]		p_data	BioEngine data structure 
 *								(contains image, and other stuffs not used for pre-processing)	
 * 
 *	@return				ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise		
*/
INT32 preProcess( BioEngineData *p_data );

/**	Evaluate a roughly a first mask for the image.
 *  processing is based on differences of std deviation in the image
 *  p_src is left unchanged, p_dst contains a mask, SEGM_BLOC_SIZE times smaller (with 255 = image and 0 = background )
 *  p_dst must be allocated before calling firstMask, its size is  (p_width/SEGM_BLOC_SIZE)*(p_height/SEGM_BLOC_SIZE) 
 *
 *	@param[in]		p_src		Source image	
 *	@param[out]		p_dst		Image mask
 *	@param[in]		p_width		Widht of p_src
 *	@param[in]		p_height	Height of p_src
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32 firstMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );


/**	Binarize the first mask
 * 
 *	@param[in]		p_src		Source image (image mask)
 *	@param[out]		p_dst		Binarized image mask
 *	@param[in]		p_width		Widht of p_src
 *	@param[in]		p_height	Height of p_src	
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise	
 *
 *	@see			firstMask
*/
INT32 binarizeMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );

/**	Cleans the image mask
 *  If p_src is a binarized mask with values 0 and 255
 *  this function keeps the biggest 8-connected region and put it in p_dst
 *  p_src and p_dst can be the same
 * 
 *	@param[in]		p_src		Source image (binarized image mask)
 *	@param[out]		p_dst		Cleaned binarized image mask
 *	@param[in]		p_width		Widht of p_src
 *	@param[in]		p_height	Height of p_src		
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise
 *
 *	@see			firstMask
*/
INT32 cleanMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );


/**	Analyses the mask obtained with firstMask and other functions to determine how the source image can be cropped
 * 
 *	@param[in]		p_mask		Cleaned binarized image mask
 *	@param[in]		p_width		Widht of p_src
 *	@param[in]		p_height	Height of p_src		
 *	@param[out]		p_crop		Information describing how to crop p_src to keep only usefull data
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise			
*/
INT32 calculateCropSize( UINT8 *p_mask, UINT32 p_width, UINT32 p_height, CropStruct *p_crop );

/**	Crops the source image using information processed in function calculateCropSize
 * 
 *	@param[in]		p_src		Source image 
 *	@param[out]		p_dst		Destination (cropped) image
 *	@param[in]		p_width		Widht of p_src
 *	@param[in]		p_height	Height of p_src		
 *	@param[in]		p_crop		Information describing how to crop p_src to keep only usefull data
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise
 *
 *	@see			calculateCropSize
*/
INT32 cropFromMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, CropStruct *p_crop );

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_PRE_PROCESSINGS_H
