/*
Company :		ZEFYR Technologies
File :			staticParameters.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	05/05/2003

Description : compile time parameters for fingerprint processing

Features : defines some constants		
    
Functions : none

Necessary package : none

Comments :none

*/

/*------------------ DEFINES FOR CUSTOM BIOENGINE GENERATION -------------------------------*/

// These compile time options allow to build BioEngines for different target (Fixed or 
// floating point for example)

//#define MEM_DEBUG               /**< use some custom malloc and free functions to track memory leaks */
//#define DP_DEBUG                /**< output debuging images at each stage of the template generation process */
//#define ZEF_FIXED_POINT         /**< if defined : all processing done using fixed point arithmetics. Otherwise floating point is used */
//#define	USE_FX_PRECOMP        /**< only meaningfull when ZEF_FIXED_POINT is defined. Use some precompiled tables for fixed point calculation. If not used, fixed point tables are computed when the bioEngine is initialized */
#define ZEF_SPLIT_RADIX           /**< use split radix FFTs */
//#define ZEF_RADIX_2_4           /**< use radix 2/4 FFTs */
#define ZEF_CORRECT_DEFORMATION   /**< enable thin plate spline deformation correction during the matching process */
//#define TIME_REPORT             /**< outputs processing timings to the console */
#define THREAD_SAFE               /**< use thread safe export functions (need mutex support by the host : see "portableMutex.h" for supported platforms) */

/*---------------- END DEFINES FOR CUSTOM BIOENGINE GENERATION -----------------------------*/

#ifndef __ZEFYR_STATIC_PARAMETERS_H
#define __ZEFYR_STATIC_PARAMETERS_H

#include "customTypes.h"
#include "fixedPoint.h"
#include "memMonitor.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Basic BioEngine version (will be completed with information concerning the compile time options) */
#define		DP_BIO_VER			"1.1.1"

/** Needed for some tables pre-computing */
//#define		PI						3.1415926535897932384626433832795f

/** Version of our template structure.
 *  For the moment the original structure is still used (version 1.0) */
#define		TEMPLATE_VERSION		(UINT32)(1.0*1000)

// Some defines needed to compile on a win32 platform
#ifdef WIN32
	#ifndef WINAPI
		#define WINAPI __stdcall
	#endif
#else // if not on a win32 platform ignore WINAPI (ie. STDCALL) define
	#define WINAPI
#endif

/*-------------------------- DEFINE COMPLETE BIOENGINE VERSION -----------------------------*/

// debug static const
// if we are debugging DP_DEBUG is defined else it is not
#ifdef DP_DEBUG
#define DP_DEBUG_VER "-d"
#else
#define DP_DEBUG_VER ""
#endif

// memory monitoring !
#ifdef MEM_DEBUG
	#define MEM_DEBUG_VER "-md"
#else
	#define MEM_DEBUG_VER ""
#endif

// for fixed point compilation
#ifdef ZEF_FIXED_POINT
#define ZEF_FIXED_POINT_VER "-fx"
#else
#define ZEF_FIXED_POINT_VER ""
#endif

// do not calculate fixed point precomp data, read it in a file
// with this option you don't need any maths library
#ifdef USE_FX_PRECOMP
#define USE_FX_PRECOMP_VER "-p"
#else
#define USE_FX_PRECOMP_VER ""
#endif

// FFT type
#ifdef ZEF_RADIX_2_4
#define ZEF_FFT_VER "-r24"
#else
#define ZEF_FFT_VER ""
#endif

// plastic distortion correction
#ifndef ZEF_CORRECT_DEFORMATION
#define ZEF_CORRECT_DEFORMATION_VER "-ndc"
#else
#define ZEF_CORRECT_DEFORMATION_VER ""
#endif

// Time report
#ifdef TIME_REPORT
	#include <stdio.h>
	#include <time.h>
	#define TIME_REPORT_VER "-tr"
#else
	#define TIME_REPORT_VER ""
#endif

// thread safe bioengine
#ifdef THREAD_SAFE
	#define THREAD_SAFE_VER ""
#else
	#define THREAD_SAFE_VER "-nts"
#endif

/** Complete BioEngine version that will indicate which options have been set during compilation */
#define DP_BIO_VERSION DP_BIO_VER DP_DEBUG_VER ZEF_FIXED_POINT_VER USE_FX_PRECOMP_VER ZEF_FFT_VER ZEF_CORRECT_DEFORMATION_VER TIME_REPORT_VER MEM_DEBUG_VER THREAD_SAFE_VER

/*------------------------- END DEFINE COMPLETE BIOENGINE VERSION --------------------------*/

/* ----------------------------- COMPILE TIME PARAMETERS -----------------------------------*/

// NOTE : most of the following constants are used in differents functions and file. Do a 
// "Find in files" in you prefered  programming IDE to see in which function(s) they are used

/** Security option to avoid hill climbing attacks
 *  if set to false, the matching score is returned only if it is greater than the  security threshold
 * otherwise it is always returned (can be dangerous...) */
#define		RETURN_NON_MATCH_SCORE		FALSE

//	******** TEMPLATE EXTRACTION PARAMETERS ********

/** size of a FFT bloc when the image is filtered (Gabor). If set to 64 we will use 64x64 blocs */
#define		FFT_BLOC_SIZE				64

/** used to compute "wheigts" during gabor filtering */
#define		MOD_FFT_BLOC_SIZE			(FFT_BLOC_SIZE-1)

/** used to compute "wheigts" during gabor filtering */
#define		MOD_FFT_BLOC_SIZE2			(FFT_BLOC_SIZE/2-1)

/** Bloc size used for orientation map processing. By default we use 16x16 blocs */
#define		ORI_BLOC_SIZE				16 

/** Bloc size used by pre-processing */
#define		SEGM_BLOC_SIZE				8

/** Bloc size used when processing the FFT spectrum in polar coordinates */
#define		SPECTRUM_RESOLUTION			32

/** Bloc size used when trying to find the correct orientation map from the N (usually n=8) images filtered by different gabor filters (wih different orentations) */
#define		FUSION_BLOC_SIZE			4

/** Used to define ORI_BLOC_SIZE2 */
#define		ORI_PROCESSING_DIVIDE		4

/** Number of pixels that are too close from the border that won't be taken into account for some processings */
#define		IMAGE_BORDER_LIMIT			8

/** max number of neighbouring minutiae used when cleaning the minuatiae found from the fingerprint skeleton */
#define		MAX_MINUTIAE_NEIGHBOR		10

/** max number of cores and delats that will be searched for */
#define		MAX_CORES_DELTAS			16

/** max number of connected point that will be merged when looking for a core or delta */
#define		MAX_PTS_CORE_DELTA			256

/** step size usez in orientation processing (before gabor filtering).*/
#define		ORI_BLOC_SIZE2				(ORI_BLOC_SIZE/ORI_PROCESSING_DIVIDE)

/** used in curvature processing */
#define		CURVATURE_MULT_FACTOR		4

/** max number of minutiae that will be stored when cleaning minutae (recursion depth of "checkMinutiae") */
#define		MAX_NB_OLD_MINUTIAE			50

/** scaling factor used to normalise the gabor filtered images */
#define		GABOR_IMAGE_INTENSITY_SCALE	FTOFIX32(40.0f,Q15_16)

/** used for frequency estimation */
#define		MIN_NB_BLOCS_FREQ			4

/** scaling factor used only for fixed point FFT. Needed to avoid underflow or overflow during FFT processing */
#define		FIXED_POINT_FFT_SCALE		8

/** scaling fator used only for fixed point gabor filtering. Needed to get output values in the range -128 +128 */
#define		FX_GABOR_IMAGE_INTENSITY_SCALE 3

/** image scaling factor used when speed paramter is set to one or two
 *  image will be scaled by 205/256 = 0.8 so that processings are faster */
#define		FAST_SCALE_FACTOR			205


// ********	MATCHER PARAMETERS ********

/** Security constant used to prevent the BioEngine from looping infinitely in some functions */
#define		MAX_LOOPS					100

/** Security constant used to prevent the BioEngine from looping infinitely in some functions */
#define		BIG_MAX_LOOPS				1000

/* --------------------------- END COMPILE TIME PARAMETERS ---------------------------------*/


/* ------------------------------------ DYNAMIC PARAMETERS ---------------------------------*/

// Default values for dynamic parameters : these parameters are stored in a structure
// declared in "dynamicParameters.h". This structure is then avalable to all functions 
// as a global variable.

//	******** TEMPLATE EXTRACTION PARAMETERS ********

/** Max number of minutiae that will be search int the fingerprint skeleton */
#define		MAX_NUMBER_OF_DETECTED_MINUTAE	500

/** Max number of minutiae that will be stored in the template structure */
#define		MAX_NUMBER_OF_MINUTIAE			50

/** Max number of local infos (bloc orientation + quality) that will be stored in the template structure */
#define		MAX_NUMBER_OF_LOCAL_INFOS		256

/** Max numbner of core and delta that will be stored in the template structure */
#define		MAX_NUMBER_OF_CORES				2

/** Mean used when normalising the image during its pre-processing */
#define		NORMALISATION_MEAN				128

/** Standard deviation used when normalising the image during its pre-processing */
#define		NORMALISATION_STD_DEV			50

/** Threshold under which quality will be set to zero (in quality map) */
#define		QUALITY_THRESHOLD				64

/** Threshold used for frequency estimation. Only frequencies which quality is above this threshold will be kept
 *  other bloc frequencies will be interpolated from the good ones */
#define		FREQ_QUAL_THRESHOLD				FTOFIX32(0.5f,Q15_16) 

/** Uses in the freqnecy selection loop to bee sure we have at least MIN_NB_BLOCS_FREQ good quality blocs */
#define		FREQ_QUAL_THRESHOLD_INC			FTOFIX32(0.01f,Q15_16)

/** Scaling factor for spectrum estimation (to avoid underflows / overflows) */
#define		SPECTRUM_MULT_FACTOR			FTOFIX32(0.25f,Q15_16)

/** width of gabor filters used for iamge enhancement */
#define		GABOR_WIDTH						FTOFIX32(0.5f,Q15_16)

/** correction factor used to generate gabor filters from frequency estimation of image blocs */
#define		GABOR_FREQ_MULT					FTOFIX32(0.9f,Q15_16)

/** threashold used in different function of gabor images fusion to decide which pixels need to be reprocessed */
#define		GABOR_FUSION_THRESHOLD			FTOFIX32(6.0f,Q15_16)

/** used for gabor images fusion. Needed to detect region that have to be reprocessed */
#define		REGION_DIV_COEF					FTOFIX32(10.0f,Q15_16) 

/** used during pre-processing. Number of classes used in kMean classification used prior to mask binarisation */
#define		NB_CLASSES_SEGMENTATION			4

/** used during pre-processing. Number of classes we need to keep after kMean classification in order to binarize the mask */
#define		NB_CLASSES_REJECT_SEGMENTATION	1

/** size of the median filter used for mask generation during pre-processing */
#define		SEGM_MEDIAN_FILTER_SIZE			3

/** used for gabor images fusion. Needed to detect region that have to be reprocessed */
#define		REGION_PERIMETER_COEF			10

/** used for gabor images fusion. Needed to detect region that have to be reprocessed */
#define		REGION_SURFACR_COEF				1

/** size of the median filter used at the first stages of gabor image fusion in order to obtain a smother orientation map*/
#define		ORI_FUSION_MEDIAN_SIZE			3

/** thresahold used in a function used to "guess" the correct orinetation of a pixel that had to bee 
 *  reprocessed in the gabor image fusion function */
#define		NB_PTS_FUSION_THRESHOLD			2

/** Correction factor applyed to GABOR_FUSION_THRESHOLD when looking for pixels to reprocess during gabor images fusion  */
#define		MULT_FACTOR_REGION_FUSION		FTOFIX32(1.25f,Q15_16)

/** Erosion filter size used to find if some orientation points must be reprocessed during gabor images fusion */
#define		FUSION_ERODE_SIZE				3

/** Median filter size used to generation the orientation quality map at the last stage of gabor images fusion */
#define		FINAL_FUSION_MEDIAN_SIZE		5

/** correction factor used to modify the quality of pixels that needed correction during gabor images fusion */
#define		FUSION_INTENSITY_MULT_FACTOR	FTOFIX32(0.5f,Q15_16)

/** binary median size for binary fingerprint image obtained after gabor filtering and fusion (end of enhancement process) */
#define		BINARY_MEDIAN_SIZE				3

/** Minimum curvature value needed to detect a core for curvature value.
 *  If the max curvature found is less than MINIMUM_CURVATURE_THRESHOLD we set the core position as the center of the image... */
#define		MINIMUM_CURVATURE_THRESHOLD		10

/** size of blocs used for local infomations (quality + orientation) stored in the template structure */
#define		LOCAL_INFOS_BLOC_SIZE			24

/** size of the neighbourhoud that will be analysed when trying to detect minutia from the fingerprint skeleton. */
#define		MINUTIAE_NEIBORHOOD_SIZE		16

/** size under which a branch wil be removed during fingerprint skeleton cleaning */
#define		BRANCH_MIN_SIZE					8

/** maximum distance between two ridge ending for automatic reconnection during skeleton cleaning */
#define		MAX_RECONNECT_SIZE				16

/** minimum reconnexion angle between two ridge ending for automatic reconnection during skeleton cleaning */
#define		MIN_RECONNECTION_ANGLE			(256/8) // PI/8

/** maximum angle for minutiae deletion during skeleton cleaning */
#define		MAX_SUPPR_ANGLE					(256/3) // PI/3

/** size of the digital curve used for poincar indexe processing. defautl = 3x3 square curve */
#define		POINCARE_SIZE					3

/** when set to true : if some image blocs have been detected with very poor image quality (during frequency estimation)
 *  they will be specially treated to improve their quality. */
#define		POW_FFT_FLAG					TRUE

/** distance used for the digital curve used for core angle detection */
#define		DIST_FIND_CORE_ANGLE			24

/** number of frequency bands used for bloc frequency estimation */
#define		NB_FREQ_POS						8

/** max image width. If image is larger, it will be cropped 
 *  SHOULD NOT BE > 512 for fixed poitn version of the algorithm !!! (because of TPS warper)
 */
#define		MAX_IMG_WIDTH					512 

/** max image height. If image is larger, it will be cropped 
 *  SHOULD NOT BE > 512 for fixed poitn version of the algorithm !!! (because of TPS warper)
 */
#define		MAX_IMG_HEIGHT					512

/** image scaling : 256 = 256/256 = 1.0 = good scale for 500 dpi images. (should be set to 512 for 250 dpi images) */
#define		GLOBAL_IMG_SCALING				256

/* correction factor for minutiae positions in the templace structure when GLOBAL_IMG_SCALING is modified
 *  usually set to : (256*256)/GLOBAL_IMG_SCALING  */
#define		GLOBAL_TEMPLATE_SCALING			256

/** Bloc used to process image FFT are overlapping. This parameters sets how much overlap we need */
#define		FFT_BLOC_OVERLAP				32

/** number of orientation used for gabor filtering */
#define		NUM_OF_ORIENTATION				8

/** speed factor for orienation map processing (before frequency estimation) */
#define		ORI_STEP_SIZE					1

/** Used to speed up a little gabor filtering. If the Gabor filtered frequency domain bloc mean magnitude is below this threashold
 *  the bloc will be ignored (no FFT-1) and so processing will be faster. */
#define		FILTER_SELECTION_THRESHOLD		0

//	******* MATCHING PARAMETERS ********

/** max angle between two templates for which registration will work.
 * the higher the registration angle, the slower the registration process !!! */
#define		MAX_REGISTRATION_ANGLE			48

/** orientation step for the first pass of the registratio process
 *  higher values means faster registration, but can leab to errors in registration */
#define		FIRST_PASS_ANGLE_STEP			4

/** minutiae translation tolerance for the registration process */
#define		REGISTER_TRANSLATION_TOLERANCE	12

/** bloc size used when trying to detect the translation value between two templates
 *  this is the value of the shift used to speed up some operations */
#define		TRANSLATION_SHIFT				4	// 1<<4 = 16 pixels

/** bloc size used when trying to detect the translation value between two templates
 *  this is the real bloc size */
#define		TRANSLATION						(1<<TRANSLATION_SHIFT)	// 1<<4 = 16 pixels

/** used for templates translation estimation : see source code for more details */
#define		MAX_TRANSLATION_WIDTH_SHIFT_1	6

/** used for templates translation estimation : see source code for more details */
#define		MAX_TRANSLATION_WIDTH_1			(1<<MAX_TRANSLATION_WIDTH_SHIFT_1)	// (-32)<<4 < dx < (+32)<<4 ==> -512 < dx < 512

/** used for templates translation estimation : see source code for more details */
#define		MAX_TRANSLATION_WIDTH_SHIFT_2	6

/** used for templates translation estimation : see source code for more details */
#define		MAX_TRANSLATION_WIDTH_2			(1<<MAX_TRANSLATION_WIDTH_SHIFT_2)	// (-16)<<4 < dx < (+16)<<4 ==> -256 < dx < 256

/** parameter used to give more "tolerance" to the translation voting algorithm used for templates translation detection. */
#define		ACCU_FILL						8

/** Threshold used to find the best translation between two templates */
#define		TRANSLATION_MAX_THRESHOLD		FTOFIX32(0.8f,Q15_16)

/** Minuitae orientation tolerance for score processing */
#define		ANGLE_TOLERANCE					24

/** Minuitae translation tolerance for score processing */
#define		TRANSLATION_TOLERANCE			12

/** spline stiffness by thin plate spline plastic deformation correction algorithm */
#define		SPLINE_STIFFNESS				FTOFIX32(0.5f,Q15_16)

/** mean number of minutiae necessary for deformation processing */
#define		MIN_NUMBER_OF_MINUTIAE			3

/** normalisation factor for bloc orientation comparison during score processing */
#define		ORI_MATCH_NORMALIZE				FTOFIX32(256.0f,Q15_16)

/** maxumum score value */
#define		MAX_SCORE						1000

/** score normalizing factor */
#define		SCORE_NORMALIZE					FTOFIX32(1.6f,Q15_16)

/** maximum deformation allow during thin plate spline deformation correction */
#define		MAX_DEFORMATION					1500

/** Correction factor used for minutiae score matching */
#define		DIV_DIST_COEF					FTOFIX32(3.0f,Q15_16)

/** Boost score in regions where good minutae density in detected */
#define		MINUTIAE_BOOST					TRUE

/** used for identification : max percentage of the template database that will be used (after templates sorting)  */
#define		MAX_DB_SEARCH					100

//	******** SECURITY THRESHOLD ********

#define  	DEFAULT_SECURITY				75

// ******** VALUES USED FOR FREQUENCY ESTIMATION ********

#define		MAX_NB_OF_FREQ					10
#define		FREQ_POS1						3
#define		FREQ_POS2						4
#define		FREQ_POS3						5
#define		FREQ_POS4						6
#define		FREQ_POS5						8
#define		FREQ_POS6						10
#define		FREQ_POS7						13	
#define		FREQ_POS8						24
#define		FREQ_POS9						0
#define		FREQ_POS10						0

/* ------------------------------- END DYNAMIC PARAMETERS ---------------------------------*/

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_STATIC_PARAMETERS_H
