/*
Company :		ZEFYR Technologies
File :			tpsWarper.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	04/07/2003

Description : calculates a non rigid deformation between two sets of points using thin plate splines

Features : estimates local deformation of fingerprints
    
Functions : 

	initTpsWarper : initialize a new estimation with new data
	initMatrixA : build tyhe A matrix for TPS system resolution
	unInitTpsWarper : free memory used during TPS warping
	warp : calculate the new position of a poirn according to the deformation calculated with initTpsWarper
	getDiffs : get the "energy" needed for the deformation

Necessary package : 

	customtypes.h
	errorCodes.h
	linAlgebra.h
	zMaths.h

Comments : none

*/

#ifndef __ZEFYR_TPS_WARPER_H
#define __ZEFYR_TPS_WARPER_H

#ifndef ZEF_FIXED_POINT

#include "customTypes.h"

#ifdef __cplusplus
extern "C" {
#endif

/**	Calculates a Thin Plate Spline model from a list of source and destination points.
 *  All TPS model data are stored in p_data
 * 
 *	@param[in]		p_nbPts		Number of source points (and destination points)
 *	@param[in]		p_srcX		Array of X coordinates of the source points
 *	@param[in]		p_srcY		Array of Y coordinates of the source points
 *	@param[in]		p_dstX		Array of X coordinates of the destination points
 *	@param[in]		p_dstY		Array of Y coordinates of the destination points
 *	@param[in]		p_alpha		Spline Stiffnes. The lower the value, the more elastic the model
 *	@param[out]		p_data		Computed TPS data
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32	initTpsWarper( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, INT32 *p_dstX, INT32 *p_dstY, REAL32 p_alpha, TpsData *p_data);

/**	Low level function used in initTpsWarper to initialise the A matrix
 * 
 *	@param[in]		p_nbPts		Number of source points (and destination points)
 *	@param[in]		p_srcX		Array of X coordinates of the source points
 *	@param[in]		p_srcY		Array of Y coordinates of the source points
 *	@param[in]		p_alpha		Spline Stiffnes. The lower the value, the more elastic the model
 *	@param[out]		p_data		Computed TPS data
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32	initMatrixA( INT32 p_nbPts, REAL32 *p_srcX, REAL32 *p_srcY, REAL32 p_alpha, TpsData *p_data );

/**	Uninit and frees TPS data	
 * 
 *	@param[in,out]	p_data		Computed TPS data htat need to be freed
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32	unInitTpsWarper( TpsData *p_data );

/**	Calculate the new coordinates of a point using the deformation information stored in TPS model data.
 * 
 *	@param[in,out]	p_x			[in] x coordinate of the source point, [out] x corrdinate of the point after TPS deformation
 *	@param[in,out]	p_y			[in] y coordinate of the source point, [out] y corrdinate of the point after TPS deformation
 *	@param[in]		p_data		Computed TPS data
 * 
 *	@return			ZEF_OK if everything was OK. An error code listed in "errorCodes.h" otherwise				
*/
INT32	warp( INT32 *p_x, INT32 *p_y, TpsData *p_data );

/**	The getDiffs function
 * 
 *	@param[in]		p_data		Computed TPS data
 * 
 *	@return			A value indicating the amount of deformation applied to all point treated with the warp function.
 *
 *	@see			warp
*/
UINT32  getDiffs( TpsData *p_data );

#ifdef __cplusplus
}
#endif

#endif // ZEF_FIXED_POINT

#endif //__ZEFYR_TPS_WARPER_H
