/*
Company :		ZEFYR Technologies
File :			zMaths.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	06/05/2003

Description : custom maths functions

Features : provides functions that are faster or more compatible with the architecture used		
    
Functions : 

	zSqrt : processes the square root of an integer number
	zfSqrt : processes the square root of a real number
	zSin : Sinus
	zCos : Cosinus
	zAbs : Absolute value
	zfAbs : Real Absolute Value
	zSinSpectrum : fast Sinus for spectrum processing (table based)
	zCosSpectrum : fast Cosinus for spectrum processing (table based)
	zReal2Int32 : fast real to int conversions for x86
	zAtan : fast Atan
	zAtanSpec : fast atan for spectrum processing (table based)
	zSqrtSpec : fast sqrt for spectrum processing (table based)
	zLog : logarithm
	zCos127 : fast integar cosinus
	zSin127 : fast integer sinus


Function prototypes : 

	INT16	zAtan(INT32 y, INT32 x);
	REAL32	zLog( UINT32 p_nb );
	INT8	zCos127( INT32 p_angle );
	INT8	zSin127( INT32 p_angle );
	INT32	Real2Int32(REAL32 fFloat)

	All other mathematical functions are macros

Necessary package : 

	math.h
	errorCodes.h
	customTypes.h

Comments : none

*/

#ifndef __ZEFYR_Z_MATHS_H
#define __ZEFYR_Z_MATHS_H

#include "math.h"
#include "errorCodes.h"
#include "customTypes.h"
#include "fixedPoint.h"

#ifdef __cplusplus
extern "C" {
#endif

// ******** Precomputed tables ********

extern INT32	g_precSinSpectrum[SPECTRUM_RESOLUTION];
extern INT32	g_precCosSpectrum[SPECTRUM_RESOLUTION];
extern INT8		g_precSin[256];
extern INT8		g_precCos[256];
extern UINT8	g_sqrtSpec[FFT_BLOC_SIZE][FFT_BLOC_SIZE/2];
extern UINT8	g_atanSpec[FFT_BLOC_SIZE][FFT_BLOC_SIZE/2];

extern UINT8	g_atanTable[257];

#ifndef ZEF_FIXED_POINT
	extern REAL32	g_invExp[512];
	extern REAL32	g_log2Table[1025];
#else
	extern FIXED16	g_invExp[512];
	extern FIXED32	g_log2Table[1025];
#endif // ZEF_FIXED_POINT

// ******** Constants ********

/** Square root of 2 */
#define	SQRT2		1.4142135623730950488016887242097f

/** Square root of two, multiplied by 65536 and cast to an integer */
#define	SQRT2_16	92681

/** floating point value of PI */
#define PI			3.1415926535897932384626433832795f

/** floating point value of PI/2 */
#define PI2			1.5707963267948966192313216916398f

/** floating point value of PI/4 */
#define	PI4			0.78539816339744830961566084581988f

/** floating point value of PI/8 */
#define	PI8			0.39269908169872415480783042290994f

/** value of PI when using the integer transofmation : PI=256 */
#define	PI_256		256

/** value of PI/2 when using the integer transofmation : PI=256 */
#define PI2_256		128

/** value of PI/4 when using the integer transofmation : PI=256 */
#define	PI4_256		64

#define	E_BASE_LN	2.7182818284590452353602874713527f

#ifndef ZEF_FIXED_POINT
	#define INVLOG2E	(1.0f/1.4426950408889634073599246810019f)
#else
	#define INVLOG2E	1420
#endif

#define	LOGDIV		(1.0f/4194304.0f)

// ******** Macros and function prototypes ********

/** Macro that allows us to define floating point SIN function as any custom function.
 *  Here we use the standard sin function.
*/ 
#define	zSin( p_num ) (REAL32)sin( (p_num) )

/** Macro that allows us to define floating point COS function as any custom function.
 *  Here we use the standard cos function.
*/ 
#define	zCos( p_num ) (REAL32)cos( (p_num) )

#ifdef _M_IX86
	// on x86 platforms we use the builtin sqrt function (it is the fastest)
	/** Macro that allows us to define SQRT function as any custom function.
	 *  Here we use the standard sqrt function.
	*/ 
	#define zSqrt( p_num )	(UINT32)sqrt( (p_num) )
#else
	// but on other platforms we use our own (theorically) fast function.
	/** Our own fast integer sqrt */
	UINT16 zSqrt( UINT32 r);
#endif // _M_IX86

/** Macro that allows us to define floating point SQRT function as any custom function.
 *  Here we use the standard sqrt function.
*/ 
#define zfSqrt( p_num )	(REAL32)sqrt( (p_num) )

/** Macro that allows us to define ABS function as any custom function.
 *  Here we use a fast if implmentation.
*/ 
#define zAbs( p_num )	(((p_num) > 0 ) ? (p_num):(-(p_num)))

/** Macro that allows us to define floating point ABS function as any custom function.
 *  Here we use the standard fabs functions
*/
#define zfAbs( p_num )	(REAL32)fabs( (p_num) )

/** Macro that returns a special sinus value used for spectrum processing 
 *  These sinus values are stored in a pre-computed table
*/
#define zSinSpectrum( p_num ) g_precSinSpectrum[ (p_num) ] // special Sin for polar spectrum processing 

/** Macro that returns a special cosinus value used for spectrum processing 
 *  These cosinus values are stored in a pre-computed table
*/
#define zCosSpectrum( p_num ) g_precCosSpectrum[ (p_num) ] // special Cos for polar spectrum processing

/** Macro that define a function that casts a floating point value to an integer value 
 *  The goal of this macro  is to provide a fast float to int cast on all processors and OS
*/
#define zReal2Int32( p_num ) Real2Int32( (p_num) )

/** Our fast arc-tangent function */
INT16	zAtan(INT32 y, INT32 x);

/** Macro that returns a special atan value used for spectrum processing 
 *  The atan values are stored in a pre-computed table
*/
#define	zAtanSpec( y, x ) g_atanSpec[(y)][(x)]

/** Macro that returns a special sqrt value used for spectrum processing 
 *  The sqrt values are stored in a pre-computed table
*/
#define	zSqrtSpec( y, x ) g_sqrtSpec[(y)][(x)]

// exp(-x) for 0.0 <= x <= 8.0
#ifndef ZEF_FIXED_POINT 
	/** Fast inverse exponential function */
	#define zInvExp( x ) g_invExp[FTOFIX16((x),6)]

	/** Fast log function */
	REAL32	zLog( UINT32 p_nb );
#else
	// x must be in Q9_6 format 
	/** Fast inverse exponential function */
	#define zInvExp( x ) g_invExp[(x)] 

	/** Fast log function */
	FIXED32	zLog( UINT32 p_nb );
#endif // ZEF_FIXED_POINT 

/** Intiallizes the maths functions. Fill the pre-computed tables */
INT32	zInitMaths( void );


/** Table based cosinus
 *  returns cosinus premultiplied by 127
 *  so range -1 -> +1 become -127 -> +127
 *  input values are integers where 256 = 2*PI and -256 = -2*PI
*/
#define zCos127(p_angle) g_precCos[(p_angle)&255]

/** Table based sinus
 *  returns sinus premultiplied by 127
 *  so range -1 -> +1 become -127 -> +127
 *  input values are integers where 256 = 2*PI and -256 = -2*PI
*/
#define zSin127(p_angle) g_precSin[(p_angle)&255]

// Casting a float value to an int value is a relatively slow opration on some architecture and compilers (due to rounding issues)
// So to speed things up we define some fast casting replacement for this operation
#ifdef _M_IX86
	#ifdef WIN32 // MSVC
		__inline INT32 Real2Int32(REAL32 fFloat) // __inline is not ansi but gcc and vc6.0 understand it
		{
			static INT32 nTemp;
			_asm {
				fld fFloat
				fistp nTemp
				mov eax,nTemp
			}
		}
	#else // GCC
		#define Real2Int32(in,out) __asm__ __volatile__ ("fistpl %0" : "=m" (out) : "t" (in) :"st") ;
	#endif
#else
	#define Real2Int32( fFloat ) (INT32)(fFloat)
#endif

#ifdef __cplusplus
}
#endif

#endif // __ZEFYR_Z_MATHS_H
