/*
Company :		ZEFYR Technologies
File :			BMPfiles.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	15/04/2003

Description : BMP load/save support

Features : 8bit greyscale non compressed BMP files		
    
Functions : 

	getBMPsize: get the size of a BMP file (necessary for buffer allocation )
    loadBMP:	load a BMP file into an array of unsigned char
    saveBMP:	save a BMP file to disk from an array of unsigned char
	saveBmp32:  save a BMP file to disk from an array of unsigned long
	saveBmpR:   save a BMP file to disk from an array of float
	saveBmpC:   save a BMP file to disk from an array of complexe numbers (two floats)
	saveBmpS:	save a BMP file to disk from an array of char
	saveBmpS16:	save a BMP file to disk from an array of short

Function prototypes : 

	INT32 getBmpSize( char *p_fileName, INT32 *p_size );
	INT32 loadBmp( char *p_fileName, UINT8 *p_buffer, INT32 p_bufferSize, INT32 *p_width, INT32 *p_height );
	INT32 saveBmp( char *p_fileName, UINT8 *p_buffer, INT32 p_width, INT32 p_height );
	INT32 saveBmp32( char *p_fileName, UINT32 *p_buffer, INT32 p_width, INT32 p_height );
	INT32 saveBmpR( char *p_fileName, REAL32 *p_buffer, INT32 p_width, INT32 p_height );
	INT32 saveBmpC( char *p_fileName, REAL32 *p_buffer, INT32 p_width, INT32 p_height );
	INT32 saveBmpS( char *p_fileName, INT8 *p_buffer, INT32 p_width, INT32 p_height, UINT32 p_nbPlans, UINT32 p_plan );
	INT32 saveBmpS16( char *p_fileName, INT16 *p_buffer, INT32 p_width, INT32 p_height, UINT32 p_nbPlans, UINT32 p_plan );

Necessary package :
	customTypes.h 
	errorCodes.h

Comments :none

*/ 

#include <stdio.h>
#include <limits.h> 
#include <float.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "zMaths.h"
#include "BMPfiles.h"

INT32 getBmpSize( char *p_fileName, INT32 *p_size )
{
	UINT8			header[54];
	FILE			*file;
	INT32			error;
	INT32			width,height;

	// init
	file = NULL;
	*p_size = 0;

	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"rb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// header reading
	error = fread(header,sizeof(UINT8),54,file);
	if( error == 0 && ferror(file) )
	{
		error = ZEF_FILE_READ;
		goto errorHandler;
	}

	// is it a BMP ?
	if( header[0] != 'B' || header[1] != 'M' )
	{	
		error = ZEF_NOT_BMP;
		goto errorHandler;
	}
	if( *((INT32*)(header+28)) != 8 )
	{
		error = ZEF_BMP_BAD_FORMAT;
		goto errorHandler;
	}

	// we read width and height of the image
	width = (*(header+18))+((*(header+19))<<8)+((*(header+20))<<16)+((*(header+21))<<24);
	height = (*(header+22))+((*(header+23))<<8)+((*(header+24))<<16)+((*(header+25))<<24);

	// then close the file 
	fclose(file);

	*p_size = (width+(width*3)%4)*height*sizeof(UINT8);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 loadBmp( char *p_fileName, UINT8 *p_buffer, INT32 p_bufferSize, INT32 *p_width, INT32 *p_height )
{
	UINT8			header[54];
	FILE			*file;
	INT32			i,error;
	INT32			width,height;

	// init
	file = NULL;

	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"rb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// header reading
	error = fread(header,sizeof(UINT8),54,file);
	if( error == 0 && ferror(file) )
	{
		error = ZEF_FILE_READ;
		goto errorHandler;
	}

	// is it a BMP ?
	if( header[0] != 'B' || header[1] != 'M' )
	{	
		error = ZEF_NOT_BMP;
	}
	if( *((INT32*)(header+28)) != 8 )
	{
		error = ZEF_BMP_BAD_FORMAT;
		goto errorHandler;
	}

	// we read width and height of the image
	width = (*(header+18))+((*(header+19))<<8)+((*(header+20))<<16)+((*(header+21))<<24);
	height = (*(header+22))+((*(header+23))<<8)+((*(header+24))<<16)+((*(header+25))<<24);

	// we check if the memory allocated for the buffer is large enough
	if( p_bufferSize < (width+(width*3)%4)*height*(INT32)sizeof(UINT8) )
	{
		error = ZEF_BUFFER_TO_SMALL;
		goto errorHandler;
	}

	// seeking over the LUT
	error = fseek(file,1024,SEEK_CUR);
	if( error != 0 )
	{
		error = ZEF_FILE_SEEK;
		goto errorHandler;
	}

	// and then we read the BMP
	for(i=height-1;i>=0;i--)
	{
		fread(p_buffer+i*width,sizeof(UINT8),width,file); // we don't do any error checking in the loop
		fseek(file,(width*3)%4,SEEK_CUR);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_READ;
		goto errorHandler;
	}

	// set width and height
	*p_width = width;
	*p_height = height;

	// then close the file 
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 saveBmp( char *p_fileName, UINT8 *p_buffer, INT32 p_width, INT32 p_height )
{
	UINT8			header[2],lut[1024];
	UINT8			fill[4] = { 0, 0, 0, 0 };
	FILE			*file;
	INT32			i,error,val;

	// init
	file = NULL;

	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"wb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// we fill the header
	header[0] = 'B';
	header[1] = 'M';
  fwrite(header,1,2,file);
	val = 54+p_height*(p_width+(p_width*3)%4)+1024; fwrite(&val,4,1,file); // file size
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 1078; fwrite(&val,4,1,file);	//offset until data
	val = 40; fwrite(&val,4,1,file);		//size of BITMAPINFOHEADER structure
	val = p_width; fwrite(&val,4,1,file);  // no comment
	val = p_height; fwrite(&val,4,1,file); // no comment
	val = 1; fwrite(&val,2,1,file);	// # of plans
	val = 8; fwrite(&val,2,1,file); // bpp
	val = 0; fwrite(&val,4,1,file);		// compression 
	val = 0; fwrite(&val,4,1,file);		//size of image (can be 0 if not compressed)
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for X
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for Y
	val = 0; fwrite(&val,4,1,file);		//ClrUsed
	val = 0; fwrite(&val,4,1,file);		//ClrImportant
	
	// we generate the LUT
	for( i=0; i<256; i++ )
	{
		lut[i*4] = (UINT8)i;
		lut[i*4+1] = (UINT8)i;
		lut[i*4+2] = (UINT8)i;
		lut[i*4+3] = 0;
	}

	// and write it to disk
	error = fwrite( lut, sizeof(UINT8), 1024, file );
	if( error != 1024 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}
	
	// write the data
	for(i=p_height-1;i>=0;i--)
	{
		fwrite(p_buffer+i*p_width,sizeof(UINT8),p_width,file); // we don't do any error checking in the loop
		fwrite(fill,sizeof(UINT8),(p_width*3)%4,file);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	// close the file
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 saveBmp32( char *p_fileName, UINT32 *p_buffer, INT32 p_width, INT32 p_height )
{
	UINT8			header[2],lut[1024];
	UINT8			fill[4] = { 0, 0, 0, 0 };
	UINT8			tmp;
	FILE			*file;
	INT32			i,j,error,val;
	UINT32			max,min,div;

	// init
	file = NULL;
	max = 0;
	min = UINT_MAX;

	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"wb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// we fill the header
	header[0] = 'B';
	header[1] = 'M';
  fwrite(header,1,2,file);
	val = 54+p_height*(p_width+(p_width*3)%4)+1024; fwrite(&val,4,1,file); // file size
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 1078; fwrite(&val,4,1,file);	//offset until data
	val = 40; fwrite(&val,4,1,file);		//size of BITMAPINFOHEADER structure
	val = p_width; fwrite(&val,4,1,file);  // no comment
	val = p_height; fwrite(&val,4,1,file); // no comment
	val = 1; fwrite(&val,2,1,file);	// # of plans
	val = 8; fwrite(&val,2,1,file); // bpp
	val = 0; fwrite(&val,4,1,file);		// compression
	val = 0; fwrite(&val,4,1,file);		//size of image (can be 0 if not compressed)
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for X
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for Y
	val = 0; fwrite(&val,4,1,file);		//ClrUsed
	val = 0; fwrite(&val,4,1,file);		//ClrImportant
	
	// we generate the LUT
	for( i=0; i<256; i++ )
	{
		lut[i*4] = (UINT8)i;
		lut[i*4+1] = (UINT8)i;
		lut[i*4+2] = (UINT8)i;
		lut[i*4+3] = 0;
	}

	// and write it to disk
	error = fwrite( lut, sizeof(UINT8), 1024, file );
	if( error != 1024 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	for(i=0;i<p_height*p_width;i++)
	{
		if( p_buffer[i] > max ) 
			max = p_buffer[i];
		if( p_buffer[i] < min ) 
			min = p_buffer[i];
	}

	div = (max-min);
	if( div == 0 ) div = 1;
	
	// write the data
	for(i=p_height-1;i>=0;i--)
	{
		for( j=0; j<p_width; j++ )
		{
			tmp = 255*(p_buffer[i*p_width+j]-min)/div;
			fwrite(&tmp,sizeof(UINT8),1,file); // we don't do any error checking in the loop
		}	
		fwrite(fill,sizeof(UINT8),(p_width*3)%4,file);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	// close the file
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 saveBmp32S( char *p_fileName, INT32 *p_buffer, INT32 p_width, INT32 p_height )
{
	UINT8			header[2],lut[1024];
	UINT8			fill[4] = { 0, 0, 0, 0 };
	UINT8			tmp;
	FILE			*file;
	INT32			i,j,error,val;
	INT32			max,min,div;

	// init
	file = NULL;
	max = 0;
	min = UINT_MAX;

	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"wb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// we fill the header
	header[0] = 'B';
	header[1] = 'M';
  fwrite(header,1,2,file);
	val = 54+p_height*(p_width+(p_width*3)%4)+1024; fwrite(&val,4,1,file); // file size
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 1078; fwrite(&val,4,1,file);	//offset until data
	val = 40; fwrite(&val,4,1,file);		//size of BITMAPINFOHEADER structure
	val = p_width; fwrite(&val,4,1,file);  // no comment
	val = p_height; fwrite(&val,4,1,file); // no comment
	val = 1; fwrite(&val,2,1,file);	// # of plans
	val = 8; fwrite(&val,2,1,file); // bpp
	val = 0; fwrite(&val,4,1,file);		// compression
	val = 0; fwrite(&val,4,1,file);		//size of image (can be 0 if not compressed)
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for X
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for Y
	val = 0; fwrite(&val,4,1,file);		//ClrUsed
	val = 0; fwrite(&val,4,1,file);		//ClrImportant
	
	// we generate the LUT
	for( i=0; i<256; i++ )
	{
		lut[i*4] = (UINT8)i;
		lut[i*4+1] = (UINT8)i;
		lut[i*4+2] = (UINT8)i;
		lut[i*4+3] = 0;
	}

	// and write it to disk
	error = fwrite( lut, sizeof(UINT8), 1024, file );
	if( error != 1024 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	for(i=0;i<p_height*p_width;i++)
	{
		if( p_buffer[i] > max ) 
			max = p_buffer[i];
		if( p_buffer[i] < min ) 
			min = p_buffer[i];
	}

	div = (max-min);
	if( div == 0 ) div = 1;
	
	// write the data
	for(i=p_height-1;i>=0;i--)
	{
		for( j=0; j<p_width; j++ )
		{
			tmp = 255*(p_buffer[i*p_width+j]-min)/div;
			fwrite(&tmp,sizeof(UINT8),1,file); // we don't do any error checking in the loop
		}	
		fwrite(fill,sizeof(UINT8),(p_width*3)%4,file);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	// close the file
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 saveBmpR( char *p_fileName, REAL32 *p_buffer, INT32 p_width, INT32 p_height )
{
	UINT8			header[2],lut[1024];
	UINT8			fill[4] = { 0, 0, 0, 0 };
	UINT8			tmp;
	FILE			*file;
	INT32			i,j,error,val;
	REAL32			max,min,div;

	// init
	file = NULL;
	max = FLT_MIN;
	min = FLT_MAX;
	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"wb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// we fill the header
	header[0] = 'B';
	header[1] = 'M';
  fwrite(header,1,2,file);
	val = 54+p_height*(p_width+(p_width*3)%4)+1024; fwrite(&val,4,1,file); // file size
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 1078; fwrite(&val,4,1,file);	//offset until data
	val = 40; fwrite(&val,4,1,file);		//size of BITMAPINFOHEADER structure
	val = p_width; fwrite(&val,4,1,file);  // no comment
	val = p_height; fwrite(&val,4,1,file); // no comment
	val = 1; fwrite(&val,2,1,file);	// # of plans
	val = 8; fwrite(&val,2,1,file); // bpp
	val = 0; fwrite(&val,4,1,file);		// compression
	val = 0; fwrite(&val,4,1,file);		//size of image (can be 0 if not compressed)
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for X
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for Y
	val = 0; fwrite(&val,4,1,file);		//ClrUsed
	val = 0; fwrite(&val,4,1,file);		//ClrImportant
	
	// we generate the LUT
	for( i=0; i<256; i++ )
	{
		lut[i*4] = (UINT8)i;
		lut[i*4+1] = (UINT8)i;
		lut[i*4+2] = (UINT8)i;
		lut[i*4+3] = 0;
	}

	// and write it to disk
	error = fwrite( lut, sizeof(UINT8), 1024, file );
	if( error != 1024 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	for(i=0;i<p_height*p_width;i++)
	{
		if( p_buffer[i] > max ) max = p_buffer[i];
		if( p_buffer[i] < min ) min = p_buffer[i];
	}

	div = (max-min);
	if( div == 0 ) div = 1;
	
	// write the data
	for(i=p_height-1;i>=0;i--)
	{
		for( j=0; j<p_width; j++ )
		{
			tmp = (UINT8)(255*(p_buffer[i*p_width+j]-min)/(div));
			fwrite(&tmp,sizeof(UINT8),1,file); // we don't do any error checking in the loop
		}	
		fwrite(fill,sizeof(UINT8),(p_width*3)%4,file);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	// close the file
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 saveBmpC( char *p_fileName, REAL32 *p_buffer, INT32 p_width, INT32 p_height )
{
	UINT8			header[2],lut[1024];
	UINT8			fill[4] = { 0, 0, 0, 0 };
	UINT8			tmp;
	FILE			*file;
	INT32			i,j,error,val;
	REAL32			max,min,div;

	// init
	file = NULL;
	max = FLT_MIN;
	min = FLT_MAX;
	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"wb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// we fill the header
	header[0] = 'B';
	header[1] = 'M';
  fwrite(header,1,2,file);
	val = 54+p_height*(p_width+(p_width*3)%4)+1024; fwrite(&val,4,1,file); // file size
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 1078; fwrite(&val,4,1,file);	//offset until data
	val = 40; fwrite(&val,4,1,file);		//size of BITMAPINFOHEADER structure
	val = p_width; fwrite(&val,4,1,file);  // no comment
	val = p_height; fwrite(&val,4,1,file); // no comment
	val = 1; fwrite(&val,2,1,file);	// # of plans
	val = 8; fwrite(&val,2,1,file); // bpp
	val = 0; fwrite(&val,4,1,file);		// compression
	val = 0; fwrite(&val,4,1,file);		//size of image (can be 0 if not compressed)
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for X
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for Y
	val = 0; fwrite(&val,4,1,file);		//ClrUsed
	val = 0; fwrite(&val,4,1,file);		//ClrImportant
	
	// we generate the LUT
	for( i=0; i<256; i++ )
	{
		lut[i*4] = (UINT8)i;
		lut[i*4+1] = (UINT8)i;
		lut[i*4+2] = (UINT8)i;
		lut[i*4+3] = 0;
	}

	// and write it to disk
	error = fwrite( lut, sizeof(UINT8), 1024, file );
	if( error != 1024 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	for(i=0;i<p_height*p_width*2;i+=2)
	{
		if( p_buffer[i]*p_buffer[i]+p_buffer[i+1]*p_buffer[i+1] > max ) max = p_buffer[i]*p_buffer[i]+p_buffer[i+1]*p_buffer[i+1];
		if( p_buffer[i]*p_buffer[i]+p_buffer[i+1]*p_buffer[i+1] < min ) min = p_buffer[i]*p_buffer[i]+p_buffer[i+1]*p_buffer[i+1];
	}

	div = max-min;
	if( div == 0 ) div = 1;
	
	// write the data
	for(i=p_height-1;i>=0;i--)
	{
		for( j=0; j<p_width*2; j+=2 )
		{
			tmp = (UINT8)(255*(p_buffer[i*p_width*2+j]*p_buffer[i*p_width*2+j]+p_buffer[i*p_width*2+j+1]*p_buffer[i*p_width*2+j+1]-min)/(div));
			fwrite(&tmp,sizeof(UINT8),1,file); // we don't do any error checking in the loop
		}	
		fwrite(fill,sizeof(UINT8),(p_width*3)%4,file);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	// close the file
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 saveBmpS( char *p_fileName, INT8 *p_buffer, INT32 p_width, INT32 p_height, UINT32 p_nbPlans, UINT32 p_plan )
{
	UINT8			header[2],lut[1024];
	UINT8			fill[4] = { 0, 0, 0, 0 };
	UINT8			tmp;
	FILE			*file;
	INT32			i,j,error,val;

	// init
	file = NULL;

	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"wb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// we fill the header
	header[0] = 'B';
	header[1] = 'M';
  fwrite(header,1,2,file);
	val = 54+p_height*(p_width+(p_width*3)%4)+1024; fwrite(&val,4,1,file); // file size
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 1078; fwrite(&val,4,1,file);	//offset until data
	val = 40; fwrite(&val,4,1,file);		//size of BITMAPINFOHEADER structure
	val = p_width; fwrite(&val,4,1,file);  // no comment
	val = p_height; fwrite(&val,4,1,file); // no comment
	val = 1; fwrite(&val,2,1,file);	// # of plans
	val = 8; fwrite(&val,2,1,file); // bpp
	val = 0; fwrite(&val,4,1,file);		// compression
	val = 0; fwrite(&val,4,1,file);		//size of image (can be 0 if not compressed)
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for X
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for Y
	val = 0; fwrite(&val,4,1,file);		//ClrUsed
	val = 0; fwrite(&val,4,1,file);		//ClrImportant
	
	// we generate the LUT
	for( i=0; i<256; i++ )
	{
		lut[i*4] = (UINT8)i;
		lut[i*4+1] = (UINT8)i;
		lut[i*4+2] = (UINT8)i;
		lut[i*4+3] = 0;
	}

	// and write it to disk
	error = fwrite( lut, sizeof(UINT8), 1024, file );
	if( error != 1024 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}
	
	// write the data
	for(i=p_height-1;i>=0;i--)
	{
		for( j=0; j<p_width*(INT32)p_nbPlans; j+=p_nbPlans )
		{
			tmp = 127+p_buffer[i*p_width*p_nbPlans+j+p_plan];
			fwrite(&tmp,sizeof(UINT8),1,file); // we don't do any error checking in the loop
		}	
		fwrite(fill,sizeof(UINT8),(p_width*3)%4,file);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	// close the file
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;
}

INT32 saveBmpS16( char *p_fileName, INT16 *p_buffer, INT32 p_width, INT32 p_height, UINT32 p_nbPlans, UINT32 p_plan )
{
	UINT8			header[2],lut[1024];
	UINT8			fill[4] = { 0, 0, 0, 0 };
	UINT8			tmp;
	FILE			*file;
	INT32			i,j,error,val;
	INT32			max,min;

	// init
	file = NULL;
	max = min = 0;

	// checking filename
	if( p_fileName == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// file opening
	file=fopen(p_fileName,"wb");
	if( file == NULL )
	{
		error = ZEF_FILE_OPEN;
		goto errorHandler;
	}

	// we fill the header
	header[0] = 'B';
	header[1] = 'M';
  fwrite(header,1,2,file);
	val = 54+p_height*(p_width+(p_width*3)%4)+1024; fwrite(&val,4,1,file); // file size
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 0; fwrite(&val,2,1,file);		//reserved
	val = 1078; fwrite(&val,4,1,file);	//offset until data
	val = 40; fwrite(&val,4,1,file);		//size of BITMAPINFOHEADER structure
	val = p_width; fwrite(&val,4,1,file);  // no comment
	val = p_height; fwrite(&val,4,1,file); // no comment
	val = 1; fwrite(&val,2,1,file);	// # of plans
	val = 8; fwrite(&val,2,1,file); // bpp
	val = 0; fwrite(&val,4,1,file);		// compression
	val = 0; fwrite(&val,4,1,file);		//size of image (can be 0 if not compressed)
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for X
	val = 2834; fwrite(&val,4,1,file);	//pixels per meter for Y
	val = 0; fwrite(&val,4,1,file);		//ClrUsed
	val = 0; fwrite(&val,4,1,file);		//ClrImportant
	
	// we generate the LUT
	for( i=0; i<256; i++ )
	{
		lut[i*4] = (UINT8)i;
		lut[i*4+1] = (UINT8)i;
		lut[i*4+2] = (UINT8)i;
		lut[i*4+3] = 0;
	}

	// and write it to disk
	error = fwrite( lut, sizeof(UINT8), 1024, file );
	if( error != 1024 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	for(i=p_plan;i<(INT32)(p_height*p_width*p_nbPlans);i+=p_nbPlans)
	{
		if( p_buffer[i] > max ) max = p_buffer[i];
		if( p_buffer[i] < min ) min = p_buffer[i];
	}

	if( zAbs(min) > max ) max = zAbs(min);
	if( max == 0 ) max = 1;
	
	// write the data
	for(i=p_height-1;i>=0;i--)
	{
		for( j=0; j<p_width*(INT32)p_nbPlans; j+=p_nbPlans )
		{
			tmp = 127+127*p_buffer[i*p_width*p_nbPlans+j+p_plan]/max;
			fwrite(&tmp,sizeof(UINT8),1,file); // we don't do any error checking in the loop
		}	
		fwrite(fill,sizeof(UINT8),(p_width*3)%4,file);
	}
	// error checking
	if( ferror(file) != 0 )
	{
		error = ZEF_FILE_WRITE;
		goto errorHandler;
	}

	// close the file
	fclose(file);

	return ZEF_OK;

	//error handling
errorHandler:
	if( file != NULL )
		fclose(file);

	return error;

}
