/*
Company :		ZEFYR Technologies
File :			FFT2D.c
Author :		Matthieu Perreira Da Silva (based on Takuya OOURA (email: ooura@kurims.kyoto-u.ac.jp) algorithms) 
Fisrt release :	15/04/2003

Description : Fast Fourier Transform

Features :
    dimension   :two
    data length :power of 2
    decimation  :frequency
    radix       :split-radix, row-column
    data        :inplace
    table       :use

Functions :
    cdft2d: Complex Discrete Fourier Transform
    rdft2d: Real Discrete Fourier Transform

Function prototypes :
    void cdft2d(INT32, INT32, INT32, REAL32 **, REAL32 *, INT32 *, REAL32 *);
    void rdft2d(INT32, INT32, INT32, REAL32 **, REAL32 *, INT32 *, REAL32 *);
    void rdft2dsort(INT32, INT32, INT32, REAL32 **);

Necessary package
    FFT1D.c / FFT1D.h  : 1D-FFT package

*/

#include <stdlib.h>

#include "FFT1D.h"
#include "FFT1Db.h"
#include "FFT2D.h"

#if !defined(ZEF_FIXED_POINT)

INT32 zLog2( INT32 p_number )
{
	INT32 i;

	for( i=0; i<32; i++ )
	{
		if( p_number&1 ) 
			return i;		// return i where p_number = 2^i
		p_number >>= 1;
	}

	return 0; // return 0 if p_number == 0
}

void cdft2d(INT32 p_sizeY, INT32 p_sizeX, INT32 p_sign, REAL32 *p_data, REAL32 *p_temp, INT32 *p_inPlace, REAL32 *p_trigo)
{
    INT32		n, itnull, nthread, nt, i;
	INT32		xShift;
    
	xShift = zLog2(p_sizeX);

    n = p_sizeY << 1;
    if (n < p_sizeX) {
        n = p_sizeX;
    }
    if (n > (p_inPlace[0] << 2)) {
        makewt(n >> 2, p_inPlace, p_trigo);
    }
    itnull = 0;
    if (p_temp == NULL) {
        itnull = 1;
        nthread = 1;
        nt = 8 * nthread * p_sizeY;
        if (p_sizeX == 4 * nthread) {
            nt >>= 1;
        } else if (p_sizeX < 4 * nthread) {
            nt >>= 2;
        }
        p_temp = (REAL32 *) MALLOC(sizeof(REAL32) * nt);
    }

    {
        for (i = 0; i < p_sizeY; i++) {
            cdft(p_sizeX, p_sign, p_data+(i<<xShift), p_inPlace, p_trigo);
        }
        cdft2d_sub(p_sizeY, p_sizeX, p_sign, p_data, p_temp, p_inPlace, p_trigo);
    }
    if (itnull != 0) {
        FREE(p_temp);
    }
}


void rdft2d(INT32 p_sizeY, INT32 p_sizeX, INT32 p_sign, REAL32 *p_data, REAL32 *p_temp, INT32 *p_inPlace, REAL32 *p_trigo)
{
    INT32		n, nw, nc, itnull, nthread, nt, i;
	INT32		xShift;
    
	xShift = zLog2(p_sizeX);
    
    n = p_sizeY << 1;
    if (n < p_sizeX) {
        n = p_sizeX;
    }
    nw = p_inPlace[0];
    if (n > (nw << 2)) {
        nw = n >> 2;
        makewt(nw, p_inPlace, p_trigo);
    }
    nc = p_inPlace[1];
    if (p_sizeX > (nc << 2)) {
        nc = p_sizeX >> 2;
        makect(nc, p_inPlace, p_trigo + nw);
    }
    itnull = 0;
    if (p_temp == NULL) {
        itnull = 1;
        nthread = 1;
        nt = 8 * nthread * p_sizeY;
        if (p_sizeX == 4 * nthread) {
            nt >>= 1;
        } else if (p_sizeX < 4 * nthread) {
            nt >>= 2;
        }
        p_temp = (REAL32 *) MALLOC(sizeof(REAL32) * nt);
    }
    {
        if (p_sign < 0) {
            rdft2d_sub(p_sizeY, p_sizeX, p_sign, p_data);
            cdft2d_sub(p_sizeY, p_sizeX, p_sign, p_data, p_temp, p_inPlace, p_trigo);
        }
        for (i = 0; i < p_sizeY; i++) {
            rdft(p_sizeX, p_sign, p_data+(i<<xShift), p_inPlace, p_trigo);
        }
        if (p_sign >= 0) {
            cdft2d_sub(p_sizeY, p_sizeX, p_sign, p_data, p_temp, p_inPlace, p_trigo);
            rdft2d_sub(p_sizeY, p_sizeX, p_sign, p_data);
        }
    }
    if (itnull != 0) {
        FREE(p_temp);
    }
}


void rdft2dsort(INT32 p_sizeY, INT32 p_sizeX, INT32 isgn, REAL32 *a)
{
    INT32		n1h, i;
    REAL32	x, y;
	INT32		xShift;
    
	xShift = zLog2(p_sizeX);
    
    n1h = p_sizeY >> 1;
    if (isgn < 0) {
        for (i = n1h + 1; i < p_sizeY; i++) {
            a[(i<<xShift)] = a[(i<<xShift)+p_sizeX + 1];
            a[(i<<xShift)+1] = a[(i<<xShift)+p_sizeX];
        }
        a[1] = a[p_sizeX];
        a[(n1h<<xShift)+1] = a[(n1h<<xShift)+p_sizeX];
    } else {
        for (i = n1h + 1; i < p_sizeY; i++) {
            y = a[(i<<xShift)];
            x = a[(i<<xShift)+1];
            a[(i<<xShift)+p_sizeX] = x;
            a[(i<<xShift)+p_sizeX + 1] = y;
            a[((p_sizeY - i)<<xShift)+p_sizeX] = x;
            a[((p_sizeY - i)<<xShift)+p_sizeX + 1] = -y;
            a[(i<<xShift)] = a[(p_sizeY - i)<<xShift];
            a[(i<<xShift)+1] = -a[((p_sizeY - i)<<xShift)+1];
        }
        a[p_sizeX] = a[1];
        a[p_sizeX + 1] = 0;
        a[1] = 0;
        a[(n1h<<xShift)+p_sizeX] = a[(n1h<<xShift)+1];
        a[(n1h<<xShift)+p_sizeX + 1] = 0;
        a[(n1h<<xShift)+1] = 0;
    }
}

/* -------- child routines -------- */


void cdft2d_sub(INT32 p_sizeY, INT32 p_sizeX, INT32 isgn, REAL32 *p_data, REAL32 *p_temp, INT32 *p_inPlace, REAL32 *p_trigo)
{
    INT32		i, j;
	INT32		xShift;
    
	xShift = zLog2(p_sizeX);
    
    if (p_sizeX > 4) {
        for (j = 0; j < p_sizeX; j += 8) {
            for (i = 0; i < p_sizeY; i++) {
                p_temp[2 * i] = p_data[(i<<xShift) + j];
                p_temp[2 * i + 1] = p_data[(i<<xShift) + j + 1];
                p_temp[2 * p_sizeY + 2 * i] = p_data[(i<<xShift) + j + 2];
                p_temp[2 * p_sizeY + 2 * i + 1] = p_data[(i<<xShift) + j + 3];
                p_temp[4 * p_sizeY + 2 * i] = p_data[(i<<xShift) + j + 4];
                p_temp[4 * p_sizeY + 2 * i + 1] = p_data[(i<<xShift) + j + 5];
                p_temp[6 * p_sizeY + 2 * i] = p_data[(i<<xShift) + j + 6];
                p_temp[6 * p_sizeY + 2 * i + 1] = p_data[(i<<xShift) + j + 7];
            }
            cdft(2 * p_sizeY, isgn, p_temp, p_inPlace, p_trigo);
            cdft(2 * p_sizeY, isgn, &p_temp[2 * p_sizeY], p_inPlace, p_trigo);
            cdft(2 * p_sizeY, isgn, &p_temp[4 * p_sizeY], p_inPlace, p_trigo);
            cdft(2 * p_sizeY, isgn, &p_temp[6 * p_sizeY], p_inPlace, p_trigo);
            for (i = 0; i < p_sizeY; i++) {
                p_data[(i<<xShift) + j] = p_temp[2 * i];
                p_data[(i<<xShift) + j + 1] = p_temp[2 * i + 1];
                p_data[(i<<xShift) + j + 2] = p_temp[2 * p_sizeY + 2 * i];
                p_data[(i<<xShift) + j + 3] = p_temp[2 * p_sizeY + 2 * i + 1];
                p_data[(i<<xShift) + j + 4] = p_temp[4 * p_sizeY + 2 * i];
                p_data[(i<<xShift) + j + 5] = p_temp[4 * p_sizeY + 2 * i + 1];
                p_data[(i<<xShift) + j + 6] = p_temp[6 * p_sizeY + 2 * i];
                p_data[(i<<xShift) + j + 7] = p_temp[6 * p_sizeY + 2 * i + 1];
            }
        }
    } else if (p_sizeX == 4) {
        for (i = 0; i < p_sizeY; i++) {
            p_temp[2 * i] = p_data[(i<<xShift)];
            p_temp[2 * i + 1] = p_data[(i<<xShift) + 1];
            p_temp[2 * p_sizeY + 2 * i] = p_data[(i<<xShift) + 2];
            p_temp[2 * p_sizeY + 2 * i + 1] = p_data[(i<<xShift) + 3];
        }
        cdft(2 * p_sizeY, isgn, p_temp, p_inPlace, p_trigo);
        cdft(2 * p_sizeY, isgn, &p_temp[2 * p_sizeY], p_inPlace, p_trigo);
        for (i = 0; i < p_sizeY; i++) {
            p_data[(i<<xShift)] = p_temp[2 * i];
            p_data[(i<<xShift) + 1] = p_temp[2 * i + 1];
            p_data[(i<<xShift) + 2] = p_temp[2 * p_sizeY + 2 * i];
            p_data[(i<<xShift) + 3] = p_temp[2 * p_sizeY + 2 * i + 1];
        }
    } else if (p_sizeX == 2) {
        for (i = 0; i < p_sizeY; i++) {
            p_temp[2 * i] = p_data[(i<<xShift)];
            p_temp[2 * i + 1] = p_data[(i<<xShift) + 1];
        }
        cdft(2 * p_sizeY, isgn, p_temp, p_inPlace, p_trigo);
        for (i = 0; i < p_sizeY; i++) {
            p_data[(i<<xShift)] = p_temp[2 * i];
            p_data[(i<<xShift) + 1] = p_temp[2 * i + 1];
        }
    }
}


void rdft2d_sub(INT32 p_sizeY, INT32 p_sizeX, INT32 p_sign, REAL32 *p_data)
{
    INT32		n1h, i, j;
    REAL32	xi;
	INT32		xShift;
    
	xShift = zLog2(p_sizeX);
    
    n1h = p_sizeY >> 1;
    if (p_sign < 0) {
        for (i = 1; i < n1h; i++) {
            j = p_sizeY - i;
            xi = p_data[(i<<xShift) + 0] - p_data[(j<<xShift) + 0];
            p_data[(i<<xShift) + 0] += p_data[(j<<xShift) + 0];
            p_data[(j<<xShift) + 0] = xi;
            xi = p_data[(j<<xShift) + 1] - p_data[(i<<xShift) + 1];
            p_data[(i<<xShift) + 1] += p_data[(j<<xShift) + 1];
            p_data[(j<<xShift) + 1] = xi;
        }
    } else {
        for (i = 1; i < n1h; i++) {
            j = p_sizeY - i;
            p_data[(j<<xShift) + 0] = 0.5f * (p_data[(i<<xShift) + 0] - p_data[(j<<xShift) + 0]);
            p_data[(i<<xShift) + 0] -= p_data[(j<<xShift) + 0];
            p_data[(j<<xShift) + 1] = 0.5f * (p_data[(i<<xShift) + 1] + p_data[(j<<xShift) + 1]);
            p_data[(i<<xShift) + 1] -= p_data[(j<<xShift) + 1];
        }
    }
}

#endif
