/*
Company :		ZEFYR Technologies
File :			coresDeltas.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : core and deltas position determination

Features : routines to process cores position
    
Functions : 

	findCoresDeltas	: main function that allow to find cores and deltas in an image
	getCoreDeltasXY	: retreives the coordinates of the cores ans deltas from a coreMap processed with poincare function	
	processOriMap	: calculate orientation map from a gradient image
	scaleOriMap		: scale an orientation map
	findCoreOrientation	: processes the orientation of the main core of the image
	getVerticalAxis : Searches the position of the vertical axis of symmetry.
	getHorizontalAxis : Searches the position of the horizontal axis the "core" of an arch image.

Function prototypes : 

	INT32 findCoresDeltas( BioEngineData *p_data );
	INT32 getCoreDeltasXY( INT8 *p_src, UINT8 *p_quality, UINT32 p_width, UINT32 p_height, Template *p_template );
	INT32 processOriMap( INT16 *p_gradient, UINT8 *p_ori, UINT8 *p_quality, UINT32 p_width, UINT32 p_height );
	INT32 scaleOriMap( UINT8 *p_src, UINT8 *p_dst, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 p_scale );
	INT32 findCoreOrientation( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template );
	INT32 getVerticalAxis( UINT8 *p_ori, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 *p_symAxis )
	INT32 getHorizontalAxis( UINT8 *p_ori, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 p_yAxis, UINT32 *p_symAxis, UINT32 p_filterSize )


Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	filters.h
	zMaths.h
	coresDeltas.h
	geometry.h

Comments : none

*/


#include <stdlib.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "filters.h"
#include "zMaths.h"
#include "coresDeltas.h"
#include "geometry.h"
#include "BMPfiles.h"

INT32 findCoresDeltas( BioEngineData *p_data )
{
	INT32		error;
	INT8		*coreMap;
	UINT8		*oriMap;
	INT16		*gradient;
	UINT32		poincareSize,xCore,yCore;
#ifdef TIME_REPORT
	clock_t timeReport;
#endif

	// init
	coreMap = NULL;
	oriMap = NULL;
	gradient = NULL;
	LOCK_MUTEX(g_paramsMutex);
	poincareSize = g_params.m_poincareSize;
	UNLOCK_MUTEX(g_paramsMutex);

	// Gradient memory allocations
	gradient = (INT16*)MALLOC( 2*p_data->m_width*p_data->m_height*sizeof(INT16) );
	if( gradient == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// calculate gradient
	error = sobelGradient( p_data->m_workImg, gradient, p_data->m_width, p_data->m_height );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas - sobelGradient" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		saveBmpS16( "CoreDelta-01gradientX.bmp", gradient, p_data->m_width, p_data->m_height,2,1 );
		saveBmpS16( "CoreDelta-01gradientY.bmp", gradient, p_data->m_width, p_data->m_height,2,2 );
	#endif

	// OriMap memory allocations
	oriMap = (UINT8*)MALLOC( (p_data->m_width/ORI_BLOC_SIZE2)*(p_data->m_height/ORI_BLOC_SIZE2)*sizeof(UINT8) );
	if( oriMap == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// process orientation map
	error = processOriMap( gradient, oriMap, p_data->m_quality, p_data->m_width, p_data->m_height );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas - processOriMap" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		saveBmp( "CoreDelta-02oriMap.bmp",oriMap, p_data->m_width/ORI_BLOC_SIZE2, p_data->m_height/ORI_BLOC_SIZE2 );
	#endif

	// FREE gradient memory
	FREE(gradient);
	gradient = NULL;

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// scale orientation map
	error = scaleOriMap( oriMap, p_data->m_ori, p_data->m_quality, p_data->m_width/ORI_BLOC_SIZE2, p_data->m_height/ORI_BLOC_SIZE2, ORI_BLOC_SIZE2 );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas - scaleOriMap" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		saveBmp( "CoreDelta-03scaledOriMap.bmp",p_data->m_ori, p_data->m_width, p_data->m_height );
	#endif
	
	// FREE temporary orientation map
	FREE(oriMap);
	oriMap = NULL;

	// Core memory allocations
	coreMap = (INT8*)MALLOC( p_data->m_width*p_data->m_height*sizeof(INT8) );
	if( coreMap == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// process poincare index
	error = poincare( p_data->m_ori, coreMap, p_data->m_width, p_data->m_height, poincareSize );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas - poincare" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		saveBmpS( "CoreDelta-04coreMap.bmp",coreMap, p_data->m_width, p_data->m_height, 1, 1 );
	#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// find cores and deltas location
	error = getCoreDeltasXY( coreMap, p_data->m_quality, p_data->m_width, p_data->m_height, &p_data->m_template );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas - getCoreDeltasXY" ,clock()-timeReport );
#endif

	// FREE coreMap
	FREE(coreMap);
	coreMap = NULL;

	// if couldn't find core with poincare, try another technique
	if( p_data->m_template.m_nbCores == 0 )
	{
#ifdef TIME_REPORT
	timeReport = clock();
#endif
		// Search for the vertical axis of symmetry
		error = getVerticalAxis( p_data->m_ori, p_data->m_quality, p_data->m_width, p_data->m_height, &xCore );
		if( error < ZEF_OK )
			goto errorHandler;

		// THen search for the point that have maximum cruvature allong the vertical axis of symmetry
		error = getHorizontalAxis( p_data->m_ori, p_data->m_quality, p_data->m_width, p_data->m_height, xCore, &yCore, 16 );
		if( error < ZEF_OK )
			goto errorHandler;

		// strore these coordinates as the core of the fingerprint
		p_data->m_template.m_cores[0].m_x = xCore;
		p_data->m_template.m_cores[0].m_y = yCore;

#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas - curvature" ,clock()-timeReport );
#endif
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// Find 1st core orientation
	error = findCoreOrientation( p_data->m_ori, p_data->m_width, p_data->m_height, &(p_data->m_template) );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas - findCoreOrientation" ,clock()-timeReport );
#endif

	return ZEF_OK;

	// error Handler
errorHandler:
	if( coreMap != NULL )
		FREE( coreMap );
	if( oriMap != NULL )
		FREE( oriMap );
	if( gradient != NULL )
		FREE( gradient );

	return error;
}

INT32 getCoreDeltasXY( INT8 *p_src, UINT8 *p_quality, UINT32 p_width, UINT32 p_height, Template *p_template )
{
	UINT32		i,j;
	UINT32		nbPtsCore[MAX_CORES_DELTAS];
	UINT32		nbPtsDelta[MAX_CORES_DELTAS];
	Coord16		*corePts[MAX_CORES_DELTAS];
	Coord16		*deltaPts[MAX_CORES_DELTAS];
	UINT32		sumX,sumY;
	BOOL8		insert;
	UINT32		offset,width2,decX,decY;
	INT32		error,x,y,dec;
	UINT32		fftBlocOverlap,distFindCoreAngle,maxNumberOfCores;

	// init
	LOCK_MUTEX( g_paramsMutex );
		fftBlocOverlap = g_params.m_fftBlocOverlap;
		distFindCoreAngle = g_params.m_distFindCoreAngle;
		maxNumberOfCores = g_params.m_maxNumberOfCores;
	UNLOCK_MUTEX( g_paramsMutex );

	for( i=0; i<MAX_CORES_DELTAS; i++ )
	{
		nbPtsCore[i] = 0;
		nbPtsDelta[i] = 0;
		corePts[i] = NULL;
		deltaPts[i] = NULL;
	}

	if( fftBlocOverlap/2 > IMAGE_BORDER_LIMIT )
		dec = fftBlocOverlap/2;
	else
		dec = IMAGE_BORDER_LIMIT;

	width2 = p_width/FUSION_BLOC_SIZE;

	decX = IMAGE_BORDER_LIMIT/FUSION_BLOC_SIZE;  // +1 because of roundings...
	decY = (IMAGE_BORDER_LIMIT/FUSION_BLOC_SIZE)*width2;

	// look for pixels with values different from 0
	// if > 0 => core
	// if < 0 => delta
	for( y=dec; y<(INT32)p_height-dec; y++ )
	{
		for( x=dec; x<(INT32)p_width-dec; x++ )
		{
			offset = (y/FUSION_BLOC_SIZE)*width2+x/FUSION_BLOC_SIZE;
			// if the pixel is part of a core or delta and is far enough from the border
			if( p_src[y*p_width+x] != 0 && p_quality[offset-decX-decY] != 0 &&
				p_quality[offset+decX-decY] != 0 && 
				p_quality[offset-decX+decY] != 0 &&
				p_quality[offset+decX+decY] != 0 )
			{
				if( p_src[y*p_width+x] > 0 ) // core
				{
					insert = FALSE;
					for( i=0; i<MAX_CORES_DELTAS; i++ )
					{
						if( nbPtsCore[i] != 0 )
						{
							for( j=0; j<nbPtsCore[i]; j++ )
							{
								// if the points are 8-connected
								if( zAbs( x-corePts[i][j].m_x ) <= 1 &&  zAbs( y-corePts[i][j].m_y ) <= 1 )
								{
									corePts[i][nbPtsCore[i]].m_x = x;
									corePts[i][nbPtsCore[i]].m_y = y;
									if( nbPtsCore[i] < MAX_PTS_CORE_DELTA ) // nb pts per core is limited
										nbPtsCore[i]++;
									insert = TRUE;
									break;
								}
							}
						}
						else
						{
							// if the point is 8-connected with any other point
							if( !insert )
							{
								corePts[i] = (Coord16*)MALLOC(MAX_PTS_CORE_DELTA*sizeof(Coord16) );
								if( corePts[i] == NULL )
									return ZEF_OUT_OF_MEMORY;
								corePts[i][0].m_x = x;
								corePts[i][0].m_y = y;
								nbPtsCore[i] = 1;
							}
							break;
						}
					}
				}
				else // delta
				{
					insert = FALSE;
					for( i=0; i<MAX_CORES_DELTAS; i++ )
					{
						if( nbPtsDelta[i] != 0 )
						{
							for( j=0; j<nbPtsDelta[i]; j++ )
							{
								// if the points are 8-connected
								if( zAbs( x-deltaPts[i][j].m_x ) <= 1 &&  zAbs( y-deltaPts[i][j].m_y ) <= 1 )
								{
									deltaPts[i][nbPtsDelta[i]].m_x = x;
									deltaPts[i][nbPtsDelta[i]].m_y = y;
									if( nbPtsDelta[i] < MAX_PTS_CORE_DELTA ) // nb pts per core is limited
										nbPtsDelta[i]++;
									insert = TRUE;
									break;
								}
							}
						}
						else
						{
							// if the point is 8-connected with any other point
							if( !insert )
							{
								deltaPts[i] = (Coord16*)MALLOC(MAX_PTS_CORE_DELTA*sizeof(Coord16) );
								if( deltaPts[i] == NULL )
								{
									error = ZEF_OUT_OF_MEMORY;
									goto errorHandler;
								}
								deltaPts[i][0].m_x = x;
								deltaPts[i][0].m_y = y;
								nbPtsDelta[i] = 1;
							}
							break;
						}
					}
				}
			}
			/*else // for debug only
				p_src[y*p_width+x] = 0;*/
		}
	}

	// we calculate the center of each "core region"
	for( i=0; i<MAX_CORES_DELTAS; i++ )
	{
		sumX=0;sumY=0;
		if( nbPtsCore[i] > 0 )
		{
			for( j=0; j<nbPtsCore[i]; j ++ )
			{
				sumX += corePts[i][j].m_x;
				sumY += corePts[i][j].m_y;
			}
			corePts[i][0].m_x = sumX/nbPtsCore[i];
			corePts[i][0].m_y = sumY/nbPtsCore[i];
		}

		sumX=0;sumY=0;
		if( nbPtsDelta[i] > 0 )
		{
			for( j=0; j<nbPtsDelta[i]; j ++ )
			{
				sumX += deltaPts[i][j].m_x;
				sumY += deltaPts[i][j].m_y;
			}
			deltaPts[i][0].m_x = sumX/nbPtsDelta[i];
			deltaPts[i][0].m_y = sumY/nbPtsDelta[i];
		}
	}

	// check for fake core/deltas
	// We can't have core and deltas to close
	for( i=0; i<MAX_CORES_DELTAS; i++ )
	{
		if( nbPtsCore[i] > 0 )
		{
			for( j=0; j<MAX_CORES_DELTAS; j++ )
			{
				if( nbPtsDelta[j] > 0 )
				{
					if( zAbs( corePts[i][0].m_x - deltaPts[j][0].m_x ) < (INT32)distFindCoreAngle &&
						zAbs( corePts[i][0].m_y - deltaPts[j][0].m_y ) < (INT32)distFindCoreAngle )
					{
						nbPtsCore[i] = 0;
						nbPtsDelta[j] = 0;
					}
				}
			}
		}
	}

	// init of template data
	p_template->m_nbCores = 0;
	p_template->m_nbDeltas = 0;
	p_template->m_cores	 = (Coord16*)MALLOC(maxNumberOfCores*sizeof(Coord16) );
	if( p_template->m_cores == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_template->m_deltas = (Coord16*)MALLOC(maxNumberOfCores*sizeof(Coord16) );
	if( p_template->m_deltas == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// write the cores into the template
	// now the strategy is : we write the m_maxNumberOfCores first cores and deltas
	for( i=0; i<MAX_CORES_DELTAS; i++ )
	{
		if( nbPtsCore[i] > 0 && p_template->m_nbCores < maxNumberOfCores )
		{
			p_template->m_cores[p_template->m_nbCores].m_x = corePts[i][0].m_x;
			p_template->m_cores[p_template->m_nbCores].m_y = corePts[i][0].m_y;
			p_template->m_nbCores++;
		}
		if( nbPtsDelta[i] > 0 && p_template->m_nbDeltas < maxNumberOfCores )
		{
			p_template->m_deltas[p_template->m_nbDeltas].m_x = deltaPts[i][0].m_x;
			p_template->m_deltas[p_template->m_nbDeltas].m_y = deltaPts[i][0].m_y;
			p_template->m_nbDeltas++;
		}
	}

	// deallocation
	for( i=0; i<MAX_CORES_DELTAS; i++ )
	{
		if( corePts[i] != NULL )
			FREE(corePts[i]);

		if( deltaPts[i] != NULL )
			FREE(deltaPts[i]);
	}

	return ZEF_OK;

	// error Handler
errorHandler:

	for( i=0; i<MAX_CORES_DELTAS; i++ )
	{
		if( corePts[i] != NULL )
			FREE(corePts[i]);
		if( deltaPts[i] != NULL )
			FREE(deltaPts[i]);
	}

	return error;
}

INT32 processOriMap( INT16 *p_gradient, UINT8 *p_ori, UINT8 *p_quality, UINT32 p_width, UINT32 p_height )
{
	UINT32		x,y,i,j,offset;
	UINT32		widthQual,widthOri;
	INT32		dx,dy,vx,vy,val;
	UINT32		oriStepSize;

	//init
	LOCK_MUTEX( g_paramsMutex);
		oriStepSize = g_params.m_oriStepSize;
	UNLOCK_MUTEX(g_paramsMutex);
	widthOri = p_width/ORI_BLOC_SIZE2;
	widthQual = p_width/FUSION_BLOC_SIZE;

	// estimation of local orientation by a least mean square method
	for( y=ORI_BLOC_SIZE/2; y<p_height-ORI_BLOC_SIZE/2-1; y+= ORI_BLOC_SIZE2 )
	{
		for( x=ORI_BLOC_SIZE/2; x<p_width-ORI_BLOC_SIZE/2-1; x+= ORI_BLOC_SIZE2 )
		{
			if( p_quality[(y/FUSION_BLOC_SIZE)*widthQual+(x/FUSION_BLOC_SIZE)] )
			{
				vx=0;
				vy=0;
				// least mean square evaluation of the orientation 
				// based on local gradient
				offset = (y-ORI_BLOC_SIZE/2)*p_width;
				for( j=y-ORI_BLOC_SIZE/2; j<=y+ORI_BLOC_SIZE/2; j+=oriStepSize )
				{
					for( i=x-ORI_BLOC_SIZE/2; i<=x+ORI_BLOC_SIZE/2; i+=oriStepSize )
					{	
						dx = p_gradient[(offset+i)<<1];
						dy = p_gradient[((offset+i)<<1)+1];
						vx += 2*dx*dy;
						vy += dx*dx-dy*dy;
					}
					offset+=p_width;
				}

				// convert to an angle between 0 and PI (0 and 255 )
				vx /= ORI_BLOC_SIZE*ORI_BLOC_SIZE;
				vy /= ORI_BLOC_SIZE*ORI_BLOC_SIZE;
					
				val = (zAtan(vy,vx)+PI2_256)/2; // may need to be optimized
				p_ori[y/ORI_BLOC_SIZE2*widthOri+x/ORI_BLOC_SIZE2] = val;
			}
			else
				p_ori[y/ORI_BLOC_SIZE2*widthOri+x/ORI_BLOC_SIZE2] = 0;
		}
	}

	return ZEF_OK;
}

INT32 scaleOriMap( UINT8 *p_src, UINT8 *p_dst, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 p_scale )
{
	INT32	error;
	INT8	*dx,*dy,*dxS,*dyS;
	INT32	val;
	UINT32	i;

	// init
	dx = NULL;
	dy = NULL;
	dxS = NULL;
	dyS = NULL;
	
	// memory allocation
	dx = (INT8*)MALLOC( p_width*p_height*sizeof(INT8) );
	if( dx == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	dy = (INT8*)MALLOC( p_width*p_height*sizeof(INT8) );
	if( dy == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	dxS = (INT8*)MALLOC( p_scale*p_scale*p_width*p_height*sizeof(INT8) );
	if( dxS == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	dyS = (INT8*)MALLOC( p_scale*p_scale*p_width*p_height*sizeof(INT8) );
	if( dyS == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}


	// process x and y vectors of the oriMap
	for( i=0; i<p_width*p_height; i++ )
	{
		if( p_qual[i] )
		{
			dx[i] = zCos127(p_src[i]);
			dy[i] = zSin127(p_src[i]);
		}
		else
		{
			dx[i] = 0;
			dy[i] = 0;
		}
	}

	// scale dx and dy
	error = integerScaleS( dx, dxS, p_width, p_height, p_scale );
	if( error < ZEF_OK )
		goto errorHandler;

	error = integerScaleS( dy, dyS, p_width, p_height, p_scale );
	if( error < ZEF_OK )
		goto errorHandler;

	// rebuild oriMap
	for( i=0; i<p_width*p_height*p_scale*p_scale; i++ )
	{
		val = (zAtan(dyS[i],dxS[i]))/2; // may need to be optimized
		p_dst[i] = val;
	}

	FREE(dx);
	FREE(dy);
	FREE(dxS);
	FREE(dyS);

	return ZEF_OK;

	// error Handler
errorHandler:

	if( dx != NULL )
		FREE(dx);
	if( dy != NULL )
		FREE(dy);
	if( dxS != NULL )
		FREE(dxS);
	if( dyS != NULL )
		FREE(dyS);

	return error;
}

INT32 findCoreOrientation( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template )
{
	INT32	x,y,theta;
	INT32	diffs,angle,cpt;
	INT32	error;
	UINT32	distFindCoreAngle;

	// init
	LOCK_MUTEX( g_paramsMutex );
		distFindCoreAngle = g_params.m_distFindCoreAngle;
	UNLOCK_MUTEX( g_paramsMutex );
	cpt = 0;
	angle = 0;

	// check if the core is not to close from the border
	if( p_template->m_cores[0].m_x >= (INT32)distFindCoreAngle && p_template->m_cores[0].m_x < (INT32)p_width-(INT32)distFindCoreAngle &&
		p_template->m_cores[0].m_y >= (INT32)distFindCoreAngle && p_template->m_cores[0].m_y < (INT32)p_height-(INT32)distFindCoreAngle )
	{

		// loop on ori to find core direction
		for( theta=0; theta<512; theta++ )
		{
			#if !defined(ZEF_FIXED_POINT)
				x = p_template->m_cores[0].m_x+(INT32)(distFindCoreAngle*zCos(PI*theta/256.0f));
				y = p_template->m_cores[0].m_y+(INT32)(distFindCoreAngle*zSin(PI*theta/256.0f));
			#else
				x = p_template->m_cores[0].m_x+((INT32)(distFindCoreAngle*zCos127(theta))>>7);
				y = p_template->m_cores[0].m_y+((INT32)(distFindCoreAngle*zSin127(theta))>>7);
			#endif


			diffs = zAbs(p_ori[y*p_width+x]-(theta&255));
			if( diffs > 128 ) diffs  = 256 - diffs;

			if( diffs < 8 ) 
			{
				angle += theta;
				cpt++;
			}

		}

	}
	else
	{
		// if we've got an error, it's that the core is too close from the border
		p_template->m_nbCores = 0; // core is not really reliable...
		p_template->m_oriCore = 0;
		error = ZEF_CORE_CLOSE_TO_BORDER;
		goto errorHandler;
	}

	if( cpt != 0 )
		angle /= cpt;

	p_template->m_oriCore = angle/2;

	return ZEF_OK;

	// error Handler
errorHandler:
	return error;

}

INT32 getVerticalAxis( UINT8 *p_ori, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 *p_symAxis )
{
	UINT32	x,y,x2,y2, width2,symAxis,val,cptProfile,maxProfile;
	INT32	*profile;

	// inits
	width2 = p_width/FUSION_BLOC_SIZE;
	symAxis = -1;
	maxProfile = 0;
	profile = (UINT32*)MALLOC( p_width * sizeof(UINT32));

	// loop over all vertical lines of the image
	for( x=0; x<p_width; x++ )
	{
		// and process the average value for each vertical line
		profile[x] = 0;
		cptProfile = 0;
		x2 = x/FUSION_BLOC_SIZE;
		for( y=0; y<p_height; y++ )
		{
			// checkl if the point has to be taken into account
			y2 = y/FUSION_BLOC_SIZE;
			if( p_qual[y2*width2+x2] )
			{
				// process mean value
				profile[x] += p_ori[y*p_width+x];
				cptProfile++;
			}
		}
		if( cptProfile != 0 )
		{
			profile[x] /= cptProfile;
		}
	}

	// now process the gradient of this 1D array of mean value
	// and remember the position of the point with maximum value.
	// its our axis of symmetry.
	for( x=0; x<p_width-1; x++ )
	{
		if( profile[x] != 0 && profile[x+1] != 0 )
		{
			val = zAbs(profile[x]-profile[x+1]);
			if( val > maxProfile )
			{
				maxProfile = val;
				symAxis = x;
			}
		}
	}

	*p_symAxis = symAxis;

	FREE(profile);

	return ZEF_OK;
}

INT32 getHorizontalAxis( UINT8 *p_ori, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 p_yAxis, UINT32 *p_symAxis, UINT32 p_filterSize )
{
	INT32	y,x1,x2,x1mini,x2mini,xAxis,i;
	UINT32	width2;
	UINT32	profile[1024],maxProfile;

	// inits
	xAxis = -1;
	maxProfile = 0;
	width2 = p_width/FUSION_BLOC_SIZE;

	x1 = p_yAxis - p_filterSize;
	x2 = p_yAxis + p_filterSize;

	if( x1 <= 0 ) x1 = 0;
	if( x2 >= (INT32)p_width-1 ) x2 = p_width-1;

	x1mini = x1 / FUSION_BLOC_SIZE;
	x2mini = x2 / FUSION_BLOC_SIZE;

	// loop over all horizontal lines of the image
	for( y=0; y<(INT32)p_height; y++ )
	{
		// checkl if the point has to be taken into account (is it masked or not ?)
		profile[y] = 0;
		if( p_qual[y/FUSION_BLOC_SIZE*width2+x1mini] != 0 && p_qual[y/FUSION_BLOC_SIZE*width2+x2mini] != 0 )
		{
			// for each pixel belonging to the zone going from p_yAxis-p_filterSize to p_yAxis+p_filterSize
			// taht is to say a zone of size 2xp_filterSize
			for( i=0; i<(INT32)p_filterSize; i++ )
			{
				// if left pixel has an orientation between 90 and 180 and
				// right pixel has an orientation between 0 and 90
				if( p_ori[y*p_width+x1+i] > 127 && p_ori[y*p_width+x2-i] <= 127 )
				{
					// add angles to see where the max curvature is
					profile[y] += 255-p_ori[y*p_width+x1+i] + p_ori[y*p_width+x2-i];
				}
			}
			// get max value
			if( profile[y] > maxProfile )
			{
				maxProfile = profile[y];
				xAxis = y;
			}
		}
	}

	*p_symAxis = xAxis;

	return ZEF_OK;
}
