/*
Company :		ZEFYR Technologies
File :			dynamicParameters.h
Author :		Matthieu Perreira Da Silva 
Fisrt release :	05/05/2003

Description : run time parameters for fingerprint processing

Features : defines some parameters for all the processing routines		
    
Functions : 

  	saveParameters : save the parameters contained in BioEngineParams structure
	loadParameters : load the parameters in BioEngineParams structure

Function prototypes : 

	INT32 saveParameters( UINT8 *p_buffer, UINT32 *p_size );
	INT32 loadParameters( UINT8 *p_buffer, UINT32 p_size );

Necessary package : 

	customTypes.h
	staticParameters.h

Comments :none

*/

#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <string.h> 

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "zMaths.h"

// global structures
BioEngineParams g_params;
MUTEX_OBJECT	g_paramsMutex;
BOOL8			g_initialized=FALSE;
// global data to abort identification
MUTEX_OBJECT	g_idCancelMutex;
UINT8			*g_idCancelPointer;

INT32 initParameters( void )
{
	LOCK_MUTEX( g_paramsMutex );
	// set default parameters values 
	g_params.m_maxNumberOfDetectedMinutiae  = MAX_NUMBER_OF_DETECTED_MINUTAE;
	g_params.m_maxNumberOfMinutiae			= MAX_NUMBER_OF_MINUTIAE;
	g_params.m_maxNumberOfLocalInfos		= MAX_NUMBER_OF_LOCAL_INFOS;
	g_params.m_maxNumberOfCores				= MAX_NUMBER_OF_CORES;
	g_params.m_normalizationMean			= NORMALISATION_MEAN;
	g_params.m_normalizationStdDev			= NORMALISATION_STD_DEV;
	g_params.m_qualityThreshold				= QUALITY_THRESHOLD;
	g_params.m_freqQualThreshold			= FREQ_QUAL_THRESHOLD;
	g_params.m_freqQualThresholdInc			= FREQ_QUAL_THRESHOLD_INC;
	g_params.m_spectrumMultFactor			= SPECTRUM_MULT_FACTOR;
	g_params.m_gaborWidth					= GABOR_WIDTH;
	g_params.m_gaborFreqMult				= GABOR_FREQ_MULT;
	g_params.m_gaborFusionThreshold			= GABOR_FUSION_THRESHOLD;
	g_params.m_nbClassesSegmentation		= NB_CLASSES_SEGMENTATION;
	g_params.m_nbClassesRejectSegmentation	= NB_CLASSES_REJECT_SEGMENTATION;
	g_params.m_segmMedianFilterSize			= SEGM_MEDIAN_FILTER_SIZE;
	g_params.m_regionDivCoef				= REGION_DIV_COEF;
	g_params.m_regionPerimeterCoef			= REGION_PERIMETER_COEF;
	g_params.m_regionSurfaceCoef			= REGION_SURFACR_COEF;
	g_params.m_oriFusionMedianSize			= ORI_FUSION_MEDIAN_SIZE;
	g_params.m_nbPtsFusionThreshold			= NB_PTS_FUSION_THRESHOLD;
	g_params.m_multFactorRegionFusion		= MULT_FACTOR_REGION_FUSION;
	g_params.m_fusionErodeSize				= FUSION_ERODE_SIZE;
	g_params.m_finalFusionMedianSize		= FINAL_FUSION_MEDIAN_SIZE;
	g_params.m_fusionIntensityMultFactor	= FUSION_INTENSITY_MULT_FACTOR;
	g_params.m_binaryMedianSize				= BINARY_MEDIAN_SIZE;
	g_params.m_minimumCurvatureThreshold	= MINIMUM_CURVATURE_THRESHOLD;
	g_params.m_localInfosBlocSize			= LOCAL_INFOS_BLOC_SIZE;
	g_params.m_minutiaeNeiborhoodSize		= MINUTIAE_NEIBORHOOD_SIZE;
	g_params.m_branchMinSize				= BRANCH_MIN_SIZE;
	g_params.m_maxReconnectSize				= MAX_RECONNECT_SIZE;
	g_params.m_minReconnectionAngle			= MIN_RECONNECTION_ANGLE;
	g_params.m_maxSupprAngle				= MAX_SUPPR_ANGLE;
	g_params.m_poincareSize					= POINCARE_SIZE;
	g_params.m_maxRegistrationAngle			= MAX_REGISTRATION_ANGLE;
	g_params.m_fisrtPassAngleStep			= FIRST_PASS_ANGLE_STEP;
	g_params.m_registerTranslationTolerance	= REGISTER_TRANSLATION_TOLERANCE;
	g_params.m_translationShift				= TRANSLATION_SHIFT;
	g_params.m_translation					= TRANSLATION;
	g_params.m_maxTranslationWidthShift1	= MAX_TRANSLATION_WIDTH_SHIFT_1;
	g_params.m_maxTranslationWidth1			= MAX_TRANSLATION_WIDTH_1;
	g_params.m_maxTranslationWidthShift2	= MAX_TRANSLATION_WIDTH_SHIFT_2;
	g_params.m_maxTranslationWidth2			= MAX_TRANSLATION_WIDTH_2;
	g_params.m_accuFill						= ACCU_FILL;
	g_params.m_translationMaxThreshold		= TRANSLATION_MAX_THRESHOLD;
	g_params.m_angleTolerance				= ANGLE_TOLERANCE;
	g_params.m_translationTolerance			= TRANSLATION_TOLERANCE;
	g_params.m_splineStiffness				= SPLINE_STIFFNESS;
	g_params.m_minNumberOfMinutiae			= MIN_NUMBER_OF_MINUTIAE;
	g_params.m_oriMatchNormalize			= ORI_MATCH_NORMALIZE;
	g_params.m_maxScore						= MAX_SCORE;
	g_params.m_scoreNormalize				= SCORE_NORMALIZE;
	g_params.m_powFftFlag					= POW_FFT_FLAG;
	g_params.m_maxDeformation				= MAX_DEFORMATION;
	g_params.m_nbFreqPos					= NB_FREQ_POS;
	g_params.m_distFindCoreAngle			= DIST_FIND_CORE_ANGLE;
	g_params.m_distDivCoef					= DIV_DIST_COEF;
	g_params.m_minutiaeBoost				= MINUTIAE_BOOST;
	g_params.m_defaultSecurity				= DEFAULT_SECURITY;
	g_params.m_maxDbSearch					= MAX_DB_SEARCH;
	g_params.m_max_img_width				= MAX_IMG_WIDTH;
	g_params.m_max_img_height				= MAX_IMG_HEIGHT;
	g_params.m_globalImgScaling				= GLOBAL_IMG_SCALING;
	g_params.m_globalTemplateScaling		= GLOBAL_TEMPLATE_SCALING;
	g_params.m_fftBlocOverlap				= FFT_BLOC_OVERLAP;
	g_params.m_numOfOrientation				= NUM_OF_ORIENTATION;
	g_params.m_oriStepSize					= ORI_STEP_SIZE;
	g_params.m_filterSelectionThreshold		= FILTER_SELECTION_THRESHOLD;

	// init freqPos
	g_params.m_freqPos[0] = FREQ_POS1;
	g_params.m_freqPos[1] = FREQ_POS2;
	g_params.m_freqPos[2] = FREQ_POS3;
	g_params.m_freqPos[3] = FREQ_POS4;
	g_params.m_freqPos[4] = FREQ_POS5;
	g_params.m_freqPos[5] = FREQ_POS6;
	g_params.m_freqPos[6] = FREQ_POS7;
	g_params.m_freqPos[7] = FREQ_POS8;
	g_params.m_freqPos[8] = FREQ_POS9;
	g_params.m_freqPos[9] = FREQ_POS10;

	UNLOCK_MUTEX( g_paramsMutex );

	return ZEF_OK;
}

INT32 saveParameters( UINT8 *p_buffer, UINT32 *p_size )
{
	UINT32	maxSize,i;
	UINT16	checksum;
	UINT32	offset;
	INT32	error;

		//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// test for NULL pointer in arguments
	if( (p_buffer == NULL && *p_size != 0) || p_size == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// init
	maxSize = sizeof(BioEngineParams)+10;
	checksum = 0;
	offset =0;

	// write size
	if( *p_size < maxSize )
	{
		*p_size = maxSize;
		return ZEF_BUFFER_TO_SMALL;
	}
	// size
	*((UINT32*)(p_buffer+offset)) = 0;
	offset += sizeof(UINT32);

	// version
	*((UINT32*)(p_buffer+offset)) = BIO_ENGINE_PARAM_STRUCT_VERSION;
	offset += sizeof(UINT32);

	LOCK_MUTEX( g_paramsMutex );
	memcpy(p_buffer+offset,&g_params,sizeof(BioEngineParams));
	UNLOCK_MUTEX( g_paramsMutex );
	offset += sizeof(BioEngineParams);

	// update size
	*((UINT32*)(p_buffer)) = offset+sizeof(UINT16);

	//checksum calculation
	for( i=0; i<offset; i++ )
		checksum += p_buffer[i];
	*((UINT16*)(p_buffer+offset)) = checksum;
	offset += sizeof(UINT16);

	return ZEF_OK;

	//error Handling
errorHandler:
	return error;
}

INT32 WINAPI dpLoadParameters( UINT8 *p_buffer, UINT32 p_size )
{
	UINT32				i;
	UINT16				checksum;
	UINT32				offset,size,version;	
	INT32				error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// test for NULL pointer in arguments
	if( p_buffer == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}
	
	// test for zero size
	if( p_size == 0 )
	{
		error = ZEF_BAD_SIZE;
		goto errorHandler;
	}

	// init
	offset=0;
	checksum=0;

	// read size
	size = *((UINT32*)(p_buffer+offset));
	offset += sizeof(UINT32);

	// version
	version = *((UINT32*)(p_buffer+offset));
	offset += sizeof(UINT32);

	// check params version
	if( version != BIO_ENGINE_PARAM_STRUCT_VERSION )
	{
		error = ZEF_BAD_VERSION;
		goto errorHandler;
	}

	// checksum checking
	for( i=0; i<size-sizeof(UINT16); i++ )
		checksum += p_buffer[i];
	if( checksum != *((UINT16*)(p_buffer+size-sizeof(UINT16))) )
	{
		error = ZEF_CHECKSUM_FAILED;
		goto errorHandler;
	}
	
	// if everything is ok we can copy the params to the structure
	LOCK_MUTEX( g_paramsMutex );
	memcpy(&g_params, p_buffer+offset, sizeof(BioEngineParams) ); 
	UNLOCK_MUTEX( g_paramsMutex );
	offset += sizeof(BioEngineParams);


	return ZEF_OK;

	//error Handling
errorHandler:
	return error;
}
