/*
Company :		ZEFYR Technologies
File :			enhancements.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : fingerprint enhancements

Features : differents fingerprint specific enhancements
    
Functions : 

	enhance : main enhancement routine
	fillBlocArray :	cuts the source image into overlapping blocs and store those blocs in a array
	fftBlocArray :	compute the FFT of all blocs in a bloc array
	polarSpectrum : compute the polar spectrum of each bloc of a bloc array
	freqSpectrum :	compute the main frequency of each spectrum bloc of an array of blocs
	powBlocFFT :	enhance some blocs of an array of bloc by multipl
	applyGaborFilter : filter each bloc with N oriented gabor filters tuned the the frequency processed in freqSpectrum function
	gaborFusion : use the N filterd image to find the best orientation map for the image
	findRegions : find regions of connected pixels with the same orientation
	fillRegions : replace regions that are to different from their neighboors
	findPixels :  find pixels that are too different from their neighboors
	imageFusion : reconstruct the final image based on the orientation map
	qualityThreshold : threshold the quality map to define a mask
	maskImage : removes used part of the image, defined by a mask.

Function prototypes : 

	INT32 enhance( BioEngineData *p_data );
	INT32 fillBlocArray( UINT8 *p_src, REAL32 **p_blocArray, UINT32 p_width, UINT32 p_height );
	INT32 fftBlocArray( REAL32 **p_blocArray, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY );
	INT32 polarSpectrum( REAL32 **p_blocArray, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32	**p_spectrum );
	INT32 freqSpectrum( REAL32 **p_spectrum, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_freq, REAL32 *p_qual, REAL32 *p_meanFreq );
	INT32 powBlocFFT( REAL32 **p_spectrum, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_qual );
	INT32 applyGaborFilter(  REAL32 **p_blocArray, INT8 *p_images, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_freq );
	INT32 gaborFusion( INT8 *p_images, UINT32 p_width, UINT32 p_height, UINT8 *p_quality, UINT8 *p_orient );
	INT32 findRegions( UINT8 *p_orient, UINT32 *p_regions, UINT32 p_width, UINT32 p_height, REAL32 p_threshold  );
	INT32 fillRegions( UINT8 *p_orient, UINT8 *p_quality, UINT32 *intensities, UINT32 *regions, UINT32 miniWidth, UINT32 miniHeight );
	INT32 findPixels( UINT8 *p_img, UINT32 *p_pixels, UINT32 p_width, UINT32 p_height, REAL32 p_threshold );
	INT32 imageFusion( INT8 *p_images, UINT8 *p_orient, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );
	INT32 qualityThreshold( UINT8 *p_quality, UINT32 p_width, UINT32 p_height, UINT32 p_threshold );
	INT32 maskImage( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT8 *p_mask );


Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	FFT2D.h
	zMaths.h
	filters.h
	greyLevelTrans.h
	geometry.h
	quicksort.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <string.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "FFT2D.h"
#include "zMaths.h"
#include "filters.h"
#include "greyLevelTrans.h"
#include "geometry.h"
#include "quicksort.h"
#include "BMPfiles.h"

#include "enhancements.h"

#ifndef ZEF_FIXED_POINT

INT32 initEnhance( void ) 
{
	INT32	j,i,k,x,y;
	REAL32	*temp;
	INT32	error;

	// init pointers
	temp = NULL;
	g_preCompData.m_butterworth = NULL;
	g_preCompData.m_fftInPlace = NULL;
	g_preCompData.m_fftTemp = NULL;
	g_preCompData.m_fftTrigo = NULL;

	// initialize bioEngine parameters
	error = initParameters();
	if( error != ZEF_OK )
		goto errorHandler;

	// allocated preCompDatas
	g_preCompData.m_fftTemp = (REAL32*)MALLOC( (8*FFT_BLOC_SIZE)*sizeof(REAL32) );
	if( g_preCompData.m_fftTemp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	g_preCompData.m_fftInPlace = (INT32*)MALLOC( (2+zSqrt(FFT_BLOC_SIZE))*sizeof(INT32) );
	if( g_preCompData.m_fftInPlace == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	g_preCompData.m_fftTrigo = (REAL32*)MALLOC( (3*FFT_BLOC_SIZE/4)*sizeof(REAL32) );
	if( g_preCompData.m_fftTrigo == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	g_preCompData.m_butterworth = (REAL32*)MALLOC( (FFT_BLOC_SIZE*(FFT_BLOC_SIZE/2))*sizeof(REAL32) );
	if( g_preCompData.m_butterworth == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// execute one FFT to preCompile trigo data
	temp = (REAL32*)MALLOC( (FFT_BLOC_SIZE*FFT_BLOC_SIZE)*sizeof(REAL32) );
	if( temp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	g_preCompData.m_fftInPlace[0] = 0; // to force calculation of preCompiled table
    rdft2d(FFT_BLOC_SIZE, FFT_BLOC_SIZE, 1, temp, g_preCompData.m_fftTemp , g_preCompData.m_fftInPlace, g_preCompData.m_fftTrigo);
	FREE(temp);
	temp = NULL;

	k=0;
	// precalc butterworth filter
	for( j=0; j<FFT_BLOC_SIZE; j++ )
	{
		for( i=0; i<FFT_BLOC_SIZE/2; i++ )
		{
			x = i;
			y = (FFT_BLOC_SIZE/2)-zAbs( j-(FFT_BLOC_SIZE/2) );
			
			g_preCompData.m_butterworth[k++] = 
				2.0f*(1-1.0f/(1.0f+(x*x+y*y)*(x*x+y*y)))/(FFT_BLOC_SIZE*FFT_BLOC_SIZE);
		}
	}

	return ZEF_OK;

	//error Handling
errorHandler:

	// FREE local pointers
	if( temp != NULL )
		FREE( temp );

	// FREE memory 
	unInitEnhance();

	return error;
}

INT32 unInitEnhance( void )
{
	// FREE memory
	if( g_preCompData.m_butterworth != NULL )
		FREE( g_preCompData.m_butterworth );
	if( g_preCompData.m_fftTemp != NULL )
		FREE( g_preCompData.m_fftTemp );
	if( g_preCompData.m_fftInPlace != NULL )
		FREE( g_preCompData.m_fftInPlace );
	if( g_preCompData.m_fftTrigo != NULL )
		FREE( g_preCompData.m_fftTrigo );

	g_preCompData.m_butterworth = NULL;
	g_preCompData.m_fftInPlace = NULL;
	g_preCompData.m_fftTemp = NULL;
	g_preCompData.m_fftTrigo = NULL;

	return ZEF_OK;
}


INT32 enhance( BioEngineData *p_data )
{
	INT32		error;
	UINT32		i;
	UINT32		nbBlocsX,nbBlocsY;
	REAL32		**blocArray;
	REAL32		**blocSpectrum;
	REAL32		*freq = NULL,*qual;
	REAL32		meanFreq;
	UINT8		*tmpOri;
	INT8		*images;
	UINT32		fftBlocOverlap,gaborFreqMult,globalImgScaling,powFftFlag;
	UINT32		numOfOrientation,binaryMedianSize,qualityThresholdVar;

	//init
	LOCK_MUTEX(g_paramsMutex);
		fftBlocOverlap = g_params.m_fftBlocOverlap;
		gaborFreqMult = g_params.m_gaborFreqMult;
		globalImgScaling = g_params.m_globalImgScaling;
		powFftFlag = g_params.m_powFftFlag;
		numOfOrientation = g_params.m_numOfOrientation;
		binaryMedianSize = g_params.m_binaryMedianSize;
		qualityThresholdVar = g_params.m_qualityThreshold;
	UNLOCK_MUTEX(g_paramsMutex);
	blocArray = NULL;
	blocSpectrum = NULL;
	freq = NULL;
	qual = NULL;
	tmpOri = NULL;
	images = NULL;

	//compute the number of blocs and allocate the array
	nbBlocsX = (p_data->m_width-fftBlocOverlap)/(FFT_BLOC_SIZE-fftBlocOverlap);
	nbBlocsY = (p_data->m_height-fftBlocOverlap)/(FFT_BLOC_SIZE-fftBlocOverlap);

	blocArray = (REAL32**)MALLOC( (nbBlocsX*nbBlocsY)*sizeof(REAL32*) );
	if( blocArray == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	for( i=0; i<nbBlocsX*nbBlocsY; i++ )
		blocArray[i] = NULL;
	for( i=0; i<nbBlocsX*nbBlocsY; i++ )
	{
		blocArray[i] = (REAL32*)MALLOC( (FFT_BLOC_SIZE*FFT_BLOC_SIZE)*sizeof(REAL32) );
		if( blocArray[i] == NULL )
		{
			error = ZEF_OUT_OF_MEMORY;
			goto errorHandler;
		}
	}

	blocSpectrum = (REAL32**)MALLOC( (nbBlocsX*nbBlocsY)*sizeof(REAL32*) );
	if( blocSpectrum == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	for( i=0; i<nbBlocsX*nbBlocsY; i++ )
		blocSpectrum[i] = NULL;
	for( i=0; i<nbBlocsX*nbBlocsY; i++ )
	{
		blocSpectrum[i] = (REAL32*)MALLOC( (SPECTRUM_RESOLUTION*SPECTRUM_RESOLUTION)*sizeof(REAL32) );
		if( blocSpectrum[i] == NULL )
		{
			error = ZEF_OUT_OF_MEMORY;
			goto errorHandler;
		}
	}

	// fill the bloc arrays with little parts of images
	error = fillBlocArray( p_data->m_workImg, blocArray, p_data->m_width, p_data->m_height );
	if( error < ZEF_OK )
		goto errorHandler;

	// compute real fft of each bloc
	error = fftBlocArray( blocArray, nbBlocsX, nbBlocsY );
	if( error < ZEF_OK )
		goto errorHandler;

	// and polar spectrum of each bloc
	error = polarSpectrum( blocArray, nbBlocsX, nbBlocsY, blocSpectrum );
	if( error < ZEF_OK )
		goto errorHandler;

	//	allocate memory for freq and qual
	freq = (REAL32*) MALLOC( (nbBlocsX*nbBlocsY)*sizeof(REAL32) );
	if( freq == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	qual = (REAL32*) MALLOC( (nbBlocsX*nbBlocsY)*sizeof(REAL32) );
	if( qual == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// estimate frequency for each bloc
	error = freqSpectrum( blocSpectrum, nbBlocsX, nbBlocsY, freq, qual, &meanFreq );
	if( error < ZEF_OK )
		goto errorHandler;

	// write mean freq to template struct
	if( meanFreq != 0 )
		p_data->m_template.m_meanFreq = (UINT16)Real2Int32(1000.0f/(meanFreq*TOFLT(gaborFreqMult,Q15_16)));
	else
		p_data->m_template.m_meanFreq = 0;

	p_data->m_template.m_meanFreq = (p_data->m_template.m_meanFreq*globalImgScaling)>>8;

	// FREE spectrum memory
	for( i=0; i<nbBlocsX*nbBlocsY; i++ )
		FREE( blocSpectrum[i] );
	FREE( blocSpectrum );
	blocSpectrum = NULL;

	// enhance some very noisy blocs
	if( powFftFlag )
	{
		error = powBlocFFT( blocArray, nbBlocsX, nbBlocsY, qual ); 
		if( error < ZEF_OK )
			goto errorHandler;
	}

	// FREE quality
	FREE( qual );
	qual = NULL;

	// allocate memory for filtered images
	images = (INT8*) MALLOC( (numOfOrientation*p_data->m_width*p_data->m_height)*sizeof(INT8) );
	if( images == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// enhance image with well tuned Gabor filters
	error = applyGaborFilter( blocArray, images, nbBlocsX, nbBlocsY, freq );
	if( error < ZEF_OK )
		goto errorHandler;

	// FREE bloc array memory
	for( i=0; i<nbBlocsX*nbBlocsY; i++ )
		FREE( blocArray[i] );
	FREE( blocArray );
	blocArray = NULL;

	// FREE freq
	FREE( freq );
	freq = NULL;

	// allocate memory for qual and ori
	p_data->m_quality = (UINT8*)MALLOC( ((p_data->m_width/FUSION_BLOC_SIZE)*(p_data->m_height/FUSION_BLOC_SIZE))*sizeof(UINT8) );
	if( p_data->m_quality == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	p_data->m_ori = (UINT8*)MALLOC( (p_data->m_width)*(p_data->m_height)*sizeof(UINT8) );
	if( p_data->m_ori == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	tmpOri = (UINT8*)MALLOC( ((p_data->m_width/FUSION_BLOC_SIZE)*(p_data->m_height/FUSION_BLOC_SIZE))*sizeof(UINT8) );
	if( tmpOri == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// process orientation map and quality
	error = gaborFusion( images, p_data->m_width, p_data->m_height, p_data->m_quality, tmpOri );
	if( error < ZEF_OK )
		goto errorHandler;

	// scale the orientation map
	error = scaleOri( tmpOri, p_data->m_ori, p_data->m_width/FUSION_BLOC_SIZE, p_data->m_height/FUSION_BLOC_SIZE, FUSION_BLOC_SIZE );
	if( error < ZEF_OK )
		goto errorHandler;
	
	// FREE tmpOri
	FREE(tmpOri);
	tmpOri = NULL;

	// generation of the fused image
	error = imageFusion( images, p_data->m_ori, p_data->m_workImg, p_data->m_width, p_data->m_height );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp( "Enhance-20fusion.bmp", p_data->m_workImg, p_data->m_width, p_data->m_height );
	#endif

	// process mask for later use
	error = qualityThreshold( p_data->m_quality, p_data->m_width/FUSION_BLOC_SIZE, p_data->m_height/FUSION_BLOC_SIZE, qualityThresholdVar );	
	if( error < ZEF_OK )
		goto errorHandler;
	
	#ifdef DP_DEBUG
		saveBmp( "Enhance-22qualThresholded.bmp", p_data->m_quality, p_data->m_width/FUSION_BLOC_SIZE, p_data->m_height/FUSION_BLOC_SIZE );
	#endif

	// convert to black and white image
	error = thresholdImage( p_data->m_workImg, p_data->m_workImg, p_data->m_width, p_data->m_height, 127 ); 
	if( error < ZEF_OK )
		goto errorHandler;

	// mask binary image
	error = maskImage( p_data->m_workImg, p_data->m_workImg, p_data->m_width, p_data->m_height, p_data->m_quality );
	if( error < ZEF_OK )
		goto errorHandler;

	// apply median filter to black and white image
	error = binaryMedianFilter( p_data->m_workImg, p_data->m_workImg, p_data->m_width, p_data->m_height, binaryMedianSize );  
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp( "Enhance-21binary.bmp", p_data->m_workImg, p_data->m_width, p_data->m_height );
	#endif

	// FREE memory
	FREE( images );
	images = NULL;

	return ZEF_OK;

	//error Handling
errorHandler:
	// FREE bloc array memory
	if( blocArray != NULL )
	{
		for( i=0; i<nbBlocsX*nbBlocsY; i++ )
		{
			if( blocArray[i] != NULL )
				FREE( blocArray[i] );
		}
		FREE( blocArray );
	}

	// FREE spectrum memory
	if( blocSpectrum != NULL )
	{
		for( i=0; i<nbBlocsX*nbBlocsY; i++ )
		{
			if( blocSpectrum[i] != NULL )
				FREE( blocSpectrum[i] );
		}
		FREE( blocSpectrum );
	}

	// FREE images
	if( images != NULL )
		FREE( images );

	// FREE data buffers
	if( p_data->m_quality != NULL )
	{
		FREE(p_data->m_quality);
		p_data->m_quality = NULL;
	}
	if( p_data->m_ori != NULL )
	{
		FREE(p_data->m_ori);
		p_data->m_ori = NULL;
	}

	// FREE tmpOri
	if( tmpOri != NULL ) 
		FREE(tmpOri);
	
	return error;
}

INT32 fillBlocArray( UINT8 *p_src, REAL32 **p_blocArray, UINT32 p_width, UINT32 p_height )
{
	UINT32		i,j,x,y;
	UINT32		nbBlocsX,nbBlocsY;
	UINT32		offset,offset2;
	UINT32		usefulBlocSize;
	UINT32		fftBlocOverlap;

	// inits
	LOCK_MUTEX( g_paramsMutex );
		fftBlocOverlap = g_params.m_fftBlocOverlap;
	UNLOCK_MUTEX( g_paramsMutex );

	//compute the number of blocs
	usefulBlocSize = FFT_BLOC_SIZE - fftBlocOverlap;
	nbBlocsX = (p_width-fftBlocOverlap)/(usefulBlocSize);
	nbBlocsY = (p_height-fftBlocOverlap)/(usefulBlocSize);

	// fill the array
	for( j=0; j<nbBlocsY; j++ )
	{
		for( i=0; i<nbBlocsX; i++ )
		{
			offset = 0;
			offset2 = (j*usefulBlocSize)*p_width+(i*usefulBlocSize);
			for( y=0; y<FFT_BLOC_SIZE; y++ )
			{
				for( x=0; x<FFT_BLOC_SIZE; x++ )
				{
					p_blocArray[j*nbBlocsX+i][offset++] = p_src[offset2++]; 
				}
				offset2+=p_width-FFT_BLOC_SIZE;
			}

		}
	}

	#ifdef DP_DEBUG
		saveBmpR( "Enhance-01sampleBloc.bmp",p_blocArray[nbBlocsY/2*nbBlocsX+nbBlocsX/2], FFT_BLOC_SIZE, FFT_BLOC_SIZE );
	#endif

	return ZEF_OK;

}

INT32 fftBlocArray( REAL32 **p_blocArray, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY )
{
	UINT32		i,j,k,l;

	// for all blocs :
	k=0;
	for( j=0; j<p_nbBlocsY; j++ )
	{
		for( i=0; i<p_nbBlocsX; i++ )
		{
			// processes FFT of a bloc
			rdft2d( FFT_BLOC_SIZE, FFT_BLOC_SIZE, 1, p_blocArray[k], g_preCompData.m_fftTemp, 
				g_preCompData.m_fftInPlace, g_preCompData.m_fftTrigo );

			// apply butterworth filter
			for( l=0; l<FFT_BLOC_SIZE*FFT_BLOC_SIZE; l++ )
				p_blocArray[k][l] *= g_preCompData.m_butterworth[l>1];

			k++;
		}
	}

	#ifdef DP_DEBUG
		saveBmpC( "Enhance-02sampleBlocFFT.bmp", p_blocArray[p_nbBlocsY/2*p_nbBlocsX+p_nbBlocsX/2], FFT_BLOC_SIZE/2, FFT_BLOC_SIZE );
	#endif

	return ZEF_OK;
}

INT32 polarSpectrum( REAL32 **p_blocArray, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32	**p_spectrum )
{
	UINT32	k,y,x;
	INT32	i,j,rho,theta;
	REAL32	re,im;
	UINT32	acc[SPECTRUM_RESOLUTION*SPECTRUM_RESOLUTION];


	// init
	for( j=0; j<SPECTRUM_RESOLUTION; j++ )
	{
		for( i=0; i<SPECTRUM_RESOLUTION; i++ )
		{
			acc[j*SPECTRUM_RESOLUTION+i] = 0;
		}
	}

	// for each bloc
	k=0;
	for( y=0; y<p_nbBlocsY; y++ )
	{
		for( x=0; x<p_nbBlocsX; x++ )
		{
			for( j=0; j<SPECTRUM_RESOLUTION; j++ )
			{
				for( i=0; i<SPECTRUM_RESOLUTION; i++ )
				{
					p_spectrum[k][j*SPECTRUM_RESOLUTION+i] = 0;
				}
			}
			// fill the sprectrum in a polar coordinate array
			for( j = 0; j < FFT_BLOC_SIZE; j++ )
			{
				for( i = 0; i < FFT_BLOC_SIZE/2; i++ )
				{
					rho = zSqrtSpec(j,i);   //zSqrt(i*i+(j-32)*(j-32));
					if( rho < FFT_BLOC_SIZE/2 )
					{
						theta = zAtanSpec(j,i);  //zAtan(j-32,i)*SPECTRUM_RESOLUTION/PI_256;
						if( theta < 0 ) theta += SPECTRUM_RESOLUTION;
						re = p_blocArray[k][((j+FFT_BLOC_SIZE/2)&(FFT_BLOC_SIZE-1))*FFT_BLOC_SIZE+i*2];
						im = p_blocArray[k][((j+FFT_BLOC_SIZE/2)&(FFT_BLOC_SIZE-1))*FFT_BLOC_SIZE+i*2+1];	
						p_spectrum[k][rho*SPECTRUM_RESOLUTION+theta] += re*re+im*im;
						acc[rho*SPECTRUM_RESOLUTION+theta]++;
					}
				}
			}
			for( j=0; j<SPECTRUM_RESOLUTION; j++ )
			{
				for( i=0; i<SPECTRUM_RESOLUTION; i++ )
				{
					if( acc[j*SPECTRUM_RESOLUTION+i] != 0 )
					{
						p_spectrum[k][j*SPECTRUM_RESOLUTION+i] /= acc[j*SPECTRUM_RESOLUTION+i];
						acc[j*SPECTRUM_RESOLUTION+i] = 0;
					}
				}
			}
			k++;
		}
	}
	
	#ifdef DP_DEBUG
		saveBmpR( "Enhance-03sampleBlocSpectrum.bmp", p_spectrum[p_nbBlocsY/2*p_nbBlocsX+p_nbBlocsX/2], SPECTRUM_RESOLUTION, SPECTRUM_RESOLUTION );
	#endif

	return ZEF_OK;

}

INT32 freqSpectrum( REAL32 **p_spectrum, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_freq, REAL32 *p_qual, REAL32 *p_meanFreq )
{
	UINT32		i,j,k,l,acc,index,loop,offset;
	REAL32		threshold,meanFreq,blocMeanFreq,meanQual;
	REAL32		accu[MAX_NB_OF_FREQ-1],accR;
	UINT32		coefs[MAX_NB_OF_FREQ-1];
	REAL32		val,meanOri,stdDevOri,max;
	REAL32		bandMean;
	BOOL8		changed;
	INT32		x,y;
	UINT32		nbFreqPos,freqPos[MAX_NB_OF_FREQ];
	INT32		freqQualThreshold,freqQualThresholdInc;

	// fast lookup table final filtering
	INT8		dirX[8] = {  0,  1, -1,  1, -1,  1, -1,  0 };
	INT8		dirY[8] = { -1, -1, -1,  0,  0,  1,  1,  1 };

	// init values
	LOCK_MUTEX( g_paramsMutex );
		nbFreqPos = g_params.m_nbFreqPos;
		for( i=0; i<MAX_NB_OF_FREQ; i++)
			freqPos[i] = g_params.m_freqPos[i];
		freqQualThreshold = g_params.m_freqQualThreshold;
		freqQualThresholdInc = g_params.m_freqQualThresholdInc;
	UNLOCK_MUTEX( g_paramsMutex );
	blocMeanFreq = 0.0f;
	acc = 0;
	for(i=0; i<nbFreqPos-1; i++ )
		coefs[i] = 1+freqPos[i+1]-freqPos[i];

	// evaluate frequency for each bloc
	l=0;
	for( y=0; y<(INT32)p_nbBlocsY; y++ )
	{
		for( x=0; x<(INT32)p_nbBlocsX; x++ )
		{
			// init accu
			for(i=0; i<nbFreqPos-1; i++ )
				accu[i] = 0.0f;

			// find the band with the highest energy
			for( k=0; k<nbFreqPos-1; k++ )
			{
				meanOri=0.0f;
				stdDevOri=1.0f;
				// compute mean Orientation 
				for( j=freqPos[k]; j<=freqPos[k+1]; j++)
				{
					for( i=0; i<SPECTRUM_RESOLUTION; i++ )
					{
						val = p_spectrum[l][j*SPECTRUM_RESOLUTION+i];
						accu[k] += val;
						meanOri += val*i;
					}
				}
				if( accu[k] != 0 )
					meanOri /= accu[k];
				accu[k] /= coefs[k];
				
				// compute stdDev of frequencies around  meanOri 
				acc = 0;
				for( j=freqPos[k]; j<=freqPos[k+1]; j++)
				{
					for( i=0; i<SPECTRUM_RESOLUTION; i++ )
					{
						val = zfAbs(meanOri-i);;
						if( val > (REAL32)(SPECTRUM_RESOLUTION/2) ) val = (REAL32)SPECTRUM_RESOLUTION-val;
						stdDevOri += val*val*p_spectrum[l][j*SPECTRUM_RESOLUTION+i];
						acc++;
					}
				}
				if( acc != 0 )
					stdDevOri = zfSqrt( stdDevOri/acc );


				// divide energy by stdDev of freq arround mean Ori
				if( stdDevOri != 0 )
					accu[k] /= stdDevOri;
				/*else
					accu[k] = 0;*/
			}

			// find band with the highest energy
			max = -1.0;
			index = 0;
			for(i=0; i<nbFreqPos-1; i++ )
			{
				if( accu[i] > max )
				{
					max = accu[i];
					index = i;
				}
			}

			// calculate the mean energy value of the other bands
			bandMean = -max;
			for(i=0; i<nbFreqPos-1; i++ )
				bandMean += accu[i];

			bandMean /= (REAL32)(nbFreqPos-2);

			// then process the mean freq value of the band with max energy
			accR = 0.0;
			meanFreq = 0.0;
			for( j=freqPos[index]; j<=freqPos[index+1]; j++ )
			{
				for( i=0; i<SPECTRUM_RESOLUTION; i++ )
				{
					val = p_spectrum[l][j*SPECTRUM_RESOLUTION+i];
					accR += val;
					meanFreq += j*val; 
				}
			}

			if( accR != 0 )
				meanFreq /= accR;

			// compute final values
			if( meanFreq != 0)
				p_freq[l] = (FFT_BLOC_SIZE)/meanFreq;
			else
				p_freq[l] = 0;

			if( bandMean != 0 )
				p_qual[l] = max/bandMean;
			else
				p_qual[l] = 0;

			l++;
		}
	}

	// choose only good quality frequencies 
	// bad quality ones will be interpolated later...
	val = TOFLT(freqQualThreshold,Q15_16);
	loop=0;
	
	do
	{
		blocMeanFreq = 0.0;
		meanQual = 0.0;
		acc = 0;	
		threshold = 1/val;
		for( y=1; y<(INT32)p_nbBlocsY-1; y++ )
		{
			for( x=1; x<(INT32)p_nbBlocsX-1; x++ )
			{
				offset = y*p_nbBlocsX+x;
				p_freq[offset] = zfAbs(p_freq[offset]);
				p_qual[offset] = zfAbs(p_qual[offset]);
				if( p_qual[offset] > threshold )
				{
					blocMeanFreq += p_freq[offset]*p_qual[offset];
					meanQual += p_qual[offset];
					acc++;
				}
				else
				{
					p_freq[offset] = -p_freq[offset];
					p_qual[offset] = -p_qual[offset];
				}	
			}
		}
		loop++;
		val += TOFLT(freqQualThresholdInc,Q15_16);
	}
 	while( acc < MIN_NB_BLOCS_FREQ && loop < MAX_LOOPS );

	// set borders to zero...
	for( i=0; i<p_nbBlocsX; i++ )
	{
		p_freq[i] = -zfAbs(p_freq[i]);
		p_qual[i] = -zfAbs(p_qual[i]);

		p_freq[p_nbBlocsX*p_nbBlocsY-1-i] = -zfAbs(p_freq[p_nbBlocsX*p_nbBlocsY-1-i]);
		p_qual[p_nbBlocsX*p_nbBlocsY-1-i] = -zfAbs(p_qual[p_nbBlocsX*p_nbBlocsY-1-i]);
	}

	for( i=0; i<p_nbBlocsY; i++ )
	{
		p_freq[i*p_nbBlocsX] = -zfAbs(p_freq[i*p_nbBlocsX]);
		p_qual[i*p_nbBlocsX] = -zfAbs(p_qual[i*p_nbBlocsX]);

		p_freq[i*p_nbBlocsX+p_nbBlocsX-1] = -zfAbs(p_freq[i*p_nbBlocsX+p_nbBlocsX-1]);
		p_qual[i*p_nbBlocsX+p_nbBlocsX-1] = -zfAbs(p_qual[i*p_nbBlocsX+p_nbBlocsX-1]);
	}

	// ajust meanValues;
	if( meanQual != 0 )
		blocMeanFreq /= meanQual;

	// now we interpolate the missing freq values
	// we loop until everything is filled
	loop=0;
	do
	{
		changed = FALSE;
		for( y=0; y<(INT32)p_nbBlocsY; y++ )
		{
			for( x=0; x<(INT32)p_nbBlocsX; x++ )
			{
				if( p_freq[y*p_nbBlocsX+x] <= 0 )
				{
					meanFreq = 0.0f;
					accR = 0.0f;
					acc = 0;
					for( i=0; i<8; i++ )
					{
						if( (UINT32)(y+dirY[i]) < p_nbBlocsY  && (UINT32)(x+dirX[i]) < p_nbBlocsX )
						{
							offset = (y+dirY[i])*p_nbBlocsX+(x+dirX[i]);
							if( p_freq[offset] > 0 )
							{
								val = zfAbs(p_qual[offset]);
								meanFreq += p_freq[offset]*val;
								accR += val;
								acc++;
							}
						}
					}
					if( acc > 2 )
					{
						changed = TRUE;
						if( accR != 0.0 )
							p_freq[y*p_nbBlocsX+x] = meanFreq/accR;
						p_qual[y*p_nbBlocsX+x] = -accR / acc;
					}
				}
			}
		}
		loop++;

	}
	while( changed && loop < MAX_LOOPS );


	// one last pass
	l=0;
	for( y=0; y<(INT32)p_nbBlocsY; y++ )
	{
		for( x=0; x<(INT32)p_nbBlocsX; x++ )
		{
			if( p_freq[l] <= 0 )
				p_freq[l] = blocMeanFreq;
			l++;
		}
	}

	#ifdef DP_DEBUG
		saveBmpR( "Enhance-04freqMap.bmp", p_freq, p_nbBlocsX,p_nbBlocsY );
		saveBmpR( "Enhance-05qualSpectrum.bmp", p_qual, p_nbBlocsX,p_nbBlocsY );
	#endif

	*p_meanFreq = blocMeanFreq;
	
	return ZEF_OK;
}

INT32 powBlocFFT( REAL32 **p_blocArray, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_qual )
{
	UINT32		x,y,i,k;
	REAL32		val,max,re,im;
	//INT32		spectrumMultFactor;

	// Inits
	//LOCK_MUTEX(g_paramsMutex);
	//	spectrumMultFactor = g_params.m_spectrumMultFactor;
	//UNLOCK_MUTEX(g_paramsMutex);

	// search for blocs with p_qual[k] != 0
	k=0;
	for( y=0 ;y<p_nbBlocsY; y++ )
	{	
		for( x=0; x<p_nbBlocsX; x++ )
		{
			// if p_qual != 0 then we have to process the bloc
			if( p_qual[k] < 0)
			{
				max = FLT_MIN;
				// raise FFT to power g_params.m_spectrumMultFactor (0.25 by default)
				for( i=2; i<FFT_BLOC_SIZE*FFT_BLOC_SIZE; i+=2 )
				{
					re = p_blocArray[k][i];
					im = p_blocArray[k][i+1];
					// slower but more flexible
					//val = (REAL32)pow(re*re+im*im,TOFLT(g_params.m_spectrumMultFactor));
					// faster for pow(x;0.25) ... but less flexible
					val = (REAL32)zfSqrt(zfSqrt(re*re+im*im) ) ;
					p_blocArray[k][i] *= val;
					p_blocArray[k][i+1] *= val;
					if( val > max ) max = val;
				}
				// divide by max to have something smooth 
				// when we compare to non processed blocs
				if( max != 0 )
					max = 1.0f/max;
			
				for( i=2; i<FFT_BLOC_SIZE*FFT_BLOC_SIZE; i+=2 )
				{
					p_blocArray[k][i] *= max;
					p_blocArray[k][i+1] *= max;
				}
			}
			k++;
		}
	}

	return ZEF_OK;
}

INT32 applyGaborFilter(  REAL32 **p_blocArray, INT8 *p_images, UINT32 p_nbBlocsX, UINT32 p_nbBlocsY, REAL32 *p_freq )
{
	INT32	error;
	UINT32	x,y,k,newWidth,offset,xU,yU,max,min,usefulBlocSize,a;
	REAL32	*filter,**temp,*weights,*sum;
	REAL32	angle,size,freq,div,sumMax;
	INT32	val,x1,y1,i,j;
	UINT32	fftBlocOverlap,normalizationStdDev,numOfOrientation,filterSelectionThreshold;
	INT32	gaborWidth,gaborFreqMult;

	// init
	LOCK_MUTEX( g_paramsMutex );
		fftBlocOverlap = g_params.m_fftBlocOverlap;
		normalizationStdDev = g_params.m_normalizationStdDev;
		numOfOrientation = g_params.m_numOfOrientation;
		gaborWidth = g_params.m_gaborWidth;
		gaborFreqMult = g_params.m_gaborFreqMult;
		filterSelectionThreshold = g_params.m_filterSelectionThreshold;
	UNLOCK_MUTEX( g_paramsMutex );

	filter = NULL;
	temp = NULL;
	weights = NULL;
	sum = NULL;
	usefulBlocSize = FFT_BLOC_SIZE-fftBlocOverlap;
	newWidth = p_nbBlocsX*usefulBlocSize+fftBlocOverlap;
	div = TOFLT(GABOR_IMAGE_INTENSITY_SCALE,Q15_16)/(fftBlocOverlap*fftBlocOverlap*FFT_BLOC_SIZE*FFT_BLOC_SIZE*normalizationStdDev/4);

	// allocations
	filter = (REAL32*)MALLOC( (FFT_BLOC_SIZE*FFT_BLOC_SIZE/2)*sizeof(REAL32) );
	if( filter == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
	}

	temp = (REAL32**)MALLOC( numOfOrientation*sizeof(REAL32*) );
	if( temp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
	}

	for( j=0; j<(INT32)numOfOrientation; j++ )
	{
		temp[j] = (REAL32*)MALLOC( (FFT_BLOC_SIZE*FFT_BLOC_SIZE)*sizeof(REAL32) );
		if( temp[j] == NULL )
		{
			error = ZEF_OUT_OF_MEMORY;
		}
	}

	weights = (REAL32*)MALLOC( (FFT_BLOC_SIZE*FFT_BLOC_SIZE)*sizeof(REAL32) );
	if( weights == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
	}

	sum = (REAL32*)MALLOC( numOfOrientation*sizeof(REAL32) );
	if( sum == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
	}

	// init of weights
	k=0;
	for( j=0; j<FFT_BLOC_SIZE; j++ )
	{
		for( i=0; i<FFT_BLOC_SIZE; i++ )
		{
			x = 1+zAbs((i&MOD_FFT_BLOC_SIZE2)-((FFT_BLOC_SIZE/2-1)*(i/(FFT_BLOC_SIZE/2))));
			y = 1+zAbs((j&MOD_FFT_BLOC_SIZE2)-((FFT_BLOC_SIZE/2-1)*(j/(FFT_BLOC_SIZE/2))));
			
			if( x>y )
			{	max = x; min = y; }
			else 
			{	max = y; min = x; }

			//weights[k++] = div*x*y/(REAL32)(FFT_BLOC_SIZE*FFT_BLOC_SIZE/4.0f);
			if( max<fftBlocOverlap )
			{
				weights[k++] = x*y*div;
			}
			else if( min>=fftBlocOverlap)
			{
				weights[k++] = fftBlocOverlap*fftBlocOverlap*div;
			}
			else
			{
				weights[k++] = fftBlocOverlap*min*div;
			}
		}
	}

	// init of images
	memset(p_images,0,sizeof(INT8)*numOfOrientation*(p_nbBlocsX*usefulBlocSize+fftBlocOverlap)*(p_nbBlocsY*usefulBlocSize+fftBlocOverlap)); 

	k=0;
	for(y=0; y<p_nbBlocsY; y++)
	{
		yU = y*usefulBlocSize*numOfOrientation;
		for(x=0; x<p_nbBlocsX; x++ )
		{
			if( p_freq[k] != 0 )
			{
				xU = x*usefulBlocSize*numOfOrientation;
				// set some Gabor filter parameters
				size = TOFLT(gaborWidth,Q15_16)*p_freq[k];	
				freq = 1.0f/(p_freq[k]*TOFLT(gaborFreqMult,Q15_16));

				sumMax = 0.0;
				// for all orientations
				for( i=0; i<(INT32)numOfOrientation; i++ )
				{
					// set some of the gabor parameters
					angle = i*PI/numOfOrientation;

					// generate the appropriate Gabor filter
					error = generateGaborFFT( filter, size, freq, angle );
					if( error < ZEF_OK )
						goto errorHandler;
					// convolution in freq space
					error = freqConvolution( p_blocArray[k], temp[i], filter );
					if( error < ZEF_OK )
						goto errorHandler;

					// process "max energy" for all filters
					sum[i] = 0;
					if( filterSelectionThreshold != 0 )
					{
						for( a=0; a<FFT_BLOC_SIZE*FFT_BLOC_SIZE; a++ )
						{
							sum[i] += zfAbs(temp[i][a]);
						}
						if( sum[i] > sumMax )
							sumMax = sum[i];
					}
				}

				// for all orientations
				for( i=0; i<(INT32)numOfOrientation; i++ )
				{
					// only do inverse FFT for usefull filters
					if( filterSelectionThreshold == 0 || 
						sum[i] >= filterSelectionThreshold*sumMax/256 )
					{
						// back from time/freq representation
						rdft2d( FFT_BLOC_SIZE, FFT_BLOC_SIZE, -1, temp[i], g_preCompData.m_fftTemp, 
						g_preCompData.m_fftInPlace, g_preCompData.m_fftTrigo );

						// write images in the images array
						j=FFT_BLOC_SIZE*FFT_BLOC_SIZE-1;
						for( y1=(FFT_BLOC_SIZE-1)*numOfOrientation; y1>=0; y1-=numOfOrientation )
						{
							offset = i+xU+(yU+y1)*newWidth;
							for( x1=(FFT_BLOC_SIZE-1)*numOfOrientation; x1>=0; x1-=numOfOrientation )
							{
								val = Real2Int32(temp[i][j]*weights[j]);
								p_images[offset+x1] += (INT8)val;
								j--;
							}
						}
					}
				}
			}
			k++;
		}
	}

	#ifdef DP_DEBUG
		saveBmpS( "Enhance-07filtered01.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,0 );
		saveBmpS( "Enhance-07filtered02.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,1 );
		saveBmpS( "Enhance-07filtered03.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,2 );
		saveBmpS( "Enhance-07filtered04.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,3 );
		saveBmpS( "Enhance-07filtered05.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,4 );
		saveBmpS( "Enhance-07filtered06.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,5 );
		saveBmpS( "Enhance-07filtered07.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,6 );
		saveBmpS( "Enhance-07filtered08.bmp",p_images, newWidth, p_nbBlocsY*usefulBlocSize+fftBlocOverlap,numOfOrientation,7 );
	#endif

	FREE(weights);
	FREE(filter);
	for( j=0; j<(INT32)numOfOrientation; j++ )
		FREE(temp[j]);
	FREE(temp);
	FREE(sum);

	return ZEF_OK;

	//error Handling
errorHandler:

	if( sum != NULL )
		FREE( sum );

	if( weights != NULL )
		FREE(weights);

	if( filter != NULL )
		FREE(filter);

	if( temp != NULL )
		FREE(temp);

	return error;
}

INT32 gaborFusion( INT8 *p_images, UINT32 p_width, UINT32 p_height, UINT8 *p_quality, UINT8 *p_orient )
{
	INT32		error;
	UINT32		i,j,k,l,m;
	UINT32		*intensities,max;
	UINT32		*regions,*pixels;
	UINT32		miniWidth,miniHeight,miniWidth2;
	UINT32		offset1,offset2;
	UINT32		numOfOrientation,oriFusionMedianSize,fusionErodeSize,finalFusionMedianSize;
	INT32		gaborFusionThreshold,multFactorRegionFusion,fusionIntensityMultFactor;

	// init
	LOCK_MUTEX( g_paramsMutex );
		numOfOrientation = g_params.m_numOfOrientation;
		oriFusionMedianSize = g_params.m_oriFusionMedianSize;
		gaborFusionThreshold = g_params.m_gaborFusionThreshold;
		multFactorRegionFusion = g_params.m_multFactorRegionFusion;
		fusionErodeSize = g_params.m_fusionErodeSize;
		fusionIntensityMultFactor = g_params.m_fusionIntensityMultFactor;
		finalFusionMedianSize = g_params.m_finalFusionMedianSize;
	UNLOCK_MUTEX( g_paramsMutex );
	intensities = NULL;
	regions = NULL;
	pixels = NULL;
	miniWidth  = p_width/FUSION_BLOC_SIZE;
	miniHeight = p_height/FUSION_BLOC_SIZE;

	miniWidth2 = miniWidth*numOfOrientation;	

	// allocations
	intensities = (UINT32*)MALLOC(numOfOrientation*miniWidth*miniHeight*sizeof(UINT32));
	if( intensities == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	regions = (UINT32*)MALLOC(miniWidth*miniHeight*sizeof(UINT32)); 
	if( regions == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	pixels = (UINT32*)MALLOC(miniWidth*miniHeight*sizeof(UINT32)); 
	if( pixels == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	// init intensities
	for( i=0; i<miniWidth2*miniHeight; i++ )
		intensities[i] = 0;

	// fill the intensities array with local mean intensity
	l=0;
	for( j=0; j<p_height; j++ )
	{
		offset1 = (j/FUSION_BLOC_SIZE)*miniWidth2;
		for( i=0; i<p_width; i++ )
		{
			offset2 = offset1+i/FUSION_BLOC_SIZE*numOfOrientation;
			for( k=0; k<numOfOrientation; k++ )
			{
				intensities[offset2+k] += zAbs(p_images[l]);
				l++;
			}
		}
	}

	// find max
	max=0;
	for( i=0; i<miniWidth2*miniHeight; i++ )
		if( intensities[i] > max )
			max = intensities[i];

	if( max == 0 ) max = 1;

	for( i=0; i<miniWidth2*miniHeight; i++ )
		intensities[i] = 255*intensities[i]/max;

	// Set a first estimation of the orientation map using
	//  the orientations that have the highest intensity
	l=0;
	m=0;
	for( j=0; j<miniHeight; j++ )
	{
		for( i=0; i<miniWidth; i++ )
		{
			p_quality[m] = 0;
			for( k=0; k<numOfOrientation; k++ )
			{
				if( intensities[l] >= p_quality[m] ) 
				{
					p_orient[m] = k;
					p_quality[m] = intensities[l];
				}
				l++;
			}
			m++;
		}
	}

	#ifdef DP_DEBUG
		saveBmp("Enhance-08FirstQual.bmp", p_quality, miniWidth, miniHeight );
		saveBmp("Enhance-09FirstOriMap.bmp", p_orient,  miniWidth, miniHeight ); 
	#endif

	// then apply median filter on the orientation map
	error = medianFilter( p_orient, p_orient, miniWidth, miniHeight, oriFusionMedianSize );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp("Enhance-10FirstOriMedian.bmp", p_orient,  miniWidth, miniHeight ); 
	#endif

	// find regions which have a different orientation from their neighbors
	error = findRegions( p_orient, regions, miniWidth, miniHeight, TOFLT(gaborFusionThreshold,Q15_16) );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp32("Enhance-11Regions1.bmp",regions, miniWidth, miniHeight );
	#endif

	// replace orientation and quality in regions that are to much differents from the others
	error = fillRegions( p_orient, p_quality, intensities, regions, miniWidth, miniHeight );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp("Enhance-12SecondQual.bmp", p_quality, miniWidth, miniHeight );
		saveBmp("Enhance-13SecondOriMap.bmp", p_orient,  miniWidth, miniHeight ); 
	#endif

	// find pixels which have orientations too different from their neighborhood
	error = findPixels( p_orient, pixels, miniWidth, miniHeight, 
					TOFLT(gaborFusionThreshold,Q15_16)*TOFLT(multFactorRegionFusion,Q15_16) );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp32("Enhance-14Pixels.bmp",pixels, miniWidth, miniHeight );
	#endif

	// apply erosion on this map
	error = erosionFilter32( pixels, pixels, miniWidth, miniHeight, fusionErodeSize );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp32("Enhance-15ErodedPixels.bmp",pixels, miniWidth, miniHeight );
	#endif

	// and replace pixels
	error = fillRegions( p_orient, p_quality, intensities, pixels, miniWidth, miniHeight );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp("Enhance-16ThirdQual.bmp", p_quality, miniWidth, miniHeight );
		saveBmp("Enhance-17FinalOriMap.bmp", p_orient,  miniWidth, miniHeight ); 
	#endif

	// update quality
	for( i=0; i<miniWidth*miniHeight; i++ )
	{
		if( regions[i] == 127 )
			p_quality[i] = zReal2Int32(p_quality[i]*TOFLT(fusionIntensityMultFactor,Q15_16));
	}

	//one last search for regions
	error = findRegions( p_orient, regions, miniWidth, miniHeight, 
					TOFLT(gaborFusionThreshold,Q15_16)*TOFLT(multFactorRegionFusion,Q15_16) );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp32("Enhance-18Regions2.bmp",regions, miniWidth, miniHeight );
	#endif

	// adapt quality according to regions found
	for( i=0; i<miniWidth*miniHeight; i++ )
	{
		if( regions[i] == 0 )
			p_quality[i] = 0;
	}

	// median filter on quality
	error = medianFilter( p_quality, p_quality, miniWidth, miniHeight, finalFusionMedianSize );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp("Enhance-19FinalQual.bmp", p_quality, miniWidth, miniHeight );
	#endif

	// FREE memory
	FREE( intensities );
	FREE( regions );
	FREE( pixels );

	return ZEF_OK;

	//errorHandling
errorHandler:

	if( intensities != NULL )
		FREE( intensities );
	if( regions != NULL )
		FREE( regions );
	if( pixels != NULL )
		FREE( pixels );

	return error;
}

INT32 findRegions( UINT8 *p_img, UINT32 *p_regions, UINT32 p_width, UINT32 p_height, REAL32 p_threshold )
{
	UINT32		x,y,xx,yy,i,acc;
	UINT32		val,oldVal,*size,*angleVar;
	UINT32		offset1,offset2,offset3;
	BOOL8		match;
	INT32		error;
	INT32		regionDivCoef;
	UINT32		regionPerimeterCoef,regionSurfaceCoef,numOfOrientation;
	
	// fast lookup table region  searching
	INT8		dirX[8] = {  0,  1, -1,  1, -1,  1, -1,  0 };
	INT8		dirY[8] = { -1, -1, -1,  0,  0,  1,  1,  1 };

	// init
	LOCK_MUTEX(g_paramsMutex);
		regionDivCoef = g_params.m_regionDivCoef;
		regionPerimeterCoef = g_params.m_regionPerimeterCoef;
		regionSurfaceCoef = g_params.m_regionSurfaceCoef;
		numOfOrientation = g_params.m_numOfOrientation;
	UNLOCK_MUTEX(g_paramsMutex);
	size = NULL;
	angleVar = NULL;
	acc = 1;
	p_threshold /= TOFLT(regionDivCoef,Q15_16);

	// init regions
	for( i=0; i<p_width*p_height; i++ )
		p_regions[i] = 0;

	// search for connected pixels -> regions
	for( y=1; y<p_height-1; y++ )
	{
		for( x=1; x<p_width-1; x++ )
		{
			match = FALSE;
			offset1 = y*p_width+x;
			// for each pixel look for 8-connected pixel with the same value
			for( i=0; i<8; i++ )
			{
				offset2 = (y+dirY[i])*p_width + x+dirX[i];
				// if we find one
				if( p_img[ offset2 ] == p_img[offset1] )
				{
					// if the region doesn't exist we create it
					// else we copy the region value
					if( p_regions[ offset1 ] == 0 )
					{
						if( p_regions[ offset2 ] == 0 )
							p_regions[ offset1 ] = acc++;
						else
							p_regions[ offset1] = p_regions[ offset2 ];
					}

					// find the good region value for the other point
					if( p_regions[ offset2 ] == 0 )
						p_regions[ offset2 ] = p_regions[ offset1 ];
					else
					{
						// if we have more than one region around the point
						// we change the values of the other regions
						if( p_regions[ offset2 ] != p_regions[ offset1 ] )
						{
							oldVal = p_regions[ offset2 ];
							for( yy=0; yy<= y; yy++ )
							{
								offset3 = yy*p_width;
								for( xx=0; xx<p_width; xx++ )
								{
									if( p_regions[offset3+xx] == oldVal )
										p_regions[offset3+xx] = p_regions[ offset1 ];
								}
							}
						}
					}
					match = TRUE;

				}
			}
			// if nothing matches we create a new region
			if( !match )
				p_regions[offset1] = acc++;
		}
	}

	// allocate the size array and angleVar array
	size = (UINT32*)MALLOC( acc * sizeof(UINT32) );
	if( size == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	angleVar = (UINT32*)MALLOC( acc * sizeof(UINT32) );
	if( angleVar == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// init of size array
	for( i=0; i<acc; i++ )
	{
		angleVar[i] = 0;
		size[i] = 0;
	}
	// fill size array with surface and perimeter of the regions
	for( y=1; y<p_height-1; y++ )
	{
		for( x=1; x<p_width-1; x++ )
		{
			acc=0;
			offset1 = y*p_width+x;
			// for each pixel look for 8-connected pixel
			for( i=0; i<8; i++ )
			{
				offset2 = (y+dirY[i])*p_width + x+dirX[i];
				val = zAbs( p_img[ offset2 ] - p_img[ offset1 ] );
				if( val > 0 )
					acc++;
			}
			if( acc != 0 )
				size[ p_regions[ offset1 ] ] += regionPerimeterCoef; // for perimeter
			size[ p_regions[ offset1 ] ] += regionSurfaceCoef; // for surface
		}
	}

	// now fill the angleVar array
	for( y=1; y<p_height-1; y++ )
	{
		for( x=1; x<p_width-1; x++ )
		{
			offset1 = y*p_width+x;
			// for each pixel look for 8-connected pixel
			for( i=0; i<8; i++ )
			{
				offset2 = (y+dirY[i])*p_width + x+dirX[i];
				val = zAbs( p_img[ offset2 ] - p_img[ offset1 ] );
				if( val > numOfOrientation/2 ) 
					val = numOfOrientation-val;
				angleVar[ p_regions[ offset1 ] ] += val;
			}
		}
	}

	// process final region values :
	// if == 0 : to much changes in the region, orientation has to be reprocessed
	// if == 255 : ok
	for( y=1; y<p_height-1; y++ )
	{
		for( x=1; x<p_width-1; x++ )
		{
			offset1 = y*p_width+x;
			if( angleVar[ p_regions[ offset1 ] ] > p_threshold*size[ p_regions[ offset1 ] ] )
				p_regions[ offset1 ] = 0;
			else
				p_regions[ offset1 ] = 255;
		}
	}

	// FREE memory
	FREE( size );
	FREE( angleVar );

	return ZEF_OK;

	// error handling
errorHandler:

	if( size != NULL )
		FREE( size );
	if( angleVar != NULL )
		FREE( angleVar );

	return error;
}

INT32 fillRegions( UINT8 *p_orient, UINT8 *p_quality, UINT32 *p_intensities, UINT32 *p_regions, UINT32 p_width, UINT32 p_height )
{
	UINT32		x,y,i,j,k,offset1,offset2;
	BOOL8		loop;
	UINT32		loopCnt,val,max,maxIndex,acc;
	UINT32		*values;
	UINT32		numOfOrientation,nbPtsFusionThreshold;

	// Init
	LOCK_MUTEX( g_paramsMutex );
		numOfOrientation = g_params.m_numOfOrientation;	
		nbPtsFusionThreshold = g_params.m_nbPtsFusionThreshold;
	UNLOCK_MUTEX( g_paramsMutex );
	loopCnt=0;
	values = (UINT32*)MALLOC(numOfOrientation*sizeof(UINT32));


	// Loop until no change occurs or we do too much loops
	do
	{
		loop = FALSE;

		// For all pixel values
		for( j=1; j<p_height-1; j++ )
		{
			for( i=1; i<p_width-1; i++ )
			{
				offset1 = j*p_width+i;
				// If the pixel is part of a region marked to be processed
				if( p_regions[ offset1 ] == 0 )
				{
					for( k=0; k<numOfOrientation; k++ )
						values[k] = 0;
					max = 0;
					acc = 0;
					maxIndex = p_orient[ offset1 ];
					// Loop for the most probable orientation in the neigborhood
					for( y=j-1; y<=j+1; y++ )
					{
						for( x=i-1; x<=i+1; x++ )
						{
							offset2 = y*p_width+x;
							val = p_orient[ offset2 ];
							if( p_regions[ offset2 ] != 0 )
							{
								values[val] += p_quality[ offset2 ]*zSqrt( 255*p_intensities[offset1*numOfOrientation+val] );
								if( values[ val ] > max )
								{
									max = values[ val ];
									maxIndex = val;
								}
								acc++;
							}
						}
					}
					if( maxIndex != p_orient[ offset1 ] )
						loop = TRUE;

					// Set the new values
					if( acc > nbPtsFusionThreshold )
					{
						p_orient[ offset1 ] = maxIndex;
						p_regions[ offset1 ] = 127;
						p_quality[ offset1 ] = p_intensities[ offset1*numOfOrientation+maxIndex ];
					}

				}
			}
		}
		loopCnt++;
	}
	while( loop &&  loopCnt < MAX_LOOPS );

	FREE( values );

	return ZEF_OK;
}

INT32 findPixels( UINT8 *p_img, UINT32 *p_pixels, UINT32 p_width, UINT32 p_height, REAL32 p_threshold )
{	
	UINT32		x,y,i,offset1,offset2;
	UINT32		val2;
	INT32		val;
	UINT32		numOfOrientation;

	// fast lookup table
	INT8		dirX[8] = {  0,  1, -1,  1, -1,  1, -1,  0 };
	INT8		dirY[8] = { -1, -1, -1,  0,  0,  1,  1,  1 };

	// inits
	LOCK_MUTEX(g_paramsMutex );
		numOfOrientation = g_params.m_numOfOrientation;
	UNLOCK_MUTEX(g_paramsMutex );

	for( y=1; y<p_height-1; y++ )
	{
		for( x=1; x<p_width-1; x++ )
		{
			offset1 = y*p_width+x;
			val = p_img[ offset1 ];
			for( i=0; i<8; i++ )
			{
				offset2 = (y+dirY[i])*p_width+x+dirX[i];
				val2 = zAbs( val - p_img[ offset2 ] );
				if( val2 > numOfOrientation/2 ) 
					val2 = numOfOrientation-val2;
				p_pixels[ offset1 ] += val2;
			}
			if( p_pixels[ offset1 ] > p_threshold  )
				p_pixels[ offset1 ] = 0;
			else
				p_pixels[ offset1 ] = 255;
		}
	}
	

	return ZEF_OK;
}

INT32 imageFusion( INT8 *p_images, UINT8 *p_orient, UINT8 *p_dst, UINT32 p_width, UINT32 p_height )
{
	UINT32	i,j,k,x,y,offset1,offset2;
	UINT32	index;
	UINT32	weightSize,weightSize2;
	REAL32	weight[(FUSION_BLOC_SIZE+FUSION_BLOC_SIZE/2)*(FUSION_BLOC_SIZE+FUSION_BLOC_SIZE/2)];
	UINT32	numOfOrientation;
	
	// init 
	LOCK_MUTEX( g_paramsMutex );
		numOfOrientation = g_params.m_numOfOrientation;
	UNLOCK_MUTEX( g_paramsMutex );
	weightSize = (FUSION_BLOC_SIZE+FUSION_BLOC_SIZE/2);
	weightSize2 = weightSize/2;	

	// init of weights
	k=0;
	for( j=0; j<weightSize; j++ )
	{
		for( i=0; i<weightSize; i++ )
		{
			x = 1+zAbs((INT32)(i%(weightSize2))-(INT32)((weightSize2-1)*(i/(weightSize2))));
			y = 1+zAbs((INT32)(j%(weightSize2))-(INT32)((weightSize2-1)*(j/(weightSize2))));

			weight[k++] = x*y/(REAL32)(weightSize*weightSize/4);
		}
	}

	// clear image
	for( i=0; i<p_width*p_height; i++ )
		p_dst[i] = 127;

	// then fusion !
	
	for( j=0; j<p_height-weightSize; j+=FUSION_BLOC_SIZE )
	{
		for( i=0; i<p_width-weightSize; i+=FUSION_BLOC_SIZE )
		{
			index = p_orient[ j*p_width + i ];
			k=0;
			for( y=0; y<weightSize; y++ )
			{
				offset1 = (j+y)*p_width+i;
				for( x=0; x<weightSize; x++ )
				{
					offset2 = offset1+x;
					p_dst[offset2] += zReal2Int32(p_images[offset2*numOfOrientation+index]*weight[k++]);
				}
			}
		}
	}

	return ZEF_OK;
}

INT32 qualityThreshold( UINT8 *p_quality, UINT32 p_width, UINT32 p_height, UINT32 p_threshold )
{
	UINT32		i,x,y;
	UINT32		max,val;
	UINT32		border;
	UINT32		fftBlocOverlap;

	// init
	LOCK_MUTEX( g_paramsMutex );
		fftBlocOverlap = g_params.m_fftBlocOverlap;	
	UNLOCK_MUTEX( g_paramsMutex );
	max=0;
	border = fftBlocOverlap/2/FUSION_BLOC_SIZE;

	// search for max value
	for( i=0; i<p_width*p_height; i++ )
	{
		if( p_quality[i] > max )
			max = p_quality[i];
	}

	if( max == 0 ) max = 1;
	

	// then calculate new values and apply threshold
	for( i=0; i<p_width*p_height; i++ )
	{
		val = 255*p_quality[i]/max;
		if( val > p_threshold )
			p_quality[i] = val;
		else
			p_quality[i] = 0;
	}

	// set border to zero
	for( y=0 ;y<border; y++ )
	{
		for( x=0; x<p_width; x++ )
		{
			p_quality[y*p_width+x] = 0;
			p_quality[(p_height-1-y)*p_width+x] = 0;
		}
	}
	for( y=0 ;y<p_height; y++ )
	{
		for( x=0; x<border; x++ )
		{
			p_quality[y*p_width+x] = 0;
			p_quality[y*p_width+p_width-1-x] = 0;
		}
	}

	return ZEF_OK;
}

INT32 maskImage( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT8 *p_mask )
{
	UINT32		x,y,y2,width2;

	// inits
	width2 = p_width / FUSION_BLOC_SIZE;


	for( y=0; y<p_height; y++ )
	{
		y2 = y/FUSION_BLOC_SIZE;
		for( x=0; x<p_width; x++ )
		{
			if( p_mask[y2*width2+x/FUSION_BLOC_SIZE] != 0 ) 
				p_dst[y*p_width+x] = p_src[y*p_width+x];
			else
				p_dst[y*p_width+x] = 255;	
		}
	}

	return ZEF_OK;
}

#endif // ZEF_FIXED_POINT
