/*
Company :		ZEFYR Technologies
File :			enrollExport.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : SDK enrolment exported functions

Features : external interface functions		

Functions : 

	dpInitEnroll : initialize enrolment part
	dpUnInitEnroll : uninitialize enrolment part
	dpGetMaxTemplateSize : return the max size a template can be (with the actual parameters)		
	dpGetTemplateSize	:	return the size of a given template
	dpGenerateTemplate : generate a template from an image
	dpGetTemplateQuality : return the quality of a given template

Function prototypes : 

	INT32	dpInitEnroll( void )
	INT32   dpUnInitEnroll()
	UINT32	getMaxTemplateSize( void );
	UINT32	getTemplateSize( UINT8 *p_templateBuffer );
	INT32	generateTemplate( UINT8 *p_imgData, UINT32 p_width, UINT32 p_height, UINT8 *p_templateBuffer ); 
	UINT32  getTemplateQuality( UINT8 *p_templateBuffer );

Necessary package : 

  	zMaths.h
	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	preProcessings.h
	FFT2D.h
	enhancements.h
	coresDeltas.h
	localParameters.h
	skeleton.h
	minutiae.h
	tools.h

Comments : none

*/

#include <stdlib.h>
#include <memory.h> 

#include "zMaths.h"
#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "preProcessings.h"
#include "FFT2D.h"
#include "enhancements.h"
#include "fxEnhancements.h"
#include "coresDeltas.h"
#include "localParameters.h"
#include "skeleton.h"
#include "minutiae.h"
#include "tools.h"

#include "geometry.h"
#include "BMPfiles.h"

#include "compatibility.h"

#include "enrollExport.h"

#include "greyLevelTrans.h"
#include "filters.h"

// global structures
PreCompData		g_preCompData;
extern BOOL8	g_initialized;

INT32 WINAPI dpInitEnroll( void )
{
	INT32	error;

	// create the mutex that will protect parms structure access
	CREATE_MUTEX( g_paramsMutex );

	// initialize enhancement temp data and FFT
	error = initEnhance();
	if( error != ZEF_OK )
		goto errorHandler;

	// initialize maths precalculated arrays
	error = zInitMaths();
	if( error != ZEF_OK )
		goto errorHandler;

	// mark initialisation flag to true
	g_initialized = TRUE;

	return ZEF_OK;

	//error Handling
errorHandler:

	// FREE memory 
	dpUnInitEnroll();

	return error;
}

INT32 WINAPI dpUnInitEnroll()
{
	INT32 error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// mark initialisation flag to false
	g_initialized = FALSE;

	// FREE enhance buffers
	error = unInitEnhance();
	if( error != ZEF_OK )
		goto errorHandler;

	return ZEF_OK;

	//error Handling
errorHandler:

	return error;
}

INT32 WINAPI dpGetMaxTemplateSize( void )
{
	INT32	error,size;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	LOCK_MUTEX( g_paramsMutex );
		size = 52+8*g_params.m_maxNumberOfCores+6*g_params.m_maxNumberOfMinutiae+3*g_params.m_maxNumberOfLocalInfos/2;
	UNLOCK_MUTEX( g_paramsMutex );

	// not really the best way to calculate max size... need to be changed to something better....
	return size;

	//error handling
errorHandler:
	return error;
}

INT32	WINAPI dpGetTemplateSize( UINT8 *p_templateBuffer )
{
	INT32		error;
	Template	templ;
	UINT32		size;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// init template
	error = templateInit( &templ );
	if( error < ZEF_OK )
		goto errorHandler; 

	// test for NULL pointer in arguments
	if( p_templateBuffer == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// unpack template to get its size
	error = templateUnpacker( p_templateBuffer, &templ);
	if( error < ZEF_OK )
		goto errorHandler; 

	// retreive size
	size = templ.m_size;

	// FREE memory
	error = templateUninit( &templ );
	if( error < ZEF_OK )
		goto errorHandler; 

	return size;

	//error handling
errorHandler:

	if( g_initialized )
	{
		templateUninit( &templ );
	}

	return error;
}

INT32	WINAPI dpGenerateTemplate( UINT8 *p_imgData, UINT32 p_width, UINT32 p_height, UINT8 *p_templateBuffer )
{
	INT32			error;
	BioEngineData	data;
#ifdef TIME_REPORT
	clock_t timeReport;
#endif
	UINT32	globalImgScaling;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// params init
	LOCK_MUTEX( g_paramsMutex );
		globalImgScaling = g_params.m_globalImgScaling;
	UNLOCK_MUTEX( g_paramsMutex );

	// init of template datas
	error = templateInit( &(data.m_template) );
	if( error < ZEF_OK )
		goto errorHandler;

	// Scalling
#ifdef TIME_REPORT
	timeReport = clock();
#endif
	data.m_image = (UINT8*)MALLOC(p_width*p_height*sizeof(UINT8));
	scale(p_imgData,data.m_image,p_width,p_height,globalImgScaling);
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "scale" ,clock()-timeReport );
#endif

	// init of data structure
	data.m_width = (globalImgScaling*p_width)>>8;
	data.m_height = (globalImgScaling*p_height)>>8;
	data.m_workImg = NULL;
	data.m_ori = NULL;
	data.m_quality = NULL;

	// test for NULL pointer in arguments
	// must be done after "data" initialization, otherwise, problems when FREEing memory in error handler
	if( p_imgData == NULL || p_templateBuffer == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// test for zero size width and height
	if( p_width == 0 || p_height == 0 )
	{
		error = ZEF_BAD_SIZE;
		goto errorHandler;
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// pre-processings
	error = preProcess( &data );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "preProcess" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// enhancements
	error = enhance( &data );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "enhance" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// find cores and deltas
	error = findCoresDeltas( &data );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "findCoresDeltas" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// process local Parameters
	error = processLocalParameters( &data );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "processLocalParameters" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// extract minutiae and store then in the template
	error = extractMinutiae( &data );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "extractMinutiae" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// pack data into templateBuffer
	error = templatePacker( &(data.m_template), p_templateBuffer );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "templatePacker" ,clock()-timeReport );
#endif

	// FREEing temporary buffers
	if( data.m_workImg != NULL )
	{
		FREE(data.m_workImg);
		data.m_workImg = NULL;
	}
	if( data.m_quality != NULL )
	{
		FREE(data.m_ori);
		data.m_ori = NULL;
	}
	if( data.m_quality != NULL )
	{
		FREE(data.m_quality);
		data.m_quality = NULL;
	}

	// FREE template memory
	error = templateUninit( &(data.m_template) );
	if( error < ZEF_OK )
		goto errorHandler;

	// Scalling test !!!
	FREE(data.m_image);

	return ZEF_OK;

	// error Handling
errorHandler:

	if( g_initialized )
	{
		// FREEing temporary buffers
		if( data.m_workImg != NULL )
		FREE(data.m_workImg);
		if( data.m_ori != NULL )
		FREE(data.m_ori);
		if( data.m_quality != NULL )
		FREE(data.m_quality);
		// FREE template memory
		templateUninit( &(data.m_template) );
	}

	return error;
}


INT32 WINAPI dpGetTemplateQuality( UINT8 *p_templateBuffer )
{
	INT32		error;
	INT32		i,j;
	INT32		val,cpt=0;
	INT32		quality1=0,quality2=0,finalQuality;
	Template	templ;
	UINT32		maxNumberOfMinutiae;

	//test bioEngin initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// params init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfMinutiae = g_params.m_maxNumberOfMinutiae;
	UNLOCK_MUTEX( g_paramsMutex );

	// init template
	error = templateInit( &templ );
	if( error < ZEF_OK )
	goto errorHandler;

	// test for NULL pointer in arguments
	// must be done after template init
	if( p_templateBuffer == NULL  )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// unpack template data
	error = templateUnpacker( p_templateBuffer, &templ );
	if( error < ZEF_OK )
		goto errorHandler;

	// process template quality

	for( i=0; i<templ.m_nbMinutiae; i++ )
	{
		val = templ.m_minutiae[i].m_qual;
		quality1 += val*val;
		if( val > 180 )
			cpt++;
	}
	if( templ.m_nbMinutiae != 0 )
		quality1 /= templ.m_nbMinutiae;

	quality1 = zSqrt(quality1);
	quality1 += quality1*cpt/maxNumberOfMinutiae;

	if( templ.m_nbCores > 0 )
		quality1 += 32;

	cpt=0;
	for( j=0; j<templ.m_localH; j++ )
	{
		for( i=0; i<templ.m_localW; i++ )
		{
			val = 16*templ.m_qual[j*templ.m_localW+i];
			if( val != 0 )
			{
				cpt++;
				quality2+=val*val;
			}
		}
	}

	if( cpt != 0)
		quality2 /= cpt;
	quality2 = zSqrt(quality2);

	finalQuality = 100*(quality1+2*quality2)/768;
	if( finalQuality > 100 ) finalQuality = 100;

	// FREE template
	error = templateUninit( &templ );
	if( error < ZEF_OK )
		goto errorHandler;

	return finalQuality;

	// error Handling
errorHandler:

	if( g_initialized )
	{
		templateUninit( &templ );
	}

	return error;
}

INT32	WINAPI dpProcessPreview( UINT8 *p_imgData, UINT32 p_width, UINT32 p_height, UINT8 *p_preview, UINT32 *p_flags )
{
	INT32	error,globMean,globStdDev,indicator;
	UINT8	*buffer=NULL;
	UINT32  width,height,x,y,offset;	
	UINT32	miniWidth, miniHeight;
	UINT32	i,j,max,val;
	UINT8	*temp;
	UINT32	*mean,*stdDev;
	UINT32	sumX,sumY,cpt;
	BOOL8	moved;
	static UINT32 maxScore=0;
	static UINT32 maxCpt=0;
	UINT32	normalizationMean,normalizationStdDev,segmMedianFilterSize;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// inits
	LOCK_MUTEX( g_paramsMutex );
		normalizationMean = g_params.m_normalizationMean;
		normalizationStdDev = g_params.m_normalizationStdDev;
		segmMedianFilterSize = g_params.m_segmMedianFilterSize;
	UNLOCK_MUTEX( g_paramsMutex );
	*p_flags = 0;
	width = p_width;
	height = p_height;
	max = 0;

	// allocations 1st part
	buffer = (UINT8*)MALLOC(width*height*sizeof(UINT8));
	if( buffer == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// copy image
	memcpy( buffer, p_imgData, width*height*sizeof(UINT8) );

	// resize image to SEGM_BLOC_SIZE multiples
	error = checkBlocMultiple( buffer , &width, &height );
	if( error < ZEF_OK )
		goto errorHandler;

	// set miniWidth and height
	miniWidth = width / SEGM_BLOC_SIZE;
	if( width%SEGM_BLOC_SIZE) 
	{
		error = ZEF_NOT_BLOC_MULTIPLE;
		goto errorHandler;
	}
	miniHeight = height / SEGM_BLOC_SIZE;
	if( height%SEGM_BLOC_SIZE) 
	{
		error = ZEF_NOT_BLOC_MULTIPLE;
		goto errorHandler;
	}

	// allocations 2nd part
	temp = (UINT8*) MALLOC( width*height*sizeof( UINT8 ) );
	if( temp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	mean = (UINT32*) MALLOC( miniWidth*miniHeight*sizeof( UINT32 ) );
	if( mean == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	stdDev = (UINT32*) MALLOC( miniWidth*miniHeight*sizeof( UINT32 ) );
	if( stdDev == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// get mean and stdDev
	error = meanStdDev( buffer, width, height, &globMean, &globStdDev );
	if( error < ZEF_OK )
		goto errorHandler;

	if( globStdDev == 0 ) globStdDev = 1;

	// calculate finger detection indicator
	indicator = (globMean-127)*256/globStdDev; 

	// global normalization
	error = normalize( buffer, temp, width, height, normalizationMean, normalizationStdDev );
	if( error < ZEF_OK )
		goto errorHandler;

	// calculate stdDev of the input image
	error = localMeanStdDev( temp, width, height, mean, stdDev );
	if( error < ZEF_OK )
		goto errorHandler;

	FREE(temp);
	temp=NULL;
	
	// we convert stdDev to a UINT8 image
	for( j=2; j<miniHeight-2; j++ )
	{
		for( i=2; i<miniWidth-2; i++ )
		{
 			if( stdDev[j*miniWidth+i] > max ) 
				max = stdDev[j*miniWidth+i];
		}
	}

	if( max == 0 ) max = 1; 
	max = (1L<<16)/max; // to avoid division in loops

	for( i=0; i<miniWidth*miniHeight; i++ )
	{
		val = (stdDev[i]*max)>>8;  // p_preview[i] = 255*stdDev[i]/max;
		if( val > 255 ) val = 255;
		p_preview[i] = val;
	}

	// apply a recursive median filter to smooth everything
	error = medianFilter( p_preview, p_preview, miniWidth, miniHeight, segmMedianFilterSize );	 
	if( error < ZEF_OK )
		goto errorHandler;

	// and give memory back
	FREE(mean);
	mean = NULL;
	FREE(stdDev);
	stdDev = NULL;

	if( indicator < 127 )
	{
		*p_flags |= ZEF_FLAG_FINGER_DETECTED;
	
		// find centroid
		sumX=sumY=cpt=0;
		for( y=0; y<miniHeight; y++ )
		{
			for( x=0; x<miniWidth; x++ )
			{
				sumY += y*p_preview[y*miniWidth+x];
				sumX += x*p_preview[y*miniWidth+x];
				cpt += p_preview[y*miniWidth+x];
			}
		}

		// if the image is not all black
		if( cpt != 0 )
		{
			// check position of finger on the sensor
			sumX /= cpt;
			sumY /= cpt;
			moved = FALSE;
			if( sumY > miniHeight/2 + miniHeight*26/256)
			{
				*p_flags |= ZEF_FLAG_MOVE_UP;
				moved = TRUE;
			}
			else if( sumY < miniHeight/2 - miniHeight*26/256)
			{
				*p_flags |= ZEF_FLAG_MOVE_DOWN;
				moved = TRUE;
			}

			if( sumX > miniWidth/2 + miniWidth*26/256)
			{
				*p_flags |= ZEF_FLAG_MOVE_LEFT;
				moved = TRUE;
			}
			else if( sumX < miniWidth/2 - miniWidth*26/256)
			{
				*p_flags |= ZEF_FLAG_MOVE_RIGHT;
				moved = TRUE;
			}
			// if position is ok
			if( !moved )
			{
				// check image quality (pressure ?)
				if( cpt/(miniWidth*miniHeight) < 100 ) 
				{
					*p_flags |= ZEF_FLAG_PRESS_STRONGER;
				}
				else
				{
					// do we need to remember the image
					if( cpt > maxScore )
					{
						maxScore = cpt;
						*p_flags |= ZEF_FLAG_BEST_IMG;
					}

					// image is ok : remove finger
					if( maxCpt > 3 )
					{
						*p_flags |= ZEF_FLAG_REMOVE_FINGER;
					}
					maxCpt++;
				}
			}
		}
	}
	else
	{
		maxCpt = 0;
		maxScore=0;
	}


	// scale img
	error = scale( p_preview, buffer, miniWidth, miniHeight, SEGM_BLOC_SIZE*256 );
	if( error < ZEF_OK )
		goto errorHandler;

	// copy img
	offset=0;
	for( y=0; y<height; y++ )
	{
		for( x=0; x<width; x++ )
		{
			p_preview[offset++] = buffer[y*width+x];
		}
		offset += (p_width-width);
	}
	for( y=height; y<p_height; y++ )
	{
		for( x=0; x<p_width; x++ )
		{
			p_preview[y*width+x] = 0;
		}
	}

	FREE(buffer);

	return ZEF_OK;

errorHandler:

	if( buffer != NULL )
		FREE(buffer);

	if( temp != NULL )
		FREE( temp );

	if( mean != NULL )
		FREE( mean );

	if( stdDev != NULL )
		FREE( stdDev );

	return error;
}
