/*
Company :		ZEFYR Technologies
File :			exported.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : SDK exported functions

Features : external interface functions		

Functions : 

	dpInit :	initialize everything
	dpUnInit :	FREE everything
	dpGetEngineVersion : get bioEngine version
	dpSetSpeed : sets algorithm speed (for both extraction and matching)

Function prototypes : 

	INT32	dpInit( void );
	INT32	dpUnInit( void );
	char*   dpGetEngineVersion( void )
	INT32   dpSetSpeed( UINT8 p_type, UINT8 p_speed )


Necessary package : 

  	zMaths.h
	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	enhancements.h

Comments : none

*/

#include <stdlib.h>

#include "zMaths.h"
#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "enhancements.h"
#include "fxEnhancements.h"
#include "portableMutex.h"
#include "exported.h"


extern BOOL8	g_initialized;

INT32 WINAPI dpInit( void )
{
	INT32	error;

	if( !g_initialized )
	{
		// create the mutex that will protect params structure access
		CREATE_MUTEX( g_paramsMutex );
		// initialize ideniticaiton cancellation...
		CREATE_MUTEX( g_idCancelMutex );
		g_idCancelPointer = NULL;

		// initialize enhancement temp data and FFT
		error = initEnhance();
		if( error != ZEF_OK )
			goto errorHandler;

		// initialize maths precalculated arrays
		error = zInitMaths();
		if( error != ZEF_OK )
			goto errorHandler;

		// mark initialisation flag to true
		// this flag doesn't need to be protected by a mutex since this function is
		// not thread safe and that all other accesses will be read only
		g_initialized = TRUE;
	}

	return ZEF_OK;

	//error Handling
errorHandler:

	// FREE memory 
	dpUnInit();

	return error;
}

INT32 WINAPI dpUnInit()
{
	INT32 error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// mark initialisation flag to false
	// this flag doesn't need to be protected by a mutex since this function is
	// not thread safe and that all other accesses will be read only
	g_initialized = FALSE;

	// destroy the mutex that protected parms structure access
	DESTROY_MUTEX( g_paramsMutex );

	// FREE enhance buffers
	error = unInitEnhance();
	if( error != ZEF_OK )
		goto errorHandler;

	return ZEF_OK;

	//error Handling
errorHandler:

	return error;
}

char* WINAPI dpGetEngineVersion( void )
{
	return DP_BIO_VERSION;
}

INT32 WINAPI dpSetSpeed( UINT32 p_type, INT32 p_speed )
{
	INT32		error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error =  ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// check parameters
	if( p_type > 2 || p_speed > 2 || p_speed < 0)
	{
		error =  ZEF_BAD_SPEED_PARAM;
		goto errorHandler;
	}

	LOCK_MUTEX( g_paramsMutex );

	// template generation speed
	if( p_type == 0 || p_type == 1 )
	{
		switch( p_speed )
		{
		case 0: // normal speed
			g_params.m_globalImgScaling = GLOBAL_IMG_SCALING;
			g_params.m_globalTemplateScaling = GLOBAL_TEMPLATE_SCALING;
			g_params.m_gaborWidth = GABOR_WIDTH;
			g_params.m_oriStepSize = ORI_STEP_SIZE;
			g_params.m_numOfOrientation = NUM_OF_ORIENTATION;
			g_params.m_fftBlocOverlap = FFT_BLOC_OVERLAP;
			g_params.m_filterSelectionThreshold = FILTER_SELECTION_THRESHOLD;

			break;

		case 1: // scale by 80%
			g_params.m_globalImgScaling = (FAST_SCALE_FACTOR*GLOBAL_IMG_SCALING)/256; // *= 0.8f
			g_params.m_globalTemplateScaling = (256*256)/g_params.m_globalImgScaling+1; // +1 to avoid rounding issues
			g_params.m_gaborWidth = (FAST_SCALE_FACTOR*GABOR_WIDTH)/256;
			g_params.m_fftBlocOverlap = FFT_BLOC_OVERLAP;
			g_params.m_numOfOrientation = NUM_OF_ORIENTATION;
			g_params.m_oriStepSize = ORI_STEP_SIZE;
			g_params.m_filterSelectionThreshold = FILTER_SELECTION_THRESHOLD;

			break;

		case 2:  //scale image by 80% and use 6 ori and overlap of 16 + ori step size = 2
			g_params.m_globalImgScaling = (FAST_SCALE_FACTOR*GLOBAL_IMG_SCALING)/256; // *= 0.8f
			g_params.m_globalTemplateScaling = (256*256)/g_params.m_globalImgScaling+1; // +1 to avoid rounding issues
			g_params.m_gaborWidth = (FAST_SCALE_FACTOR*GABOR_WIDTH)/256;
			g_params.m_fftBlocOverlap = FFT_BLOC_OVERLAP/2;
			g_params.m_numOfOrientation = (NUM_OF_ORIENTATION*6)/8;
			g_params.m_oriStepSize = ORI_STEP_SIZE*2; // gagne on beaucoup  en l'enlevant ???
			g_params.m_filterSelectionThreshold = 128;
			
			break;

		default:
			break;
		}
	}

	// template matching speed
	if( p_type == 0 || p_type == 2 )
	{
		switch( p_speed )
		{
		case 0:
			g_params.m_maxRegistrationAngle = MAX_REGISTRATION_ANGLE;
			g_params.m_fisrtPassAngleStep = FIRST_PASS_ANGLE_STEP;
			break;

		case 1:
			g_params.m_maxRegistrationAngle = MAX_REGISTRATION_ANGLE/2;
			g_params.m_fisrtPassAngleStep = FIRST_PASS_ANGLE_STEP;
			break;

		case 2:
			g_params.m_maxRegistrationAngle = MAX_REGISTRATION_ANGLE/2;
			g_params.m_fisrtPassAngleStep = FIRST_PASS_ANGLE_STEP*2;
			break;

		default:
			break;
		}
	}

	UNLOCK_MUTEX( g_paramsMutex );

	return ZEF_OK;

//error Handling
errorHandler:

	return error;
}
