/*
Company :		ZEFYR Technologies
File :			filters.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : filters like median, erosion, dilation, etc.

Features : recursive and non recursive filtering		
    
Functions : 

	medianFind : helper function that find the median element of an array of n elements
	medianFilter : recursive and non recursive median filter
	medianFind32 : helper function that find the median element of an array of n elements (for long)
	medianFilter32 : recursive and non recursive median filter (for long)
	binaryMedianFilter : recursive and non recursive median filter (binary)
	erosionFilter32 : morphological filter : grayscale erosion (for long) 
	generateGaborFFT : generates a gabor filter in the frequency domain
	freqConvolution : double convolution in frequancy domain (used with real filters)
	poincare : process poincare index of pixels of an orientation map
	sobelGradient : process X and Y gradient with sobel filters
	curvature : process curvature of an orientation map

Function prototypes : 

	UINT8 medianFind(UINT8 p_arr[], UINT32 p_n);
	INT32 medianFilter( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );	
	UINT32 medianFind32(UINT32 p_arr[], UINT32 p_n);
	INT32 medianFilter32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );
	INT32 binaryMedianFilter( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );
	INT32 erosionFilter32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );
	INT32 generateGaborFFT( REAL32 *p_filter, REAL32 p_size, REAL32 p_freq, REAL32 p_angle );
	INT32 freqConvolution( REAL32 *p_src, REAL32 *p_dst, REAL32 *p_filter );
	INT32 poincare( UINT8 *p_src, INT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );
	INT32 sobelGradient( UINT8 *p_src, INT16 *p_gradient, UINT32 p_width, UINT32 p_height );
	INT32 curvature( UINT8 *p_src, UINT8 *p_dst, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize );

Necessary package : 

	customTypes.h
	staticParameters.h
	errorCodes.h
	zMaths.h
	limits.h

Comments : none

*/

#include <stdlib.h>

#include "customTypes.h"
#include "staticParameters.h"
#include "errorCodes.h"
#include "zMaths.h"
#include "limits.h"

#include "filters.h"

// This findMedian routine is based on the algorithm described in
// "Numerical recipes in C", Second Edition,
// It returns the median value of an array of n elements
// Elements passed in array p_arr are moved during the processing
// you need to take it into account when using this function
#define SWAP(a,b) { UINT8 t=(a);(a)=(b);(b)=t; }

UINT8 medianFind(UINT8 p_array[], UINT32 p_size)
{
	UINT32 low, high ;
	UINT32 median;
	UINT32 middle, ll, hh;

	low = 0 ; high = p_size-1 ; median = (low + high) / 2;
	while(1)
	{
		if (high <= low) // One element only
			return p_array[median] ;

		if (high == low + 1) { // Two elements only
			if (p_array[low] > p_array[high])
				SWAP(p_array[low], p_array[high]) ;
			return p_array[median] ;
		}

		// Find median of low, middle and high items; swap into position low
		middle = (low + high) / 2;
		if (p_array[middle] > p_array[high]) SWAP(p_array[middle], p_array[high]) ;
		if (p_array[low] > p_array[high]) SWAP(p_array[low], p_array[high]) ;
		if (p_array[middle] > p_array[low]) SWAP(p_array[middle], p_array[low]) ;

		// Swap low item (now in position middle) into position (low+1)
		SWAP(p_array[middle], p_array[low+1]) ;

		// Nibble from each end towards middle, swapping items when stuck
		ll = low + 1;
		hh = high;
		while(1)
		{
			do ll++; while (p_array[low] > p_array[ll]) ;
			do hh--; while (p_array[hh] > p_array[low]) ;
			if (hh < ll)
				break;
			SWAP(p_array[ll], p_array[hh]) ;
		}

		// Swap middle item (in position low) back into correct position
		SWAP(p_array[low], p_array[hh]) ;

		// Re-set active partition
		if (hh <= median)
			low = ll;
		if (hh >= median)
			high = hh - 1;
	}
}
#undef ELEM_SWAP

// median filter
// if you pass the same image as source and destination you will have a recursif filter
// you will have a normal filter otherwise
// the p_filterSize/2 pixel near the borders of the image will not be filtered
INT32 medianFilter( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize )
{
	UINT8		*tmp;
	INT32		i,j,k;
	INT32		x,y;
	UINT32		dec,dec2,halfSize,size2;
	UINT32		srcOffset,dstOffset;
	INT32		error;

	// init 
	tmp = NULL;

	// set some helpful constants
	halfSize	= p_filterSize/2;
	size2	= p_filterSize*p_filterSize;
	dec		= halfSize*p_width+halfSize;
	dec2	= p_width-p_filterSize;

	// memory allocation for the temporary array
	tmp = (UINT8*)MALLOC( size2*sizeof(UINT8) );
	if( tmp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// main image processing
	dstOffset = dec;
	for( j=0; j<(INT32)(p_height-(p_filterSize-1)); j++ )
	{
		for( i=0; i<(INT32)(p_width-(p_filterSize-1)); i++ )
		{
			srcOffset = dstOffset-dec;
			k=0;
			for( y=0; y<(INT32)p_filterSize; y++ )
			{
				for( x=0; x<(INT32)p_filterSize; x++ )
				{
					tmp[k++] = p_src[srcOffset++];
				}
				srcOffset += dec2;
			}
			p_dst[dstOffset++] = medianFind( tmp, size2 );
		}
		dstOffset += p_filterSize-1;
	}

	// border processing
	//top
	for( j=0; j<(INT32)halfSize; j++ )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind( tmp, k );
		}
	}
	//bottoom
	for( j=(INT32)p_height-1; j>(INT32)p_height-1-(INT32)halfSize; j-- )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind( tmp, k );
		}
	}
	// left
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=0; i<(INT32)halfSize; i++ )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind( tmp, k );
		}
	}

	// right
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=(INT32)p_width-1; i>(INT32)p_width-1-(INT32)halfSize; i-- )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind( tmp, k );
		}
	}

	FREE( tmp );

	return ZEF_OK;

	// error Handling
errorHandler:
	
	FREE( tmp);

	return error;
}

UINT32 medianFind32(UINT32 p_array[], UINT32 p_size)
{
	UINT32 low, high ;
	UINT32 median;
	UINT32 middle, ll, hh;

	low = 0 ; high = p_size-1 ; median = (low + high) / 2;
	while(1)
	{
		if (high <= low) // One element only
			return p_array[median] ;

		if (high == low + 1) { // Two elements only
			if (p_array[low] > p_array[high])
				SWAP(p_array[low], p_array[high]) ;
			return p_array[median] ;
		}

		// Find median of low, middle and high items; swap into position low
		middle = (low + high) / 2;
		if (p_array[middle] > p_array[high]) SWAP(p_array[middle], p_array[high]) ;
		if (p_array[low] > p_array[high]) SWAP(p_array[low], p_array[high]) ;
		if (p_array[middle] > p_array[low]) SWAP(p_array[middle], p_array[low]) ;

		// Swap low item (now in position middle) into position (low+1)
		SWAP(p_array[middle], p_array[low+1]) ;

		// Nibble from each end towards middle, swapping items when stuck
		ll = low + 1;
		hh = high;
		while(1)
		{
			do ll++; while (p_array[low] > p_array[ll]) ;
			do hh--; while (p_array[hh] > p_array[low]) ;
			if (hh < ll)
				break;
			SWAP(p_array[ll], p_array[hh]) ;
		}

		// Swap middle item (in position low) back into correct position
		SWAP(p_array[low], p_array[hh]) ;

		// Re-set active partition
		if (hh <= median)
			low = ll;
		if (hh >= median)
			high = hh - 1;
	}
}
#undef ELEM_SWAP

// median filter
// if you pass the same image as source and destination you will have a recursif filter
// you will have a normal filter otherwise
// the p_filterSize/2 pixel near the borders of the image will not be filtered
INT32 medianFilter32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize )
{
	UINT32		*tmp;
	INT32		i,j,k;
	INT32		x,y;
	UINT32		dec,dec2,halfSize,size2;
	UINT32		srcOffset,dstOffset;
	INT32		error;

	// init 
	tmp = NULL;

	// set some helpful constants
	halfSize	= p_filterSize/2;
	size2	= p_filterSize*p_filterSize;
	dec		= halfSize*p_width+halfSize;
	dec2	= p_width-p_filterSize;

	// memory allocation for the temporary array
	tmp = (UINT32*)MALLOC( size2*sizeof(UINT32) );
	if( tmp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// main image processing
	dstOffset = dec;
	for( j=0; j<(INT32)(p_height-(p_filterSize-1)); j++ )
	{
		for( i=0; i<(INT32)(p_width-(p_filterSize-1)); i++ )
		{
			srcOffset = dstOffset-dec;
			k=0;
			for( y=0; y<(INT32)p_filterSize; y++ )
			{
				for( x=0; x<(INT32)p_filterSize; x++ )
				{
					tmp[k++] = p_src[srcOffset++];
				}
				srcOffset += dec2;
			}
			p_dst[dstOffset++] = medianFind32( tmp, size2 );
		}
		dstOffset += p_filterSize-1;
	}

	// border processing
	//top
	for( j=0; j<(INT32)halfSize; j++ )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind32( tmp, k );
		}
	}
	//bottoom
	for( j=(INT32)p_height-1; j>(INT32)p_height-1-(INT32)halfSize; j-- )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind32( tmp, k );
		}
	}
	// left
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=0; i<(INT32)halfSize; i++ )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind32( tmp, k );
		}
	}

	// right
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=(INT32)p_width-1; i>(INT32)p_width-1-(INT32)halfSize; i-- )
		{
			// top
			k=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						tmp[k++] = p_src[y*p_width+x];
				}
			}
			if( k != 0 )
				p_dst[j*p_width+i] = medianFind32( tmp, k );
		}
	}

	FREE( tmp );

	return ZEF_OK;

	// error Handling
errorHandler:
	
	FREE( tmp);

	return error;
}

INT32 binaryMedianFilter( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize )
{
	UINT32		acc;		
	INT32		i,j;
	INT32		x,y;
	UINT32		dec,dec2,halfSize;
	UINT32		srcOffset,dstOffset,midVal;
	
	// set some helpful constants
	halfSize	= p_filterSize/2;
	dec		= halfSize*p_width+halfSize;
	dec2	= p_width-p_filterSize;
	midVal  = 255*p_filterSize*p_filterSize/2;
	
	
	// main image processing
	dstOffset = dec;
	for( j=0; j<(INT32)(p_height-(p_filterSize-1)); j++ )
	{
		for( i=0; i<(INT32)(p_width-(p_filterSize-1)); i++ )
		{
			srcOffset = dstOffset-dec;
			acc=0;
			for( y=(INT32)p_filterSize-1; y>=0; y-- )
			{
				for( x=(INT32)p_filterSize-1; x>=0; x-- )
				{
					acc += p_src[srcOffset++];
				}
				srcOffset += dec2;
			}
			if( acc >= midVal )
				p_dst[dstOffset++] = 255;
			else
				p_dst[dstOffset++] = 0;
		}
		dstOffset += p_filterSize-1;
	}

	midVal  = 255*halfSize*halfSize/2;
	
	// border processing
	//top
	for( j=0; j<(INT32)halfSize; j++ )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			acc=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
						acc += p_src[y*p_width+x];
				}
			}
			if( acc >= midVal )
				p_dst[j*p_width+i] = 255;
			else
				p_dst[j*p_width+i] = 0;
		}
	}
	//bottoom
	for( j=(INT32)p_height-1; j>(INT32)p_height-1-(INT32)halfSize; j-- )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			acc=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
					{
						acc += p_src[y*p_width+x];
					}
				}
			}
			if( acc >= midVal )
				p_dst[j*p_width+i] = 255;
			else
				p_dst[j*p_width+i] = 0;
		}
	}
	// left
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=0; i<(INT32)halfSize; i++ )
		{
			// top
			acc=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
					{
						acc += p_src[y*p_width+x];
					}
				}
			}
			if( acc >= midVal )
				p_dst[j*p_width+i] = 255;
			else
				p_dst[j*p_width+i] = 0;
		}
	}
	
	// right
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=(INT32)p_width-1; i>(INT32)p_width-1-(INT32)halfSize; i-- )
		{
			// top
			acc=0;
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
					{
						acc += p_src[y*p_width+x];
					}
				}
			}
			if( acc >= midVal )
				p_dst[j*p_width+i] = 255;
			else
				p_dst[j*p_width+i] = 0;
		}
	}
	
	return ZEF_OK;
}

// erosion
INT32 erosionFilter32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize )
{
	UINT32		min;
	INT32		i,j;
	INT32		x,y;
	UINT32		dec,dec2,halfSize;
	UINT32		srcOffset,dstOffset;

	// set some helpful constants
	halfSize	= p_filterSize/2;
	dec		= halfSize*p_width+halfSize;
	dec2	= p_width-p_filterSize;


	// main image processing
	min		=INT_MAX;
	dstOffset = dec;
	for( j=0; j<(INT32)(p_height-(p_filterSize-1)); j++ )
	{
		for( i=0; i<(INT32)(p_width-(p_filterSize-1)); i++ )
		{
			srcOffset = dstOffset-dec;
			for( y=0; y<(INT32)p_filterSize; y++ )
			{
				for( x=0; x<(INT32)p_filterSize; x++ )
				{
					if( p_src[srcOffset] < min)
						min = p_src[srcOffset];
					srcOffset++;
				}
				srcOffset += dec2;
			}
			p_dst[dstOffset++] = min;
		}
		dstOffset += p_filterSize-1;
	}

	// border processing
	//top
	min		=INT_MAX;
	for( j=0; j<(INT32)halfSize; j++ )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
					{
						if( p_src[y*p_width+x] < min)
							min = p_src[y*p_width+x];
					}
				}
			}
			p_dst[j*p_width+i] = min;
		}
	}
	//bottoom
	min		=INT_MAX;
	for( j=(INT32)p_height-1; j>(INT32)p_height-1-(INT32)halfSize; j-- )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			// top
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
					{
						if( p_src[y*p_width+x] < min)
							min = p_src[y*p_width+x];
					}
				}
			}
			p_dst[j*p_width+i] = min;
		}
	}
	// left
	min		=INT_MAX;
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=0; i<(INT32)halfSize; i++ )
		{
			// top
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
					{
						if( p_src[y*p_width+x] < min)
							min = p_src[y*p_width+x];
					}
				}
			}
			p_dst[j*p_width+i] = min;
		}
	}

	// right
	min		=INT_MAX;
	for( j=0; j<(INT32)p_height; j++ )
	{
		for( i=(INT32)p_width-1; i>(INT32)p_width-1-(INT32)halfSize; i-- )
		{
			// top
			for( y=j-(INT32)halfSize; y<=j+(INT32)halfSize; y++ )
			{
				for( x=i-(INT32)halfSize; x<=i+(INT32)halfSize; x++ )
				{
					if( (UINT32)y<p_height && (UINT32)x<p_width )
					{
						if( p_src[y*p_width+x] < min)
							min = p_src[y*p_width+x];
					}
				}
			}
			p_dst[j*p_width+i] = min;
		}
	}

	return ZEF_OK;
	
}

#ifndef ZEF_FIXED_POINT

// gabor filter genereration in Fourier space
INT32 generateGaborFFT( REAL32 *p_filter, REAL32 p_size, REAL32 p_freq, REAL32 p_angle )
{
	REAL32		sTheta,cTheta,xc,yc;
	REAL32		amp,xp,yp,G;
	UINT32		i,j,j2;
	REAL32		xpTab[FFT_BLOC_SIZE];
	
	// initializations
	// parameters
	p_freq *= FFT_BLOC_SIZE;
	sTheta = (REAL32)sin(p_angle); 
	cTheta = (REAL32)cos(p_angle);
	p_size *= p_size;
	xc = FFT_BLOC_SIZE/2.0f+p_freq*sTheta;
	yc = FFT_BLOC_SIZE/2.0f+p_freq*cTheta;
	amp = 2*p_size*PI*PI/(FFT_BLOC_SIZE*FFT_BLOC_SIZE);

	// filter
	for(i=0; i<FFT_BLOC_SIZE*FFT_BLOC_SIZE/2; i++ )
		p_filter[i] = 0;

	// speedup for xp processing
	for( i=0; i<FFT_BLOC_SIZE; i++ )
		xpTab[i] = amp*(((i+FFT_BLOC_SIZE/2)%FFT_BLOC_SIZE)-xc)*(((i+FFT_BLOC_SIZE/2)%FFT_BLOC_SIZE)-xc);

	// now generate the FFT gabor filter
	for( j=0; j<FFT_BLOC_SIZE; j++ )
	{
		j2 = (j+FFT_BLOC_SIZE/2)&(FFT_BLOC_SIZE-1);
		yp = amp*(j-yc)*(j-yc);
		for( i=0; i<FFT_BLOC_SIZE/2; i++ )
		{
			xp = xpTab[i];

			G = xp+yp;
			// speedup for exponential processing
			if( G < 8.0f )
				G = FFT_BLOC_SIZE*FFT_BLOC_SIZE*zInvExp(G);
			else
				G = 0.0f;
			
			p_filter[j2*(FFT_BLOC_SIZE/2)+i] = G;
		}
		for( i=FFT_BLOC_SIZE/2; i<FFT_BLOC_SIZE; i++ )
		{
			xp = xpTab[i];

			G = xp+yp;
			// speedup for exponential processing
			if( G < 8.0f )
				G = FFT_BLOC_SIZE*FFT_BLOC_SIZE*zInvExp(G);
			else
				G = 0.0f;
			
			p_filter[(FFT_BLOC_SIZE-1-j2)*(FFT_BLOC_SIZE/2)+FFT_BLOC_SIZE-1-i] += G;
		}
	}

	

	return ZEF_OK;
}

// Time/freq space convolution with a real filter (non imaginary part)
INT32 freqConvolution( REAL32 *p_src, REAL32 *p_dst, REAL32 *p_filter )
{
	UINT32		i,j;

	j=0;
	// perform the convolution
	// since the filter is real we only have : 
	// dest = filterRe*(srcRe+i*srcIm) with i the imaginary operator (i*i=-1)
	for( i=0; i<FFT_BLOC_SIZE*FFT_BLOC_SIZE; i+=2 )
	{
		p_dst[i] = p_src[i]*p_filter[j];
		p_dst[i+1] = p_src[i+1]*p_filter[j];
		j++;
	}

	return ZEF_OK;
}

INT32 sobelGradient( UINT8 *p_src, INT16 *p_gradient, UINT32 p_width, UINT32 p_height )
{
	UINT32		offset;
	UINT32		y,x;
	INT16		val;

	for( x=0; x<2*p_width; x++ )
		p_gradient[x] = 0;

	offset = p_width+1;
	for( y=1; y<p_height-1; y++ )
	{	
		p_gradient[(offset-1)<<1]=0;
		p_gradient[((offset-1)<<1)+1]=0;
		for( x=1; x<p_width-1; x++ )
		{
			val = zReal2Int32(SQRT2*((INT16)p_src[offset+1]-(INT16)p_src[offset-1]));
			p_gradient[offset<<1] = val+(p_src[offset-p_width+1]+p_src[offset+p_width+1])-
									(p_src[offset-p_width-1]+p_src[offset+p_width-1]); //dx 
			val = zReal2Int32(SQRT2*((INT16)p_src[offset-p_width]-(INT16)p_src[offset+p_width]));
			p_gradient[(offset<<1)+1] = val+(p_src[offset-p_width+1]+p_src[offset-p_width-1])-
									(p_src[offset+p_width+1]+p_src[offset+p_width-1]); //dy
			offset++;
		}
		p_gradient[offset<<1]=0;
		p_gradient[(offset<<1)+1]=0;
		offset+=2;
	}
	offset--;
	offset<<=1;
	for( x=0; x<2*p_width; x++ )
		p_gradient[offset+x] = 0;

	return ZEF_OK;
}

#endif// ZEF_FIXED_POINT

INT32 poincare( UINT8 *p_src, INT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_filterSize )
{
	UINT32	x,y,i,x1,y1;
	INT32	val;
	INT8	angle; // must be 8 bits because of angles
	UINT32	size2;
	UINT32	offset;

	//init
	size2 = p_filterSize/2;

	// init dest image
	for( i=0; i<p_width*p_height; i++ )
		p_dst[i] = 0;	
	
	for( y=size2; y<p_height-size2; y++ )
	{
		for( x=size2; x<p_width-size2; x++ )
		{
			val = 0;
			// top raw
			for( x1=x-size2+1; x1<=x+size2; x1++ )
			{
				offset = (y-size2)*p_width+x1;
				angle = p_src[offset]-p_src[offset-1];
				// no need to check angles if angle = 8bits
				//if( angle <= -256/2 ) angle += 256;
				//if( angle >= 256/2 ) angle -= 256;
				val += angle;
			}

			// right column
			for( y1=y-size2+1; y1<=y+size2; y1++ )
			{
				offset = y1*p_width+x+size2;
				angle = p_src[offset]-p_src[offset-p_width];
				//if( angle <= -256/2 ) angle += 256;
				//if( angle >= 256/2 ) angle -= 256;
				val += angle;
			}

			// bottom raw
			for( x1=x-size2; x1<x+size2; x1++ )
			{
				offset = (y+size2)*p_width+x1;
				angle = p_src[offset]-p_src[offset+1];
				//if( angle <= -256/2 ) angle += 256;
				//if( angle >= 256/2 ) angle -= 256;
				val += angle;
			}

			// left column
			for( y1=y-size2; y1<y+size2; y1++ )
			{
				offset = y1*p_width+x-size2;
				angle = p_src[offset]-p_src[offset+p_width];
				//if( angle <= -256/2 ) angle += 256;
				//if( angle >= 256/2 ) angle -= 256;
				val += angle;
			}
			val = 127*val/256;
			p_dst[y*p_width+x] = val;
		}
	}
	

	return ZEF_OK;
}
