/*
Company :		ZEFYR Technologies
File :			fxFFT1D.c
Author :		Matthieu Perreira Da Silva (based on Takuya OOURA (email: ooura@kurims.kyoto-u.ac.jp) algorithms) 
Fisrt release :	15/04/2003

Description : Fixed PoINT32 Fast Fourier Transform

Features : 
    dimension   :one
    data length :power of 2
    decimation  :frequency
    radix       :split-radix
    data        :inplace
    table       :use

Functions : 

    fxcdft: Complex Discrete Fourier Transform
    fxrdft: Real Discrete Fourier Transform

Function prototypes : 

    void fxcdft(INT32 p_size, INT32 p_sign, FIXED16 *p_data, INT32 *p_inPlace, FIXED16 *p_trigo);
	void fxrdft(INT32 p_size, INT32 p_sign, FIXED16 *p_data, INT32 *p_inPlace, FIXED16 *p_trigo);

Comments : see file FFTs.txt

Necessary package : none

Appendix :
    The cos/sin table is recalculated when the larger table required.
    w[] and ip[] are compatible with all routines.
*/

#include <math.h>

#include "fxFFT1D.h"

#if defined(ZEF_FIXED_POINT) && defined(ZEF_SPLIT_RADIX)

void fxcdft(INT32 n, INT32 isgn, FIXED16 *a, INT32 *ip, FIXED16 *w)
{
    INT32 nw;
    
    nw = ip[0];
    if (n > (nw << 2)) {
        nw = n >> 2;
        fxmakewt(nw, ip, w);
    }
    if (isgn >= 0) {
        fxcftfsub(n, a, ip, nw, w);
    } else {
        fxcftbsub(n, a, ip, nw, w);
    }
}


void fxrdft(INT32 n, INT32 isgn, FIXED16 *a, INT32 *ip, FIXED16 *w)
{
    INT32 nw, nc;
    FIXED32 xi;
    
    nw = ip[0];
    if (n > (nw << 2)) {
        nw = n >> 2;
        fxmakewt(nw, ip, w);
    }
    nc = ip[1];
    if (n > (nc << 2)) {
        nc = n >> 2;
        fxmakect(nc, ip, w + nw);
    }
    if (isgn >= 0) {
        /*if (n > 4) {*/
            fxcftfsub(n, a, ip, nw, w);
            fxrftfsub(n, a, nc, w + nw);
        /*} else if (n == 4) {
            fxcftfsub(n, a, ip, nw, w);
        }*/
        xi = FSUB(a[0],a[1]); // a[0] - a[1];
        a[0] = FADD(a[0],a[1]); // += a[1];
        a[1] = xi;
    } else {
        a[1] = FSUB(a[0],a[1])>>1; // 0.5f * (a[0] - a[1]);
        a[0] = FSUB(a[0],a[1]);  // -= a[1];
        /*if (n > 4) {*/
            fxrftbsub(n, a, nc, w + nw);
            fxcftbsub(n, a, ip, nw, w);
        /*} else if (n == 4) {
            fxcftbsub(n, a, ip, nw, w);
        }*/
    }
}



/* -------- initializing routines -------- */


void fxmakewt(INT32 nw, INT32 *ip, FIXED16 *w)
{
#if !defined(USE_FX_PRECOMP)
    INT32		j, nwh, nw0, nw1;
    REAL32	delta;
	FIXED32	wn4r, wk1r, wk1i, wk3r, wk3i;
    
    ip[0] = nw;
    ip[1] = 1;
    if (nw > 2) {
        nwh = nw >> 1;
        delta = (REAL32)atan(1.0) / nwh;
        wn4r = FTOFIX32(cos(delta * nwh),Q1_14);
        w[0] = ITOFIX16(1,Q1_14);
        w[1] = wn4r;
        if (nwh == 4) {
            w[2] = FTOFIX16((REAL32)cos(delta * 2),Q1_14);
            w[3] = FTOFIX16((REAL32)sin(delta * 2),Q1_14);
        } else if (nwh > 4) {
            fxmakeipt(nw, ip);
            w[2] = FTOFIX16(0.5f / (REAL32)cos(delta * 2),Q1_14);
            w[3] = FTOFIX16(0.5f / (REAL32)cos(delta * 6),Q1_14);
            for (j = 4; j < nwh; j += 4) {
                w[j] = FTOFIX16((REAL32)cos(delta * j),Q1_14);
                w[j + 1] = FTOFIX16(sin(delta * j),Q1_14);
                w[j + 2] = FTOFIX16(cos(3 * delta * j),Q1_14);
                w[j + 3] = FTOFIX16(-sin(3 * delta * j),Q1_14);
            }
        }
        nw0 = 0;
        while (nwh > 2) {
            nw1 = nw0 + nwh;
            nwh >>= 1;
            w[nw1] = ITOFIX16(1,Q1_14);
            w[nw1 + 1] = wn4r;
            if (nwh == 4) {
                wk1r = w[nw0 + 4];
                wk1i = w[nw0 + 5];
                w[nw1 + 2] = wk1r;
                w[nw1 + 3] = wk1i;
            } else if (nwh > 4) {
                wk1r = w[nw0 + 4];
                wk3r = w[nw0 + 6];
                w[nw1 + 2] = FDIV(FTOFIX16(0.5f,Q1_14),wk1r,Q1_14);
                w[nw1 + 3] = FDIV(FTOFIX16(0.5f,Q1_14),wk3r,Q1_14);
                for (j = 4; j < nwh; j += 4) {
                    wk1r = w[nw0 + 2 * j];
                    wk1i = w[nw0 + 2 * j + 1];
                    wk3r = w[nw0 + 2 * j + 2];
                    wk3i = w[nw0 + 2 * j + 3];
                    w[nw1 + j] = wk1r;
                    w[nw1 + j + 1] = wk1i;
                    w[nw1 + j + 2] = wk3r;
                    w[nw1 + j + 3] = wk3i;
                }
            }
            nw0 = nw1;
        }
    }
#endif // USE_FX_PRECOMP
}


void fxmakeipt(INT32 nw, INT32 *ip)
{
    INT32 j, l, m, m2, p, q;
    
    ip[2] = 0;
    ip[3] = 16;
    m = 2;
    for (l = nw; l > 32; l >>= 2) {
        m2 = m << 1;
        q = m2 << 3;
        for (j = m; j < m2; j++) {
            p = ip[j] << 2;
            ip[m + j] = p;
            ip[m2 + j] = p + q;
        }
        m = m2;
    }
}


void fxmakect(INT32 nc, INT32 *ip, FIXED16 *c)
{
#if !defined(USE_FX_PRECOMP)
    INT32 j, nch;
    REAL32 delta;
    
    ip[1] = nc;
    if (nc > 1) {
        nch = nc >> 1;
        delta = (REAL32)atan(1.0) / nch;
        c[0] = FTOFIX16(cos(delta * nch),Q1_14);
        c[nch] = c[0]>>1; // 0.5f * c[0];
        for (j = 1; j < nch; j++) {
            c[j] = FTOFIX16(cos(delta * j),Q1_14)>>1; // 0.5f * (FIXED16)cos(delta * j);
            c[nc - j] = FTOFIX16(sin(delta * j),Q1_14)>>1; // 0.5f * (FIXED16)sin(delta * j);
        }
    }
#endif
}


/* -------- child routines -------- */



void fxcftfsub(INT32 n, FIXED16 *a, INT32 *ip, INT32 nw, FIXED16 *w)
{  
    /*if (n > 8) {
        if (n > 32) {*/
            fxcftf1st(n, a, &w[nw - (n >> 2)]);
            /*if (n > 512) {
                fxcftrec4(n, a, nw, w);
            } else if (n > 128) {
                fxcftleaf(n, 1, a, nw, w);
            } else {*/
                fxcftfx41(n, a, nw, w);
            /*}*/
            fxbitrv2(n, ip, a);
        /*} else if (n == 32) {
            fxcftf161(a, &w[nw - 8]);
            fxbitrv216(a);
        } else {
            fxcftf081(a, w);
            fxbitrv208(a);
        }
    } else if (n == 8) {
        fxcftf040(a);
    } else if (n == 4) {
        fxcftx020(a);
    }*/
}


void fxcftbsub(INT32 n, FIXED16 *a, INT32 *ip, INT32 nw, FIXED16 *w)
{  
	/*if (n > 8) {
        if (n > 32) {*/
            fxcftb1st(n, a, &w[nw - (n >> 2)]);
            /*if (n > 512) {
                fxcftrec4(n, a, nw, w);
            } else if (n > 128) {
                fxcftleaf(n, 1, a, nw, w);
            } else {*/
                fxcftfx41(n, a, nw, w);
            /*}*/
            fxbitrv2conj(n, ip, a);
        /*} else if (n == 32) {
            fxcftf161(a, &w[nw - 8]);
            fxbitrv216neg(a);
        } else {
            fxcftf081(a, w);
            fxbitrv208neg(a);
        }
    } else if (n == 8) {
        fxcftb040(a);
    } else if (n == 4) {
        fxcftx020(a);
    }*/
}


void fxbitrv2(INT32 n, INT32 *ip, FIXED16 *a)
{
    INT32 j, j1, k, k1, l, m, nh, nm;
    FIXED32 xr, xi, yr, yi;
    
    m = 1;
    for (l = n >> 2; l > 8; l >>= 2) {
        m <<= 1;
    }
    nh = n >> 1;
    nm = 4 * m;
    if (l == 8) {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 4 * j + 2 * ip[m + k];
                k1 = 4 * k + 2 * ip[m + j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 -= nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nh;
                k1 += 2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 += nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += 2;
                k1 += nh;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 -= nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nh;
                k1 -= 2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 += nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 4 * k + 2 * ip[m + k];
            j1 = k1 + 2;
            k1 += nh;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 += nm;
            k1 += 2 * nm;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 += nm;
            k1 -= nm;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 -= 2;
            k1 -= nh;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 += nh + 2;
            k1 += nh + 2;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 -= nh - nm;
            k1 += 2 * nm - 2;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
        }
    } else {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 4 * j + ip[m + k];
                k1 = 4 * k + ip[m + j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nh;
                k1 += 2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += 2;
                k1 += nh;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nh;
                k1 -= 2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= nm;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 4 * k + ip[m + k];
            j1 = k1 + 2;
            k1 += nh;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 += nm;
            k1 += nm;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
        }
    }
}


void fxbitrv2conj(INT32 n, INT32 *ip, FIXED16 *a)
{
    INT32 j, j1, k, k1, l, m, nh, nm;
    FIXED32 xr, xi, yr, yi;
    
    m = 1;
    for (l = n >> 2; l > 8; l >>= 2) {
        m <<= 1;
    }
    nh = n >> 1;
    nm = 4 * m;
    if (l == 8) {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 4 * j + 2 * ip[m + k];
                k1 = 4 * k + 2 * ip[m + j];
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 -= nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nh;
                k1 += 2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 += nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += 2;
                k1 += nh;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 -= nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nh;
                k1 -= 2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 += nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= 2 * nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 4 * k + 2 * ip[m + k];
            j1 = k1 + 2;
            k1 += nh;
            a[j1 - 1] = -a[j1 - 1];
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            a[k1 + 3] = -a[k1 + 3];
            j1 += nm;
            k1 += 2 * nm;
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 += nm;
            k1 -= nm;
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 -= 2;
            k1 -= nh;
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 += nh + 2;
            k1 += nh + 2;
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            j1 -= nh - nm;
            k1 += 2 * nm - 2;
            a[j1 - 1] = -a[j1 - 1];
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            a[k1 + 3] = -a[k1 + 3];
        }
    } else {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 4 * j + ip[m + k];
                k1 = 4 * k + ip[m + j];
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nh;
                k1 += 2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += 2;
                k1 += nh;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += nm;
                k1 += nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nh;
                k1 -= 2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 -= nm;
                k1 -= nm;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 4 * k + ip[m + k];
            j1 = k1 + 2;
            k1 += nh;
            a[j1 - 1] = -a[j1 - 1];
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            a[k1 + 3] = -a[k1 + 3];
            j1 += nm;
            k1 += nm;
            a[j1 - 1] = -a[j1 - 1];
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            a[k1 + 3] = -a[k1 + 3];
        }
    }
}


/*void fxbitrv216(FIXED16 *a)
{
    FIXED32 x1r, x1i, x2r, x2i, x3r, x3i, x4r, x4i, 
        x5r, x5i, x7r, x7i, x8r, x8i, x10r, x10i, 
        x11r, x11i, x12r, x12i, x13r, x13i, x14r, x14i;
    
    x1r = a[2];
    x1i = a[3];
    x2r = a[4];
    x2i = a[5];
    x3r = a[6];
    x3i = a[7];
    x4r = a[8];
    x4i = a[9];
    x5r = a[10];
    x5i = a[11];
    x7r = a[14];
    x7i = a[15];
    x8r = a[16];
    x8i = a[17];
    x10r = a[20];
    x10i = a[21];
    x11r = a[22];
    x11i = a[23];
    x12r = a[24];
    x12i = a[25];
    x13r = a[26];
    x13i = a[27];
    x14r = a[28];
    x14i = a[29];
    a[2] = x8r;
    a[3] = x8i;
    a[4] = x4r;
    a[5] = x4i;
    a[6] = x12r;
    a[7] = x12i;
    a[8] = x2r;
    a[9] = x2i;
    a[10] = x10r;
    a[11] = x10i;
    a[14] = x14r;
    a[15] = x14i;
    a[16] = x1r;
    a[17] = x1i;
    a[20] = x5r;
    a[21] = x5i;
    a[22] = x13r;
    a[23] = x13i;
    a[24] = x3r;
    a[25] = x3i;
    a[26] = x11r;
    a[27] = x11i;
    a[28] = x7r;
    a[29] = x7i;
}*/


/*void fxbitrv216neg(FIXED16 *a)
{
    FIXED32 x1r, x1i, x2r, x2i, x3r, x3i, x4r, x4i, 
        x5r, x5i, x6r, x6i, x7r, x7i, x8r, x8i, 
        x9r, x9i, x10r, x10i, x11r, x11i, x12r, x12i, 
        x13r, x13i, x14r, x14i, x15r, x15i;
    
    x1r = a[2];
    x1i = a[3];
    x2r = a[4];
    x2i = a[5];
    x3r = a[6];
    x3i = a[7];
    x4r = a[8];
    x4i = a[9];
    x5r = a[10];
    x5i = a[11];
    x6r = a[12];
    x6i = a[13];
    x7r = a[14];
    x7i = a[15];
    x8r = a[16];
    x8i = a[17];
    x9r = a[18];
    x9i = a[19];
    x10r = a[20];
    x10i = a[21];
    x11r = a[22];
    x11i = a[23];
    x12r = a[24];
    x12i = a[25];
    x13r = a[26];
    x13i = a[27];
    x14r = a[28];
    x14i = a[29];
    x15r = a[30];
    x15i = a[31];
    a[2] = x15r;
    a[3] = x15i;
    a[4] = x7r;
    a[5] = x7i;
    a[6] = x11r;
    a[7] = x11i;
    a[8] = x3r;
    a[9] = x3i;
    a[10] = x13r;
    a[11] = x13i;
    a[12] = x5r;
    a[13] = x5i;
    a[14] = x9r;
    a[15] = x9i;
    a[16] = x1r;
    a[17] = x1i;
    a[18] = x14r;
    a[19] = x14i;
    a[20] = x6r;
    a[21] = x6i;
    a[22] = x10r;
    a[23] = x10i;
    a[24] = x2r;
    a[25] = x2i;
    a[26] = x12r;
    a[27] = x12i;
    a[28] = x4r;
    a[29] = x4i;
    a[30] = x8r;
    a[31] = x8i;
}*/


/*void fxbitrv208(FIXED16 *a)
{
    FIXED32 x1r, x1i, x3r, x3i, x4r, x4i, x6r, x6i;
    
    x1r = a[2];
    x1i = a[3];
    x3r = a[6];
    x3i = a[7];
    x4r = a[8];
    x4i = a[9];
    x6r = a[12];
    x6i = a[13];
    a[2] = x4r;
    a[3] = x4i;
    a[6] = x6r;
    a[7] = x6i;
    a[8] = x1r;
    a[9] = x1i;
    a[12] = x3r;
    a[13] = x3i;
}*/


/*void fxbitrv208neg(FIXED16 *a)
{
    FIXED32 x1r, x1i, x2r, x2i, x3r, x3i, x4r, x4i, 
        x5r, x5i, x6r, x6i, x7r, x7i;
    
    x1r = a[2];
    x1i = a[3];
    x2r = a[4];
    x2i = a[5];
    x3r = a[6];
    x3i = a[7];
    x4r = a[8];
    x4i = a[9];
    x5r = a[10];
    x5i = a[11];
    x6r = a[12];
    x6i = a[13];
    x7r = a[14];
    x7i = a[15];
    a[2] = x7r;
    a[3] = x7i;
    a[4] = x3r;
    a[5] = x3i;
    a[6] = x5r;
    a[7] = x5i;
    a[8] = x1r;
    a[9] = x1i;
    a[10] = x6r;
    a[11] = x6i;
    a[12] = x2r;
    a[13] = x2i;
    a[14] = x4r;
    a[15] = x4i;
}*/


void fxcftf1st(INT32 n, FIXED16 *a, FIXED16 *w)
{
    INT32 j, j0, j1, j2, j3, k, m, mh;
    FIXED32 wn4r, csc1, csc3, wk1r, wk1i, wk3r, wk3i, 
        wd1r, wd1i, wd3r, wd3i;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i;
    
    mh = n >> 3;
    m = 2 * mh;
    j1 = m;
    j2 = j1 + m;
    j3 = j2 + m;		
    x0r = FADD(a[0],a[j2]);				// a[0] + a[j2];	
    x0i = FADD(a[1],a[j2 + 1]);			// a[1] + a[j2 + 1];
    x1r = FSUB(a[0],a[j2]);				// a[0] - a[j2];
    x1i = FSUB(a[1],a[j2 + 1]);			// a[1] - a[j2 + 1];
    x2r = FADD(a[j1],a[j3]);			// a[j1] + a[j3];
    x2i = FADD(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] + a[j3 + 1];
    x3r = FSUB(a[j1],a[j3]);			// a[j1] - a[j3];
    x3i = FSUB(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] - a[j3 + 1];
    a[0] = FADD(x0r,x2r)>>1;				// x0r + x2r;
    a[1] = FADD(x0i,x2i)>>1;				// x0i + x2i;
    a[j1] = FSUB(x0r,x2r)>>1;				// x0r - x2r;
    a[j1 + 1] = FSUB(x0i,x2i)>>1;			// x0i - x2i;
    a[j2] = FSUB(x1r,x3i)>>1;				// x1r - x3i;
    a[j2 + 1] = FADD(x1i,x3r)>>1;			// x1i + x3r;
    a[j3] = FADD(x1r,x3i)>>1;				// x1i + x3r;
    a[j3 + 1] = FSUB(x1i,x3r)>>1;			// x1i - x3r;
    wn4r = w[1];					// w[1];
    csc1 = w[2];					// w[2];
    csc3 = w[3];					// w[3];
    wd1r = FTOFIX16(1,Q1_14);						// 1;
    wd1i = FTOFIX16(0,Q1_14);						// 0;
    wd3r = FTOFIX16(1,Q1_14);						// 1;
    wd3i = FTOFIX16(0,Q1_14);						// 0;
    k = 0;
    for (j = 2; j < mh - 2; j += 4) {
        k += 4;
        wk1r = FMUL(csc1,FADD(wd1r,w[k]),Q1_14);	// csc1 * (wd1r + w[k]);
        wk1i = FMUL(csc1,FADD(wd1i,w[k + 1]),Q1_14);	// csc1 * (wd1i + w[k + 1]);
        wk3r = FMUL(csc3,FADD(wd3r,w[k + 2]),Q1_14);	// csc3 * (wd3r + w[k + 2]);
        wk3i = FMUL(csc3,FADD(wd3i,w[k + 3]),Q1_14);	// csc3 * (wd3i + w[k + 3]);
        wd1r = w[k];					// w[k];
        wd1i = w[k + 1];				// w[k + 1];
        wd3r = w[k + 2];				// w[k + 2];
        wd3i = w[k + 3];				// w[k + 3];
        j1 = j + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FADD(a[j],a[j2]);				// a[j] + a[j2];
        x0i = FADD(a[j + 1],a[j2 + 1]);		// a[j + 1] + a[j2 + 1];
        x1r = FSUB(a[j],a[j2]);				// a[j] - a[j2];	
        x1i = FSUB(a[j + 1],a[j2 + 1]);		// a[j + 1] - a[j2 + 1];	
        y0r = FADD(a[j + 2], a[j2 + 2]);		// a[j + 2] + a[j2 + 2];
        y0i = FADD(a[j + 3], a[j2 + 3]);		// a[j + 3] + a[j2 + 3];
        y1r = FSUB(a[j + 2],a[j2 + 2]);		// a[j + 2] - a[j2 + 2];
        y1i = FSUB(a[j + 3],a[j2 + 3]);		// a[j + 3] - a[j2 + 3];	
        x2r = FADD(a[j1], a[j3]);			// a[j1] + a[j3];
        x2i = FADD(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] + a[j3 + 1];
        x3r = FSUB(a[j1], a[j3]);			// a[j1] - a[j3];
        x3i = FSUB(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] - a[j3 + 1];
        y2r = FADD(a[j1 + 2],a[j3 + 2]);	// a[j1 + 2] + a[j3 + 2];
        y2i = FADD(a[j1 + 3], a[j3 + 3]);	// a[j1 + 3] + a[j3 + 3];
        y3r = FSUB(a[j1 + 2], a[j3 + 2]);	// a[j1 + 2] - a[j3 + 2];
        y3i = FSUB(a[j1 + 3],a[j3 + 3]);	// a[j1 + 3] - a[j3 + 3];
        a[j] = FADD(x0r,x2r)>>1;				// x0r + x2r;	
        a[j + 1] = FADD(x0i,x2i)>>1;			// x0i + x2i;	
        a[j + 2] = FADD(y0r,y2r)>>1;			// y0r + y2r;	
        a[j + 3] = FADD(y0i,y2i)>>1;			// y0i + y2i;
        a[j1] = FSUB(x0r,x2r)>>1;				// x0r - x2r;
        a[j1 + 1] = FSUB(x0i,x2i)>>1;			// x0i - x2i;
        a[j1 + 2] = FSUB(y0r, y2r)>>1;			// y0r - y2r;
        a[j1 + 3] = FSUB(y0i,y2i)>>1;			// y0i - y2i;
        x0r = FSUB(x1r,x3i);				// x1r - x3i;
        x0i = FADD(x1i,x3r);				// x1i + x3r;
        a[j2] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14))>>1;		// wk1r * x0r - wk1i * x0i;
        a[j2 + 1] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14))>>1;	// wk1r * x0i + wk1i * x0r;
        x0r = FSUB(y1r,y3i);				// y1r - y3i;
        x0i = FADD(y1i,y3r);				// y1i + y3r;
        a[j2 + 2] = FSUB(FMUL(wd1r,x0r,Q1_14),FMUL(wd1i,x0i,Q1_14))>>1;	// wd1r * x0r - wd1i * x0i;
        a[j2 + 3] = FADD(FMUL(wd1r,x0i,Q1_14),FMUL(wd1i,x0r,Q1_14))>>1;	// wd1r * x0i + wd1i * x0r;
        x0r = FADD(x1r,x3i);				// x1r + x3i;
        x0i = FSUB(x1i,x3r);				// x1i - x3r;
        a[j3] = FADD(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14))>>1;		// wk3r * x0r + wk3i * x0i;
        a[j3 + 1] = FSUB(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14))>>1;	// wk3r * x0i - wk3i * x0r;
        x0r = FADD(y1r,y3i);				// y1r + y3i;	
        x0i = FSUB(y1i,y3r);				// y1i - y3r;
        a[j3 + 2] = FADD(FMUL(wd3r,x0r,Q1_14),FMUL(wd3i,x0i,Q1_14))>>1;	// wd3r * x0r + wd3i * x0i;
        a[j3 + 3] = FSUB(FMUL(wd3r,x0i,Q1_14),FMUL(wd3i,x0r,Q1_14))>>1;	// wd3r * x0i - wd3i * x0r;
        j0 = m - j;
        j1 = j0 + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FADD(a[j0],a[j2]);			// a[j0] + a[j2];	
        x0i = FADD(a[j0 + 1],a[j2 + 1]);	// a[j0 + 1] + a[j2 + 1];
        x1r = FSUB(a[j0],a[j2]);			// a[j0] - a[j2];
        x1i = FSUB(a[j0 + 1],a[j2 + 1]);	// a[j0 + 1] - a[j2 + 1];
        y0r = FADD(a[j0 - 2],a[j2 - 2]);	// a[j0 - 2] + a[j2 - 2];
        y0i = FADD(a[j0 - 1],a[j2 - 1]);	// a[j0 - 1] + a[j2 - 1];
        y1r = FSUB(a[j0 - 2],a[j2 - 2]);	// a[j0 - 2] - a[j2 - 2];
        y1i = FSUB(a[j0 - 1],a[j2 - 1]);	// a[j0 - 1] - a[j2 - 1];
        x2r = FADD(a[j1],a[j3]);			// a[j1] + a[j3];
        x2i = FADD(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] + a[j3 + 1];
        x3r = FSUB(a[j1],a[j3]);			// a[j1] - a[j3];
        x3i = FSUB(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] - a[j3 + 1];
        y2r = FADD(a[j1 - 2],a[j3 - 2]);	// a[j1 - 2] + a[j3 - 2];
        y2i = FADD(a[j1 - 1],a[j3 - 1]);	// a[j1 - 1] + a[j3 - 1];
        y3r = FSUB(a[j1 - 2],a[j3 - 2]);	// a[j1 - 2] - a[j3 - 2];
        y3i = FSUB(a[j1 - 1],a[j3 - 1]);	// a[j1 - 1] - a[j3 - 1];
        a[j0] = FADD(x0r,x2r)>>1;				// x0r + x2r;	
        a[j0 + 1] = FADD(x0i,x2i)>>1;			// x0i + x2i;	
        a[j0 - 2] = FADD(y0r,y2r)>>1;			// y0r + y2r;	
        a[j0 - 1] = FADD(y0i,y2i)>>1;			// y0i + y2i;
        a[j1] = FSUB(x0r,x2r)>>1;				// x0r - x2r;		
        a[j1 + 1] = FSUB(x0i,x2i)>>1;			// x0i - x2i;	
        a[j1 - 2] = FSUB(y0r,y2r)>>1;			// y0r - y2r;	
        a[j1 - 1] = FSUB(y0i,y2i)>>1;			// y0i - y2i;	
        x0r = FSUB(x1r,x3i);				// x1r - x3i;
        x0i = FADD(x1i,x3r);				// x1i + x3r;
        a[j2] = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14))>>1;		// wk1i * x0r - wk1r * x0i;
        a[j2 + 1] = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14))>>1;	// wk1i * x0i + wk1r * x0r;
        x0r = FSUB(y1r,y3i);				// y1r - y3i;
        x0i = FADD(y1i,y3r);				// y1i + y3r;
        a[j2 - 2] = FSUB(FMUL(wd1i,x0r,Q1_14),FMUL(wd1r,x0i,Q1_14))>>1;	// wd1i * x0r - wd1r * x0i;
        a[j2 - 1] = FADD(FMUL(wd1i, x0i,Q1_14),FMUL(wd1r,x0r,Q1_14))>>1;	// wd1i * x0i + wd1r * x0r;
        x0r = FADD(x1r,x3i);				// x1r + x3i;
        x0i = FSUB(x1i,x3r);				// x1i - x3r;
        a[j3] = FADD(FMUL(wk3i,x0r,Q1_14),FMUL(wk3r, x0i,Q1_14))>>1;		// wk3i * x0r + wk3r * x0i;
        a[j3 + 1] = FSUB(FMUL(wk3i,x0i,Q1_14),FMUL(wk3r,x0r,Q1_14))>>1;	// wk3i * x0i - wk3r * x0r;
        x0r = FADD(y1r,y3i);				// y1r + y3i;
        x0i = FSUB(y1i,y3r);				// y1i - y3r;
        a[j3 - 2] = FADD(FMUL(wd3i,x0r,Q1_14),FMUL(wd3r,x0i,Q1_14))>>1;	// wd3i * x0r + wd3r * x0i;
        a[j3 - 1] = FSUB(FMUL(wd3i,x0i,Q1_14),FMUL(wd3r,x0r,Q1_14))>>1;	// wd3i * x0i - wd3r * x0r;
    }
    wk1r = FMUL(csc1,FADD(wd1r,wn4r),Q1_14);	// csc1 * (wd1r + wn4r);
    wk1i = FMUL(csc1,FADD(wd1i,wn4r),Q1_14);	// csc1 * (wd1i + wn4r);
    wk3r = FMUL(csc3,FSUB(wd3r,wn4r),Q1_14);	// csc3 * (wd3r - wn4r);
    wk3i = FMUL(csc3,FSUB(wd3i,wn4r),Q1_14);	// csc3 * (wd3i - wn4r);
    j0 = mh;						
    j1 = j0 + m;
    j2 = j1 + m;
    j3 = j2 + m;
    x0r = FADD(a[j0 - 2],a[j2 - 2]);	// a[j0 - 2] + a[j2 - 2];
    x0i = FADD(a[j0 - 1],a[j2 - 1]);	// a[j0 - 1] + a[j2 - 1];
    x1r = FSUB(a[j0 - 2],a[j2 - 2]);	// a[j0 - 2] - a[j2 - 2];
    x1i = FSUB(a[j0 - 1],a[j2 - 1]);	// a[j0 - 1] - a[j2 - 1];
    x2r = FADD(a[j1 - 2],a[j3 - 2]);	// a[j1 - 2] + a[j3 - 2];
    x2i = FADD(a[j1 - 1],a[j3 - 1]);	// a[j1 - 1] + a[j3 - 1];
    x3r = FSUB(a[j1 - 2],a[j3 - 2]);	// a[j1 - 2] - a[j3 - 2];
    x3i = FSUB(a[j1 - 1],a[j3 - 1]);	// a[j1 - 1] - a[j3 - 1];
    a[j0 - 2] = FADD(x0r,x2r)>>1;			// x0r + x2r;
    a[j0 - 1] = FADD(x0i,x2i)>>1;			// x0i + x2i;
    a[j1 - 2] = FSUB(x0r,x2r)>>1;			// x0r - x2r;
    a[j1 - 1] = FSUB(x0i,x2i)>>1;			// x0i - x2i;
    x0r = FSUB(x1r,x3i);				// x1r - x3i;
    x0i =FADD( x1i,x3r);				// x1i + x3r;
    a[j2 - 2] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14))>>1;	// wk1r * x0r - wk1i * x0i;
    a[j2 - 1] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14))>>1;	// wk1r * x0i + wk1i * x0r;
    x0r = FADD(x1r,x3i);				// x1r + x3i;
    x0i = FSUB(x1i,x3r);				// x1i - x3r;
    a[j3 - 2] = FADD(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14))>>1;	// wk3r * x0r + wk3i * x0i;
    a[j3 - 1] = FSUB(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14))>>1;	// wk3r * x0i - wk3i * x0r;
    x0r = FADD(a[j0],a[j2]);			// a[j0] + a[j2];
    x0i = FADD(a[j0 + 1],a[j2 + 1]);	// a[j0 + 1] + a[j2 + 1];
    x1r = FSUB(a[j0],a[j2]);			// a[j0] - a[j2];	
    x1i = FSUB(a[j0 + 1],a[j2 + 1]);	// a[j0 + 1] - a[j2 + 1];
    x2r = FADD(a[j1],a[j3]);			// a[j1] + a[j3];	
    x2i = FADD(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] + a[j3 + 1];
    x3r = FSUB(a[j1],a[j3]);			// a[j1] - a[j3];
    x3i = FSUB(a[j1 + 1],a[j3 + 1]);	// a[j1 + 1] - a[j3 + 1];
    a[j0] = FADD(x0r,x2r)>>1;				// x0r + x2r;	
    a[j0 + 1] = FADD(x0i,x2i)>>1;			// x0i + x2i;	
    a[j1] = FSUB(x0r,x2r)>>1;				// x0r - x2r;	
    a[j1 + 1] = FSUB(x0i,x2i)>>1;			// x0i - x2i;
    x0r = FSUB(x1r,x3i);				// x1r - x3i;
    x0i = FADD(x1i,x3r);				// x1i + x3r;
    a[j2] = FMUL(wn4r,FSUB(x0r,x0i),Q1_14)>>1;		// wn4r * (x0r - x0i);
    a[j2 + 1] = FMUL(wn4r,FADD(x0i,x0r),Q1_14)>>1;	// wn4r * (x0i + x0r);
    x0r = FADD(x1r,x3i);				// x1r + x3i;
    x0i = FSUB(x1i,x3r);				// x1i - x3r;
    a[j3] = FMUL(-wn4r,FADD(x0r,x0i),Q1_14)>>1;	// -wn4r * (x0r + x0i);
    a[j3 + 1] = FMUL(-wn4r,FSUB(x0i,x0r),Q1_14)>>1;	// -wn4r * (x0i - x0r);
    x0r = FADD(a[j0 + 2],a[j2 + 2]);	// a[j0 + 2] + a[j2 + 2];
    x0i = FADD(a[j0 + 3],a[j2 + 3]);	// a[j0 + 3] + a[j2 + 3];
    x1r = FSUB(a[j0 + 2],a[j2 + 2]);	// a[j0 + 2] - a[j2 + 2];
    x1i = FSUB(a[j0 + 3],a[j2 + 3]);	// a[j0 + 3] - a[j2 + 3];
    x2r = FADD(a[j1 + 2],a[j3 + 2]);	// a[j1 + 2] + a[j3 + 2];
    x2i = FADD(a[j1 + 3],a[j3 + 3]);	// a[j1 + 3] + a[j3 + 3];
    x3r = FSUB(a[j1 + 2],a[j3 + 2]);	// a[j1 + 2] - a[j3 + 2];
    x3i = FSUB(a[j1 + 3],a[j3 + 3]);	// a[j1 + 3] - a[j3 + 3];
    a[j0 + 2] = FADD(x0r,x2r)>>1;			// = x0r + x2r;
    a[j0 + 3] = FADD(x0i,x2i)>>1;			// = x0i + x2i;	
    a[j1 + 2] = FSUB(x0r,x2r)>>1;			// = x0r - x2r;
    a[j1 + 3] = FSUB(x0i,x2i)>>1;			// = x0i - x2i;
    x0r = FSUB(x1r,x3i);				// x1r - x3i;
    x0i = FADD(x1i,x3r);				// x1i + x3r;
    a[j2 + 2] = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14))>>1;	// wk1i * x0r - wk1r * x0i;
    a[j2 + 3] = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14))>>1;	// wk1i * x0i + wk1r * x0r;
    x0r = FADD(x1r,x3i);				// x1r + x3i;
    x0i =FSUB( x1i,x3r);				// x1i - x3r;
    a[j3 + 2] = FADD(FMUL(wk3i,x0r,Q1_14),FMUL(wk3r,x0i,Q1_14))>>1;	// wk3i * x0r + wk3r * x0i;
    a[j3 + 3] = FSUB(FMUL(wk3i,x0i,Q1_14),FMUL(wk3r,x0r,Q1_14))>>1;	// wk3i * x0i - wk3r * x0r;
}


void fxcftb1st(INT32 n, FIXED16 *a, FIXED16 *w)
{
    INT32 j, j0, j1, j2, j3, k, m, mh;
    FIXED32 wn4r, csc1, csc3, wk1r, wk1i, wk3r, wk3i, 
        wd1r, wd1i, wd3r, wd3i;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i;
    
    mh = n >> 3;
    m = 2 * mh;
    j1 = m;
    j2 = j1 + m;
    j3 = j2 + m;
    x0r = FADD(a[0],a[j2]);			// a[0] + a[j2];
    x0i = FSUB(-a[1],a[j2 + 1]);	// -a[1] - a[j2 + 1];
    x1r = FSUB(a[0],a[j2]);			// a[0] - a[j2];
    x1i = FADD(-a[1],a[j2 + 1]);	// -a[1] + a[j2 + 1];
    x2r = FADD(a[j1],a[j3]);		// a[j1] + a[j3];
    x2i = FADD(a[j1 + 1],a[j3 + 1]); // a[j1 + 1] + a[j3 + 1];
    x3r = FSUB(a[j1],a[j3]);		// a[j1] - a[j3];
    x3i = FSUB(a[j1 + 1],a[j3 + 1]); // a[j1 + 1] - a[j3 + 1];
    a[0] = FADD(x0r,x2r)>>1;				// x0r + x2r;
    a[1] = FSUB(x0i,x2i)>>1;				// x0i - x2i;
    a[j1] = FSUB(x0r,x2r)>>1;				// x0r - x2r;
    a[j1 + 1] = FADD(x0i,x2i)>>1;			// x0i + x2i;
    a[j2] = FADD(x1r,x3i)>>1;				// x1r + x3i;
    a[j2 + 1] = FADD(x1i,x3r)>>1;			// x1i + x3r;
    a[j3] = FSUB(x1r,x3i)>>1;				// x1r - x3i;
    a[j3 + 1] = FSUB(x1i,x3r)>>1;			// x1i - x3r;
    wn4r = w[1];
    csc1 = w[2];
    csc3 = w[3];
    wd1r = ITOFIX16(1,Q1_14);
    wd1i = ITOFIX16(0,Q1_14);
    wd3r = ITOFIX16(1,Q1_14);
    wd3i = ITOFIX16(0,Q1_14);
    k = 0;
    for (j = 2; j < mh - 2; j += 4) {
        k += 4;
        wk1r = FMUL(csc1,FADD(wd1r,w[k]),Q1_14);		// csc1 * (wd1r + w[k]);
        wk1i = FMUL(csc1,FADD(wd1i,w[k + 1]),Q1_14);	// csc1 * (wd1i + w[k + 1]);
        wk3r = FMUL(csc3,FADD(wd3r,w[k + 2]),Q1_14);	// csc3 * (wd3r + w[k + 2]);
        wk3i = FMUL(csc3,FADD(wd3i,w[k + 3]),Q1_14);	// csc3 * (wd3i + w[k + 3]);
        wd1r = w[k];
        wd1i = w[k + 1];
        wd3r = w[k + 2];
        wd3i = w[k + 3];
        j1 = j + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FADD(a[j],a[j2]);					// a[j] + a[j2];	
        x0i = FSUB(-a[j + 1],a[j2 + 1]);		// -a[j + 1] - a[j2 + 1];
        x1r = FSUB(a[j],a[j2]);					// a[j] - a[j2];	
        x1i = FADD(-a[j + 1],a[j2 + 1]);		// -a[j + 1] + a[j2 + 1];
        y0r = FADD(a[j + 2],a[j2 + 2]);			// a[j + 2] + a[j2 + 2];	
        y0i = FSUB(-a[j + 3],a[j2 + 3]);		// -a[j + 3] - a[j2 + 3];
        y1r = FSUB(a[j + 2],a[j2 + 2]);			// a[j + 2] - a[j2 + 2];	
        y1i = FADD(-a[j + 3],a[j2 + 3]);		// -a[j + 3] + a[j2 + 3];
        x2r = FADD(a[j1],a[j3]);				// a[j1] + a[j3];	
        x2i = FADD(a[j1 + 1],a[j3 + 1]);		// a[j1 + 1] + a[j3 + 1];	
        x3r = FSUB(a[j1],a[j3]);				// a[j1] - a[j3];	
        x3i = FSUB(a[j1 + 1],a[j3 + 1]);		// a[j1 + 1] - a[j3 + 1];
        y2r = FADD(a[j1 + 2],a[j3 + 2]);		// a[j1 + 2] + a[j3 + 2];
        y2i = FADD(a[j1 + 3],a[j3 + 3]);		// a[j1 + 3] + a[j3 + 3];
        y3r = FSUB(a[j1 + 2],a[j3 + 2]);		// a[j1 + 2] - a[j3 + 2];
        y3i = FSUB(a[j1 + 3],a[j3 + 3]);		// a[j1 + 3] - a[j3 + 3];
        a[j] = FADD(x0r,x2r)>>1;					// x0r + x2r;
        a[j + 1] = FSUB(x0i,x2i)>>1;				// x0i - x2i;
        a[j + 2] = FADD(y0r,y2r)>>1;				// y0r + y2r;
        a[j + 3] = FSUB(y0i,y2i)>>1;				// y0i - y2i;
        a[j1] = FSUB(x0r,x2r)>>1;					// x0r - x2r;	
        a[j1 + 1] = FADD(x0i,x2i)>>1;				// x0i + x2i;
        a[j1 + 2] = FSUB(y0r,y2r)>>1;				// y0r - y2r;
        a[j1 + 3] = FADD(y0i,y2i)>>1;				// y0i + y2i;
        x0r = FADD(x1r,x3i);					// x1r + x3i;
        x0i = FADD(x1i,x3r);					// x1i + x3r;		
        a[j2] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14))>>1;		// wk1r * x0r - wk1i * x0i;
        a[j2 + 1] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14))>>1;	// wk1r * x0i + wk1i * x0r;
        x0r = FADD(y1r,y3i);						// y1r + y3i;
        x0i = FADD(y1i,y3r);						// y1i + y3r;
        a[j2 + 2] = FSUB(FMUL(wd1r,x0r,Q1_14),FMUL(wd1i,x0i,Q1_14))>>1;	// wd1r * x0r - wd1i * x0i;
        a[j2 + 3] = FADD(FMUL(wd1r,x0i,Q1_14),FMUL(wd1i,x0r,Q1_14))>>1;	// wd1r * x0i + wd1i * x0r;
        x0r = FSUB(x1r,x3i);						// x1r - x3i;
        x0i = FSUB(x1i,x3r);						// x1i - x3r;
        a[j3] = FADD(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14))>>1;		// wk3r * x0r + wk3i * x0i;
        a[j3 + 1] = FSUB(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14))>>1;	// wk3r * x0i - wk3i * x0r;
        x0r = FSUB(y1r,y3i);						// y1r - y3i;
        x0i = FSUB(y1i,y3r);						// y1i - y3r;
        a[j3 + 2] = FADD(FMUL(wd3r,x0r,Q1_14),FMUL(wd3i,x0i,Q1_14))>>1;	// wd3r * x0r + wd3i * x0i;
        a[j3 + 3] = FSUB(FMUL(wd3r,x0i,Q1_14),FMUL(wd3i,x0r,Q1_14))>>1;	// wd3r * x0i - wd3i * x0r;
        j0 = m - j;
        j1 = j0 + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FADD(a[j0],a[j2]);				// a[j0] + a[j2];
        x0i = FSUB(-a[j0 + 1],a[j2 + 1]);		// -a[j0 + 1] - a[j2 + 1];
        x1r = FSUB(a[j0],a[j2]);				// a[j0] - a[j2];
        x1i = FADD(-a[j0 + 1],a[j2 + 1]);		// -a[j0 + 1] + a[j2 + 1];
        y0r = FADD(a[j0 - 2],a[j2 - 2]);		// a[j0 - 2] + a[j2 - 2];
        y0i = FSUB(-a[j0 - 1],a[j2 - 1]);		// -a[j0 - 1] - a[j2 - 1];
        y1r = FSUB(a[j0 - 2],a[j2 - 2]);		// a[j0 - 2] - a[j2 - 2];
        y1i = FADD(-a[j0 - 1],a[j2 - 1]);		// -a[j0 - 1] + a[j2 - 1];
        x2r = FADD(a[j1],a[j3]);				// a[j1] + a[j3];
        x2i = FADD(a[j1 + 1],a[j3 + 1]);		// a[j1 + 1] + a[j3 + 1];
        x3r = FSUB(a[j1],a[j3]);				// a[j1] - a[j3];
        x3i = FSUB(a[j1 + 1],a[j3 + 1]);		// a[j1 + 1] - a[j3 + 1];
        y2r = FADD(a[j1 - 2],a[j3 - 2]);		// a[j1 - 2] + a[j3 - 2];
        y2i = FADD(a[j1 - 1],a[j3 - 1]);		// a[j1 - 1] + a[j3 - 1];
        y3r = FSUB(a[j1 - 2],a[j3 - 2]);		// a[j1 - 2] - a[j3 - 2];
        y3i = FSUB(a[j1 - 1],a[j3 - 1]);		// a[j1 - 1] - a[j3 - 1];
        a[j0] = FADD(x0r,x2r)>>1;					// x0r + x2r;
        a[j0 + 1] = FSUB(x0i,x2i)>>1;				// x0i - x2i;
        a[j0 - 2] = FADD(y0r,y2r)>>1;				// y0r + y2r;
        a[j0 - 1] = FSUB(y0i,y2i)>>1;				// y0i - y2i;
        a[j1] = FSUB(x0r,x2r)>>1;					// x0r - x2r;
        a[j1 + 1] = FADD(x0i,x2i)>>1;				// x0i + x2i;
        a[j1 - 2] = FSUB(y0r,y2r)>>1;				// y0r - y2r;
        a[j1 - 1] = FADD(y0i,y2i)>>1;				// y0i + y2i;
        x0r = FADD(x1r,x3i);					// x1r + x3i;
        x0i = FADD(x1i,x3r);					// x1i + x3r;
        a[j2] = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14))>>1;	// wk1i * x0r - wk1r * x0i;
        a[j2 + 1] = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14))>>1;	// wk1i * x0i + wk1r * x0r;
        x0r = FADD(y1r,y3i);					// y1r + y3i;
        x0i = FADD(y1i,y3r);					// y1i + y3r;
        a[j2 - 2] = FSUB(FMUL(wd1i,x0r,Q1_14),FMUL(wd1r,x0i,Q1_14))>>1;	// wd1i * x0r - wd1r * x0i;
        a[j2 - 1] = FADD(FMUL(wd1i,x0i,Q1_14),FMUL(wd1r,x0r,Q1_14))>>1;	// wd1i * x0i + wd1r * x0r;
        x0r = FSUB(x1r,x3i);					// x1r - x3i;
        x0i = FSUB(x1i,x3r);					// x1i - x3r;
        a[j3] = FADD(FMUL(wk3i,x0r,Q1_14),FMUL(wk3r,x0i,Q1_14))>>1;	// wk3i * x0r + wk3r * x0i;
        a[j3 + 1] = FSUB(FMUL(wk3i,x0i,Q1_14),FMUL(wk3r,x0r,Q1_14))>>1;	// wk3i * x0i - wk3r * x0r;
        x0r = FSUB(y1r,y3i);					// y1r - y3i;
        x0i = FSUB(y1i,y3r);					// y1i - y3r;
        a[j3 - 2] = FADD(FMUL(wd3i,x0r,Q1_14),FMUL(wd3r,x0i,Q1_14))>>1;	// wd3i * x0r + wd3r * x0i;
        a[j3 - 1] = FSUB(FMUL(wd3i,x0i,Q1_14),FMUL(wd3r,x0r,Q1_14))>>1;	// wd3i * x0i - wd3r * x0r;
    }
    wk1r = FMUL(csc1,FADD(wd1r,wn4r),Q1_14);		// csc1 * (wd1r + wn4r);
    wk1i = FMUL(csc1,FADD(wd1i,wn4r),Q1_14);		// csc1 * (wd1i + wn4r);
    wk3r = FMUL(csc3,FSUB(wd3r,wn4r),Q1_14);		// csc3 * (wd3r - wn4r);
    wk3i = FMUL(csc3,FSUB(wd3i,wn4r),Q1_14);		// csc3 * (wd3i - wn4r);
    j0 = mh;
    j1 = j0 + m;
    j2 = j1 + m;
    j3 = j2 + m;
    x0r = FADD(a[j0 - 2],a[j2 - 2]);		// a[j0 - 2] + a[j2 - 2];
    x0i = FSUB(-a[j0 - 1],a[j2 - 1]);		// -a[j0 - 1] - a[j2 - 1];
    x1r = FSUB(a[j0 - 2],a[j2 - 2]);		// a[j0 - 2] - a[j2 - 2];
    x1i = FADD(-a[j0 - 1],a[j2 - 1]);		// -a[j0 - 1] + a[j2 - 1];
    x2r = FADD(a[j1 - 2],a[j3 - 2]);		// a[j1 - 2] + a[j3 - 2];
    x2i = FADD(a[j1 - 1],a[j3 - 1]);		// a[j1 - 1] + a[j3 - 1];
    x3r = FSUB(a[j1 - 2],a[j3 - 2]);		// a[j1 - 2] - a[j3 - 2];
    x3i = FSUB(a[j1 - 1],a[j3 - 1]);		// a[j1 - 1] - a[j3 - 1];
    a[j0 - 2] = FADD(x0r,x2r)>>1;				// x0r + x2r;	
    a[j0 - 1] = FSUB(x0i,x2i)>>1;				// x0i - x2i;
    a[j1 - 2] = FSUB(x0r,x2r)>>1;				// x0r - x2r;
    a[j1 - 1] = FADD(x0i,x2i)>>1;				// x0i + x2i;	
    x0r = FADD(x1r,x3i);					// x1r + x3i;
    x0i = FADD(x1i,x3r);					// x1i + x3r;
    a[j2 - 2] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14))>>1;	// wk1r * x0r - wk1i * x0i;
    a[j2 - 1] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14))>>1;	// wk1r * x0i + wk1i * x0r;
    x0r = FSUB(x1r,x3i);					// x1r - x3i;
    x0i = FSUB(x1i,x3r);					// x1i - x3r;
    a[j3 - 2] = FADD(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14))>>1;	// wk3r * x0r + wk3i * x0i;
    a[j3 - 1] = FSUB(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14))>>1;	// wk3r * x0i - wk3i * x0r;
    x0r = FADD(a[j0],a[j2]);				//	a[j0] + a[j2];
    x0i = FSUB(-a[j0 + 1],a[j2 + 1]);		// -a[j0 + 1] - a[j2 + 1];
    x1r = FSUB(a[j0],a[j2]);				// a[j0] - a[j2];	
    x1i = FADD(-a[j0 + 1],a[j2 + 1]);		// -a[j0 + 1] + a[j2 + 1];
    x2r = FADD(a[j1],a[j3]);				// a[j1] + a[j3];
    x2i = FADD(a[j1 + 1],a[j3 + 1]);		// a[j1 + 1] + a[j3 + 1];
    x3r = FSUB(a[j1],a[j3]);				// a[j1] - a[j3];
    x3i = FSUB(a[j1 + 1],a[j3 + 1]);		// a[j1 + 1] - a[j3 + 1];
    a[j0] = FADD(x0r,x2r)>>1;					// x0r + x2r;	
    a[j0 + 1] = FSUB(x0i,x2i)>>1;				// x0i - x2i;	
    a[j1] = FSUB(x0r,x2r)>>1;					// x0r - x2r;		
    a[j1 + 1] = FADD(x0i,x2i)>>1;				// x0i + x2i;
    x0r = FADD(x1r,x3i);					// x1r + x3i;
    x0i = FADD(x1i,x3r);					// x1i + x3r;	
    a[j2] = FMUL(wn4r,FSUB(x0r,x0i),Q1_14)>>1;			// wn4r * (x0r - x0i);	
    a[j2 + 1] = FMUL(wn4r,FADD(x0i,x0r),Q1_14)>>1;		// wn4r * (x0i + x0r);
    x0r = FSUB(x1r,x3i);					// x1r - x3i;
    x0i = FSUB(x1i,x3r);					// x1i - x3r;
    a[j3] = FMUL(-wn4r,FADD(x0r,x0i),Q1_14)>>1;		// -wn4r * (x0r + x0i);
    a[j3 + 1] = FMUL(-wn4r,FSUB(x0i,x0r),Q1_14)>>1;	// -wn4r * (x0i - x0r);
    x0r = FADD(a[j0 + 2],a[j2 + 2]);		// a[j0 + 2] + a[j2 + 2];
    x0i = FSUB(-a[j0 + 3],a[j2 + 3]);		// -a[j0 + 3] - a[j2 + 3];
    x1r = FSUB(a[j0 + 2],a[j2 + 2]);		// a[j0 + 2] - a[j2 + 2];
    x1i = FADD(-a[j0 + 3],a[j2 + 3]);		// -a[j0 + 3] + a[j2 + 3];
    x2r = FADD(a[j1 + 2],a[j3 + 2]);		// a[j1 + 2] + a[j3 + 2];
    x2i = FADD(a[j1 + 3],a[j3 + 3]);		// a[j1 + 3] + a[j3 + 3];
    x3r = FSUB(a[j1 + 2],a[j3 + 2]);		// a[j1 + 2] - a[j3 + 2];
    x3i = FSUB(a[j1 + 3],a[j3 + 3]);		// a[j1 + 3] - a[j3 + 3];
    a[j0 + 2] = FADD(x0r,x2r)>>1;				// x0r + x2r;	
    a[j0 + 3] = FSUB(x0i,x2i)>>1;				// x0i - x2i;	
    a[j1 + 2] = FSUB(x0r,x2r)>>1;				// x0r - x2r;	
    a[j1 + 3] = FADD(x0i,x2i)>>1;				// x0i + x2i;	
    x0r = FADD(x1r,x3i);					// x1r + x3i;
    x0i = FADD(x1i,x3r);					// x1i + x3r;	
    a[j2 + 2] = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14))>>1;	// wk1i * x0r - wk1r * x0i;
    a[j2 + 3] = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14))>>1;	// wk1i * x0i + wk1r * x0r;
    x0r = FSUB(x1r,x3i);					// x1r - x3i;	
    x0i = FSUB(x1i,x3r);					// x1i - x3r;
    a[j3 + 2] = FADD(FMUL(wk3i,x0r,Q1_14),FMUL(wk3r,x0i,Q1_14))>>1;	// wk3i * x0r + wk3r * x0i;
    a[j3 + 3] = FSUB(FMUL(wk3i,x0i,Q1_14),FMUL(wk3r,x0r,Q1_14))>>1;	// wk3i * x0i - wk3r * x0r;
}


/*void fxcftrec4(INT32 n, FIXED16 *a, INT32 nw, FIXED16 *w)
{
    INT32 isplt, j, k, m;
    
    m = n;
    while (m > 512) {
        m >>= 2;
        fxcftmdl1(m, &a[n - m], &w[nw - (m >> 1)]);
    }
    fxcftleaf(m, 1, &a[n - m], nw, w);
    k = 0;
    for (j = n - m; j > 0; j -= m) {
        k++;
        isplt = fxcfttree(m, j, k, a, nw, w);
        fxcftleaf(m, isplt, &a[j - m], nw, w);
    }
}*/


/*INT32 fxcfttree(INT32 n, INT32 j, INT32 k, FIXED16 *a, INT32 nw, FIXED16 *w)
{
    INT32 i, isplt, m;
    
    if ((k & 3) != 0) {
        isplt = k & 1;
        if (isplt != 0) {
            fxcftmdl1(n, &a[j - n], &w[nw - (n >> 1)]);
        } else {
            fxcftmdl2(n, &a[j - n], &w[nw - n]);
        }
    } else {
        m = n;
        for (i = k; (i & 3) == 0; i >>= 2) {
            m <<= 2;
        }
        isplt = i & 1;
        if (isplt != 0) {
            while (m > 128) {
                fxcftmdl1(m, &a[j - m], &w[nw - (m >> 1)]);
                m >>= 2;
            }
        } else {
            while (m > 128) {
                fxcftmdl2(m, &a[j - m], &w[nw - m]);
                m >>= 2;
            }
        }
    }
    return isplt;
}*/


/*void fxcftleaf(INT32 n, INT32 isplt, FIXED16 *a, INT32 nw, FIXED16 *w)
{
   
    if (n == 512) {
        fxcftmdl1(128, a, &w[nw - 64]);
        fxcftf161(a, &w[nw - 8]);
        fxcftf162(&a[32], &w[nw - 32]);
        fxcftf161(&a[64], &w[nw - 8]);
        fxcftf161(&a[96], &w[nw - 8]);
        fxcftmdl2(128, &a[128], &w[nw - 128]);
        fxcftf161(&a[128], &w[nw - 8]);
        fxcftf162(&a[160], &w[nw - 32]);
        fxcftf161(&a[192], &w[nw - 8]);
        fxcftf162(&a[224], &w[nw - 32]);
        fxcftmdl1(128, &a[256], &w[nw - 64]);
        fxcftf161(&a[256], &w[nw - 8]);
        fxcftf162(&a[288], &w[nw - 32]);
        fxcftf161(&a[320], &w[nw - 8]);
        fxcftf161(&a[352], &w[nw - 8]);
        if (isplt != 0) {
            fxcftmdl1(128, &a[384], &w[nw - 64]);
            fxcftf161(&a[480], &w[nw - 8]);
        } else {
            fxcftmdl2(128, &a[384], &w[nw - 128]);
            fxcftf162(&a[480], &w[nw - 32]);
        }
        fxcftf161(&a[384], &w[nw - 8]);
        fxcftf162(&a[416], &w[nw - 32]);
        fxcftf161(&a[448], &w[nw - 8]);
    } else {
        fxcftmdl1(64, a, &w[nw - 32]);
        fxcftf081(a, &w[nw - 8]);
        fxcftf082(&a[16], &w[nw - 8]);
        fxcftf081(&a[32], &w[nw - 8]);
        fxcftf081(&a[48], &w[nw - 8]);
        fxcftmdl2(64, &a[64], &w[nw - 64]);
        fxcftf081(&a[64], &w[nw - 8]);
        fxcftf082(&a[80], &w[nw - 8]);
        fxcftf081(&a[96], &w[nw - 8]);
        fxcftf082(&a[112], &w[nw - 8]);
        fxcftmdl1(64, &a[128], &w[nw - 32]);
        fxcftf081(&a[128], &w[nw - 8]);
        fxcftf082(&a[144], &w[nw - 8]);
        fxcftf081(&a[160], &w[nw - 8]);
        fxcftf081(&a[176], &w[nw - 8]);
        if (isplt != 0) {
            fxcftmdl1(64, &a[192], &w[nw - 32]);
            fxcftf081(&a[240], &w[nw - 8]);
        } else {
            fxcftmdl2(64, &a[192], &w[nw - 64]);
            fxcftf082(&a[240], &w[nw - 8]);
        }
        fxcftf081(&a[192], &w[nw - 8]);
        fxcftf082(&a[208], &w[nw - 8]);
        fxcftf081(&a[224], &w[nw - 8]);
    }
}*/


/*void fxcftmdl1(INT32 n, FIXED16 *a, FIXED16 *w)
{
    INT32 j, j0, j1, j2, j3, k, m, mh;
    FIXED32 wn4r, wk1r, wk1i, wk3r, wk3i;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    mh = n >> 3;
    m = 2 * mh;
    j1 = m;
    j2 = j1 + m;
    j3 = j2 + m;
    x0r = FADD(a[0],a[j2]);
    x0i = FADD(a[1],a[j2 + 1]);
    x1r = FSUB(a[0],a[j2]);
    x1i = FSUB(a[1], a[j2 + 1]);
    x2r = FADD(a[j1],a[j3]);
    x2i = FADD(a[j1 + 1],a[j3 + 1]);
    x3r = FSUB(a[j1],a[j3]);
    x3i = FSUB(a[j1 + 1],a[j3 + 1]);
    a[0] = FADD(x0r,x2r)>>1;
    a[1] = FADD(x0i,x2i)>>1;
    a[j1] = FSUB(x0r,x2r)>>1;
    a[j1 + 1] = FSUB(x0i,x2i)>>1;
    a[j2] = FSUB(x1r,x3i)>>1;
    a[j2 + 1] = FADD(x1i,x3r)>>1;
    a[j3] = FADD(x1r,x3i)>>1;
    a[j3 + 1] = FSUB(x1i,x3r)>>1;
    wn4r = w[1];
    k = 0;
    for (j = 2; j < mh; j += 2) {
        k += 4;
        wk1r = w[k];
        wk1i = w[k + 1];
        wk3r = w[k + 2];
        wk3i = w[k + 3];
        j1 = j + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FADD(a[j],a[j2]);
        x0i = FADD(a[j + 1],a[j2 + 1]);
        x1r = FSUB(a[j],a[j2]);
        x1i = FSUB(a[j + 1],a[j2 + 1]);
        x2r = FADD(a[j1],a[j3]);
        x2i = FADD(a[j1 + 1],a[j3 + 1]);
        x3r = FSUB(a[j1],a[j3]);
        x3i = FSUB(a[j1 + 1],a[j3 + 1]);
        a[j] = FADD(x0r,x2r)>>1;
        a[j + 1] = FADD(x0i,x2i)>>1;
        a[j1] = FSUB(x0r,x2r)>>1;
        a[j1 + 1] = FSUB(x0i,x2i)>>1;
        x0r = FSUB(x1r,x3i);
        x0i = FADD(x1i,x3r);
        a[j2] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14))>>1;
        a[j2 + 1] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14))>>1;
        x0r = FADD(x1r,x3i);
        x0i = FSUB(x1i,x3r);
        a[j3] = FADD(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14))>>1;
        a[j3 + 1] = FSUB(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14))>>1;
        j0 = m - j;
        j1 = j0 + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FADD(a[j0],a[j2]);
        x0i = FADD(a[j0 + 1],a[j2 + 1]);
        x1r = FSUB(a[j0],a[j2]);
        x1i = FSUB(a[j0 + 1],a[j2 + 1]);
        x2r = FADD(a[j1],a[j3]);
        x2i = FADD(a[j1 + 1],a[j3 + 1]);
        x3r = FSUB(a[j1],a[j3]);
        x3i = FSUB(a[j1 + 1],a[j3 + 1]);
        a[j0] = FADD(x0r,x2r)>>1;
        a[j0 + 1] = FADD(x0i,x2i)>>1;
        a[j1] = FSUB(x0r,x2r)>>1;
        a[j1 + 1] = FSUB(x0i,x2i)>>1;
        x0r = FSUB(x1r,x3i);
        x0i = FADD(x1i,x3r);
        a[j2] = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14))>>1;
        a[j2 + 1] = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14))>>1;
        x0r = FADD(x1r,x3i);
        x0i = FSUB(x1i,x3r);
        a[j3] = FADD(FMUL(wk3i,x0r,Q1_14),FMUL(wk3r,x0i,Q1_14))>>1;
        a[j3 + 1] = FSUB(FMUL(wk3i,x0i,Q1_14),FMUL(wk3r,x0r,Q1_14))>>1;
    }
    j0 = mh;
    j1 = j0 + m;
    j2 = j1 + m;
    j3 = j2 + m;
    x0r = FADD(a[j0],a[j2]);
    x0i = FADD(a[j0 + 1],a[j2 + 1]);
    x1r = FSUB(a[j0],a[j2]);
    x1i = FSUB(a[j0 + 1],a[j2 + 1]);
    x2r = FADD(a[j1],a[j3]);
    x2i = FADD(a[j1 + 1],a[j3 + 1]);
    x3r = FSUB(a[j1],a[j3]);
    x3i = FSUB(a[j1 + 1],a[j3 + 1]);
    a[j0] = FADD(x0r,x2r)>>1;
    a[j0 + 1] = FADD(x0i, x2i)>>1;
    a[j1] = FSUB(x0r,x2r)>>1;
    a[j1 + 1] = FSUB(x0i,x2i)>>1;
    x0r = FSUB(x1r,x3i);
    x0i = FADD(x1i,x3r);
    a[j2] = FMUL(wn4r,FSUB(x0r,x0i),Q1_14)>>1;
    a[j2 + 1] = FMUL(wn4r,FADD(x0i,x0r),Q1_14)>>1;
    x0r = FADD(x1r,x3i);
    x0i = FSUB(x1i,x3r);
    a[j3] = FMUL(-wn4r,FADD(x0r,x0i),Q1_14)>>1;
    a[j3 + 1] = FMUL(-wn4r,FSUB(x0i,x0r),Q1_14)>>1;
}*/


/*void fxcftmdl2(INT32 n, FIXED16 *a, FIXED16 *w)
{
    INT32 j, j0, j1, j2, j3, k, kr, m, mh;
    FIXED32 wn4r, wk1r, wk1i, wk3r, wk3i, wd1r, wd1i, wd3r, wd3i;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, y0r, y0i, y2r, y2i;
    
    mh = n >> 3;
    m = 2 * mh;
    wn4r = w[1];
    j1 = m;
    j2 = j1 + m;
    j3 = j2 + m;
    x0r = FSUB(a[0],a[j2 + 1]);
    x0i = FADD(a[1],a[j2]);
    x1r = FADD(a[0],a[j2 + 1]);
    x1i = FSUB(a[1],a[j2]);
    x2r = FSUB(a[j1],a[j3 + 1]);
    x2i = FADD(a[j1 + 1],a[j3]);
    x3r = FADD(a[j1],a[j3 + 1]);
    x3i = FSUB(a[j1 + 1],a[j3]);
    y0r = FMUL(wn4r,FSUB(x2r,x2i),Q1_14);
    y0i = FMUL(wn4r,FADD(x2i,x2r),Q1_14);
    a[0] = FADD(x0r,y0r)>>1;
    a[1] = FADD(x0i,y0i)>>1;
    a[j1] = FSUB(x0r,y0r)>>1;
    a[j1 + 1] = FSUB(x0i,y0i)>>1;
    y0r = FMUL(wn4r, FSUB(x3r,x3i),Q1_14);
    y0i = FMUL(wn4r,FADD(x3i,x3r),Q1_14);
    a[j2] = FSUB(x1r,y0i)>>1;
    a[j2 + 1] = FADD(x1i,y0r)>>1;
    a[j3] = FADD(x1r,y0i)>>1;
    a[j3 + 1] = FSUB(x1i,y0r)>>1;
    k = 0;
    kr = 2 * m;
    for (j = 2; j < mh; j += 2) {
        k += 4;
        wk1r = w[k];
        wk1i = w[k + 1];
        wk3r = w[k + 2];
        wk3i = w[k + 3];
        kr -= 4;
        wd1i = w[kr];
        wd1r = w[kr + 1];
        wd3i = w[kr + 2];
        wd3r = w[kr + 3];
        j1 = j + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FSUB(a[j],a[j2 + 1]);
        x0i = FADD(a[j + 1],a[j2]);
        x1r = FADD(a[j],a[j2 + 1]);
        x1i = FSUB(a[j + 1],a[j2]);
        x2r = FSUB(a[j1],a[j3 + 1]);
        x2i = FADD(a[j1 + 1],a[j3]);
        x3r = FADD(a[j1],a[j3 + 1]);
        x3i = FSUB(a[j1 + 1],a[j3]);
        y0r = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
        y0i = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
        y2r = FSUB(FMUL(wd1r,x2r,Q1_14),FMUL(wd1i,x2i,Q1_14));
        y2i = FADD(FMUL(wd1r,x2i,Q1_14),FMUL(wd1i,x2r,Q1_14));
        a[j] = FADD(y0r,y2r)>>1;
        a[j + 1] = FADD(y0i,y2i)>>1;
        a[j1] = FSUB(y0r,y2r)>>1;
        a[j1 + 1] = FSUB(y0i,y2i)>>1;
        y0r = FADD(FMUL(wk3r,x1r,Q1_14),FMUL(wk3i, x1i,Q1_14));
        y0i = FSUB(FMUL(wk3r,x1i,Q1_14),FMUL(wk3i,x1r,Q1_14));
        y2r = FADD(FMUL(wd3r,x3r,Q1_14),FMUL(wd3i,x3i,Q1_14));
        y2i = FSUB(FMUL(wd3r,x3i,Q1_14),FMUL(wd3i,x3r,Q1_14));
        a[j2] = FADD(y0r,y2r)>>1;
        a[j2 + 1] = FADD(y0i, y2i)>>1;
        a[j3] = FSUB(y0r,y2r)>>1;
        a[j3 + 1] = FSUB(y0i, y2i)>>1;
        j0 = m - j;
        j1 = j0 + m;
        j2 = j1 + m;
        j3 = j2 + m;
        x0r = FSUB(a[j0],a[j2 + 1]);
        x0i = FADD(a[j0 + 1], a[j2]);
        x1r = FADD(a[j0],a[j2 + 1]);
        x1i = FSUB(a[j0 + 1],a[j2]);
        x2r = FSUB(a[j1],a[j3 + 1]);
        x2i = FADD(a[j1 + 1],a[j3]);
        x3r = FADD(a[j1],a[j3 + 1]);
        x3i = FSUB(a[j1 + 1],a[j3]);
        y0r = FSUB(FMUL(wd1i,x0r,Q1_14),FMUL(wd1r,x0i,Q1_14));
        y0i = FADD(FMUL(wd1i,x0i,Q1_14),FMUL(wd1r,x0r,Q1_14));
        y2r = FSUB(FMUL(wk1i,x2r,Q1_14),FMUL(wk1r,x2i,Q1_14));
        y2i = FADD(FMUL(wk1i,x2i,Q1_14),FMUL(wk1r,x2r,Q1_14));
        a[j0] = FADD(y0r,y2r)>>1;
        a[j0 + 1] = FADD(y0i,y2i)>>1;
        a[j1] = FSUB(y0r,y2r)>>1;
        a[j1 + 1] = FSUB(y0i,y2i)>>1;
        y0r = FADD(FMUL(wd3i,x1r,Q1_14),FMUL(wd3r,x1i,Q1_14));
        y0i = FSUB(FMUL(wd3i,x1i,Q1_14),FMUL(wd3r,x1r,Q1_14));
        y2r = FADD(FMUL(wk3i,x3r,Q1_14),FMUL(wk3r,x3i,Q1_14));
        y2i = FSUB(FMUL(wk3i,x3i,Q1_14),FMUL(wk3r,x3r,Q1_14));
        a[j2] = FADD(y0r,y2r)>>1;
        a[j2 + 1] = FADD(y0i,y2i)>>1;
        a[j3] = FSUB(y0r,y2r)>>1;
        a[j3 + 1] = FSUB(y0i,y2i)>>1;
    }
    wk1r = w[m];
    wk1i = w[m + 1];
    j0 = mh;
    j1 = j0 + m;
    j2 = j1 + m;
    j3 = j2 + m;
    x0r = FSUB(a[j0],a[j2 + 1]);
    x0i = FADD(a[j0 + 1],a[j2]);
    x1r = FADD(a[j0],a[j2 + 1]);
    x1i = FSUB(a[j0 + 1],a[j2]);
    x2r = FSUB(a[j1],a[j3 + 1]);
    x2i = FADD(a[j1 + 1],a[j3]);
    x3r = FADD(a[j1],a[j3 + 1]);
    x3i = FSUB(a[j1 + 1],a[j3]);
    y0r = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i, x0i,Q1_14));
    y0i = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
    y2r = FSUB(FMUL(wk1i,x2r,Q1_14),FMUL(wk1r,x2i,Q1_14));
    y2i = FADD(FMUL(wk1i,x2i,Q1_14),FMUL(wk1r,x2r,Q1_14));
    a[j0] = FADD(y0r,y2r)>>1;
    a[j0 + 1] = FADD(y0i,y2i)>>1;
    a[j1] = FSUB(y0r,y2r)>>1;
    a[j1 + 1] = FSUB(y0i,y2i)>>1;
    y0r = FSUB(FMUL(wk1i,x1r,Q1_14),FMUL(wk1r, x1i,Q1_14));
    y0i = FADD(FMUL(wk1i,x1i,Q1_14),FMUL(wk1r,x1r,Q1_14));
    y2r = FSUB(FMUL(wk1r,x3r,Q1_14),FMUL(wk1i,x3i,Q1_14));
    y2i = FADD(FMUL(wk1r,x3i,Q1_14),FMUL(wk1i,x3r,Q1_14));
    a[j2] = FSUB(y0r,y2r)>>1;
    a[j2 + 1] = FSUB(y0i,y2i)>>1;
    a[j3] = FADD(y0r,y2r)>>1;
    a[j3 + 1] = FADD(y0i,y2i)>>1;
}*/


void fxcftfx41(INT32 n, FIXED16 *a, INT32 nw, FIXED16 *w)
{
  
    if (n == 128) {
        fxcftf161(a, &w[nw - 8]);
        fxcftf162(&a[32], &w[nw - 32]);
        fxcftf161(&a[64], &w[nw - 8]);
        fxcftf161(&a[96], &w[nw - 8]);
    } else {
        fxcftf081(a, &w[nw - 8]);
        fxcftf082(&a[16], &w[nw - 8]);
        fxcftf081(&a[32], &w[nw - 8]);
        fxcftf081(&a[48], &w[nw - 8]);
    }
}


void fxcftf161(FIXED16 *a, FIXED16 *w)
{
    FIXED32 wn4r, wk1r, wk1i, 
        x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i, 
        y4r, y4i, y5r, y5i, y6r, y6i, y7r, y7i, 
        y8r, y8i, y9r, y9i, y10r, y10i, y11r, y11i, 
        y12r, y12i, y13r, y13i, y14r, y14i, y15r, y15i;
    
    wn4r = w[1];
    wk1r = w[2];
    wk1i = w[3];
    x0r = FADD(a[0],a[16]);
    x0i = FADD(a[1],a[17]);
    x1r = FSUB(a[0],a[16]);
    x1i = FSUB(a[1],a[17]);
    x2r = FADD(a[8],a[24]);
    x2i = FADD(a[9],a[25]);
    x3r = FSUB(a[8],a[24]);
    x3i = FSUB(a[9],a[25]);
    y0r = FADD(x0r,x2r);
    y0i = FADD(x0i,x2i);
    y4r = FSUB(x0r,x2r);
    y4i = FSUB(x0i,x2i);
    y8r = FSUB(x1r,x3i);
    y8i = FADD(x1i,x3r);
    y12r = FADD(x1r,x3i);
    y12i = FSUB(x1i,x3r);
    x0r = FADD(a[2],a[18]);
    x0i = FADD(a[3],a[19]);
    x1r = FSUB(a[2],a[18]);
    x1i = FSUB(a[3],a[19]);
    x2r = FADD(a[10],a[26]);
    x2i = FADD(a[11],a[27]);
    x3r = FSUB(a[10],a[26]);
    x3i = FSUB(a[11],a[27]);
    y1r = FADD(x0r,x2r);
    y1i = FADD(x0i,x2i);
    y5r = FSUB(x0r,x2r);
    y5i = FSUB(x0i,x2i);
    x0r = FSUB(x1r,x3i);
    x0i = FADD(x1i,x3r);
    y9r = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
    y9i = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
    x0r = FADD(x1r,x3i);
    x0i = FSUB(x1i,x3r);
    y13r = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14));
    y13i = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14));
    x0r = FADD(a[4],a[20]);
    x0i = FADD(a[5],a[21]);
    x1r = FSUB(a[4],a[20]);
    x1i = FSUB(a[5],a[21]);
    x2r = FADD(a[12],a[28]);
    x2i = FADD(a[13],a[29]);
    x3r = FSUB(a[12],a[28]);
    x3i = FSUB(a[13],a[29]);
    y2r = FADD(x0r,x2r);
    y2i = FADD(x0i,x2i);
    y6r = FSUB(x0r,x2r);
    y6i = FSUB(x0i,x2i);
    x0r = FSUB(x1r,x3i);
    x0i = FADD(x1i,x3r);
    y10r = FMUL(wn4r, FSUB(x0r,x0i),Q1_14);
    y10i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    x0r = FADD(x1r,x3i);
    x0i = FSUB(x1i,x3r);
    y14r = FMUL(wn4r,FADD(x0r,x0i),Q1_14);
    y14i = FMUL(wn4r,FSUB(x0i,x0r),Q1_14);
    x0r = FADD(a[6],a[22]);
    x0i = FADD(a[7],a[23]);
    x1r = FSUB(a[6],a[22]);
    x1i = FSUB(a[7],a[23]);
    x2r = FADD(a[14],a[30]);
    x2i = FADD(a[15],a[31]);
    x3r = FSUB(a[14],a[30]);
    x3i = FSUB(a[15],a[31]);
    y3r = FADD(x0r,x2r);
    y3i = FADD(x0i,x2i);
    y7r = FSUB(x0r,x2r);
    y7i = FSUB(x0i,x2i);
    x0r = FSUB(x1r, x3i);
    x0i = FADD(x1i,x3r);
    y11r = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14));
    y11i = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14));
    x0r = FADD(x1r,x3i);
    x0i = FSUB(x1i,x3r);
    y15r = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
    y15i = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
    x0r = FSUB(y12r,y14r);
    x0i = FSUB(y12i,y14i);
    x1r = FADD(y12r,y14r);
    x1i = FADD(y12i,y14i);
    x2r = FSUB(y13r,y15r);
    x2i = FSUB(y13i,y15i);
    x3r = FADD(y13r,y15r);
    x3i = FADD(y13i,y15i);
    a[24] = FADD(x0r,x2r)>>1;
    a[25] = FADD(x0i,x2i)>>1;
    a[26] = FSUB(x0r,x2r)>>1;
    a[27] = FSUB(x0i,x2i)>>1;
    a[28] = FSUB(x1r,x3i)>>1;
    a[29] = FADD(x1i,x3r)>>1;
    a[30] = FADD(x1r,x3i)>>1;
    a[31] = FSUB(x1i,x3r)>>1;
    x0r = FADD(y8r,y10r);
    x0i = FADD(y8i,y10i);
    x1r = FSUB(y8r,y10r);
    x1i = FSUB(y8i,y10i);
    x2r = FADD(y9r,y11r);
    x2i = FADD(y9i,y11i);
    x3r = FSUB(y9r,y11r);
    x3i = FSUB(y9i,y11i);
    a[16] = FADD(x0r,x2r)>>1;
    a[17] = FADD(x0i,x2i)>>1;
    a[18] = FSUB(x0r,x2r)>>1;
    a[19] = FSUB(x0i,x2i)>>1;
    a[20] = FSUB(x1r,x3i)>>1;
    a[21] = FADD(x1i,x3r)>>1;
    a[22] = FADD(x1r,x3i)>>1;
    a[23] = FSUB(x1i,x3r)>>1;
    x0r = FSUB(y5r,y7i);
    x0i = FADD(y5i,y7r);
    x2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    x0r = FADD(y5r,y7i);
    x0i = FSUB(y5i,y7r);
    x3r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x3i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    x0r = FSUB(y4r,y6i);
    x0i = FADD(y4i,y6r);
    x1r = FADD(y4r,y6i);
    x1i = FSUB(y4i,y6r);
    a[8] = FADD(x0r,x2r)>>1;
    a[9] = FADD(x0i,x2i)>>1;
    a[10] = FSUB(x0r,x2r)>>1;
    a[11] = FSUB(x0i,x2i)>>1;
    a[12] = FSUB(x1r,x3i)>>1;
    a[13] = FADD(x1i,x3r)>>1;
    a[14] = FADD(x1r,x3i)>>1;
    a[15] = FSUB(x1i,x3r)>>1;
    x0r = FADD(y0r,y2r);
    x0i = FADD(y0i,y2i);
    x1r = FSUB(y0r,y2r);
    x1i = FSUB(y0i,y2i);
    x2r = FADD(y1r,y3r);
    x2i = FADD(y1i,y3i);
    x3r = FSUB(y1r,y3r);
    x3i = FSUB(y1i,y3i);
    a[0] = FADD(x0r,x2r)>>1;
    a[1] = FADD(x0i,x2i)>>1;
    a[2] = FSUB(x0r,x2r)>>1;
    a[3] = FSUB(x0i,x2i)>>1;
    a[4] = FSUB(x1r,x3i)>>1;
    a[5] = FADD(x1i,x3r)>>1;
    a[6] = FADD(x1r,x3i)>>1;
    a[7] = FSUB(x1i,x3r)>>1;
}


void fxcftf162(FIXED16 *a, FIXED16 *w)
{
    FIXED32 wn4r, wk1r, wk1i, wk2r, wk2i, wk3r, wk3i, 
        x0r, x0i, x1r, x1i, x2r, x2i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i, 
        y4r, y4i, y5r, y5i, y6r, y6i, y7r, y7i, 
        y8r, y8i, y9r, y9i, y10r, y10i, y11r, y11i, 
        y12r, y12i, y13r, y13i, y14r, y14i, y15r, y15i;
    
    wn4r = w[1];
    wk1r = w[4];
    wk1i = w[5];
    wk3r = w[6];
    wk3i = -w[7];
    wk2r = w[8];
    wk2i = w[9];
    x1r = FSUB(a[0],a[17]);
    x1i = FADD(a[1],a[16]);
    x0r = FSUB(a[8],a[25]);
    x0i = FADD(a[9],a[24]);
    x2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    y0r = FADD(x1r,x2r);
    y0i = FADD(x1i,x2i);
    y4r = FSUB(x1r,x2r);
    y4i = FSUB(x1i,x2i);
    x1r = FADD(a[0],a[17]);
    x1i = FSUB(a[1],a[16]);
    x0r = FADD(a[8],a[25]);
    x0i = FSUB(a[9],a[24]);
    x2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    y8r = FSUB(x1r,x2i);
    y8i = FADD(x1i,x2r);
    y12r = FADD(x1r,x2i);
    y12i = FSUB(x1i,x2r);
    x0r = FSUB(a[2],a[19]);
    x0i = FADD(a[3],a[18]);
    x1r = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
    x1i = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
    x0r = FSUB(a[10],a[27]);
    x0i = FADD(a[11],a[26]);
    x2r = FSUB(FMUL(wk3i,x0r,Q1_14),FMUL(wk3r,x0i,Q1_14));
    x2i = FADD(FMUL(wk3i,x0i,Q1_14),FMUL(wk3r,x0r,Q1_14));
    y1r = FADD(x1r,x2r);
    y1i = FADD(x1i,x2i);
    y5r = FSUB(x1r,x2r);
    y5i = FSUB(x1i,x2i);
    x0r = FADD(a[2],a[19]);
    x0i = FSUB(a[3],a[18]);
    x1r = FSUB(FMUL(wk3r,x0r,Q1_14), FMUL(wk3i,x0i,Q1_14));
    x1i = FADD(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14));
    x0r = FADD(a[10],a[27]);
    x0i = FSUB(a[11],a[26]);
    x2r = FADD(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
    x2i = FSUB(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
    y9r = FSUB(x1r,x2r);
    y9i = FSUB(x1i,x2i);
    y13r = FADD(x1r,x2r);
    y13i = FADD(x1i,x2i);
    x0r = FSUB(a[4],a[21]);
    x0i = FADD(a[5],a[20]);
    x1r = FSUB(FMUL(wk2r,x0r,Q1_14),FMUL(wk2i,x0i,Q1_14));
    x1i = FADD(FMUL(wk2r,x0i,Q1_14),FMUL(wk2i,x0r,Q1_14));
    x0r = FSUB(a[12],a[29]);
    x0i = FADD(a[13],a[28]);
    x2r = FSUB(FMUL(wk2i,x0r,Q1_14),FMUL(wk2r,x0i,Q1_14));
    x2i = FADD(FMUL(wk2i,x0i,Q1_14),FMUL(wk2r,x0r,Q1_14));
    y2r = FADD(x1r,x2r);
    y2i = FADD(x1i,x2i);
    y6r = FSUB(x1r,x2r);
    y6i = FSUB(x1i,x2i);
    x0r = FADD(a[4], a[21]);
    x0i = FSUB(a[5],a[20]);
    x1r = FSUB(FMUL(wk2i,x0r,Q1_14),FMUL(wk2r,x0i,Q1_14));
    x1i = FADD(FMUL(wk2i,x0i,Q1_14),FMUL(wk2r,x0r,Q1_14));
    x0r = FADD(a[12],a[29]);
    x0i = FSUB(a[13],a[28]);
    x2r = FSUB(FMUL(wk2r,x0r,Q1_14),FMUL(wk2i,x0i,Q1_14));
    x2i = FADD(FMUL(wk2r,x0i,Q1_14),FMUL(wk2i,x0r,Q1_14));
    y10r = FSUB(x1r,x2r);
    y10i = FSUB(x1i,x2i);
    y14r = FADD(x1r,x2r);
    y14i = FADD(x1i,x2i);
    x0r = FSUB(a[6],a[23]);
    x0i = FADD(a[7],a[22]);
    x1r = FSUB(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14));
    x1i = FADD(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14));
    x0r = FSUB(a[14],a[31]);
    x0i = FADD(a[15],a[30]);
    x2r = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14));
    x2i = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14));
    y3r = FADD(x1r,x2r);
    y3i = FADD(x1i,x2i);
    y7r = FSUB(x1r,x2r);
    y7i = FSUB(x1i,x2i);
    x0r = FADD(a[6],a[23]);
    x0i = FSUB(a[7],a[22]);
    x1r = FADD(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14));
    x1i = FSUB(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14));
    x0r = FADD(a[14],a[31]);
    x0i = FSUB(a[15],a[30]);
    x2r = FSUB(FMUL(wk3i,x0r,Q1_14),FMUL(wk3r,x0i,Q1_14));
    x2i = FADD(FMUL(wk3i,x0i,Q1_14),FMUL(wk3r,x0r,Q1_14));
    y11r = FADD(x1r,x2r);
    y11i = FADD(x1i,x2i);
    y15r = FSUB(x1r,x2r);
    y15i = FSUB(x1i,x2i);
    x1r = FADD(y0r,y2r);
    x1i = FADD(y0i,y2i);
    x2r = FADD(y1r,y3r);
    x2i = FADD(y1i, y3i);
    a[0] = FADD(x1r,x2r)>>1;
    a[1] = FADD(x1i,x2i)>>1;
    a[2] = FSUB(x1r,x2r)>>1;
    a[3] = FSUB(x1i,x2i)>>1;
    x1r = FSUB(y0r,y2r);
    x1i = FSUB(y0i,y2i);
    x2r = FSUB(y1r,y3r);
    x2i = FSUB(y1i,y3i);
    a[4] = FSUB(x1r,x2i)>>1;
    a[5] = FADD(x1i,x2r)>>1;
    a[6] = FADD(x1r,x2i)>>1;
    a[7] = FSUB(x1i,x2r)>>1;
    x1r = FSUB(y4r,y6i);
    x1i = FADD(y4i,y6r);
    x0r = FSUB(y5r,y7i);
    x0i = FADD(y5i,y7r);
    x2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    a[8] = FADD(x1r,x2r)>>1;
    a[9] = FADD(x1i, x2i)>>1;
    a[10] = FSUB(x1r,x2r)>>1;
    a[11] = FSUB(x1i,x2i)>>1;
    x1r = FADD(y4r,y6i);
    x1i = FSUB(y4i,y6r);
    x0r = FADD(y5r,y7i);
    x0i = FSUB(y5i,y7r);
    x2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    a[12] = FSUB(x1r,x2i)>>1;
    a[13] = FADD(x1i,x2r)>>1;
    a[14] = FADD(x1r,x2i)>>1;
    a[15] = FSUB(x1i,x2r)>>1;
    x1r = FADD(y8r,y10r);
    x1i = FADD(y8i,y10i);
    x2r = FSUB(y9r,y11r);
    x2i = FSUB(y9i,y11i);
    a[16] = FADD(x1r,x2r)>>1;
    a[17] = FADD(x1i,x2i)>>1;
    a[18] = FSUB(x1r,x2r)>>1;
    a[19] = FSUB(x1i,x2i)>>1;
    x1r = FSUB(y8r,y10r);
    x1i = FSUB(y8i,y10i);
    x2r = FADD(y9r,y11r);
    x2i = FADD(y9i, y11i);
    a[20] = FSUB(x1r,x2i)>>1;
    a[21] = FADD(x1i,x2r)>>1;
    a[22] = FADD(x1r,x2i)>>1;
    a[23] = FSUB(x1i,x2r)>>1;
    x1r = FSUB(y12r,y14i);
    x1i = FADD(y12i,y14r);
    x0r = FADD(y13r,y15i);
    x0i = FSUB(y13i,y15r);
    x2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    a[24] = FADD(x1r ,x2r)>>1;
    a[25] = FADD(x1i,x2i)>>1;
    a[26] = FSUB(x1r,x2r)>>1;
    a[27] = FSUB(x1i,x2i)>>1;
    x1r = FADD(y12r,y14i);
    x1i = FSUB(y12i,y14r);
    x0r = FSUB(y13r,y15i);
    x0i = FADD(y13i,y15r);
    x2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    x2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    a[28] = FSUB(x1r,x2i)>>1;
    a[29] = FADD(x1i,x2r)>>1;
    a[30] = FADD(x1r,x2i)>>1;
    a[31] = FSUB(x1i,x2r)>>1;
}


void fxcftf081(FIXED16 *a, FIXED16 *w)
{
    FIXED32 wn4r, x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i, 
        y4r, y4i, y5r, y5i, y6r, y6i, y7r, y7i;
    
    wn4r = w[1];
    x0r = FADD(a[0],a[8]);		// a[0] + a[8];
    x0i = FADD(a[1],a[9]);		// a[1] + a[9];
    x1r = FSUB(a[0],a[8]);		// a[0] - a[8];
    x1i = FSUB(a[1],a[9]);		// a[1] - a[9];
    x2r = FADD(a[4],a[12]);		// a[4] + a[12];	
    x2i = FADD(a[5],a[13]);		// a[5] + a[13];	
    x3r = FSUB(a[4],a[12]);		// a[4] - a[12];
    x3i = FSUB(a[5],a[13]);		// a[5] - a[13];	
    y0r = FADD(x0r,x2r);		// x0r + x2r;
    y0i = FADD(x0i,x2i);		// x0i + x2i;
    y2r = FSUB(x0r,x2r);		// x0r - x2r;
    y2i = FSUB(x0i,x2i);		// x0i - x2i;
    y1r = FSUB(x1r,x3i);		// x1r - x3i;
    y1i = FADD(x1i,x3r);		// x1i + x3r;
    y3r = FADD(x1r,x3i);		// x1r + x3i;
    y3i = FSUB(x1i,x3r);		// x1i - x3r;
    x0r = FADD(a[2],a[10]);		// a[2] + a[10];	
    x0i = FADD(a[3],a[11]);		// a[3] + a[11];
    x1r = FSUB(a[2],a[10]);		// a[2] - a[10];	
    x1i = FSUB(a[3],a[11]);		// a[3] - a[11];	
    x2r = FADD(a[6],a[14]);		// a[6] + a[14];
    x2i = FADD(a[7],a[15]);		// a[7] + a[15];	
    x3r = FSUB(a[6],a[14]);		// a[6] - a[14];
    x3i = FSUB(a[7],a[15]);		// a[7] - a[15];	
    y4r = FADD(x0r,x2r);		// x0r + x2r;
    y4i = FADD(x0i,x2i);		// x0i + x2i;
    y6r = FSUB(x0r,x2r);		// x0r - x2r;
    y6i = FSUB(x0i,x2i);		// x0i - x2i;
    x0r = FSUB(x1r,x3i);		// x1r - x3i;
    x0i = FADD(x1i,x3r);		// x1i + x3r;
    x2r = FADD(x1r,x3i);		// x1r + x3i;
    x2i = FSUB(x1i,x3r);		// x1i - x3r;
    y5r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);	// wn4r * (x0r - x0i);
    y5i = FMUL(wn4r,FADD(x0r,x0i),Q1_14);	// wn4r * (x0r + x0i);
    y7r = FMUL(wn4r,FSUB(x2r,x2i),Q1_14);	// wn4r * (x2r - x2i);
    y7i = FMUL(wn4r,FADD(x2r,x2i),Q1_14);	// wn4r * (x2r + x2i);
    a[8] = FADD(y1r,y5r)>>1;		// y1r + y5r;
    a[9] = FADD(y1i,y5i)>>1;		// y1i + y5i;
    a[10] = FSUB(y1r,y5r)>>1;		// y1r - y5r;	
    a[11] = FSUB(y1i,y5i)>>1;		// y1i - y5i;
    a[12] = FSUB(y3r,y7i)>>1;		// y3r - y7i;
    a[13] = FADD(y3i,y7r)>>1;		// y3i + y7r;
    a[14] = FADD(y3r,y7i)>>1;		// y3r + y7i;
    a[15] = FSUB(y3i,y7r)>>1;		// y3i - y7r;
    a[0] = FADD(y0r,y4r)>>1;		// y0r + y4r;
    a[1] = FADD(y0i,y4i)>>1;		// y0i + y4i;
    a[2] = FSUB(y0r,y4r)>>1;		// y0r - y4r;
    a[3] = FSUB(y0i,y4i)>>1;		// y0i - y4i;
    a[4] = FSUB(y2r,y6i)>>1;		// y2r - y6i;
    a[5] = FADD(y2i,y6r)>>1;		// y2i + y6r;
    a[6] = FADD(y2r,y6i)>>1;		// y2r + y6i;
    a[7] = FSUB(y2i,y6r)>>1;		// y2i - y6r;
}


void fxcftf082(FIXED16 *a, FIXED16 *w)
{
    FIXED32 wn4r, wk1r, wk1i, x0r, x0i, x1r, x1i, 
        y0r, y0i, y1r, y1i, y2r, y2i, y3r, y3i, 
        y4r, y4i, y5r, y5i, y6r, y6i, y7r, y7i;
    
    wn4r = w[1];
    wk1r = w[2];
    wk1i = w[3];
    y0r = FSUB(a[0],a[9]);
    y0i = FADD(a[1],a[8]);
    y1r = FADD(a[0],a[9]);
    y1i = FSUB(a[1],a[8]);
    x0r = FSUB(a[4],a[13]);
    x0i = FADD(a[5],a[12]);
    y2r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    y2i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    x0r = FADD(a[4],a[13]);
    x0i = FSUB(a[5],a[12]);
    y3r = FMUL(wn4r,FSUB(x0r,x0i),Q1_14);
    y3i = FMUL(wn4r,FADD(x0i,x0r),Q1_14);
    x0r = FSUB(a[2],a[11]);
    x0i = FADD(a[3],a[10]);
    y4r = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
    y4i = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
    x0r = FADD(a[2],a[11]);
    x0i = FSUB(a[3],a[10]);
    y5r = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14));
    y5i = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14));
    x0r = FSUB(a[6],a[15]);
    x0i = FADD(a[7],a[14]);
    y6r = FSUB(FMUL(wk1i,x0r,Q1_14),FMUL(wk1r,x0i,Q1_14));
    y6i = FADD(FMUL(wk1i,x0i,Q1_14),FMUL(wk1r,x0r,Q1_14));
    x0r = FADD(a[6],a[15]);
    x0i = FSUB(a[7],a[14]);
    y7r = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
    y7i = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
    x0r = FADD(y0r,y2r);
    x0i = FADD(y0i,y2i);
    x1r = FADD(y4r,y6r);
    x1i = FADD(y4i,y6i);
    a[0] = FADD(x0r,x1r)>>1;
    a[1] = FADD(x0i,x1i)>>1;
    a[2] = FSUB(x0r, x1r)>>1;
    a[3] = FSUB(x0i,x1i)>>1;
    x0r = FSUB(y0r,y2r);
    x0i = FSUB(y0i,y2i);
    x1r = FSUB(y4r,y6r);
    x1i = FSUB(y4i,y6i);
    a[4] = FSUB(x0r,x1i)>>1;
    a[5] = FADD(x0i,x1r)>>1;
    a[6] = FADD(x0r,x1i)>>1;
    a[7] = FSUB(x0i,x1r)>>1;
    x0r = FSUB(y1r,y3i);
    x0i = FADD(y1i,y3r);
    x1r = FSUB(y5r,y7r);
    x1i = FSUB(y5i,y7i);
    a[8] = FADD(x0r,x1r)>>1;
    a[9] = FADD(x0i,x1i)>>1;
    a[10] = FSUB(x0r,x1r)>>1;
    a[11] = FSUB(x0i,x1i)>>1;
    x0r = FADD(y1r,y3i);
    x0i = FSUB(y1i,y3r);
    x1r = FADD(y5r, y7r);
    x1i = FADD(y5i,y7i);
    a[12] = FSUB(x0r,x1i)>>1;
    a[13] = FADD(x0i,x1r)>>1;
    a[14] = FADD(x0r,x1i)>>1;
    a[15] = FSUB(x0i,x1r)>>1;
}


/*void fxcftf040(FIXED16 *a)
{
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    x0r = FADD(a[0],a[4]);
    x0i = FADD(a[1],a[5]);
    x1r = FSUB(a[0],a[4]);
    x1i = FSUB(a[1],a[5]);
    x2r = FADD(a[2],a[6]);
    x2i = FADD(a[3],a[7]);
    x3r = FSUB(a[2],a[6]);
    x3i = FSUB(a[3],a[7]);
    a[0] = FADD(x0r,x2r)>>1;
    a[1] = FADD(x0i,x2i)>>1;
    a[2] = FSUB(x1r,x3i)>>1;
    a[3] = FADD(x1i,x3r)>>1;
    a[4] = FSUB(x0r,x2r)>>1;
    a[5] = FSUB(x0i,x2i)>>1;
    a[6] = FADD(x1r,x3i)>>1;
    a[7] = FSUB(x1i,x3r)>>1;
}*/


/*void fxcftb040(FIXED16 *a)
{
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    x0r = FADD(a[0],a[4]);
    x0i = FADD(a[1],a[5]);
    x1r = FSUB(a[0],a[4]);
    x1i = FSUB(a[1],a[5]);
    x2r = FADD(a[2],a[6]);
    x2i = FADD(a[3],a[7]);
    x3r = FSUB(a[2],a[6]);
    x3i = FSUB(a[3],a[7]);
    a[0] = FADD(x0r,x2r)>>1;
    a[1] = FADD(x0i,x2i)>>1;
    a[2] = FADD(x1r,x3i)>>1;
    a[3] = FSUB(x1i,x3r)>>1;
    a[4] = FSUB(x0r,x2r)>>1;
    a[5] = FSUB(x0i,x2i)>>1;
    a[6] = FSUB(x1r,x3i)>>1;
    a[7] = FADD(x1i,x3r)>>1;
}*/


/*void fxcftx020(FIXED16 *a)
{
    FIXED32 x0r, x0i;
    
    x0r = FSUB(a[0],a[2]);
    x0i = FSUB(a[1],a[3]);
    a[0] = FADD(a[0],a[2])>>1;
    a[1] = FADD(a[1],a[3])>>1;
    a[2] = x0r>>1;
    a[3] = x0i>>1;
}*/


void fxrftfsub(INT32 n, FIXED16 *a, INT32 nc, FIXED16 *c)
{
    INT32 j, k, kk, ks, m;
    FIXED32 wkr, wki, xr, xi, yr, yi;
    
    m = n >> 1;
    ks = 2 * nc / m;
    kk = 0;
    for (j = 2; j < m; j += 2) {
        k = n - j;
        kk += ks;
        wkr = FSUB(CONST_0_5_Q1_14,c[nc - kk]);	// 0.5f - c[nc - kk];
        wki = c[kk];				// c[kk];
        xr = FSUB(a[j],a[k]);			// a[j] - a[k];
        xi = FADD(a[j + 1],a[k + 1]);	// a[j + 1] + a[k + 1];
        yr = FSUB(FMUL(wkr,xr,Q1_14),FMUL(wki,xi,Q1_14));	// wkr * xr - wki * xi;
        yi = FADD(FMUL(wkr,xi,Q1_14),FMUL(wki,xr,Q1_14));	// wkr * xi + wki * xr;
        a[j] = FSUB(a[j],yr);					// -= yr;		
        a[j + 1] = FSUB(a[j+1],yi);				// -= yi;		
        a[k] = FADD(a[k],yr);					// += yr;	
        a[k + 1] = FSUB(a[k+1],yi);				// -= yi;	
    }
}


void fxrftbsub(INT32 n, FIXED16 *a, INT32 nc, FIXED16 *c)
{
    INT32 j, k, kk, ks, m;
    FIXED32 wkr, wki, xr, xi, yr, yi;
    
    m = n >> 1;
    ks = 2 * nc / m;
    kk = 0;
    for (j = 2; j < m; j += 2) {
        k = n - j;
        kk += ks;
        wkr = FSUB(CONST_0_5_Q1_14,c[nc - kk]); //0.5f - c[nc - kk];
        wki = c[kk];
        xr = FSUB(a[j],a[k]); // a[j] - a[k];
        xi = FADD(a[j + 1],a[k + 1]); // a[j + 1] + a[k + 1];
        yr = FADD(FMUL(wkr,xr,Q1_14),FMUL(wki,xi,Q1_14)); // wkr * xr + wki * xi;
        yi = FSUB(FMUL(wkr,xi,Q1_14),FMUL(wki,xr,Q1_14)); // wkr * xi - wki * xr;
        a[j] = FSUB(a[j],yr);	// -= yr;
        a[j + 1] = FSUB(a[j+1],yi);	// -= yi;
        a[k] = FADD(a[k],yr);	// += yr;
        a[k + 1] = 	FSUB(a[k+1],yi); // -= yi;
    }
}

#endif // ZEF_FIXED_POINT // ZEF_SPLIT_RADIX
