/*
Company :		ZEFYR Technologies
File :			FFT1Db.c
Author :		Matthieu Perreira Da Silva (based on Takuya OOURA (email: ooura@kurims.kyoto-u.ac.jp) algorithms) 
Fisrt release :	15/04/2003

Description : Fast Fourier Transform

Features : 
    dimension   :one
    data length :power of 2
    decimation  :frequency
    radix       :4, 2
    data        :inplace
    table       :use

Functions : 

    cdft: Complex Discrete Fourier Transform
    rdft: Real Discrete Fourier Transform

Function prototypes : 

    void cdft(INT32 p_size, INT32 p_sign, FIXED16 *p_data, INT32 *p_inPlace, FIXED16 *p_trigo);
	void rdft(INT32 p_size, INT32 p_sign, FIXED16 *p_data, INT32 *p_inPlace, FIXED16 *p_trigo);

Comments : see file FFTs.txt

Necessary package : none

Appendix :
    The cos/sin table is recalculated when the larger table required.
    w[] and ip[] are compatible with all routines.
*/

#include <math.h>
#include "customTypes.h"

#include "fxFFT1Db.h"

#if defined(ZEF_FIXED_POINT) && defined(ZEF_RADIX_2_4)

void fxcdft(INT32 n, INT32 isgn, FIXED16 *a, INT32 *ip, FIXED16 *w)
{   
    if (n > (ip[0] << 2)) {
        fxmakewt(n >> 2, ip, w);
    }
    if (n > 4) {
        if (isgn >= 0) {
            fxbitrv2(n, ip + 2, a);
            fxcftfsub(n, a, w);
        } else {
            fxbitrv2conj(n, ip + 2, a);
            fxcftbsub(n, a, w);
        }
    } else if (n == 4) {
        fxcftfsub(n, a, w);
    }
}


void fxrdft(INT32 n, INT32 isgn, FIXED16 *a, INT32 *ip, FIXED16 *w)
{
    INT32 nw, nc;
    FIXED32 xi;
    
    nw = ip[0];
    if (n > (nw << 2)) {
        nw = n >> 2;
        fxmakewt(nw, ip, w);
    }
    nc = ip[1];
    if (n > (nc << 2)) {
        nc = n >> 2;
        fxmakect(nc, ip, w + nw);
    }
    if (isgn >= 0) {
        if (n > 4) {
            fxbitrv2(n, ip + 2, a);
            fxcftfsub(n, a, w);
            fxrftfsub(n, a, nc, w + nw);
        } else if (n == 4) {
            fxcftfsub(n, a, w);
        }
        xi = FSUB(a[0],a[1]);
        a[0] = FADD(a[0],a[1]);
        a[1] = xi;
    } else {
        a[1] = FSUB(a[0],a[1])>>1;
        a[0] = FSUB(a[0],a[1]);
        if (n > 4) {
            fxrftbsub(n, a, nc, w + nw);
            fxbitrv2(n, ip + 2, a);
            fxcftbsub(n, a, w);
        } else if (n == 4) {
            fxcftfsub(n, a, w);
        }
    }
}


/* -------- initializing routines -------- */

void fxmakewt(INT32 nw, INT32 *ip, FIXED16 *w)
{
    INT32 j, nwh;
    REAL32 delta;
	FIXED32 x,y;
    
    ip[0] = nw;
    ip[1] = 1;
    if (nw > 2) {
        nwh = nw >> 1;
        delta = (REAL32)atan(1.0f) / nwh;
        w[0] = ITOFIX16(1,Q1_14);
        w[1] = ITOFIX16(0,Q1_14);;
        w[nwh] = FTOFIX16(cos(delta * nwh),Q1_14);
        w[nwh + 1] = w[nwh];
        if (nwh > 2) {
            for (j = 2; j < nwh; j += 2) {
                x = FTOFIX16(cos(delta * j),Q1_14);
                y = FTOFIX16(sin(delta * j),Q1_14);
                w[j] = x;
                w[j + 1] = y;
                w[nw - j] = y;
                w[nw - j + 1] = x;
            }
            fxbitrv2(nw, ip + 2, w);
        }
    }
}


void fxmakect(INT32 nc, INT32 *ip, FIXED16 *c)
{
    INT32 j, nch;
    REAL32 delta;
    
    ip[1] = nc;
    if (nc > 1) {
        nch = nc >> 1;
        delta = (REAL32)atan(1.0f) / nch;
        c[0] = FTOFIX16(cos(delta * nch),Q1_14);
        c[nch] = c[0]>>1;
        for (j = 1; j < nch; j++) {
            c[j] = FTOFIX16(cos(delta * j),Q1_14)>>1;
            c[nc - j] = FTOFIX16(sin(delta * j),Q1_14)>>1;
        }
    }
}


/* -------- child routines -------- */


void fxbitrv2(INT32 n, INT32 *ip, FIXED16 *a)
{
    INT32 j, j1, k, k1, l, m, m2;
    FIXED32 xr, xi, yr, yi;
    
    ip[0] = 0;
    l = n;
    m = 1;
    while ((m << 3) < l) {
        l >>= 1;
        for (j = 0; j < m; j++) {
            ip[m + j] = ip[j] + l;
        }
        m <<= 1;
    }
    m2 = 2 * m;
    if ((m << 3) == l) {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 -= m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            j1 = 2 * k + m2 + ip[k];
            k1 = j1 + m2;
            xr = a[j1];
            xi = a[j1 + 1];
            yr = a[k1];
            yi = a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
        }
    } else {
        for (k = 1; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += m2;
                xr = a[j1];
                xi = a[j1 + 1];
                yr = a[k1];
                yi = a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
        }
    }
}


void fxbitrv2conj(INT32 n, INT32 *ip, FIXED16 *a)
{
    INT32 j, j1, k, k1, l, m, m2;
    FIXED32 xr, xi, yr, yi;
    
    ip[0] = 0;
    l = n;
    m = 1;
    while ((m << 3) < l) {
        l >>= 1;
        for (j = 0; j < m; j++) {
            ip[m + j] = ip[j] + l;
        }
        m <<= 1;
    }
    m2 = 2 * m;
    if ((m << 3) == l) {
        for (k = 0; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 -= m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += 2 * m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 2 * k + ip[k];
            a[k1 + 1] = -a[k1 + 1];
            j1 = k1 + m2;
            k1 = j1 + m2;
            xr = a[j1];
            xi = -a[j1 + 1];
            yr = a[k1];
            yi = -a[k1 + 1];
            a[j1] = yr;
            a[j1 + 1] = yi;
            a[k1] = xr;
            a[k1 + 1] = xi;
            k1 += m2;
            a[k1 + 1] = -a[k1 + 1];
        }
    } else {
        a[1] = -a[1];
        a[m2 + 1] = -a[m2 + 1];
        for (k = 1; k < m; k++) {
            for (j = 0; j < k; j++) {
                j1 = 2 * j + ip[k];
                k1 = 2 * k + ip[j];
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
                j1 += m2;
                k1 += m2;
                xr = a[j1];
                xi = -a[j1 + 1];
                yr = a[k1];
                yi = -a[k1 + 1];
                a[j1] = yr;
                a[j1 + 1] = yi;
                a[k1] = xr;
                a[k1 + 1] = xi;
            }
            k1 = 2 * k + ip[k];
            a[k1 + 1] = -a[k1 + 1];
            a[k1 + m2 + 1] = -a[k1 + m2 + 1];
        }
    }
}


void fxcftfsub(INT32 n, FIXED16 *a, FIXED16 *w)
{
    INT32 j, j1, j2, j3, l;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    l = 2;
    if (n > 8) {
        fxcft1st(n, a, w);
        l = 8;
        while ((l << 2) < n) {
            fxcftmdl(n, l, a, w);
            l <<= 2;
        }
    }
    if ((l << 2) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = FADD(a[j],a[j1]);
            x0i = FADD(a[j + 1],a[j1 + 1]);
            x1r = FSUB(a[j],a[j1]);
            x1i = FSUB(a[j + 1],a[j1 + 1]);
            x2r = FADD(a[j2],a[j3]);
            x2i = FADD(a[j2 + 1],a[j3 + 1]);
            x3r = FSUB(a[j2],a[j3]);
            x3i = FSUB(a[j2 + 1],a[j3 + 1]);
            a[j] = FADD(x0r, x2r)>>1;
            a[j + 1] = FADD(x0i,x2i)>>1;
            a[j2] = FSUB(x0r,x2r)>>1;
            a[j2 + 1] = FSUB(x0i, x2i)>>1;
            a[j1] = FSUB(x1r,x3i)>>1;
            a[j1 + 1] = FADD(x1i,x3r)>>1;
            a[j3] = FADD(x1r,x3i)>>1;
            a[j3 + 1] = FSUB(x1i,x3r)>>1;
        }
    } else {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = FSUB(a[j], a[j1]);
            x0i = FSUB(a[j + 1], a[j1 + 1]);
            a[j] = FADD(a[j], a[j1])>>1;
            a[j + 1] = FADD(a[j + 1], a[j1 + 1])>>1;
            a[j1] = x0r>>1;
            a[j1 + 1] = x0i>>1;
        }
    }
}


void fxcftbsub(INT32 n, FIXED16 *a, FIXED16 *w)
{
    INT32 j, j1, j2, j3, l;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    l = 2;
    if (n > 8) {
        fxcft1st(n, a, w);
        l = 8;
        while ((l << 2) < n) {
            fxcftmdl(n, l, a, w);
            l <<= 2;
        }
    }
    if ((l << 2) == n) {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = FADD(a[j],a[j1]);
            x0i = FSUB(-a[j + 1],a[j1 + 1]);
            x1r = FSUB(a[j],a[j1]);
            x1i = FADD(-a[j + 1],a[j1 + 1]);
            x2r = FADD(a[j2],a[j3]);
            x2i = FADD(a[j2 + 1],a[j3 + 1]);
            x3r = FSUB(a[j2],a[j3]);
            x3i = FSUB(a[j2 + 1],a[j3 + 1]);
            a[j] = FADD(x0r,x2r)>>1;
            a[j + 1] = FSUB(x0i,x2i)>>1;
            a[j2] = FSUB(x0r,x2r)>>1;
            a[j2 + 1] = FADD(x0i, x2i)>>1;
            a[j1] = FSUB(x1r,x3i)>>1;
            a[j1 + 1] = FSUB(x1i,x3r)>>1;
            a[j3] = FADD(x1r,x3i)>>1;
            a[j3 + 1] = FADD(x1i,x3r)>>1;
        }
    } else {
        for (j = 0; j < l; j += 2) {
            j1 = j + l;
            x0r = FSUB(a[j],a[j1]);
            x0i = FADD(-a[j + 1],a[j1 + 1]);
            a[j] = FADD(a[j],a[j1])>>1;
            a[j + 1] = FSUB(-a[j + 1],a[j1 + 1])>>1;
            a[j1] = x0r>>1;
            a[j1 + 1] = x0i>>1;
        }
    }
}


void fxcft1st(INT32 n, FIXED16 *a, FIXED16 *w)
{
    INT32 j, k1, k2;
    FIXED32 wk1r, wk1i, wk2r, wk2i, wk3r, wk3i;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    x0r = FADD(a[0],a[2]);
    x0i = FADD(a[1],a[3]);
    x1r = FSUB(a[0],a[2]);
    x1i = FSUB(a[1],a[3]);
    x2r = FADD(a[4],a[6]);
    x2i = FADD(a[5],a[7]);
    x3r = FSUB(a[4],a[6]);
    x3i = FSUB(a[5],a[7]);
    a[0] = FADD(x0r,x2r)>>1;
    a[1] = FADD(x0i,x2i)>>1;
    a[4] = FSUB(x0r,x2r)>>1;
    a[5] = FSUB(x0i, x2i)>>1;
    a[2] = FSUB(x1r,x3i)>>1;
    a[3] = FADD(x1i, x3r)>>1;
    a[6] = FADD(x1r,x3i)>>1;
    a[7] = FSUB(x1i,x3r)>>1;
    wk1r = w[2];
    x0r = FADD(a[8],a[10]);
    x0i = FADD(a[9],a[11]);
    x1r = FSUB(a[8],a[10]);
    x1i = FSUB(a[9],a[11]);
    x2r = FADD(a[12],a[14]);
    x2i = FADD(a[13], a[15]);
    x3r = FSUB(a[12], a[14]);
    x3i = FSUB(a[13], a[15]);
    a[8] = FADD(x0r,x2r)>>1;
    a[9] = FADD(x0i,x2i)>>1;
    a[12] = FSUB(x2i,x0i)>>1;
    a[13] = FSUB(x0r,x2r)>>1;
    x0r = FSUB(x1r,x3i);
    x0i = FADD(x1i,x3r);
    a[10] = FMUL(wk1r,FSUB(x0r,x0i),Q1_14)>>1;
    a[11] = FMUL(wk1r,FADD(x0r,x0i),Q1_14)>>1;
    x0r = FADD(x3i,x1r);
    x0i = FSUB(x3r,x1i);
    a[14] = FMUL(wk1r,FSUB(x0i,x0r),Q1_14)>>1;
    a[15] = FMUL(wk1r,FADD(x0i,x0r),Q1_14)>>1;
    k1 = 0;
    for (j = 16; j < n; j += 16) {
        k1 += 2;
        k2 = 2 * k1;
        wk2r = w[k1];
        wk2i = w[k1 + 1];
        wk1r = w[k2];
        wk1i = w[k2 + 1];
        wk3r = FSUB(wk1r,FMUL(wk2i,wk1i,Q1_14)<<1);
        wk3i = FSUB(FMUL(wk2i,wk1r,Q1_14)<<1,wk1i);
        x0r = FADD(a[j],a[j + 2]);
        x0i = FADD(a[j + 1],a[j + 3]);
        x1r = FSUB(a[j],a[j + 2]);
        x1i = FSUB(a[j + 1],a[j + 3]);
        x2r = FADD(a[j + 4], a[j + 6]);
        x2i = FADD(a[j + 5],a[j + 7]);
        x3r = FSUB(a[j + 4],a[j + 6]);
        x3i = FSUB(a[j + 5],a[j + 7]);
        a[j] = FADD(x0r,x2r)>>1;
        a[j + 1] = FADD(x0i,x2i)>>1;
        x0r = FSUB(x0r,x2r);
        x0i = FSUB(x0i,x2i);
        a[j + 4] = FSUB(FMUL(wk2r,x0r,Q1_14),FMUL(wk2i,x0i,Q1_14))>>1;
        a[j + 5] = FADD(FMUL(wk2r,x0i,Q1_14),FMUL(wk2i,x0r,Q1_14))>>1;
        x0r = FSUB(x1r,x3i);
        x0i = FADD(x1i,x3r);
        a[j + 2] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14))>>1;
        a[j + 3] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14))>>1;
        x0r = FADD(x1r,x3i);
        x0i = FSUB(x1i, x3r);
        a[j + 6] = FSUB(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14))>>1;
        a[j + 7] = FADD(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14))>>1;
        wk1r = w[k2 + 2];
        wk1i = w[k2 + 3];
        wk3r = FSUB(wk1r,FMUL(wk2r,wk1i,Q1_14)<<1);
        wk3i = FSUB(FMUL(wk2r,wk1r,Q1_14)<<1, wk1i);
        x0r = FADD(a[j + 8],a[j + 10]);
        x0i = FADD(a[j + 9],a[j + 11]);
        x1r = FSUB(a[j + 8],a[j + 10]);
        x1i = FSUB(a[j + 9],a[j + 11]);
        x2r = FADD(a[j + 12],a[j + 14]);
        x2i = FADD(a[j + 13],a[j + 15]);
        x3r = FSUB(a[j + 12],a[j + 14]);
        x3i = FSUB(a[j + 13],a[j + 15]);
        a[j + 8] = FADD(x0r,x2r)>>1;
        a[j + 9] = FADD(x0i,x2i)>>1;
        x0r = FSUB(x0r,x2r);
        x0i = FSUB(x0i,x2i);
        a[j + 12] = FSUB(FMUL(-wk2i,x0r,Q1_14),FMUL(wk2r,x0i,Q1_14))>>1;
        a[j + 13] = FADD(FMUL(-wk2i,x0i,Q1_14),FMUL(wk2r,x0r,Q1_14))>>1;
        x0r = FSUB(x1r,x3i);
        x0i = FADD(x1i,x3r);
        a[j + 10] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14))>>1;
        a[j + 11] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i, x0r,Q1_14))>>1;
        x0r = FADD(x1r,x3i);
        x0i = FSUB(x1i,x3r);
        a[j + 14] = FSUB(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14))>>1;
        a[j + 15] = FADD(FMUL(wk3r, x0i,Q1_14),FMUL(wk3i,x0r,Q1_14))>>1;
    }
}


void fxcftmdl(INT32 n, INT32 l, FIXED16 *a, FIXED16 *w)
{
    INT32 j, j1, j2, j3, k, k1, k2, m, m2;
    FIXED32 wk1r, wk1i, wk2r, wk2i, wk3r, wk3i;
    FIXED32 x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i;
    
    m = l << 2;
    for (j = 0; j < l; j += 2) {
        j1 = j + l;
        j2 = j1 + l;
        j3 = j2 + l;
        x0r = FADD(a[j],a[j1]);
        x0i = FADD(a[j + 1],a[j1 + 1]);
        x1r = FSUB(a[j],a[j1]);
        x1i = FSUB(a[j + 1],a[j1 + 1]);
        x2r = FADD(a[j2],a[j3]);
        x2i = FADD(a[j2 + 1],a[j3 + 1]);
        x3r = FSUB(a[j2],a[j3]);
        x3i = FSUB(a[j2 + 1], a[j3 + 1]);
        a[j] = FADD(x0r,x2r);
        a[j + 1] = FADD(x0i,x2i);
        a[j2] = FSUB(x0r,x2r);
        a[j2 + 1] = FSUB(x0i,x2i);
        a[j1] = FSUB(x1r,x3i);
        a[j1 + 1] = FADD(x1i,x3r);
        a[j3] = FADD(x1r,x3i);
        a[j3 + 1] = FSUB(x1i, x3r);
    }
    wk1r = w[2];
    for (j = m; j < l + m; j += 2) {
        j1 = j + l;
        j2 = j1 + l;
        j3 = j2 + l;
        x0r = FADD(a[j],a[j1]);
        x0i = FADD(a[j + 1],a[j1 + 1]);
        x1r = FSUB(a[j],a[j1]);
        x1i = FSUB(a[j + 1],a[j1 + 1]);
        x2r = FADD(a[j2],a[j3]);
        x2i = FADD(a[j2 + 1],a[j3 + 1]);
        x3r = FSUB(a[j2],a[j3]);
        x3i = FSUB(a[j2 + 1],a[j3 + 1]);
        a[j] = FADD(x0r,x2r);
        a[j + 1] = FADD(x0i,x2i);
        a[j2] = FSUB(x2i,x0i);
        a[j2 + 1] = FSUB(x0r,x2r);
        x0r = FSUB(x1r,x3i);
        x0i = FADD(x1i,x3r);
        a[j1] = FMUL(wk1r,FSUB(x0r,x0i),Q1_14);
        a[j1 + 1] = FMUL(wk1r,FADD(x0r, x0i),Q1_14);
        x0r = FADD(x3i,x1r);
        x0i = FSUB(x3r,x1i);
        a[j3] = FMUL(wk1r,FSUB(x0i,x0r),Q1_14);
        a[j3 + 1] = FMUL(wk1r,FADD(x0i,x0r),Q1_14);
    }
    k1 = 0;
    m2 = 2 * m;
    for (k = m2; k < n; k += m2) {
        k1 += 2;
        k2 = 2 * k1;
        wk2r = w[k1];
        wk2i = w[k1 + 1];
        wk1r = w[k2];
        wk1i = w[k2 + 1];
        wk3r = FSUB(wk1r,FMUL(wk2i,wk1i,Q1_14)<<1);
        wk3i = FSUB(FMUL( wk2i,wk1r,Q1_14)<<1,wk1i);
        for (j = k; j < l + k; j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = FADD(a[j],a[j1]);
            x0i = FADD(a[j + 1],a[j1 + 1]);
            x1r = FSUB(a[j],a[j1]);
            x1i = FSUB(a[j + 1],a[j1 + 1]);
            x2r = FADD(a[j2],a[j3]);
            x2i = FADD(a[j2 + 1],a[j3 + 1]);
            x3r = FSUB(a[j2],a[j3]);
            x3i = FSUB(a[j2 + 1],a[j3 + 1]);
            a[j] = FADD(x0r,x2r);
            a[j + 1] = FADD(x0i,x2i);
            x0r = FSUB(x0r,x2r);
            x0i = FSUB(x0i,x2i);
            a[j2] = FSUB(FMUL(wk2r,x0r,Q1_14),FMUL(wk2i,x0i,Q1_14));
            a[j2 + 1] = FADD(FMUL(wk2r,x0i,Q1_14),FMUL(wk2i,x0r,Q1_14));
            x0r = FSUB(x1r,x3i);
            x0i = FADD(x1i,x3r);
            a[j1] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
            a[j1 + 1] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
            x0r = FADD(x1r,x3i);
            x0i = FSUB(x1i,x3r);
            a[j3] = FSUB(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14));
            a[j3 + 1] = FADD(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14));
        }
        wk1r = w[k2 + 2];
        wk1i = w[k2 + 3];
        wk3r = FSUB(wk1r,FMUL(wk2r,wk1i,Q1_14)<<1);
        wk3i = FSUB(FMUL(wk2r,wk1r,Q1_14)<<1,wk1i);
        for (j = k + m; j < l + (k + m); j += 2) {
            j1 = j + l;
            j2 = j1 + l;
            j3 = j2 + l;
            x0r = FADD(a[j],a[j1]);
            x0i = FADD(a[j + 1],a[j1 + 1]);
            x1r = FSUB(a[j],a[j1]);
            x1i = FSUB(a[j + 1],a[j1 + 1]);
            x2r = FADD(a[j2],a[j3]);
            x2i = FADD(a[j2 + 1],a[j3 + 1]);
            x3r = FSUB(a[j2],a[j3]);
            x3i = FSUB(a[j2 + 1],a[j3 + 1]);
            a[j] = FADD(x0r,x2r);
            a[j + 1] = FADD(x0i,x2i);
            x0r = FSUB(x0r,x2r);
            x0i = FSUB(x0i,x2i);
            a[j2] = FSUB(FMUL(-wk2i,x0r,Q1_14),FMUL(wk2r,x0i,Q1_14));
            a[j2 + 1] = FADD(FMUL(-wk2i,x0i,Q1_14),FMUL(wk2r,x0r,Q1_14));
            x0r = FSUB(x1r,x3i);
            x0i = FADD(x1i,x3r);
            a[j1] = FSUB(FMUL(wk1r,x0r,Q1_14),FMUL(wk1i,x0i,Q1_14));
            a[j1 + 1] = FADD(FMUL(wk1r,x0i,Q1_14),FMUL(wk1i,x0r,Q1_14));
            x0r = FADD(x1r,x3i);
            x0i = FSUB(x1i,x3r);
            a[j3] = FSUB(FMUL(wk3r,x0r,Q1_14),FMUL(wk3i,x0i,Q1_14));
            a[j3 + 1] = FADD(FMUL(wk3r,x0i,Q1_14),FMUL(wk3i,x0r,Q1_14));
        }
    }
}


void fxrftfsub(INT32 n, FIXED16 *a, INT32 nc, FIXED16 *c)
{
    INT32 j, k, kk, ks, m;
    FIXED32 wkr, wki, xr, xi, yr, yi;
    
    m = n >> 1;
    ks = 2 * nc / m;
    kk = 0;
    for (j = 2; j < m; j += 2) {
        k = n - j;
        kk += ks;
        wkr = FSUB(CONST_0_5_Q1_14,c[nc - kk]);
        wki = c[kk];
        xr = FSUB(a[j],a[k]);
        xi = FADD(a[j + 1],a[k + 1]);
        yr = FSUB(FMUL(wkr,xr,Q1_14),FMUL(wki,xi,Q1_14));
        yi = FADD(FMUL(wkr,xi,Q1_14),FMUL(wki,xr,Q1_14));
        a[j] = FSUB(a[j],yr);
        a[j + 1] = FSUB(a[j + 1],yi);
        a[k] = FADD(a[k],yr);
        a[k + 1] = FSUB(a[k + 1],yi);
    }
}


void fxrftbsub(INT32 n, FIXED16 *a, INT32 nc, FIXED16 *c)
{
    INT32 j, k, kk, ks, m;
    FIXED32 wkr, wki, xr, xi, yr, yi;
    
    a[1] = -a[1];
    m = n >> 1;
    ks = 2 * nc / m;
    kk = 0;
    for (j = 2; j < m; j += 2) {
        k = n - j;
        kk += ks;
        wkr = FSUB(CONST_0_5_Q1_14,c[nc - kk]);
        wki = c[kk];
        xr = FSUB(a[j],a[k]);
        xi = FADD(a[j + 1],a[k + 1]);
        yr = FADD(FMUL(wkr,xr,Q1_14),FMUL(wki,xi,Q1_14));
        yi = FSUB(FMUL(wkr,xi,Q1_14),FMUL(wki,xr,Q1_14));
        a[j] = FSUB(a[j],yr);
        a[j + 1] = FSUB(yi,a[j + 1]);
        a[k] = FADD(a[k],yr);
        a[k + 1] = FSUB(yi,a[k + 1]);
    }
    a[m + 1] = -a[m + 1];
}

#endif // ZEF_FIXED_POINT // ZEF_RADIX_2_4

