/*
Company :		ZEFYR Technologies
File :			fxFilters.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : fixed point filters

Features :fixed point !!!		
    
Functions : 

	generateGaborFFT : generates a gabor filter in the frequency domain
	freqConvolution : double convolution in frequancy domain (used with real filters)

Function prototypes : 

	INT32 generateGaborFFT( FIXED16 *p_filter, FIXED32 p_size, FIXED32 p_freq, FIXED32 p_angle );
	INT32 freqConvolution( FIXED16 *p_src, FIXED16 *p_dst, FIXED16 *p_filter );

Necessary package : 

	customTypes.h
	staticParameters.h
	errorCodes.h
	zMaths.h
	limits.h

Comments : none

*/

#include <stdlib.h>

#include "customTypes.h"
#include "staticParameters.h"
#include "errorCodes.h"
#include "zMaths.h"
#include "limits.h"

#include "fxFilters.h"

#ifdef ZEF_FIXED_POINT

// gabor filter genereration in Fourier space
INT32 generateGaborFFT( FIXED16 *p_filter, FIXED32 p_size, FIXED32 p_freq, INT32 p_angle )
{
	FIXED32		sTheta,cTheta,xc,yc;
	FIXED32		amp,xp,yp,G,tmp;
	INT32		i,j,j2;
	FIXED32		xpTab[FFT_BLOC_SIZE];
	
	// initializations
	// parameters
	p_freq = FMULI(p_freq,FFT_BLOC_SIZE);
	sTheta = zSin127(p_angle)<<1; 
	cTheta = zCos127(p_angle)<<1;
	p_size = FMUL(p_size,p_size,Q7_8);
	xc = -FADD(FDIVI(ITOFIX32(FFT_BLOC_SIZE,Q7_8),2),FMUL(p_freq,sTheta,Q7_8));
	yc = -FADD(FDIVI(ITOFIX32(FFT_BLOC_SIZE,Q7_8),2),FMUL(p_freq,cTheta,Q7_8));
	amp = FDIVI(FMUL(FMULI(p_size,-2),FTOFIX32(PI*PI,Q7_8),Q7_8),(FFT_BLOC_SIZE*FFT_BLOC_SIZE));

	// filter
	for(i=0; i<FFT_BLOC_SIZE*FFT_BLOC_SIZE/2; i++ )
		p_filter[i] = 0;

	// speedup for xp processing
	for( i=0; i<FFT_BLOC_SIZE; i++ )
	{
		tmp = FADDI(xc,((i+FFT_BLOC_SIZE/2)&MOD_FFT_BLOC_SIZE),Q7_8);
		xpTab[i] = FMUL(amp,FMUL(tmp,tmp,Q7_8),Q7_8);
	}

	// now generate the FFT gabor filter
	for( j=0; j<FFT_BLOC_SIZE; j++ )
	{
		j2 = (j+FFT_BLOC_SIZE/2)&(FFT_BLOC_SIZE-1);
		tmp = FADDI(yc,j,Q7_8);
		yp = FMUL(amp,FMUL(tmp,tmp,Q7_8),Q7_8);
		for( i=0; i<FFT_BLOC_SIZE/2; i++ )
		{
			xp = xpTab[i];

			G = -FADD(xp,yp);
			// speedup for exponential processing
			if( (UINT32)G < ITOFIX32(8,Q7_8) )
				G = zInvExp(G>>2);
			else
				G = 0;
			
			p_filter[j2*(FFT_BLOC_SIZE/2)+i] = G;
		}
		for( i=FFT_BLOC_SIZE/2; i<FFT_BLOC_SIZE; i++ )
		{
			xp = xpTab[i];

			G = -FADD(xp,yp);
			// speedup for exponential processing
			if( (UINT32)G < ITOFIX32(8,Q7_8) )
				G = zInvExp(G>>2);
			else
				G = 0;
			
			p_filter[(FFT_BLOC_SIZE-1-j2)*(FFT_BLOC_SIZE/2)+FFT_BLOC_SIZE-1-i] += G;
		}
	}

	

	return ZEF_OK;
}

// Time/freq space convolution with a real filter (non imaginary part)
INT32 freqConvolution( FIXED16 *p_src, FIXED16 *p_dst, FIXED16 *p_filter )
{
	UINT32		i,j;

	j=0;
	// perform the convolution
	// since the filter is real we only have : 
	// dest = filterRe*(srcRe+i*srcIm) with i the imaginary operator (i*i=-1)
	for( i=0; i<FFT_BLOC_SIZE*FFT_BLOC_SIZE; i+=2 )
	{
		p_dst[i] = FMUL(p_src[i],p_filter[j],Q1_14);
		p_dst[i+1] = FMUL(p_src[i+1],p_filter[j],Q1_14);
		j++;
	}

	return ZEF_OK;
}


INT32 sobelGradient( UINT8 *p_src, INT16 *p_gradient, UINT32 p_width, UINT32 p_height )
{
	UINT32		offset;
	UINT32		y,x;
	INT32		val;

	for( x=0; x<2*p_width; x++ )
		p_gradient[x] = 0;

	offset = p_width+1;
	for( y=1; y<p_height-1; y++ )
	{	
		p_gradient[(offset-1)<<1]=0;
		p_gradient[((offset-1)<<1)+1]=0;
		for( x=1; x<p_width-1; x++ )
		{
			val = (SQRT2_16*((INT16)p_src[offset+1]-(INT16)p_src[offset-1]))>>16;
			p_gradient[offset<<1] = val+(p_src[offset-p_width+1]+p_src[offset+p_width+1])-
									(p_src[offset-p_width-1]+p_src[offset+p_width-1]); //dx 
			val = (SQRT2_16*((INT16)p_src[offset-p_width]-(INT16)p_src[offset+p_width]))>>16;
			p_gradient[(offset<<1)+1] = val+(p_src[offset-p_width+1]+p_src[offset-p_width-1])-
									(p_src[offset+p_width+1]+p_src[offset+p_width-1]); //dy
			offset++;
		}
		p_gradient[offset<<1]=0;
		p_gradient[(offset<<1)+1]=0;
		offset+=2;
	}
	offset--;
	offset<<=1;
	for( x=0; x<2*p_width; x++ )
		p_gradient[offset+x] = 0;

	return ZEF_OK;
}


#endif // ZEF_FIXED_POINT
