/*
Company :		ZEFYR Technologies
File :			fxLinAlgebra.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	04/07/2003

Description : linear system resolution (fixed point)

Features : finds the solutions of a linear sysem
    
Functions : 

	luDecompose : transform a matrix into its LU decomposition (used to solve a linear system)
	luBackSubstitute : resolve a system using its LU decomposition

Function prototypes :

	INT32 luDecompose(INT32 **p_matrix, INT32 p_size, INT32 *p_index, INT32 *p_evenOdd);
	INT32 luBackSubstitute(INT32 **p_matrix, INT32 p_size, INT32 *p_index, INT32 p_rightSide[]);

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	zMaths.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <string.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "fixedPoint.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "zMaths.h"

#include "fxLinAlgebra.h"

#if defined(ZEF_FIXED_POINT) && defined(ZEF_CORRECT_DEFORMATION)

#define TINY 1


//Given a matrix p_matrix[1..n][1..n], this routine replaces it by the LU decomposition of a rowwise
//permutation of itself. p_matrix and p_size are input. p_matrix is output
//p_index[1..n] is an output vector that records the row permutation effected by the partial
//pivoting; p_evenOdd is output as 1 depending on whether the number of row interchanges was even
//or odd, respectively. This routine is used in combination with luBackSubstitute to solve linear equations
//or invert a matrix.
INT32 luDecompose(INT32 **p_matrix, INT32 p_size, INT32 *p_index, INT32 *p_evenOdd)
{
	
	INT32	i,imax,j,k;
	INT32	big,dum,sum,temp;
	INT32	*vv;
	INT32	error;
	
	vv=(INT32*)MALLOC(p_size*sizeof(INT32));;

	*p_evenOdd=1; 
	for (i=0;i<p_size;i++)
	{ 
		big=0;
		for (j=0;j<p_size;j++)
		{
			if ((temp=zAbs(p_matrix[i][j])) > big) big=temp;
		}
		if (big == 0)
		{
			error = ZEF_MATHS_ERROR;
			goto errorHandler;
		}
		
		vv[i]=big; 
	}
	for (j=0;j<p_size;j++) { 
		for (i=0;i<j;i++) {
			sum=p_matrix[i][j];
			for (k=0;k<i;k++) {
				sum -= FMUL(p_matrix[i][k],p_matrix[k][j],Q1_14);
			}
			p_matrix[i][j]=sum;
		}
		big=0; 
		for (i=j;i<p_size;i++) { 
			sum=p_matrix[i][j];
			for (k=0;k<j;k++){
				sum -= FMUL(p_matrix[i][k],p_matrix[k][j],Q1_14);
			}
			p_matrix[i][j]=sum;
			if ( (dum=FDIV(zAbs(sum),vv[i],Q1_14)) >= big) {
				
				big=dum;
				imax=i;
			}
		}
		if (j != imax) { 
			for (k=0;k<p_size;k++) { 
				dum=p_matrix[imax][k];
				p_matrix[imax][k]=p_matrix[j][k];
				p_matrix[j][k]=dum;
			}
			*p_evenOdd = -(*p_evenOdd); 
			vv[imax]=vv[j]; 
		}
		p_index[j]=imax;
		if (p_matrix[j][j] == 0 ) 
			p_matrix[j][j]=TINY;
		
		if (j != p_size-1) {
			dum=p_matrix[j][j];
			for (i=j+1;i<p_size;i++) 
			{
				p_matrix[i][j] = FDIV(p_matrix[i][j],dum,Q1_14);
			}
		}
	} 

	FREE(vv);

	return ZEF_OK;

	// errorHandling
errorHandler:

	FREE(vv);

	return error;
}

//Solves the set of n linear equations AX = B. Here p_matrix[0..p_size-1][0..p_size-1] is input, not as the matrix
//A but rather as its LU decomposition, determined by the routine luDecompose. p_index[0..p_size-1] is input
//as the permutation vector returned by ludcmp. p_rightSide[0..p_size-1] is input as the right-hand side vector
//B, and returns with the solution vector X. p_matrix, p_size, and p_index are not modified by this routine
//and can be left in place for successive calls with different right-hand sides p_rightSide. This routine takes
//into account the possibility that p_rightSide will begin with many zero elements, so it is efficient for use
//in matrix inversion.

INT32 luBackSubstitute(INT32 **p_matrix, INT32 p_size, INT32 *p_index, INT32 p_rightSide[])
{
	INT32 i,ii=-1,ip,j;
	INT32 sum;
	
	
	for (i=0;i<p_size;i++)
	{ 
		ip=p_index[i];
		sum=p_rightSide[ip];
		p_rightSide[ip]=p_rightSide[i];
		if (ii != -1)
		{
			for (j=ii;j<=i-1;j++) 
				sum -= FMUL(p_matrix[i][j],p_rightSide[j],Q1_14);
		}
		else if (sum)
			ii=i;
		p_rightSide[i]=sum;
	}
	for (i=p_size-1;i>=0;i--)
	{ 
		sum=p_rightSide[i];
		for (j=i+1;j<p_size;j++) 
			sum -= FMUL(p_matrix[i][j],p_rightSide[j],Q1_14);
		if( p_matrix[i][i] != 0 )
			p_rightSide[i]=FDIV(sum,p_matrix[i][i],Q1_14); 
		else
			p_rightSide[i] = sum;
	} 

	// convert to good fixed point format
	for (i=0;i<p_size;i++)
	{ 
		p_rightSide[i] = FCONV(p_rightSide[i],Q1_14,Q6_9);
	}

	return ZEF_OK;
}

#endif // ZEF_FIXED_POINT && ZEF_CORRECT_DEFORMATION

