/*
Company :		ZEFYR Technologies
File :			fxMatcher.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	02/07/2003

Description : Minutiae matching algorithm (fixed point)
Features : Fast matching !!!
    
Functions : 

	matchTemplate : main function for comparing two templates
	fastVerifyTemplate : same as verifyTemplate, but using a fast core based registration algorithm
	registerMinutiae : Main registering function, align to minutiae sets
	register1stPass : first step of minutiae alignment, finds a approximate alignment
	register2ndPass : second step of minutiae alignment, refines the fisrt estimation
	applyRegistration : modifies a template using  the registration informations (rotation, translation) 
	findMatches : searches the corresponding minutiae between two aligned templates
	estimateDeformations : uses a thin plate spline model to estimate the local deformation of the fingerprints
	applyDeformations : deforms a template using the model calculated with estimateDeformations
	matchOri : calculate a matching score based on bloc orientations
	computeScore : Main function to calculate the matching score of two aligned templates 
	findRotation : Fast registration algorithm used in fastVerifyTemplate
	verifyAndRefineTemplate : Matches two templates and if the matching score is good makes a composite template
	scoreIndex : processes a distance score used for linear classification during identification
	sortTemplates : sort an array of template according to their distance score


Function prototypes : 

	INT32	matchTemplate( Template *p_template1, Template *p_template2, INT32 **p_matches, INT32 *p_nbMatches, UINT32 *p_score );
	INT32	fastVerifyTemplate( UINT8 *p_templateBuffer1, UINT8 *p_templateBuffer2, UINT32 *p_score );
	INT32	registerMinutiae( Template *templ1, Template *templ2, INT32 *dx, INT32 *dy, INT32 *dTheta );
	INT32	register1stPass( Template *templ1, Template *templ2, INT32 *dx, INT32 *dy, INT32 *dTheta );
	INT32	register2ndPass( Template *templ1, Template *templ2, INT32 *dx, INT32 *dy, INT32 *dTheta );
	INT32	applyRegistration( Template *p_templ, INT32 p_dx, INT32 p_dy, INT32 p_dTheta );
	INT32	findMatches( Template *p_templ1, Template *p_templ2, INT32 **p_matches, INT32 *p_nbMatches );
	INT32   estimateDeformations( Template *p_templ1, Template *p_templ2, INT32 **p_matches, UINT32 p_cptMatches, REAL32 p_alpha, TpsData *p_data );
	INT32	applyDeformations( Template *p_templ, UINT32 *p_diffs, TpsData *p_data );
	UINT32  matchOri( Template *p_templ1, Template *p_templ2 );
	UINT32	computeScore( Template *p_templ1, Template *p_templ2, INT32 **p_matches, INT32 p_nbMatches );
	INT32	findRotation( Template *p_templ1, Template *p_templ2, INT32 *p_dTheta );
	INT32	verifyAndRefineTemplate( UINT8 *p_templateBuffer1, UINT8 *p_templateBuffer2, UINT32 *p_score, BOOL8 *p_refined );
	UINT32	scoreIndex( Template *p_templ1, Template *p_templ2 );
	INT32	sortTemplates( UINT8 *p_templateBuffer, UINT8 **p_templateBufferArray, UINT32 p_nbTemplates,
					 UINT32 *p_coreIndexes, UINT32 *p_coreDist, UINT32 *p_noCoreIndexe, UINT32 *p_noCoreDist  )

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	zMaths.h
	tools.h
	tpsWarper.h
	quicksort.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>
#include <float.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "zMaths.h"
#include "tools.h"
#include "tpsWarper.h"
#include "fxTpsWarper.h"
#include "quicksort.h"

#include "fxMatcher.h"

#ifdef ZEF_FIXED_POINT

INT32	matchTemplate( Template *p_template1, Template *p_template2, INT32 **p_matches, INT32 *p_nbMatches, UINT32 *p_score )
{
	INT32		error;
	INT32		dx,dy,dTheta;
#ifdef ZEF_CORRECT_DEFORMATION
	INT32		oldNbMatches,cpt,ndcNbMatches;
	TpsData		data;
	UINT32		diffsTmp;
#endif
	UINT32		diffs,maxNumberOfCores,maxNumberOfDeltas;
	Template	saveTemplate2;
	UINT32		minNumberOfMinutiae,maxDeformation;
	INT32		splineStiffness;

	//init 
	LOCK_MUTEX( g_paramsMutex );
		minNumberOfMinutiae = g_params.m_minNumberOfMinutiae;
		splineStiffness = g_params.m_splineStiffness;
		maxDeformation = g_params.m_maxDeformation;
	UNLOCK_MUTEX( g_paramsMutex );
	diffs=0;
	*p_score = 0;
	maxNumberOfCores = (p_template1->m_nbCores > p_template2->m_nbCores) ? p_template1->m_nbCores : p_template2->m_nbCores;
	maxNumberOfDeltas = (p_template1->m_nbDeltas > p_template2->m_nbDeltas) ? p_template1->m_nbDeltas : p_template2->m_nbDeltas;

	// init saveTemplate2 
	error = templateInit( &saveTemplate2 );
	if( error < ZEF_OK )
		goto errorHandler;	

	// copy template2 to another location to retreive it later
	error = templateCopy( &saveTemplate2, p_template2 );
	if( error < ZEF_OK )
		goto errorHandler;

	// register minutiae in order to find translation and rotation 
	// between the two templates

	error = registerMinutiae( p_template1, p_template2, &dx, &dy, &dTheta );
	if( error < ZEF_OK )
		goto errorHandler;	


	// now apply registration to template2
	error = applyRegistration( p_template2, dx, dy, dTheta );
	if( error < ZEF_OK )
		goto errorHandler;	
	
	// find matches
	error = findMatches( p_template1, p_template2, p_matches, p_nbMatches );
	if( error < ZEF_OK )
		goto errorHandler;	
	
	// if there is enough minutiae for deformation estimation 
	if( (*p_nbMatches)-(maxNumberOfCores+maxNumberOfDeltas) > minNumberOfMinutiae )
	{
		#ifdef ZEF_CORRECT_DEFORMATION
			cpt=0;
			ndcNbMatches = *p_nbMatches;
			// find and correct deformation until there is no more deformation
			do
			{
				oldNbMatches = (*p_nbMatches);
				// deformation estimation
				error = estimateDeformations( p_template1, p_template2, p_matches, *p_nbMatches, splineStiffness, &data );
				if( error < ZEF_OK )
					goto errorHandler;	
				
				// apply the deformations found
				error = applyDeformations( p_template2, &diffsTmp, &data );
				if( error < ZEF_OK )
					goto errorHandler;	
				
				diffs+=diffsTmp;
				if( diffsTmp > maxDeformation )
				{
					// retreive template 2 
					error = templateUninit(p_template2);
					if( error < ZEF_OK )
						goto errorHandler;
					error = templateCopy( p_template2, &saveTemplate2 );
					if( error < ZEF_OK )
						goto errorHandler;
					error = applyRegistration( p_template2, dx, dy, dTheta );
					if( error < ZEF_OK )
						goto errorHandler;	
					cpt = MAX_LOOPS;
				}
				
				// FREE memory allocated in the structure
				error = unInitTpsWarper( &data );
				if( error < ZEF_OK )
					goto errorHandler;	
				
				// find matches
				error = findMatches( p_template1, p_template2, p_matches, p_nbMatches );
				if( error < ZEF_OK )
					goto errorHandler;	
				
				cpt++;
			}
			while( (*p_nbMatches) > oldNbMatches && cpt < MAX_LOOPS );

		// check if deformation correction was effective (should not happen often !!!)
		if( *p_nbMatches < ndcNbMatches ) // if not cancel it !!
		{
			// retreive template 2 
			error = templateUninit(p_template2);
			if( error < ZEF_OK )
				goto errorHandler;
			error = templateCopy( p_template2, &saveTemplate2 );
			if( error < ZEF_OK )
				goto errorHandler;
			error = applyRegistration( p_template2, dx, dy, dTheta );
			if( error < ZEF_OK )
				goto errorHandler;
			// find matches
			error = findMatches( p_template1, p_template2, p_matches, p_nbMatches );
			if( error < ZEF_OK )
				goto errorHandler;	
		}


		#endif // ZEF_CORRECT_DEFORMATION
		
		// process score
		error = computeScore( p_template1, p_template2, p_matches, *p_nbMatches, p_score );
		if( error < ZEF_OK )
			goto errorHandler;
	}

	// uninit saveTemplate2 
	error = templateUninit( &saveTemplate2 );
	if( error < ZEF_OK )
		goto errorHandler;	

	return ZEF_OK;

	//error handling
errorHandler:

	// uninit saveTemplate2 
	templateUninit( &saveTemplate2 );

	return error;
}

INT32	fastVerifyTemplate( UINT8 *p_templateBuffer1, UINT8 *p_templateBuffer2, UINT32 *p_score )
{
	INT32		error;
	Template	templ1,templ2;
	INT32		dx,dy,dTheta;
	INT32		nbMatches,i;
	INT32		**matches;
	INT32		matchArraySize;
	UINT32		diffs;
	UINT32		maxNumberOfMinutiae,maxNumberOfCores,maxNumberOfDeltas;
	UINT32		minNumberOfMinutiae,defaultSecurity;

	// init params
	LOCK_MUTEX( g_paramsMutex );
		minNumberOfMinutiae = g_params.m_minNumberOfMinutiae;
		defaultSecurity = g_params.m_defaultSecurity;
	UNLOCK_MUTEX( g_paramsMutex );

	// init template structures
	error = templateInit( &templ1 );		// optional
	if( error < ZEF_OK )
		goto errorHandler;
	error = templateInit( &templ2 );		// optional
	if( error < ZEF_OK )
		goto errorHandler;

	// unpack templates
	error = templateUnpacker( p_templateBuffer1, &templ1 );
	if( error < ZEF_OK )
		goto errorHandler;
	error = templateUnpacker( p_templateBuffer2, &templ2 );
	if( error < ZEF_OK )
		goto errorHandler;

	// init
	matches = NULL;
	maxNumberOfMinutiae = (templ1.m_nbMinutiae > templ2.m_nbMinutiae) ? templ1.m_nbMinutiae : templ2.m_nbMinutiae;
	maxNumberOfCores = (templ1.m_nbCores > templ2.m_nbCores) ? templ1.m_nbCores : templ2.m_nbCores;
	maxNumberOfDeltas = (templ1.m_nbDeltas > templ2.m_nbDeltas) ? templ1.m_nbDeltas : templ2.m_nbDeltas;		
	// init score
	*p_score = 0;
	diffs=0;
		
	// memory allocation
	matchArraySize = maxNumberOfCores+maxNumberOfDeltas+maxNumberOfMinutiae+1;
	matches = (INT32**)MALLOC( matchArraySize*sizeof(INT32*) );
	if( matches == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	for( i=0; i<matchArraySize; i++ )
	{
		matches[i] = (INT32*)MALLOC( 2*sizeof(INT32) );
		if( matches[i] == NULL )
		{
			error =  ZEF_OUT_OF_MEMORY;
			goto errorHandler;	
		}
	}
	
	if( templ1.m_nbCores != 0 && templ2.m_nbCores != 0 )
	{
		
		// register minutiae in order to find translation and rotation 
		// between the two templates
		error = findRotation( &templ1, &templ2, &dTheta );
		dx=0;
		dy=0;
		if( error < ZEF_OK )
			goto errorHandler;
		
		// now apply registration to template2
		error = applyRegistration( &templ2, dx, dy, dTheta );
		if( error < ZEF_OK )
			goto errorHandler;
		
		// find matches
		error = findMatches( &templ1, &templ2, matches, &nbMatches );
		if( error < ZEF_OK )
			goto errorHandler;
		
		// if there is enough minutiae for deformation estimation 
		if( nbMatches-(maxNumberOfCores+maxNumberOfDeltas) > minNumberOfMinutiae )
		{
			// process score
			error = computeScore( &templ1, &templ2, matches, nbMatches, p_score );
			if( error < ZEF_OK )
				goto errorHandler;
		}
		
	}
	
	//FREE matches structure
	for( i=0; i<matchArraySize; i++ )
		FREE(matches[i]);
	FREE(matches);
	
	// FREE template structures
	error = templateUninit( &templ1 );
	if( error < ZEF_OK )
		goto errorHandler;
	error = templateUninit( &templ2 );
	if( error < ZEF_OK )
		goto errorHandler;
	
	// check if match is OK
	if( *p_score >= defaultSecurity )
		return ZEF_OK;
	else
	{
		if( !RETURN_NON_MATCH_SCORE )
			*p_score = 0;
		return ZEF_NO_MATCH;
	}
	
	// error handling
errorHandler:

	//FREE matches structure
	if( matches != NULL )
	{
		for( i=0; i<matchArraySize; i++ )
		{
			if( matches[i] != NULL )
				FREE(matches[i]);
		}
		FREE(matches);
	}
	// FREE template structures
	templateUninit( &templ1 );
	templateUninit( &templ2 );

	return error;
}

INT32	registerMinutiae( Template *p_templ1, Template *p_templ2, INT32 *p_dx, INT32 *p_dy, INT32 *p_dTheta )
{
	INT32	error;
	
	// first find the transformation roughly
	error = register1stPass( p_templ1, p_templ2, p_dx, p_dy, p_dTheta );
	if( error < ZEF_OK )
		goto errorHandler;	
	// then get a more precise estimation of the transform
	error = register2ndPass( p_templ1, p_templ2, p_dx, p_dy, p_dTheta );
	if( error < ZEF_OK )
		goto errorHandler;	
	
	return ZEF_OK;

	// error handling
errorHandler:

	return error;
}

INT32	findRotation( Template *p_templ1, Template *p_templ2, INT32 *p_dTheta )
{
	INT32		ori22,*o1,x1,y1,x2,y2,i,*l1,l2;
	INT32		g1,g2;
	UINT32		max,dist;
	UINT32		*accu,*accuS;
	INT32		ori1,ori2,dOri,angle,val;
	INT32		error,lThreshold;
	UINT32		translationTolerance;
	INT32		registerTranslationTolerance,maxRegistrationAngle;

	// init
	LOCK_MUTEX( g_paramsMutex );
		translationTolerance = g_params.m_translationTolerance;
		registerTranslationTolerance = g_params.m_registerTranslationTolerance;
		maxRegistrationAngle = g_params.m_maxRegistrationAngle;
	UNLOCK_MUTEX( g_paramsMutex );
	accu = NULL;
	accuS = NULL;
	l1=NULL;

	// allocations
	accu = (UINT32*)MALLOC( 2*PI_256*sizeof(UINT32)+1 );
	if( accu == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	accuS = (UINT32*)MALLOC( 2*PI_256*sizeof(UINT32)+1 );
	if( accuS == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	l1 = (INT32*)MALLOC( p_templ1->m_nbMinutiae*sizeof(INT32) );
	if( l1 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	o1 = (INT32*)MALLOC( p_templ1->m_nbMinutiae*sizeof(INT32) );
	if( o1 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	for( i=0; i<2*PI_256; i++ )
		accu[i] = 0;

	// precalc l1
	for( g1=p_templ1->m_nbMinutiae-1; g1>=0; g1-- )
	{			
		x1 = p_templ1->m_minutiae[g1].m_coord.m_x;
		y1 = p_templ1->m_minutiae[g1].m_coord.m_y;
		l1[g1]=x1*x1+y1*y1;
		o1[g1] = zAtan(y1,x1);
	}

	lThreshold = 4*translationTolerance*translationTolerance;
	
	// loop on all minutiae of template2
	for( g2=p_templ2->m_nbMinutiae-1; g2>=0; g2--)
	{
		x2 = p_templ2->m_minutiae[g2].m_coord.m_x;
		y2 = p_templ2->m_minutiae[g2].m_coord.m_y;
		ori2 = p_templ2->m_minutiae[g2].m_ori*2;
		ori22 = zAtan(y2,x2);
		l2=x2*x2+y2*y2;
		// then loop on all minutiae of template1 in order to find the best translation for this orientation
		for( g1=p_templ1->m_nbMinutiae-1; g1>=0; g1-- )
		{	
			dist = zAbs(l2-l1[g1]);
			if( dist <= (UINT32)lThreshold )
			{
				ori1 = p_templ1->m_minutiae[g1].m_ori*2;
				dOri = ori1-ori2;
				if( dOri > 256 ) dOri -= 512;
				if( dOri < -256 ) dOri += 512;
					
				angle = o1[g1]-ori22;
				if( angle > 256 ) angle -= 512;
				if( angle < -256 ) angle += 512;
				
				dOri = zAbs(angle-dOri);
				if( dOri > 256 ) 
					dOri = 512-dOri;
				
				// best orientation voting
				if( dOri < (INT32)2*registerTranslationTolerance )
				{
					val = p_templ2->m_minutiae[g2].m_qual*p_templ1->m_minutiae[g1].m_qual;
					accu[256+angle]+=val;
				}
			}
		}			
	}
	
	// filter (mean filter) the score to avoid local max
	accuS[0]=accu[0];
	accuS[511]=accu[511];
	for( i=1; i<511; i++ )
		accuS[i] = (accu[i-1]+accu[i]+accu[i+1]);
	
	// get max ori score
	max = 0;
	for( i=256-maxRegistrationAngle; i<256+maxRegistrationAngle; i++ )
	{
		if( accuS[i] > max ) 
		{
			max = accuS[i];
			*p_dTheta = 180*(i-256)/256;
		}
	}
	
	FREE(accu);
	FREE(accuS);
	FREE(l1);
	FREE(o1);
	
	return ZEF_OK;

	// error handling
errorHandler:

	if( accu != NULL )
		FREE(accu);
	if( accuS != NULL )
		FREE(accuS);
	if( l1 != NULL )
		FREE(l1);
	if( o1 != NULL )
		FREE(o1);

	return error;
}

INT32	register1stPass( Template *p_templ1, Template *p_templ2, INT32 *p_dx, INT32 *p_dy, INT32 *p_dTheta )
{
	INT32		y,x,k,i,j,trans;
	UINT32		cpt;
	INT32		dx,dy,fx,fy;
	INT32		g1,g2;
	INT32		theta;
	FIXED32		sinTheta,cosTheta;
	INT32		ori1,ori2,dOri,index;
	INT32		val;
	UINT32		maxTmp;
	INT32		moyX,moyY,maxTmp2,max,minLocal;
	INT32		valX,valY;
	INT32		dec;
	UINT32		*tIndex;
	INT32		*tDx,*tDy;
	INT32		*tVal;
	UINT32		*accu;
	INT32		error;
	UINT32		maxNumberOfMinutiae;
	INT32		maxTranslationWidth1,maxRegistrationAngle,angleTolerance,translationMaxThreshold,registerTranslationTolerance;
	UINT32		translation,fisrtPassAngleStep,translationShift,maxTranslationWidthShift1;

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxTranslationWidth1 = g_params.m_maxTranslationWidth1;
		translation = g_params.m_translation;
		maxRegistrationAngle = g_params.m_maxRegistrationAngle;
		fisrtPassAngleStep = g_params.m_fisrtPassAngleStep;
		angleTolerance = g_params.m_angleTolerance;
		translationShift = g_params.m_translationShift;
		translationMaxThreshold = g_params.m_translationMaxThreshold;
		registerTranslationTolerance = g_params.m_registerTranslationTolerance;
		maxTranslationWidthShift1 = g_params.m_maxTranslationWidthShift1;
	UNLOCK_MUTEX( g_paramsMutex );
	tIndex = NULL;
	tDx = NULL;
	tDy = NULL;
	tVal = NULL;
	accu = NULL;
	maxNumberOfMinutiae = (p_templ1->m_nbMinutiae > p_templ2->m_nbMinutiae) ? p_templ1->m_nbMinutiae : p_templ2->m_nbMinutiae;
	
	// arrays allocation
	tIndex = (UINT32*)MALLOC( maxNumberOfMinutiae*sizeof(UINT32) );
	if( tIndex == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tDx = (INT32*)MALLOC( maxNumberOfMinutiae*maxNumberOfMinutiae*sizeof(INT32) );
	if( tDx == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tDy = (INT32*)MALLOC( maxNumberOfMinutiae*maxNumberOfMinutiae*sizeof(INT32) );
	if( tDy == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tVal = (INT32*)MALLOC( maxNumberOfMinutiae*maxNumberOfMinutiae*sizeof(INT32) );
	if( tVal == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	accu = (UINT32*)CALLOC( maxTranslationWidth1*maxTranslationWidth1,sizeof(UINT32) );
	if( accu == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// init
	max=0;
	dec = (maxTranslationWidth1/2)*translation+(translation/2); // 520 = 32*16+8
	
	// try different orientations
	for( theta = -maxRegistrationAngle; theta <=maxRegistrationAngle; theta+=fisrtPassAngleStep ) 
	{
		// apply rotation to all minutiae of template2
		sinTheta = zSin127(theta*256/360);
		cosTheta = zCos127(theta*256/360);
		
		maxTmp=0;
		cpt=0;
		// loop on all minutiae of template 2
		for( g2=p_templ2->m_nbMinutiae-1; g2>=0; g2--)
		{
			tIndex[g2]=0;
			fx=p_templ2->m_minutiae[g2].m_coord.m_x;
			fy=p_templ2->m_minutiae[g2].m_coord.m_y;
			x = TOINT(FSUB(FMULI(fx,cosTheta),FMULI(fy,sinTheta)),7);
			y = TOINT(FADD(FMULI(fx,sinTheta),FMULI(fy,cosTheta)),7);
			ori2 = p_templ2->m_minutiae[g2].m_ori*2+theta*256/180;
			if( ori2 > 512 ) ori2 -= 512;
			if( ori2 < 0 ) ori2 += 512;
			// then loop on all minutiae of template1 in order to find the best translation for this orientation
			for( g1=p_templ1->m_nbMinutiae-1; g1>=0; g1-- )
			{	
				dx = (p_templ1->m_minutiae[g1].m_coord.m_x - x);
				dy = (p_templ1->m_minutiae[g1].m_coord.m_y - y);
				ori1 = p_templ1->m_minutiae[g1].m_ori*2;
				dOri = zAbs(ori1-ori2);
				if( dOri > 256 ) dOri = 512-dOri;
				
				// best translation voting for this orientation
				if( dOri < angleTolerance )
				{
					val = ((angleTolerance-dOri)*(p_templ1->m_minutiae[g1].m_qual+p_templ2->m_minutiae[g2].m_qual))>>5; // division per 32 to avoid overflows
					tIndex[g2]++;
					tVal[cpt] = val;
					tDx[cpt] = dx;
					tDy[cpt] = dy;
					cpt++;
				}
			}			
		}
		// we fill an accumulator which will tell us what translation is the best
		cpt=0;
		for( i=p_templ2->m_nbMinutiae-1; i>=0; i-- )
		{
			for( j=tIndex[i]-1; j>=0; j-- )
			{
				dx = (tDx[cpt]+dec)>>translationShift; 
				dy = (tDy[cpt]+dec)>>translationShift;
				
				if( (UINT32)dx < (UINT32)maxTranslationWidth1 && (UINT32)dy<(UINT32)maxTranslationWidth1 )
				{
					index = dx+(dy<<maxTranslationWidthShift1);
					
					accu[index]+=tVal[cpt];
					if( accu[index] > maxTmp )
						maxTmp = accu[index]; 
				}
				cpt++;
			}
		}
		
		maxTmp = TOINT(FMULI(maxTmp,FCONV(translationMaxThreshold,Q15_16,Q7_8)),Q7_8);
		
		// then determine the best translation
		k=0;
		trans = (maxTranslationWidth1/2)<<translationShift;
		for( y=-trans; y<trans; y+=translation )
		{
			for( x=-trans; x<trans; x+=translation )
			{
				if( accu[k] >= maxTmp)
				{
					valX = x;
					valY = y;
					maxTmp2 = 0;
					moyX = 0;
					moyY = 0;
					
					// calculate the score for this proposition
					cpt=0;
					for( i=p_templ2->m_nbMinutiae-1; i>=0; i--)
					{
						minLocal = INT_MAX;
						index = -1;
						for( j=tIndex[i]-1; j>=0; j-- )
						{
							dx = zAbs(tDx[cpt]-valX );
							dy = zAbs(tDy[cpt]-valY );
							
							if( dx <= registerTranslationTolerance &&  dy <= registerTranslationTolerance && (dx+dy) < minLocal )
							{
								minLocal = (dx+dy);
								index = cpt;
							}
							cpt++;
							
						}
						if( index != -1 )
						{
							maxTmp2 += tVal[index];
							moyX+=tDx[index]*tVal[index];
							moyY+=tDy[index]*tVal[index];
						}
					}
					
					if( maxTmp2 != 0 )
					{
						moyX/=maxTmp2;
						moyY/=maxTmp2;
					}
					
					// if cores are aligned it's better
					if( moyX < 2*registerTranslationTolerance && moyY < 2*registerTranslationTolerance )
						maxTmp2 = TOINT(269*maxTmp2,Q7_8);
					
					// keep the best solution
					if( maxTmp2 > max ) 
					{
						max=maxTmp2;
						*p_dTheta = theta;
						*p_dx = moyX;
						*p_dy = moyY;
					}
				}
				accu[k] = 0;
				k++;
			}		
		}		
	}
	
	// memory release
	FREE(tIndex);
	FREE(tDx);
	FREE(tDy);
	FREE(tVal);
	FREE(accu);
	
	return ZEF_OK;

	// error handling
errorHandler:

	if( tIndex != NULL )
		FREE(tIndex);
	if( tDx != NULL )
		FREE(tDx);
	if( tDy != NULL )
		FREE(tDy);
	if( tVal != NULL )
		FREE(tVal);
	if( accu != NULL )
		FREE(accu);

	return error;
}

INT32	register2ndPass( Template *p_templ1, Template *p_templ2, INT32 *p_dx, INT32 *p_dy, INT32 *p_dTheta )
{
	INT32		y,x;
	INT32		i,j,trans;
	INT32		dx,dy,fx,fy;
	INT32		g1,g2;
	INT32		theta,dTheta;
	FIXED32		sinTheta,cosTheta;
	INT32		ori1,ori2,dOri,index,index2;
	INT32		val;
	UINT32		maxTmp;
	INT32		moyX,moyY,minLocal,maxTmp2,max;
	INT32		valX,valY;
	INT32		xC,yC,thetaC;
	INT32		cpt,fill;	
	UINT32		*tIndex;
	INT32		*tDx,*tDy;
	UINT32		*tVal;
	UINT32		*accu;
	INT32		error;
	UINT32		maxNumberOfMinutiae,maxTranslationWidthShift2;
	INT32		maxTranslationWidth2,accuFill,fisrtPassAngleStep,angleTolerance,registerTranslationTolerance;

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxTranslationWidth2 = g_params.m_maxTranslationWidth2;
		accuFill = g_params.m_accuFill;
		fisrtPassAngleStep = g_params.m_fisrtPassAngleStep;
		angleTolerance = g_params.m_angleTolerance;
		registerTranslationTolerance = g_params.m_registerTranslationTolerance;
		maxTranslationWidthShift2 = g_params.m_maxTranslationWidthShift2;
	UNLOCK_MUTEX( g_paramsMutex );
	tIndex = NULL;
	tDx = NULL;
	tDy = NULL;
	tVal = NULL;
	accu = NULL;
	maxNumberOfMinutiae = (p_templ1->m_nbMinutiae > p_templ2->m_nbMinutiae) ? p_templ1->m_nbMinutiae : p_templ2->m_nbMinutiae;

	// arrays allocation
	tIndex = (UINT32*)MALLOC( maxNumberOfMinutiae*sizeof(UINT32) );
	if( tIndex == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tDx = (INT32*)MALLOC( maxNumberOfMinutiae*maxNumberOfMinutiae*sizeof(INT32) );
	if( tDx == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tDy = (INT32*)MALLOC( maxNumberOfMinutiae*maxNumberOfMinutiae*sizeof(INT32) );
	if( tDy == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tVal = (UINT32*)MALLOC( maxNumberOfMinutiae*maxNumberOfMinutiae*sizeof(UINT32) );
	if( tVal == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	accu = (UINT32*)CALLOC( maxTranslationWidth2*maxTranslationWidth2,sizeof(UINT32) );
	if( accu == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// init
	max = 0;
	xC = *p_dx;
	yC = *p_dy;
	thetaC = *p_dTheta;
	fill = accuFill*maxTranslationWidth2;
	
	// try different orientations
	for( dTheta = -(fisrtPassAngleStep/2); dTheta <=(fisrtPassAngleStep/2); dTheta++ ) 
	{
		// init some usefull variables
		cpt=0;
		theta = (*p_dTheta)+dTheta;
		// rotate template2
		sinTheta = zSin127(theta*256/360);
		cosTheta = zCos127(theta*256/360);
		moyX=0;
		moyY=0;
		maxTmp=0;
		for( g2=p_templ2->m_nbMinutiae-1; g2>=0; g2--)
		{
			tIndex[g2] = 0;
			fx=p_templ2->m_minutiae[g2].m_coord.m_x;
			fy=p_templ2->m_minutiae[g2].m_coord.m_y;
			x = TOINT(FSUB(FMULI(fx,cosTheta),FMULI(fy,sinTheta)),7);
			y = TOINT(FADD(FMULI(fx,sinTheta),FMULI(fy,cosTheta)),7);
			ori2 = p_templ2->m_minutiae[g2].m_ori*2+theta*255/180;
			if( ori2 > 512 ) ori2 -= 512;
			if( ori2 < 0 ) ori2 += 512;
			// and loop on all minutiae of template1 to find the best trnaslation
			for( g1=p_templ1->m_nbMinutiae-1; g1>=0; g1-- )
			{
				dx = (p_templ1->m_minutiae[g1].m_coord.m_x - x)-(*p_dx);
				dy = (p_templ1->m_minutiae[g1].m_coord.m_y - y)-(*p_dy);
				ori1 = p_templ1->m_minutiae[g1].m_ori*2;
				dOri = zAbs(ori1-ori2);
				if( dOri > 256 ) dOri = 512-dOri;
				// Best trnaslation voting
				if( dOri < angleTolerance )
				{
					val = ((angleTolerance-dOri)*(p_templ1->m_minutiae[g1].m_qual+p_templ2->m_minutiae[g2].m_qual))>>5;
					tIndex[g2]++;
					tVal[cpt] = val;
					tDx[cpt] = dx;
					tDy[cpt] = dy;
					cpt++;
				}
			}			
		}
		
		// we fill the accumulator
		cpt=0;
		for( i=p_templ2->m_nbMinutiae-1; i>=0; i-- )
		{
			for( j=tIndex[i]-1; j>=0; j-- )
			{
				dx = (tDx[cpt]+(maxTranslationWidth2/2)); 
				dy = (tDy[cpt]+(maxTranslationWidth2/2));
				
				if( dx > accuFill && dx < (maxTranslationWidth2-accuFill)
					&& dy>accuFill && dy< (maxTranslationWidth2-accuFill) )
				{
					index = dx+(dy<<maxTranslationWidthShift2);
					
					accu[index]+=tVal[cpt];
					for( y=-fill; y<=fill; y+=maxTranslationWidth2 )
					{
						index2 = index+y;
						for( x=-accuFill; x<=accuFill; x++ )
						{
							accu[index2+x]+=tVal[cpt];	
						}
					}	
				}
				cpt++;
			}
		}
		
		// search for the best translation
		trans = maxTranslationWidth2*maxTranslationWidth2;
		for( i=trans-1; i>=0; i--)
		{
			if( accu[i] > maxTmp )
			{
				maxTmp = accu[i]; 
				valX = ((i&(maxTranslationWidth2-1))-(maxTranslationWidth2/2));
				valY = ((i>>maxTranslationWidthShift2)-(maxTranslationWidth2/2));
			}
			accu[i] = 0;
		}
		
		maxTmp2 = 0;
		
		// determine coordinates of the translation
		cpt=0;
		for( i=p_templ2->m_nbMinutiae-1; i>=0; i--)
		{
			minLocal = INT_MAX;
			index = -1;
			for( j=tIndex[i]-1; j>=0; j-- )
			{
				dx = zAbs(tDx[cpt] - valX );
				dy = zAbs(tDy[cpt] - valY );
				
				if( dx <= registerTranslationTolerance && dy <= registerTranslationTolerance && (dx+dy) < minLocal)
				{
					minLocal = dx+dy;
					index = cpt;
				}
				cpt++;
			}
			
			if( index != -1 )
			{
				moyX+= tDx[index]*tVal[index];
				moyY+= tDy[index]*tVal[index];
				maxTmp2 += tVal[index];
			}
		}
		if( maxTmp2 != 0 )
		{
			moyX/=maxTmp2;
			moyY/=maxTmp2;
		}
		if( maxTmp2 > max )
		{
			max = maxTmp2;
			xC = (*p_dx)+moyX;
			yC = (*p_dy)+moyY;
			thetaC = theta;
		}
		
	}
	
	// set return values
	*p_dx = xC;
	*p_dy = yC;
	*p_dTheta = thetaC;
	
	// memory release
	FREE(tIndex);
	FREE(tDx);
	FREE(tDy);
	FREE(tVal);
	FREE(accu);
	
	return ZEF_OK;

	// error handling
errorHandler:

	if( tIndex != NULL )
		FREE(tIndex);
	if( tDx != NULL )
		FREE(tDx);
	if( tDy != NULL )
		FREE(tDy);
	if( tVal != NULL )
		FREE(tVal);
	if( accu != NULL )
		FREE(accu);

	return error;
}

INT32	applyRegistration( Template *p_templ, INT32 p_dx, INT32 p_dy, INT32 p_dTheta )
{
	INT32			i,x1,y1,incX,incY;
	INT32			angle;
	INT32			x,y,newX,newY;
	FIXED32			cosA,sinA,cY,sY;
	UINT8			*qual;
	UINT8			*ori;
	INT32			xMin,xMax,yMin,yMax;
	INT32			newCenterX,newCenterY;
	INT32			newNbBlocX,newNbBlocY;
	INT32			error;
	UINT32			maxNumberOfLocalInfos;

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfLocalInfos = g_params.m_maxNumberOfLocalInfos;
	UNLOCK_MUTEX( g_paramsMutex );
	qual = NULL;
	ori = NULL;
	
	// memory allocations
	qual = (UINT8*)MALLOC(maxNumberOfLocalInfos*sizeof(UINT8) );
	if( qual == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	ori = (UINT8*)MALLOC(maxNumberOfLocalInfos*sizeof(UINT8) );
	if( ori == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// init
	
	xMin = SHRT_MAX; 
	yMin = SHRT_MAX;
	xMax = SHRT_MIN;
	yMax = SHRT_MIN;
	
	p_templ->m_boundBox.m_left = SHRT_MAX; 
	p_templ->m_boundBox.m_top = SHRT_MAX;
	p_templ->m_boundBox.m_right = SHRT_MIN;
	p_templ->m_boundBox.m_bottom = SHRT_MIN;
	
	for( i=0; i<(INT32)(maxNumberOfLocalInfos); i++ )
	{
		qual[i] = 0;
		ori[i]  = 0;
	}
	
	// Local infos transform
	cosA = zCos127( 256*p_dTheta/360 );
	sinA = zSin127( 256*p_dTheta/360 );
	
	incX = p_templ->m_localW;
	incY = p_templ->m_localH;
	if( incX == 0 ) incX = 1;
	if( incY == 0 ) incY = 1;
	
	// compute new local infos boundaries
	for( y=0; y<=p_templ->m_localH; y+= incY )
	{
		y1 = (y-(INT32)p_templ->m_localH/2)*p_templ->m_blocSize+p_templ->m_localShiftY;
		sY = FMULI(y1,sinA);
		cY = FMULI(y1,cosA);
		for( x=0; x<=p_templ->m_localW; x+= incX )
		{
			x1 = (x-(INT32)p_templ->m_localW/2)*p_templ->m_blocSize+p_templ->m_localShiftX;
			
			newX = TOINT(FSUB(FMULI(x1,cosA),sY),7);
			newY = TOINT(FADD(FMULI(x1,sinA),cY),7);
			
			if( newX > xMax ) xMax = newX;
			if( newY > yMax ) yMax = newY;
			if( newX < xMin ) xMin = newX;
			if( newY < yMin ) yMin = newY;
		}
	}
	newCenterX = xMin+(xMax-xMin)/2;
	newCenterY = yMin+(yMax-yMin)/2;
	
	if( p_templ->m_blocSize != 0 )
	{
		newNbBlocX = (xMax-xMin)/p_templ->m_blocSize;
		newNbBlocY = (yMax-yMin)/p_templ->m_blocSize;
	}
	else
		newNbBlocX = newNbBlocY = 0;
	if( newNbBlocX%2 ) newNbBlocX++;
	if( newNbBlocY%2 ) newNbBlocY++;
	
	if( newNbBlocY*newNbBlocX > (INT32)maxNumberOfLocalInfos ) // If we have too much blocs
	{
		newNbBlocX -= 2; // reduce size in X a little
		if( newNbBlocX == 0 ) newNbBlocX = 1;
		newNbBlocY = maxNumberOfLocalInfos / newNbBlocX; // then compute the new height
		if( newNbBlocY%2 != 0 )  newNbBlocY--;
	}
	
	// Now we can apply the transformation to local infos
	
	cosA = zCos127( -256*p_dTheta/360 );
	sinA = zSin127( -256*p_dTheta/360 );
	
	for( y=0; y<newNbBlocY; y++)
	{
		y1 = (y-newNbBlocY/2)*p_templ->m_blocSize+newCenterY;
		sY = FMULI(y1,sinA);
		cY = FMULI(y1,cosA);
		for( x=0; x<newNbBlocX; x++ )
		{
			x1 = (x-newNbBlocX/2)*p_templ->m_blocSize+newCenterX;
			
			newX = TOINT(FSUB(FMULI(x1,cosA),sY),7);
			newY = TOINT(FADD(FMULI(x1,sinA),cY),7);
			
			newX = (newX-p_templ->m_localShiftX)/p_templ->m_blocSize+p_templ->m_localW/2;
			newY = (newY-p_templ->m_localShiftY)/p_templ->m_blocSize+p_templ->m_localH/2;
			
			if( (UINT32)newX < (UINT32)p_templ->m_localW && (UINT32)newY < p_templ->m_localH )
			{
				// on traite les orientations
				angle = p_templ->m_ori[newY*p_templ->m_localW+newX];
				angle += p_dTheta*256/180;
				if( angle > 256 ) angle -= 256;
				if( angle < 0 ) angle += 256;
				
				ori[y*newNbBlocX+x] = angle;
				qual[y*newNbBlocX+x] = p_templ->m_qual[newY*p_templ->m_localW+newX];	
			}
			else
			{
				ori[y*newNbBlocX+x] = 0;
				qual[y*newNbBlocX+x] = 0;	
			}
		}
	}
	
	// copy temporary buffers to qual and ori in template
	
	// reallocate memory for local infos
	FREE(p_templ->m_ori);
	FREE(p_templ->m_qual);
	
	p_templ->m_ori = (UINT8*)MALLOC( newNbBlocY*newNbBlocX*sizeof(UINT8) );
	if( p_templ->m_ori == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_templ->m_qual = (UINT8*)MALLOC( newNbBlocY*newNbBlocX*sizeof(UINT8) );
	if( p_templ->m_qual == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// copy !
	for( i=0; i<(INT32)(newNbBlocY*newNbBlocX); i++ )
	{
		p_templ->m_ori[i] = ori[i];
		p_templ->m_qual[i] = qual[i];
	}
	p_templ->m_localShiftX = newCenterX;
	p_templ->m_localShiftY = newCenterY;
	p_templ->m_localH = newNbBlocY;
	p_templ->m_localW = newNbBlocX;
	
	p_templ->m_localShiftX += p_dx;
	p_templ->m_localShiftY += p_dy;
	
	// now apply transform to minutiae
	
	cosA = zCos127( 256*p_dTheta/360 );
	sinA = zSin127( 256*p_dTheta/360 );
	
	// loop on all minutiae
	for( i=0; i<p_templ->m_nbMinutiae; i++ )
	{
		// Orientation correction
		angle = p_templ->m_minutiae[i].m_ori*2;
		angle += p_dTheta*256/180;
		if( angle > 512 ) angle -= 512;
		if( angle < 0 ) angle = 512+angle;
		p_templ->m_minutiae[i].m_ori = angle/2;
		
		// Rotate minutiae
		x = p_templ->m_minutiae[i].m_coord.m_x;
		y = p_templ->m_minutiae[i].m_coord.m_y;
		
		p_templ->m_minutiae[i].m_coord.m_x = (INT16)TOINT(FSUB(FMULI(x,cosA),FMULI(y,sinA)),7);
		p_templ->m_minutiae[i].m_coord.m_y = (INT16)TOINT(FADD(FMULI(x,sinA),FMULI(y,cosA)),7);
		
		// Then tranlation
		p_templ->m_minutiae[i].m_coord.m_x += p_dx;
		p_templ->m_minutiae[i].m_coord.m_y += p_dy;
		
		// reprocess min et max
		if( p_templ->m_minutiae[i].m_coord.m_x < p_templ->m_boundBox.m_left ) p_templ->m_boundBox.m_left = p_templ->m_minutiae[i].m_coord.m_x;
		if( p_templ->m_minutiae[i].m_coord.m_y < p_templ->m_boundBox.m_top ) p_templ->m_boundBox.m_top = p_templ->m_minutiae[i].m_coord.m_y;
		if( p_templ->m_minutiae[i].m_coord.m_x > p_templ->m_boundBox.m_right ) p_templ->m_boundBox.m_right = p_templ->m_minutiae[i].m_coord.m_x;
		if( p_templ->m_minutiae[i].m_coord.m_y > p_templ->m_boundBox.m_bottom ) p_templ->m_boundBox.m_bottom = p_templ->m_minutiae[i].m_coord.m_y;
	}
	
	// Then cores & deltas
	for( i=0; i<p_templ->m_nbCores; i++)
	{
		x = p_templ->m_cores[i].m_x;
		y = p_templ->m_cores[i].m_y;
		
		p_templ->m_cores[i].m_x = (INT16)TOINT(FSUB(FMULI(x,cosA),FMULI(y,sinA)),7);
		p_templ->m_cores[i].m_y = (INT16)TOINT(FADD(FMULI(x,sinA),FMULI(y,cosA)),7);
		
		// translation
		p_templ->m_cores[i].m_x += p_dx;
		p_templ->m_cores[i].m_y += p_dy;
	}
	
	// 1st core Ori 
	angle = p_templ->m_oriCore;
	angle += p_dTheta*256/180;
	if( angle > 256 ) angle -= 256;
	if( angle < 0 ) angle = 256+angle;
	p_templ->m_oriCore = angle;
	
	for( i=0; i<p_templ->m_nbDeltas; i++)
	{
		x = p_templ->m_deltas[i].m_x;
		y = p_templ->m_deltas[i].m_y;
		
		p_templ->m_deltas[i].m_x = (INT16)TOINT(FSUB(FMULI(x,cosA),FMULI(y,sinA)),7);
		p_templ->m_deltas[i].m_y =(INT16)TOINT(FADD(FMULI(x,sinA),FMULI(y,cosA)),7);
		
		// puis la translation
		p_templ->m_deltas[i].m_x += p_dx;
		p_templ->m_deltas[i].m_y += p_dy;
		
	}
	
	// release memory
	FREE(qual);
	FREE(ori);
	
	return ZEF_OK;

	// error handling
errorHandler:

	if( qual != NULL )
		FREE(qual);
	if( ori != NULL )
		FREE(ori);

	if( p_templ->m_ori != NULL )
	{
		FREE(p_templ->m_ori);
		p_templ->m_ori = NULL;
	}
	if( p_templ->m_qual != NULL )
	{
		FREE(p_templ->m_qual);
		p_templ->m_qual = NULL;
	}

	return error;
}

INT32	findMatches( Template *p_templ1, Template *p_templ2, INT32 **p_matches, INT32 *p_nbMatches )
{
	INT32		i,j,dx,dy,dTheta,jConcorde,cpt1,cpt2;
	INT32		cptConcordances = 0;
	INT32		nbConcordances;
	INT32		dL,score; 
	INT32		*scores,*inverse,*index;
	INT32		nbIndex;
	INT32		distThreshold;
	INT32		matchArraySize;
	INT32		error;
	UINT32		maxNumberOfMinutiae,maxNumberOfCores,maxNumberOfDeltas;
	UINT32		translationTolerance;
	INT32		angleTolerance;
	
	// init
	LOCK_MUTEX( g_paramsMutex );
		translationTolerance = g_params.m_translationTolerance;
		angleTolerance = g_params.m_angleTolerance;
	UNLOCK_MUTEX( g_paramsMutex );
	scores = NULL;
	inverse = NULL;
	index = NULL;
	maxNumberOfMinutiae = (p_templ1->m_nbMinutiae > p_templ2->m_nbMinutiae) ? p_templ1->m_nbMinutiae : p_templ2->m_nbMinutiae;
	maxNumberOfCores = (p_templ1->m_nbCores > p_templ2->m_nbCores) ? p_templ1->m_nbCores : p_templ2->m_nbCores;
	maxNumberOfDeltas = (p_templ1->m_nbDeltas > p_templ2->m_nbDeltas) ? p_templ1->m_nbDeltas : p_templ2->m_nbDeltas;
	distThreshold = translationTolerance*translationTolerance;
	matchArraySize = maxNumberOfCores+maxNumberOfDeltas+maxNumberOfMinutiae+1;
	nbIndex	= p_templ1->m_nbMinutiae;

	// memory allocation
	scores = (INT32*)MALLOC(matchArraySize*sizeof(INT32) );
	if( scores == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	inverse = (INT32*)MALLOC(matchArraySize*sizeof(INT32) );
	if( inverse == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	index = (INT32*)MALLOC(matchArraySize*sizeof(INT32) );
	if( index == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// arrays init
	for( i=0; i<matchArraySize; i++ )
	{
		p_matches[i][0] = -1;
		p_matches[i][1] = -1;
	}
	
	// we're looking for matching cores
	for( i=0; i<p_templ1->m_nbCores; i++ )
	{
		for( j=0; j<p_templ2->m_nbCores; j++ )
		{
			dx = p_templ1->m_cores[i].m_x - p_templ2->m_cores[j].m_x;
			dy = p_templ1->m_cores[i].m_y - p_templ2->m_cores[j].m_y;
			dL = dx*dx+dy*dy;
			if( dL <= distThreshold )
			{
				p_matches[cptConcordances][0] = i;
				p_matches[cptConcordances][1] = j;
				cptConcordances++;
			}
		}
	}
	
	// go to the end of cores
	for( i=cptConcordances; i<(INT32)maxNumberOfCores; i++ )
	{
		p_matches[i][0] = -1;
		p_matches[i][1] = -1;
		cptConcordances++;
	}
	
	// look for matching deltas
	for( i=0; i<p_templ1->m_nbDeltas; i++ )
	{
		for( j=0; j<p_templ2->m_nbDeltas; j++ )
		{
			dx = p_templ1->m_deltas[i].m_x - p_templ2->m_deltas[j].m_x;
			dy = p_templ1->m_deltas[i].m_y - p_templ2->m_deltas[j].m_y;
			dL = dx*dx+dy*dy;
			if( dL <= distThreshold )
			{
				p_matches[cptConcordances][0] = i;
				p_matches[cptConcordances][1] = j;
				cptConcordances++;
			}
		}
	}
	
	// go to the end of deltas
	for( i=cptConcordances; i<(INT32)(maxNumberOfCores+maxNumberOfDeltas); i++ )
	{
		p_matches[i][0] = -1;
		p_matches[i][1] = -1;
		cptConcordances++;
	}
	
	
	// now look for matching minutiae
	for( i=0; i<matchArraySize; i++ )
	{
		inverse[i] = -1;
		scores[i] = INT_MAX;
	}
	for( i=0; i<p_templ1->m_nbMinutiae; i++ )
		index[i] = i;
	
	index[p_templ1->m_nbMinutiae] = -1;
	
	cpt1=0;
	while( nbIndex > 0 && cpt1 < BIG_MAX_LOOPS ) // les cpt sont des protections contre les boucles sans fin
	{
		cpt1++;
		i=0;
		nbIndex=0;
		cpt2=0;
		while( index[i] != -1 && cpt2 < BIG_MAX_LOOPS )
		{
			cpt2++;
			nbConcordances = 0;
			score = INT_MAX-1;
			for( j=0; j<p_templ2->m_nbMinutiae; j++ )
			{
				dx = p_templ1->m_minutiae[index[i]].m_coord.m_x - p_templ2->m_minutiae[j].m_coord.m_x;
				dy = p_templ1->m_minutiae[index[i]].m_coord.m_y - p_templ2->m_minutiae[j].m_coord.m_y;
				dTheta = zAbs(p_templ1->m_minutiae[index[i]].m_ori - p_templ2->m_minutiae[j].m_ori)*2;
				if( dTheta > 256 ) dTheta = 512-dTheta;
				dL = dx*dx+dy*dy;
				if( dL <= distThreshold && dTheta <= angleTolerance && (dL+dTheta) <= scores[j] && 
					( inverse[j] == -1 || (inverse[j] != -1 && p_matches[inverse[j]][0] != index[nbIndex] ) ) )
				{
					if( dL+dTheta < score  )
					{
						score = dTheta + dL;
						jConcorde = j;
					}
					nbConcordances++;
				}
			}
			if( nbConcordances >= 1 )
			{
				if( inverse[jConcorde] != -1 )	
				{
					if( score <= scores[jConcorde]  )
					{
						index[nbIndex++] = p_matches[inverse[jConcorde]][0];
						p_matches[inverse[jConcorde]][0] = index[i];
						p_matches[inverse[jConcorde]][1] = jConcorde;
						scores[jConcorde] = score;
					}
				}
				else
				{
					p_matches[cptConcordances][0] = index[i];
					p_matches[cptConcordances][1] = jConcorde;
					inverse[jConcorde] = cptConcordances;
					cptConcordances++;
					scores[jConcorde] = score;
				}
				
			}
			i++;
		}
		index[nbIndex] = -1;
	}
	
	p_matches[cptConcordances][0] = -1;
	p_matches[cptConcordances][1] = -1;
	*p_nbMatches = cptConcordances;
	
	FREE( scores );
	FREE( inverse );
	FREE( index );
	
	return cptConcordances;
	

	// error handling
errorHandler:

	if( scores != NULL )
		FREE( scores );
	if( inverse != NULL )
		FREE( inverse );
	if( index != NULL )
		FREE( index );

	return error;
}

#ifdef ZEF_CORRECT_DEFORMATION

INT32 estimateDeformations( Template *p_templ1, Template *p_templ2, INT32 **p_matches, UINT32 p_cptMatches, FIXED32 p_alpha, TpsData *p_data )
{
	INT32	*srcX,*srcY,*dstX,*dstY;
	UINT32	i,acc;
	INT32	error;
	UINT32	maxNumberOfCores,maxNumberOfDeltas;
	
	// init
	srcX = NULL;
	srcY = NULL;
	dstX = NULL;
	dstY = NULL;
	acc=0;
	maxNumberOfCores = (p_templ1->m_nbCores > p_templ2->m_nbCores) ? p_templ1->m_nbCores : p_templ2->m_nbCores;
	maxNumberOfDeltas = (p_templ1->m_nbDeltas > p_templ2->m_nbDeltas) ? p_templ1->m_nbDeltas : p_templ2->m_nbDeltas;
	
	// allocations
	srcX = (INT32*)MALLOC(p_cptMatches*sizeof(INT32));
	if( srcX == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	srcY = (INT32*)MALLOC(p_cptMatches*sizeof(INT32));
	if( srcY == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	dstX = (INT32*)MALLOC(p_cptMatches*sizeof(INT32));
	if( dstX == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	dstY = (INT32*)MALLOC(p_cptMatches*sizeof(INT32));
	if( dstY == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// les cores
	for( i=0; i<maxNumberOfCores; i++)
	{
		if( p_matches[i][0] != -1 )
		{
			dstX[acc] = p_templ1->m_cores[p_matches[i][0]].m_x;
			dstY[acc] = p_templ1->m_cores[p_matches[i][0]].m_y;
			
			srcX[acc] = p_templ2->m_cores[p_matches[i][1]].m_x;
			srcY[acc] = p_templ2->m_cores[p_matches[i][1]].m_y;
			
			acc++;
		}
	}
	
	// les deltas
	for( i=maxNumberOfCores; i<maxNumberOfCores+maxNumberOfDeltas; i++)
	{
		if( p_matches[i][0] != -1 )
		{
			dstX[acc] = p_templ1->m_deltas[p_matches[i][0]].m_x;
			dstY[acc] = p_templ1->m_deltas[p_matches[i][0]].m_y;
			
			srcX[acc] = p_templ2->m_deltas[p_matches[i][1]].m_x;
			srcY[acc] = p_templ2->m_deltas[p_matches[i][1]].m_y;
			
			acc++;
		}
	}
	
	// les minuties
	acc=0;
	for( i=maxNumberOfCores+maxNumberOfDeltas; i<p_cptMatches; i++)
	{
		dstX[acc] = p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_x;
		dstY[acc] = p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_y;
		
		srcX[acc] = p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_x;
		srcY[acc] = p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_y;
		
		acc++;
	}
	
	// deformation estimate
	error = initTpsWarper( acc, srcX, srcY, dstX, dstY, p_alpha, p_data );
	if( error < ZEF_OK )
		goto errorHandler;
	
	// release memory
	FREE(srcX);
	FREE(srcY);
	FREE(dstX);
	FREE(dstY);
	
	return error;

	// error handling
errorHandler:

	if( srcX != NULL )
		FREE(srcX);
	if( srcY != NULL )
		FREE(srcY);
	if( dstX != NULL )
		FREE(dstX);
	if( dstY != NULL )
		FREE(dstY);

	return error;
	
}

INT32	applyDeformations( Template *p_templ, UINT32 *p_diffs, TpsData *p_data )
{
	INT32				i;
	INT32				x,y;
	
	p_templ->m_boundBox.m_left = SHRT_MAX; 
	p_templ->m_boundBox.m_top = SHRT_MAX;
	p_templ->m_boundBox.m_right = SHRT_MIN;
	p_templ->m_boundBox.m_bottom = SHRT_MIN;
	
	// to save processing time, there is no error checking for "warp"
	// but this is not a problem (at this time) because warp always return ZEF_OK
	
	// minutiae
	for( i=0; i<p_templ->m_nbMinutiae; i++ )
	{
		x = p_templ->m_minutiae[i].m_coord.m_x;
		y = p_templ->m_minutiae[i].m_coord.m_y;
		
		warp(&x,&y,p_data);
		
		p_templ->m_minutiae[i].m_coord.m_x = x;
		p_templ->m_minutiae[i].m_coord.m_y = y;
		
		if( p_templ->m_minutiae[i].m_coord.m_x < p_templ->m_boundBox.m_left ) p_templ->m_boundBox.m_left = p_templ->m_minutiae[i].m_coord.m_x;
		if( p_templ->m_minutiae[i].m_coord.m_y < p_templ->m_boundBox.m_top ) p_templ->m_boundBox.m_top = p_templ->m_minutiae[i].m_coord.m_y;
		if( p_templ->m_minutiae[i].m_coord.m_x > p_templ->m_boundBox.m_right ) p_templ->m_boundBox.m_right = p_templ->m_minutiae[i].m_coord.m_x;
		if( p_templ->m_minutiae[i].m_coord.m_y > p_templ->m_boundBox.m_bottom ) p_templ->m_boundBox.m_bottom = p_templ->m_minutiae[i].m_coord.m_y;
		
	}
	
	// cores
	for( i=0; i<p_templ->m_nbCores; i++)
	{
		x = p_templ->m_cores[i].m_x;
		y = p_templ->m_cores[i].m_y;
		
		warp(&x,&y,p_data);
		
		p_templ->m_cores[i].m_x = x;
		p_templ->m_cores[i].m_y = y;
		
	}
	
	// deltas
	for( i=0; i<p_templ->m_nbDeltas; i++)
	{
		x = p_templ->m_deltas[i].m_x;
		y = p_templ->m_deltas[i].m_y;
		
		warp(&x,&y,p_data);
		
		p_templ->m_deltas[i].m_x = x;
		p_templ->m_deltas[i].m_y = y;
		
	}
	
	*p_diffs = getDiffs( p_data );
	
	return ZEF_OK;
}

#endif // ZEF_CORRECT_DEFORMATION

UINT32 matchOri( Template *p_templ1, Template *p_templ2 )
{
	INT32		decX,decY;
	INT32		difAngles=0;
	INT32		i,j,y,x;
	UINT32		cpt=0,sumAngle=0;
	INT32		qual;
	
	if( p_templ1->m_blocSize == p_templ2->m_blocSize && p_templ1->m_blocSize != 0 )
	{

	decX = p_templ2->m_localShiftX-p_templ1->m_localShiftX;
	decY = p_templ2->m_localShiftY-p_templ1->m_localShiftY;
	
	decX = (decX/p_templ1->m_blocSize)+(decX%p_templ1->m_blocSize)/(p_templ1->m_blocSize/2);
	decY = (decY/p_templ1->m_blocSize)+(decY%p_templ1->m_blocSize)/(p_templ1->m_blocSize/2);
	
	decX -= ((INT32)p_templ2->m_localW-(INT32)p_templ1->m_localW)/2;
	decY -= ((INT32)p_templ2->m_localH-(INT32)p_templ1->m_localH)/2;
	
	for( j=0; j<p_templ2->m_localH; j++ )
	{
		for( i=0; i<p_templ2->m_localW; i++ )
		{
			y = j+decY;
			x = i+decX;
			if( (UINT32)y<p_templ1->m_localH && (UINT32)x<p_templ1->m_localW )
			{
				qual = (p_templ2->m_qual[j*p_templ2->m_localW+i])*(p_templ1->m_qual[y*p_templ1->m_localW+x]);
				difAngles = zAbs(p_templ2->m_ori[j*p_templ2->m_localW+i] - p_templ1->m_ori[y*p_templ1->m_localW+x]);
				if( difAngles > 127 ) difAngles = 255 - difAngles;
				cpt+=qual;
				sumAngle += qual*difAngles*difAngles;
			}
		}
	}
	
	if( cpt != 0 )
		sumAngle /= cpt;
	else
		sumAngle = 255L*255L;
	
	}
	
	return sumAngle/16;
}

INT32 computeScore( Template *p_templ1, Template *p_templ2, INT32 **p_matches, INT32 p_nbMatches, UINT32 *p_score )
{
	INT32		i,j,k,x,y;
	UINT32		*match1,*match2;
	UINT32		*nonMatch1,*nonMatch2;
	INT32		x1,x2,y1,y2,xMin,xMax,yMin,yMax;
	INT32		nonMatchScore=0,matchScore=0;
	INT32		cptNonMatch1=0,cptNonMatch2=0;
	INT32		cpt;
	BOOL8		inside;	
	INT32		qual,diffOri;
	INT32		score;
	UINT32		scoreOri = 0;
	FIXED32		txFreq,dist,mulFact;
	INT32		error;
	UINT32		maxNumberOfMinutiae,maxNumberOfCores,maxNumberOfDeltas;
	UINT32		translationTolerance,maxScore,maxNumberOfMinutiaeParam;
	INT32		distDivCoef,oriMatchNormalize,scoreNormalize;
	INT8		minutiaeBoost;

	// init
	LOCK_MUTEX( g_paramsMutex );
		translationTolerance = g_params.m_translationTolerance;
		distDivCoef = g_params.m_distDivCoef;
		oriMatchNormalize = g_params.m_oriMatchNormalize;
		maxScore = g_params.m_maxScore;
		minutiaeBoost = g_params.m_minutiaeBoost;
		maxNumberOfMinutiaeParam = g_params.m_maxNumberOfMinutiae;
		scoreNormalize = g_params.m_scoreNormalize;
	UNLOCK_MUTEX( g_paramsMutex );
	*p_score = 0;
	match1 = NULL;
	match2 = NULL;
	nonMatch1 = NULL;
	nonMatch2 = NULL;
	maxNumberOfMinutiae = (p_templ1->m_nbMinutiae > p_templ2->m_nbMinutiae) ? p_templ1->m_nbMinutiae : p_templ2->m_nbMinutiae;
	maxNumberOfCores = (p_templ1->m_nbCores > p_templ2->m_nbCores) ? p_templ1->m_nbCores : p_templ2->m_nbCores;
	maxNumberOfDeltas = (p_templ1->m_nbDeltas > p_templ2->m_nbDeltas) ? p_templ1->m_nbDeltas : p_templ2->m_nbDeltas;

	
	// memory allocations
	match1 = (UINT32*)MALLOC(maxNumberOfMinutiae*sizeof(UINT32));
	if( match1 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	match2 = (UINT32*)MALLOC(maxNumberOfMinutiae*sizeof(UINT32));
	if( match2 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	nonMatch1 = (UINT32*)MALLOC(maxNumberOfMinutiae*sizeof(UINT32));
	if( nonMatch1 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	nonMatch2 = (UINT32*)MALLOC(maxNumberOfMinutiae*sizeof(UINT32));
	if( nonMatch2 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// init
	for( i=0; i<(INT32)maxNumberOfMinutiae; i++ )
	{
		match1[i] = match2[i] = 0;
		nonMatch1[i] = nonMatch2[i] = 0;
	}
	
	// calculate the common bounding box
	xMin = (p_templ1->m_boundBox.m_left>p_templ2->m_boundBox.m_left)?p_templ1->m_boundBox.m_left:p_templ2->m_boundBox.m_left;
	yMin = (p_templ1->m_boundBox.m_top>p_templ2->m_boundBox.m_top)?p_templ1->m_boundBox.m_top:p_templ2->m_boundBox.m_top;
	xMax = (p_templ1->m_boundBox.m_right<p_templ2->m_boundBox.m_right)?p_templ1->m_boundBox.m_right:p_templ2->m_boundBox.m_right;
	yMax = (p_templ1->m_boundBox.m_bottom<p_templ2->m_boundBox.m_bottom)?p_templ1->m_boundBox.m_bottom:p_templ2->m_boundBox.m_bottom;
	
	// enlarge a little the bounding box so that it is not too small
	xMin -= 2*translationTolerance;
	yMin -= 2*translationTolerance;
	xMax += 2*translationTolerance;
	yMax += 2*translationTolerance;

	// fill the match arrays
	for( i=maxNumberOfCores+maxNumberOfDeltas; i<p_nbMatches; i++ )
	{
		match1[p_matches[i][0]] = 1;
		match2[p_matches[i][1]] = 1;
	}
	
	// process non matching score
	for( i=0; i<p_templ1->m_nbMinutiae; i++ )
	{
		if( !match1[i] )
		{
			x = p_templ1->m_minutiae[i].m_coord.m_x;
			y = p_templ1->m_minutiae[i].m_coord.m_y;
			
			if( x>=xMin && x<=xMax && y>=yMin && y<=yMax )
			{
				if( p_templ2->m_blocSize != 0 )
				{
				x2 = (x-p_templ2->m_localShiftX)/p_templ2->m_blocSize+p_templ2->m_localW/2;
				y2 = (y-p_templ2->m_localShiftY)/p_templ2->m_blocSize+p_templ2->m_localH/2;
				if( (UINT32)y2 < p_templ2->m_localH  && (UINT32)x2 < p_templ2->m_localW )
					qual = p_templ2->m_qual[y2*p_templ2->m_localW+x2];
				else
					qual = 7; // if we don't know we use a medium value
				}
				else
					qual = 7;
				
				nonMatch1[cptNonMatch1] = i;
				cptNonMatch1++;
				nonMatchScore+=qual*(p_templ1->m_minutiae[i].m_qual*p_templ1->m_minutiae[i].m_qual)/15;
			}
			
		}
	}
	for( i=0; i<p_templ2->m_nbMinutiae; i++ )
	{
		if( !match2[i] )
		{
			x = p_templ2->m_minutiae[i].m_coord.m_x;
			y = p_templ2->m_minutiae[i].m_coord.m_y;
			
			if( x>=xMin && x<=xMax && y>=yMin && y<=yMax )
			{
				
				if( p_templ1->m_blocSize != 0 )
				{
				x2 = (x-p_templ1->m_localShiftX)/p_templ1->m_blocSize+p_templ1->m_localW/2;
				y2 = (y-p_templ1->m_localShiftY)/p_templ1->m_blocSize+p_templ1->m_localH/2;
				if( (UINT32)y2 < p_templ1->m_localH  && (UINT32)x2 < p_templ1->m_localW )
					qual = p_templ1->m_qual[y2*p_templ1->m_localW+x2];
				else
					qual = 7; // if we don't know we use a medium value
				}
				else
					qual = 7;
				
				nonMatch2[cptNonMatch2] = i;
				cptNonMatch2++;
				nonMatchScore+=qual*(p_templ2->m_minutiae[i].m_qual*p_templ2->m_minutiae[i].m_qual)/15;
			}
		}
	}
	
	matchScore=0;
	
	// process matching score
	// loop on matching minutiae
	for( i=0; i<(INT32)(maxNumberOfCores+maxNumberOfDeltas); i++ )
	{
		if( p_matches[i][0] != -1 )
		{
			matchScore += 256L*256L;
		}
	}
	for( i=maxNumberOfCores+maxNumberOfDeltas; i<p_nbMatches; i++ )
	{
		x1 = (p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_x+p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_x)/2;
		y1 = (p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_y+p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_y)/2;
		// we check if there is a non matching minutiae between two matching minutiae
		// if it is not the case, the minutiae is counted
		cpt=1;
		for( j=i; j<p_nbMatches; j++ )
		{
			x2 = (p_templ1->m_minutiae[p_matches[j][0]].m_coord.m_x+p_templ2->m_minutiae[p_matches[j][1]].m_coord.m_x)/2;
			y2 = (p_templ1->m_minutiae[p_matches[j][0]].m_coord.m_y+p_templ2->m_minutiae[p_matches[j][1]].m_coord.m_y)/2;
			
			// process min and max of the rectangle made by the two minutiae
			if( x1<x2 )	{ xMin = x1;	xMax = x2; }
			else		{ xMin = x2;	xMax = x1; }
			
			if( y1<y2 ) { yMin = y1;	yMax = y2; }
			else		{ yMin = y2;	yMax = y1; }

			// if minutae are too close, we can't process a score boost as it would have no sense
			if( xMax-xMin > (INT32)translationTolerance && yMax-yMin > (INT32)translationTolerance )
			{
				// loop on non matching minutiae to test them
				inside = FALSE;
				for( k=0; k<cptNonMatch1; k++)
				{
					x = p_templ1->m_minutiae[nonMatch1[k]].m_coord.m_x;
					y = p_templ1->m_minutiae[nonMatch1[k]].m_coord.m_y;
					
					if( x > xMin && x < xMax && y > yMin && y < yMax )
					{
						inside = TRUE;
						break;
					}
				}
				// 2nd loop on non matching minutiae from template2
				if( !inside )
				{
					for( k=0; k<cptNonMatch2; k++)
					{
						x = p_templ2->m_minutiae[nonMatch2[k]].m_coord.m_x;
						y = p_templ2->m_minutiae[nonMatch2[k]].m_coord.m_y;
						
						if( x > xMin && x < xMax && y > yMin && y < yMax )
						{
							inside = TRUE;
							break;
						}
					}
				}
				
				if( !inside )
					cpt++;
			}
		}
		
		// now calculate the score for the minutia
		x = p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_x-p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_x;
		y = p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_y-p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_y;
		diffOri = zAbs(p_templ1->m_minutiae[p_matches[i][0]].m_ori-p_templ2->m_minutiae[p_matches[i][1]].m_ori);
		if( diffOri > 127 ) 
			diffOri = 255-diffOri;
		dist=(zSqrt(x*x+y*y)+diffOri)/TOINT(distDivCoef,Q15_16);
		matchScore += (cpt*p_templ1->m_minutiae[p_matches[i][0]].m_qual*p_templ2->m_minutiae[p_matches[i][1]].m_qual)/(dist+1);
		
	}

	
	if(nonMatchScore == 0 )nonMatchScore=1; 
	
	
	// process mean freq similarity
	txFreq = zAbs(p_templ1->m_meanFreq-p_templ2->m_meanFreq);
	if( p_templ1->m_meanFreq == 0 || p_templ2->m_meanFreq == 0 )
		txFreq = 1;

	txFreq = 100-txFreq;
	
	if( txFreq < 0 ) txFreq = 0;
	
	scoreOri = matchOri(p_templ1,p_templ2);

	matchScore /= 256;
	nonMatchScore /= 256;
	
	
	// process final score
	mulFact = FMUL(FADD(256,FDIV(ITOFIX32(scoreOri,Q7_8),FCONV(oriMatchNormalize,Q15_16,Q7_8),Q7_8)),matchScore+nonMatchScore,Q7_8);
	if( mulFact == 0 ) mulFact = 1;
	score = TOINT(FMULI(FDIV(txFreq*matchScore/100,mulFact,Q7_8),maxScore),Q7_8);
	
	if( minutiaeBoost )  // to be tested !!
		score =  score * (p_nbMatches-(maxNumberOfCores+maxNumberOfDeltas)) / maxNumberOfMinutiaeParam; //maxNumberOfMinutiae;

	score = TOINT(FMULI(score,FCONV(scoreNormalize,Q15_16,Q7_8)),Q7_8);
	
	if( score < 0 ) score = 0;
	if( score > (INT32)maxScore ) score = maxScore;  
	
	// memory release
	FREE(match1);
	FREE(match2);
	FREE(nonMatch1);
	FREE(nonMatch2);
	
	*p_score = score;
	return ZEF_OK;

	// error handling
errorHandler:

	if( match1 != NULL )
		FREE(match1);
	if( match2 != NULL )
		FREE(match2);
	if( nonMatch1 != NULL )
		FREE(nonMatch1);
	if( nonMatch2 != NULL )
		FREE(nonMatch2);

	return error;
	
}

UINT32 scoreIndex( Template *p_templ1, Template *p_templ2 )
{
	INT32	i;
	INT32	diffs;
	UINT32	error=0;
	
	// check for non existant signature
	error=0;
	for( i=0; i<16; i++ )
		error += p_templ1->m_sign[i];
	if( error == 0 )
		return 256*16;
	
	error=0;
	for( i=0; i<16; i++ )
		error += p_templ2->m_sign[i];
	if( error == 0 )
		return 256*16;
	
	
	error=0;
	for( i=0; i<16; i++ )
	{
		diffs = zAbs((INT32)p_templ1->m_sign[i] - (INT32)p_templ2->m_sign[i]);
		if( diffs > 128 ) diffs = 256 - diffs;
		error += diffs;
	}
	
	return error;
}


INT32	sortTemplates( UINT8 *p_templateBuffer, UINT8 **p_templateBufferArray, UINT32 p_nbTemplates,
					 UINT32 *p_coreIndexes, UINT32 *p_coreDist, UINT32 *p_noCoreIndexes, UINT32 *p_noCoreDist  )
{
	UINT32		i,score,dist1,dist2;
	INT32		x,y;
	Template	refTempl,templ;
	INT32		error;
	INT32		nbCoreDist;

	// init templ
	error = templateInit(&templ);
	if( error < ZEF_OK )
		goto errorHandler;
	
	// init refTempl
	error = templateInit(&refTempl);
	if( error < ZEF_OK )
		goto errorHandler;
	error = templateUnpacker(p_templateBuffer, &refTempl );
	if( error < ZEF_OK )
		goto errorHandler;
	
	// init faulty template index... p_nbTemplates = no faulty template
	p_noCoreIndexes[p_nbTemplates] = p_nbTemplates;

	// reinit indexes and distances
	for( i=0; i<p_nbTemplates; i++ )
		p_noCoreIndexes[i] = i;

	nbCoreDist = 0;
	
	for( i=0; i<p_nbTemplates; i++ )
	{
		// unpack template
		error = templateUnpacker( p_templateBufferArray[i], &templ );
		if( error < ZEF_OK )
		{
			// we store the index of the faulty template !!!
			p_noCoreIndexes[p_nbTemplates] = i;
			goto errorHandler;
		}
		
		// calc freq score
		if( refTempl.m_meanFreq != 0 && templ.m_meanFreq != 0 )
			p_noCoreDist[i] = 10*zAbs(refTempl.m_meanFreq-templ.m_meanFreq);
		else
			p_noCoreDist[i] = 1000;
		
		// scores only valid if refTemplate && templ have at least one core
		if( refTempl.m_nbCores != 0 && templ.m_nbCores != 0 )
		{	
			p_coreDist[nbCoreDist] = p_noCoreDist[i];
			p_coreIndexes[nbCoreDist] = i;

			// then core index score
			score = scoreIndex( &refTempl, &templ );
			if( score < 200 )
				p_coreDist[nbCoreDist] += score/2;
			else
				p_coreDist[nbCoreDist] += 100;
			
			// calc dist core score
			score = 1;
			if( refTempl.m_nbCores == 2 && templ.m_nbCores == 2 )
			{
				x = refTempl.m_cores[0].m_x - refTempl.m_cores[1].m_x;
				y = refTempl.m_cores[0].m_y - refTempl.m_cores[1].m_y;
				dist1 = zSqrt(x*x+y*y);
				
				x = templ.m_cores[0].m_x - templ.m_cores[1].m_x;
				y = templ.m_cores[0].m_y - templ.m_cores[1].m_y;
				dist2 = zSqrt(x*x+y*y);
				
				if( zAbs( dist1-dist2 < 12 ) )
					score++;
			}
			if( refTempl.m_nbDeltas > 0 && templ.m_nbDeltas > 0 )
			{
				x = refTempl.m_cores[0].m_x - refTempl.m_deltas[0].m_x;
				y = refTempl.m_cores[0].m_y - refTempl.m_deltas[0].m_y;
				dist1 = zSqrt(x*x+y*y);
				
				x = templ.m_cores[0].m_x - templ.m_deltas[0].m_x;
				y = templ.m_cores[0].m_y - templ.m_deltas[0].m_y;
				dist2 = zSqrt(x*x+y*y);
				
				if( zAbs( dist1-dist2 < 12 ) )
					score++;
			}
			
			p_coreDist[nbCoreDist] += 100/score;
			p_noCoreDist[i] = p_coreDist[nbCoreDist];
			nbCoreDist++;
		}
		else
			p_noCoreDist[i] += 150;
		
		error = templateUninit(&templ);
		if( error < ZEF_OK )
			goto errorHandler;
	}
	
	error = templateUninit(&refTempl);
	if( error < ZEF_OK )
		goto errorHandler;
	
	// we only sort the core tabs, the noCOre tabs will be sorted outside of the function
	// for speed improvement reasons
	sortUINT32( p_coreDist, p_coreIndexes, 0, nbCoreDist-1 );

	p_noCoreIndexes[p_nbTemplates] = p_nbTemplates;
	p_coreIndexes[p_nbTemplates] = nbCoreDist;
	
	return ZEF_OK;

		// error handling
errorHandler:

	templateUninit(&templ);
	templateUninit(&refTempl);

	return error;
}

#endif // ZEF_FIXED_POINT
