/*
Company :		ZEFYR Technologies
File :			fxTpsWarper.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	04/07/2003

Description : calculates a non rigid deformation between two sets of points using thin plate splines

Features : estimates local deformation of fingerprints (fixed point )
    
Functions : 

	initTpsWarper : initialize a new estimation with new data
	initMatrixA : build tyhe A matrix for TPS system resolution
	unInitTpsWarper : FREE memory used during TPS warping
	warp : calculate the new position of a poirn according to the deformation calculated with initTpsWarper
	getDiffs : get the "energy" needed for the deformation

Function prototypes : 

	INT32	initTpsWarper( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, INT32 *p_dstX, INT32 *p_dstY, FIXED32 p_alpha, TpsData *p_data);
	INT32	initMatrixA( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, FIXED32 p_alpha, TpsData *p_data );
	INT32	unInitTpsWarper( TpsData *p_data );
	INT32	warp( INT32 *p_x, INT32 *p_y, TpsData *p_data );
	UINT32  getDiffs( TpsData *p_data );

Necessary package : 

	customtypes.h
	errorCodes.h
	linAlgebra.h
	zMaths.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "fixedPoint.h"
#include "fxLinAlgebra.h"
#include "zMaths.h"

#include "fxTpsWarper.h"

#if defined(ZEF_FIXED_POINT) && defined(ZEF_CORRECT_DEFORMATION)

INT32 initTpsWarper( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, INT32 *p_dstX, INT32 *p_dstY, FIXED32 p_alpha, TpsData *p_data)
{
	INT32	error;
	INT32	i;
	INT32	even;

	// init
	p_data->m_index = NULL;
	p_data->m_matrixA = NULL;
	p_data->m_srcX = NULL;
	p_data->m_srcY = NULL;
	p_data->m_vectorBx = NULL;
	p_data->m_vectorBy = NULL;

	// initialise differences indicator
	p_data->m_error = 0;
	p_data->m_cptPts = 0;

	// then the number of control points
	p_data->m_nbPts = p_nbPts;

	// index
	p_data->m_index = (INT32*)MALLOC((p_nbPts+3)*sizeof(INT32));
	if( p_data->m_index == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	// and source points
	p_data->m_srcX = (INT32*)MALLOC((p_nbPts)*sizeof(INT32));
	if( p_data->m_srcX == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_data->m_srcY = (INT32*)MALLOC((p_nbPts)*sizeof(INT32));
	if( p_data->m_srcY == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_srcX[i] = FCONV(p_srcX[i],Q6_9,Q1_14);
		p_data->m_srcY[i] = FCONV(p_srcY[i],Q6_9,Q1_14);
	}

	// initialisation of matrix A
	error = initMatrixA( p_nbPts, p_data->m_srcX, p_data->m_srcY, p_alpha, p_data );
	if( error < ZEF_OK )
		goto errorHandler;

	// then we fill the right side with X  source coords
	p_data->m_vectorBx = (INT32*)MALLOC((p_nbPts+3)*sizeof(INT32));
	if( p_data->m_vectorBx == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_vectorBx[i] = FCONV(p_dstX[i],Q6_9,Q1_14);
	}
	for( i=p_nbPts; i<p_nbPts+3; i++ )
	{
		p_data->m_vectorBx[i] = 0;
	}
	

	p_data->m_vectorBy = (INT32*)MALLOC((p_nbPts+3)*sizeof(INT32));
	if( p_data->m_vectorBy == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// solve the system
	error = luDecompose(p_data->m_matrixA, p_data->m_nbPts+3, p_data->m_index, &even);
	if( error < ZEF_OK )
		goto errorHandler;

	error = luBackSubstitute(p_data->m_matrixA, p_data->m_nbPts+3, p_data->m_index, p_data->m_vectorBx );
	if( error < ZEF_OK )
		goto errorHandler;

	// now fill the right side with Y source coords
	
	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_vectorBy[i] = FCONV(p_dstY[i],Q6_9,Q1_14);
	}
	for( i=p_nbPts; i<p_nbPts+3; i++ )
	{
		p_data->m_vectorBy[i] = 0;
	}

	// and solve a second time
	error = luBackSubstitute(p_data->m_matrixA, p_data->m_nbPts+3, p_data->m_index, p_data->m_vectorBy );
	if( error < ZEF_OK )
		goto errorHandler;


	
	return ZEF_OK;


	// error handling
errorHandler:

	unInitTpsWarper(p_data);
	
	return error;

}
INT32 initMatrixA( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, FIXED32 p_alpha, TpsData *p_data )
{
	INT32		i,j;
	INT32		r;
	INT32		error;
	INT32		subLog;

	// init 
	subLog = zLog(1<<Q1_14);

	// allocation of matrix A
	p_data->m_matrixA = (INT32**)MALLOC( (p_nbPts+3)*sizeof(INT32*) );
	if( p_data->m_matrixA == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	for( i=0; i<p_nbPts+3; i++ )
	{
		p_data->m_matrixA[i] = (INT32*)MALLOC( (p_nbPts+3)*sizeof(INT32) );
		if( p_data->m_matrixA[i] == NULL )
		{
			error = ZEF_OUT_OF_MEMORY;
			goto errorHandler;
		}
	}

	// We fill the K part of the matrix
	for( i=0; i<p_nbPts; i++ )
	{
		for( j=0; j<=i; j++ )
		{
			r = FMUL((p_srcX[i]-p_srcX[j]),(p_srcX[i]-p_srcX[j]),Q1_14) + FMUL((p_srcY[i]-p_srcY[j]),(p_srcY[i]-p_srcY[j]),Q1_14);
			if( r == 0 ) 
			{	
				p_data->m_matrixA[i][j] = 0;
				p_data->m_matrixA[j][i] = 0;
			}
			else
			{
				r /= 2; // to avoid overflows (it doesn't really change the final result

				// here we should calculate r*ln(sqrt(r)) = r*ln(r)/2
			// but to avoid division in this loop (for speed improvement on machines where division is not cabled
			// we do not devide by two (which doesn't change very muche the final result !)
				p_data->m_matrixA[i][j] = FMUL(FCONV(zLog(r)-subLog,Q4_11,Q1_14),r,Q1_14);  // equivalent to r*ln(r)
				p_data->m_matrixA[j][i] = p_data->m_matrixA[i][j];
			}
		}
	}


	// then we add I*alpha in order to have K+I*alpha
	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_matrixA[i][i] += FCONV(p_alpha,Q15_16,Q1_14);
	}

	// and add P in the top right of the matrix
	for( i=0; i< p_nbPts; i++ )
	{
		p_data->m_matrixA[i][p_nbPts] = ITOFIX32(1,Q1_14);
		p_data->m_matrixA[i][p_nbPts+1] = p_srcX[i];
		p_data->m_matrixA[i][p_nbPts+2] = p_srcY[i];
	}

	// Add P transposed in the bottom left of the matrix
	for( j=0; j< p_nbPts; j++ )
	{
		p_data->m_matrixA[p_nbPts][j] = ITOFIX32(1,Q1_14);
		p_data->m_matrixA[p_nbPts+1][j] = p_srcX[j];
		p_data->m_matrixA[p_nbPts+2][j] = p_srcY[j];
	}

	// and zeros in the bottom right of the matrix
	for( i=0; i<3; i++)
	{
		for( j=0; j<3; j++)
		{
			p_data->m_matrixA[p_nbPts+i][p_nbPts+j] = 0;	
		}
	}

	return ZEF_OK;

	// error handling
errorHandler:

	unInitTpsWarper(p_data);

	return error;
}
INT32 unInitTpsWarper( TpsData *p_data )
{
	INT32	i;

	for( i=0; i<p_data->m_nbPts+3; i++ )
	{
		if( p_data->m_matrixA[i] != NULL )
			FREE( p_data->m_matrixA[i] );
	}

	if( p_data->m_matrixA != NULL )
		FREE(p_data->m_matrixA);
	if( p_data->m_vectorBx != NULL )
		FREE(p_data->m_vectorBx);
	if( p_data->m_vectorBy != NULL )
		FREE(p_data->m_vectorBy);
	if( p_data->m_srcX != NULL )
		FREE(p_data->m_srcX);
	if( p_data->m_srcY != NULL )
		FREE(p_data->m_srcY);
	if( p_data->m_index != NULL );
		FREE(p_data->m_index);

	p_data->m_nbPts = 0;
	p_data->m_error = 0;
	p_data->m_cptPts = 0;

	return ZEF_OK;
}
INT32 warp( INT32 *p_x, INT32 *p_y, TpsData *p_data )
{
	INT32	i;
	INT32	r;
	INT32	tmpX,tmpY,x,y,x1,y1;
	INT32	Ur,subLog;

	// init 
	x = *p_x;
	y = *p_y;
	subLog = zLog(1<<Q1_14);

	x1 = FCONV(*p_x,Q6_9,Q1_14);
	y1 = FCONV(*p_y,Q6_9,Q1_14);

	tmpX = p_data->m_vectorBx[p_data->m_nbPts] + FMUL(p_data->m_vectorBx[p_data->m_nbPts+1],(x),Q6_9) + FMUL(p_data->m_vectorBx[p_data->m_nbPts+2],(y),Q6_9);
	tmpY = p_data->m_vectorBy[p_data->m_nbPts] + FMUL(p_data->m_vectorBy[p_data->m_nbPts+1],(x),Q6_9) + FMUL(p_data->m_vectorBy[p_data->m_nbPts+2],(y),Q6_9);

	// process new coordinates
	for( i=0; i<p_data->m_nbPts; i++ )
	{
		r = FMUL((p_data->m_srcX[i]-(x1)),(p_data->m_srcX[i]-(x1)),Q1_14)+FMUL((p_data->m_srcY[i]-(y1)),(p_data->m_srcY[i]-(y1)),Q1_14);
		if( r != 0 )
		{
			r /= 2; // to avoid overflows (it doesn't really change the final result

			// here we should calculate r*ln(sqrt(r)) = r*ln(r)/2
			// but to avoid division in this loop (for speed improvement on machines where division is not cabled
			// we do not devide by two (which doesn't change very muche the final result !)
			Ur = FCONV(FMUL(FCONV(zLog(r)-subLog,Q4_11,Q1_14),r,Q1_14),Q1_14,Q6_9); // equivalent to r*ln(r)
		}
		else
			Ur = 0;

		tmpX += FMUL(Ur, p_data->m_vectorBx[i],Q6_9);
		tmpY += FMUL(Ur, p_data->m_vectorBy[i],Q6_9);
	}

	// update deformation indicator
	p_data->m_error += FMUL(((x)-tmpX),((x)-tmpX),Q6_9) + FMUL(((y)-tmpY),((y)-tmpY),Q6_9);
	(p_data->m_cptPts)++;

	*p_x = tmpX; 
	*p_y = tmpY;


	return ZEF_OK;
}
UINT32  getDiffs( TpsData *p_data )
{
	if( p_data->m_cptPts != 0 )
		return p_data->m_error/p_data->m_cptPts;
	else
		return 0;
}

#endif //ZEF_FIXED_POINT && ZEF_CORRECT_DEFORMATION
