/*
Company :		ZEFYR Technologies
File :			geometry.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	07/05/2003

Description : Geometric trnasformations functions

Features : different transformation functions : scaling, rotations, etc.		
    
Functions : 

	scale			: scale an image by a real factor (nearest neighbour )
	integerScale	: scale an image by an integer factor (bilinear interpolation)
	integerScale32	: scale an image by an integer factor (bilinear interpolation)
	integerScaleS	: scale an image by an integer factor (bilinear interpolation)
	scaleOri		: scale an orientation map
	checkBlocMultiple : crops an image to a bloc size multiple

Function prototypes : 

	INT32 scale( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, REAL32 p_scale )
	INT32 integerScale( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_scale );
	INT32 integerScale32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_scale );
	INT32 checkBlocMultiple( UINT8 *p_src, UINT32 *p_width, UINT32 *p_height );
	INT32 scaleOri( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_factor );
	INT32 integerScaleS( INT8 *p_src, INT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale );

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h

Comments : none

*/

#include <stdlib.h>
#include <string.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "zMaths.h"

#include "geometry.h"

// check if the size of the image is a multiple of SEGM_BLOC_SIZE
// if not is crops the image so that its size is a multiple of SEGM_BLOC_SIZE
// no memory allocation is done since the dest image size will be smaller or equal to 
// the original size
// p_width and p_height will contain the final size
INT32 checkBlocMultiple( UINT8 *p_src, UINT32 *p_width, UINT32 *p_height )
{
	UINT32		i,j;
	UINT32		offset;
	UINT32		newWidth,newHeight;

	newWidth = (*p_width/SEGM_BLOC_SIZE)*SEGM_BLOC_SIZE;
	newHeight = (*p_height/SEGM_BLOC_SIZE)*SEGM_BLOC_SIZE;

	// is the new size is different from the old size
	// we copy the image with its new size in the old buffer
	offset=0;
	if( newWidth != *p_width || newHeight != *p_height )
	{
		for( j=0; j<newHeight; j++ )
		{
			for( i=0; i<newWidth; i++ )
			{
				p_src[offset++] = p_src[j*(*p_width)+i];
			}
		}

		// an we update the size
		*p_width = newWidth;
		*p_height = newHeight;
	}

	return ZEF_OK;
}

// scale p_src by p_scale (scale=p_scale/256) and stores the result in p_dst
// p_dst must be allocated by the caller with enough space to hold an image
// of p_width*p_height*p_scale*p_scale bytes
// p_src and p_dst can be the same pointer (in place processing )
INT32 scale( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale )
{
	INT32	newWidht,newHeight,width,height;
	INT32	y,x,j,i,di,dj,offset,offset2;
	UINT32	scale;
	UINT8	*temp;

	if( p_scale != 256 ) // scale by 256/256=1 = no scale !!!
	{
		newWidht = (p_scale*p_width)>>8;
		newHeight = (p_scale*p_height)>>8;
		width = p_width;
		height = p_height;

		temp = (UINT8*)MALLOC( newWidht*p_height*sizeof(UINT8) );

		scale = 65536/p_scale;
		offset = 0;

		for( y=0; y<height; y++ )
		{
			offset2 = y*width;
			for( x=newWidht-2; x>=0; x-- )
			{
				i=(x*scale)>>8;
				di=(x*scale)&255;
				temp[offset++] = (p_src[i+offset2]*(256-di)+p_src[i+1+offset2]*di)>>8;
			}
			offset++;
		}

		offset = 0;
		for( y=0; y<newHeight-1; y++ )
		{
			j = (y*scale)>>8;
			dj=(y*scale)&255;
			for( x=newWidht-1; x>=0; x-- )
			{
				p_dst[offset++] = (temp[x+j*newWidht]*(256-dj)+temp[x+(j+1)*newWidht]*dj)>>8;
			}
		}

		for( x=newWidht-1; x>=0; x-- )
		{
			p_dst[offset] = p_dst[offset-newWidht];
			offset++;
		}
		for( y=0; y<newHeight-1; y++ )
		{
			p_dst[y*newWidht] = p_dst[y*newWidht+1];	
		}

		FREE(temp);
	}
	else
	{
		offset=0;
		for( i=p_height*p_width-1; i>=0; i-- )
		{
			p_dst[offset] = p_src[offset];
			offset++;
		}
	}

	return ZEF_OK;
}

// scale p_src by p_scale and stores the result in p_dst
// p_dst must be allocated by the caller with enough space to hold an image
// of p_width*p_height*p_scale*p_scale bytes
// p_src and p_dst cannot be the same pointer (no in place processing )
INT32 integerScale( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale )
{
	UINT32	i,j,x,y;
	UINT32	*p0,*p1,*p2,*p3;
	UINT32	scale2;
	UINT32	newWidth;
	UINT32	offset,offset2;
	UINT8	*tmp;
	UINT8	v1,v2,v3,v4;
	INT32	error;

	// init
	p0 = NULL;
	p1 = NULL;
	p2 = NULL;
	p3 = NULL;
	newWidth = p_width*p_scale;
	scale2 = (1<<8)/(p_scale*p_scale);

	// allocation of the coef tables
	p0 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p1 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p2 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p3 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	if( p0 == 0 || p1 == 0 || p2 == 0 || p3 == 0 )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// init of the coef tables
	i=0;
	for( y=0; y<p_scale; y++ )
	{
		for( x=0; x<p_scale; x++ )
		{
			p0[i] = (p_scale-x)*(p_scale-y);
			p1[i] = x*(p_scale-y);
			p2[i] = (p_scale-x)*y;
			p3[i] = x*y;
			i++;
		}
	}

	// we fill the new image
	tmp = p_dst+(p_scale/2)+(p_scale/2)*newWidth;
	for( j=0; j<p_height-1; j++ )
	{
		for( y=0; y<p_scale; y++ )
		{
			offset2=j*p_width;
			for( i=0; i<p_width-1; i++ )
			{
				v1 = p_src[offset2];
				v2 = p_src[offset2+1];
				v3 = p_src[offset2+p_width];
				v4 = p_src[offset2+p_width+1];
				offset = y*p_scale;
				for( x=0; x<p_scale; x++ )
				{
					*(tmp++) =	((p0[offset]*v1+p1[offset]*v2+
								p2[offset]*v3+p3[offset]*v4)*scale2)>>8;
					offset++;
				}
				offset2++;
			}
			tmp += p_scale;
		}
	}

	FREE(p0);
	FREE(p1);
	FREE(p2);
	FREE(p3);

	return ZEF_OK;

	// error Handling
errorHandler:

	if( p0 != NULL )
		FREE(p0);
	if( p1 != NULL )
		FREE(p1);
	if( p2 != NULL )
		FREE(p2);
	if( p3 != NULL )
		FREE(p3);

	return error;
}

// a version of integerScale for UINT32 data arrays
INT32 integerScale32( UINT32 *p_src, UINT32 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale )
{
	INT32	i,j,x,y;
	UINT32	*p0,*p1,*p2,*p3;
	UINT32	scale2;
	UINT32	newWidth;
	UINT32	offset,offset2;
	UINT32	*tmp;
	UINT32	v1,v2,v3,v4;
	INT32	error;

	// init
	p0 = NULL;
	p1 = NULL;
	p2 = NULL;
	p3 = NULL;
	newWidth = p_width*p_scale;
	scale2 = (1<<8)/(p_scale*p_scale);

	// allocation of the coef tables
	p0 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p1 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p2 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p3 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	if( p0 == 0 || p1 == 0 || p2 == 0 || p3 == 0 )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// init of the coef tables
	i=0;
	for( y=0; y<(INT32)p_scale; y++ )
	{
		for( x=0; x<(INT32)p_scale; x++ )
		{
			p0[i] = (p_scale-x)*(p_scale-y);
			p1[i] = x*(p_scale-y);
			p2[i] = (p_scale-x)*y;
			p3[i] = x*y;
			i++;
		}
	}

	// we fill the new image
	tmp = p_dst+(p_scale/2)+(p_scale/2)*newWidth;
	for( j=0; j<(INT32)p_height-1; j++ )
	{
		for( y=0; y<(INT32)p_scale; y++ )
		{
			offset2=j*p_width;
			for( i=0; i<(INT32)p_width-1; i++ )
			{
				v1 = p_src[offset2];
				v2 = p_src[offset2+1];
				v3 = p_src[offset2+p_width];
				v4 = p_src[offset2+p_width+1];
				offset = y*p_scale;
				for( x=0; x<(INT32)p_scale; x++ )
				{
					*(tmp++) =	((p0[offset]*v1+p1[offset]*v2+
								p2[offset]*v3+p3[offset]*v4)*scale2)>>8;
					offset++;
				}
				offset2++;
			}
			tmp += p_scale;
		}
	}

	FREE(p0);
	FREE(p1);
	FREE(p2);
	FREE(p3);

	return ZEF_OK;

	// error Handling
errorHandler:

	if( p0 != NULL )
		FREE(p0);
	if( p1 != NULL )
		FREE(p1);
	if( p2 != NULL )
		FREE(p2);
	if( p3 != NULL )
		FREE(p3);

	return error;
}

// special scaling for discrete orientation maps
INT32 scaleOri( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_factor )
{
	UINT32		i,x,y;	
	UINT32		*p,factor2;
	INT8		val1,val2,val;
	UINT32		width2;
	INT32		error;
	UINT32		numOfOrientation;
	
	// init
	LOCK_MUTEX( g_paramsMutex );
		numOfOrientation = g_params.m_numOfOrientation;
	UNLOCK_MUTEX( g_paramsMutex );
	p = NULL;

	// memory allocation
	p = (UINT32*)MALLOC( p_factor*sizeof(UINT32) );
	if( p == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	// init
	factor2=p_factor/2;
	width2 = p_width*p_factor;
	for( i=0; i<p_factor; i++ )
		p[i] = i;

	memset( p_dst, 0, p_width*p_height*p_factor*p_factor );

	// first interpolation of the raws
	for( y=0; y<p_height-1; y++ )
	{
		for( x=0; x<p_width-1; x++ )
		{
			val1 = p_src[y*p_width+x];
			val2 = p_src[y*p_width+x+1];

			if( val1 > (INT32)numOfOrientation/2+1 && val2 < (INT32)numOfOrientation/2-1 )
				val1 -= numOfOrientation;
			if( val1 < (INT32)numOfOrientation/2-1 && val2 > (INT32)numOfOrientation/2+1 )
				val2 -= numOfOrientation;

			for( i=0; i<p_factor; i++ )
			{
				val = ((p_factor-p[i])*val1+p[i]*val2)/p_factor;
				if( val < 0 ) val += numOfOrientation;
				p_dst[(factor2+y*p_factor)*width2+factor2+x*p_factor+i] = val;
			}
		}
	}

	// then interpolation of the columns
	for( y=0; y<p_height-1; y++ )
	{	
		for( x=0; x<(p_width-1)*p_factor; x++ )
		{
			val1 = p_dst[(factor2+y*p_factor)*width2+x+factor2];
			val2 = p_dst[(factor2+(y+1)*p_factor)*width2+x+factor2];

			if( val1 > (INT32)numOfOrientation/2+1 && val2 < (INT32)numOfOrientation/2-1 )
				val1 -= numOfOrientation;
			if( val1 < (INT32)numOfOrientation/2-1 && val2 > (INT32)numOfOrientation/2+1 )
				val2 -= numOfOrientation;

			for( i=0; i<p_factor; i++ )
			{
				val = ((p_factor-p[i])*val1+p[i]*val2)/p_factor;
				if( val < 0 ) val += numOfOrientation;
				p_dst[(factor2+y*p_factor+i)*width2+factor2+x] = val;
			}
		}
	}

	// deallocation
	FREE(p);

	return ZEF_OK;

	// error Handling
errorHandler:

	FREE(p);

	return error;
}

// scale p_src by p_scale and stores the result in p_dst using a mask
// p_dst must be allocated by the caller with enough space to hold an image
// of p_width*p_height*p_scale*p_scale bytes
// p_src and p_dst cannot be the same pointer (no in place processing )
INT32 integerScaleS( INT8 *p_src, INT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_scale )
{
	UINT32	i,j,x,y;
	UINT32	*p0,*p1,*p2,*p3;
	UINT32	scale2;
	UINT32	newWidth;
	UINT32	offset,offset2;
	INT8	*tmp;
	INT8	v1,v2,v3,v4;
	INT32	error;

	// init
	p0 = NULL;
	p1 = NULL;
	p2 = NULL;
	p3 = NULL;
	newWidth = p_width*p_scale;
	scale2 = (1<<8)/(p_scale*p_scale);

	// allocation of the coef tables
	p0 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p1 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p2 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	p3 = (UINT32*)MALLOC( p_scale*p_scale*sizeof(UINT32) );
	if( p0 == 0 || p1 == 0 || p2 == 0 || p3 == 0 )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// init of the coef tables
	i=0;
	for( y=0; y<p_scale; y++ )
	{
		for( x=0; x<p_scale; x++ )
		{
			p0[i] = (p_scale-x)*(p_scale-y);
			p1[i] = x*(p_scale-y);
			p2[i] = (p_scale-x)*y;
			p3[i] = x*y;
			i++;
		}
	}

	// we fill the new image
	tmp = p_dst+(p_scale/2)+(p_scale/2)*newWidth;
	for( j=0; j<p_height-1; j++ )
	{
		for( y=0; y<p_scale; y++ )
		{
			offset2=j*p_width;
			for( i=0; i<p_width-1; i++ )
			{
				v1 = p_src[offset2];
				v2 = p_src[offset2+1];
				v3 = p_src[offset2+p_width];
				v4 = p_src[offset2+p_width+1];
				offset = y*p_scale;
				
				for( x=0; x<p_scale; x++ )
				{
					*(tmp++) =	((p0[offset]*v1+p1[offset]*v2+
								p2[offset]*v3+p3[offset]*v4)*scale2)>>8;
					offset++;
				}
				offset2++;
			}
			tmp += p_scale;
		}
	}

	FREE(p0);
	FREE(p1);
	FREE(p2);
	FREE(p3);

	return ZEF_OK;

	// error Handling
errorHandler:

	if( p0 != NULL )
		FREE(p0);
	if( p1 != NULL )
		FREE(p1);
	if( p2 != NULL )
		FREE(p2);
	if( p3 != NULL )
		FREE(p3);

	return error;
}

