/*
Company :		ZEFYR Technologies
File :			greyLevelTrans.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	06/05/2003

Description : Image normalization functions, K-means, etc.

Features : different normalization functions		
    
Functions : 

	meanStdDev : evaluation the mean and standard deviation of an image
	localMeanStdDev : evaluation the mean and standard deviation of blocs of an image
	normalize : set the mean and standard deviation of an image to the given values
	normalizeLocally : set the mean and standard deviation of pixel neiborhoods to the given values
	kMeans : classify the pixels of an image using a k-means algortihms (useb for segmentation)
	thresholdImage : binarize an image, based on the threshold level

Function prototypes : 

	INT32 meanStdDev( UINT8 *p_src, UINT32 p_width, UINT32 p_height, INT32 *p_mean, INT32 *p_stdDev );
	INT32 normalize( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_mean, INT32 p_stdDev );
	INT32 localMeanStdDev( UINT8 *p_src, UINT32 p_width, UINT32 p_height, INT32 *p_mean, INT32 *p_stdDev );
	INT32 normalizeLocally( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_mean, INT32 p_stdDev );
	INT32 kMeans( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_nbClasses );
	INT32 thresholdImage( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_threshold );

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	zMaths.h
	geometry.h
	BMPfiles.h
	filters.h

Comments : none

*/

#include <stdlib.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "zMaths.h"
#include "geometry.h"
#include "BMPfiles.h"
#include "filters.h"

#include "greyLevelTrans.h"


// evaluate the mean and standard deviation of an image
// the input image is left unchanged
// image must be smaller than 1024x1024
INT32 meanStdDev( UINT8 *p_src, UINT32 p_width, UINT32 p_height, INT32 *p_mean, INT32 *p_stdDev )
{
	INT32	mean = 0;
	INT32	stdDev = 0;
	UINT32	i;
	UINT32	imageSize;

	// compute image size
	imageSize = p_width*p_height;

	// then add pixel values for mean and std dev evaluation
	for( i=0; i<imageSize; i++ )
	{
		mean	+= p_src[i];
		stdDev	+= (p_src[i]*p_src[i])>>4; // to avoid overflows
	}

	// We calculate the mean : mean = sum of pixels / number of pixels
	if( imageSize != 0 )
		mean	/= imageSize;
	// We calculate the standard deviation : stdDev = sqrt( mean(pixels*pixels) - mean(pixel)*mean(pixels) )
	if( (imageSize>>4) != 0 )
		stdDev	/= (imageSize>>4);

	stdDev	-= mean*mean;
	
	stdDev = zSqrt( stdDev );

	// set the return values 
	*p_mean = mean;
	*p_stdDev = stdDev;

	return ZEF_OK;
}

// normalizes image p_src into p_dst with p_mean mean and p_stdDev standard deviation
// if p_mean = -1 p_dst mean is left unchanged (is left to p_src mean)
// if p_stdDev = -1 p_dst standard deviation is left unchanged (is left to p_src standard deviation)
// p_dst can be the same as p_src (in place processing )
INT32 normalize( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_mean, INT32 p_stdDev )
{
	INT32	error;
	INT32	mean = 0;
	INT32	stdDev = 0;
	INT32	tmpVal;
	UINT32	i;
	UINT32	imageSize;
	INT32	divider;

	// compute image size
	imageSize = p_width*p_height;

	// evaluate mean and stdDev values
	error = meanStdDev( p_src, p_width, p_height, &mean, &stdDev );
	if( error < ZEF_OK )
		goto errorHandler;

	// we handle the case where we don't want to change the mean
	if( p_mean == -1 )
		p_mean = mean;

	// we handle the case where we don't want to change the standard deviation
	if( p_stdDev == -1 )
		p_stdDev = stdDev;

	// we handle case where stdDev = 0
	if( stdDev == 0 )
		stdDev = 1;

	divider = (p_stdDev<<16)/stdDev;

	// this loop modifies the image in order to have the mean and stdDev we want
	for( i=0; i<imageSize; i++ )
	{
		tmpVal = p_mean + (((p_src[i]-mean)*divider)>>16);
		// saturation mode for overflow
		if( tmpVal > 255 ) tmpVal = 255;
		if( tmpVal < 0 ) tmpVal = 0;
		p_dst[i] = tmpVal;
	}

	return ZEF_OK;

	// error handling
errorHandler:

	return error;
}

// evaluate the local mean and standard deviation on SEGM_BLOC_SIZExSEGM_BLOC_SIZE blocs of an image
// p_mean and p_stdDev arrays must be allocated before calling this function
// their max size is (p_width/SEGM_BLOC_SIZE)*((p_height/SEGM_BLOC_SIZE)
// p_width and p_height MUST BE multiples of SEGM_BLOC_SIZE
// p_src is left unchanged
INT32 localMeanStdDev( UINT8 *p_src, UINT32 p_width, UINT32 p_height, UINT32 *p_mean, UINT32 *p_stdDev )
{
	UINT32		i,j,k;
	UINT32		miniWidth, miniHeight;
	UINT32		miniSize,blocSize;
	UINT32		offset;

	// set width and height of intermediate arrays
	miniWidth = p_width / SEGM_BLOC_SIZE;
	if( p_width%SEGM_BLOC_SIZE) 
		return ZEF_NOT_BLOC_MULTIPLE;
	miniHeight = p_height / SEGM_BLOC_SIZE;
	if( p_height%SEGM_BLOC_SIZE) 
		return ZEF_NOT_BLOC_MULTIPLE;

	blocSize = SEGM_BLOC_SIZE*SEGM_BLOC_SIZE;
	miniSize = miniHeight*miniWidth;

	// initializations
	for(i=0; i<miniSize; i++ )
	{
		p_mean[i] = 0;
		p_stdDev[i] = 0;
	}

	// we fill the tabs
	k=0;
	for( j=0; j<p_height; j++ )
	{
		offset = j/SEGM_BLOC_SIZE*miniWidth;
		for( i=0; i<p_width; i++ )
		{		
			p_mean[offset+i/SEGM_BLOC_SIZE] += p_src[k];
			k++;
		}
	}
	
	// then divide the results in order to have the means and stdDevs
	for( i=0; i<miniSize; i++ )
		p_mean[i] /= blocSize;

	// we fill the tabs
	k=0;
	for( j=0; j<p_height; j++ )
	{
		offset = j/SEGM_BLOC_SIZE*miniWidth;
		for( i=0; i<p_width; i++ )
		{		
			p_stdDev[offset+i/SEGM_BLOC_SIZE] += (p_src[k]-p_mean[offset+i/SEGM_BLOC_SIZE])*(p_src[k]-p_mean[offset+i/SEGM_BLOC_SIZE]);
			k++;
		}
	}
	
	// then divide the results in order to have the means and stdDevs
	for( i=0; i<miniSize; i++ )
	{
		p_stdDev[i] /= blocSize;
		p_stdDev[i] = zSqrt(p_stdDev[i] );
	}


	return ZEF_OK;
}

// normalizes image p_src into p_dst with p_mean mean and p_stdDev standard deviation
// normalization is done pixel by pixel on a local neiborhood. The size of the neiborhood
// is defined by SEGM_BLOC_SIZE static parameter
// if p_mean = -1 p_dst mean is left unchanged (is left to p_src mean)
// if p_stdDev = -1 p_dst standard deviation is left unchanged (is left to p_src standard deviation)
// p_dst can be the same as p_src (in place processing )
// p_width and p_height MUST BE multiples of SEGM_BLOC_SIZE
INT32 normalizeLocally( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, INT32 p_mean, INT32 p_stdDev )
{
	INT32		error;
	UINT32		i,j;
	UINT32		miniWidth, miniHeight;
	UINT32		miniSize,blocSize;
	UINT32		*miniMean,*miniStdDev;
	UINT32		*mean,*stdDev;
	INT32		tmpVal;

	// init
	miniMean = NULL;
	miniStdDev = NULL;
	mean = NULL;
	stdDev = NULL;

	// set width and height of intermediate arrays
	miniWidth = p_width / SEGM_BLOC_SIZE;
	if( p_width%SEGM_BLOC_SIZE)
		return ZEF_NOT_BLOC_MULTIPLE;
	miniHeight = p_height / SEGM_BLOC_SIZE;
	if( p_height%SEGM_BLOC_SIZE)
		return ZEF_NOT_BLOC_MULTIPLE;

	blocSize = SEGM_BLOC_SIZE*SEGM_BLOC_SIZE;
	miniSize = miniHeight*miniWidth;

	// memory allocation
	miniMean	= (UINT32*) MALLOC( miniSize*sizeof(UINT32) );
	if( miniMean == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	miniStdDev  = (UINT32*) MALLOC( miniSize*sizeof(UINT32) );
	if( miniStdDev == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	mean	= (UINT32*) MALLOC( miniSize*sizeof(UINT32)*blocSize );
	if( mean == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	stdDev  = (UINT32*) MALLOC( miniSize*sizeof(UINT32)*blocSize );
	if( stdDev == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// initializations
	for( j=0; j<miniSize*blocSize ; j++ )
		mean[j] = stdDev[j] = 0;

	// evaluate local mean and standard deviation
	error = localMeanStdDev( p_src, p_width, p_height, miniMean, miniStdDev );
	if( error < ZEF_OK )
		goto errorHandler;

	// then divide the results in order to have the means and stdDevs
	for( i=0; i<miniSize; i++ )
	{
		if( miniStdDev[i] != 0 )
			miniStdDev[i] = (p_stdDev<<16)/miniStdDev[i];
		else
			miniStdDev[i] = 0;
	}
	
	// we scale the means and stdDevs
	
	error = integerScale32( miniMean, mean, miniWidth, miniHeight, SEGM_BLOC_SIZE );
	if( error < ZEF_OK )
		goto errorHandler;
	
	error = integerScale32( miniStdDev, stdDev, miniWidth, miniHeight, SEGM_BLOC_SIZE );
	if( error < ZEF_OK )
		goto errorHandler;

	// this loop modifies the image in order to have the mean and stdDev we want
	for( i=0; i<p_width*p_height; i++ )
	{
		tmpVal = p_mean + ((((INT32)p_src[i]-(INT32)mean[i])*(INT32)stdDev[i])>>16);
		// saturation mode for overflow
		if( tmpVal > 255 ) 
			tmpVal = 255;
		if( tmpVal < 0 ) 
			tmpVal = 0;
		p_dst[i] = tmpVal;
	}

	FREE( mean);
	FREE( stdDev);
	FREE( miniMean );
	FREE( miniStdDev );

	return ZEF_OK;

	//error Handling
errorHandler:

	if( mean != NULL )
		FREE( mean);
	if( stdDev != NULL )
		FREE( stdDev);
	if( miniMean != NULL )
		FREE( miniMean );
	if( miniStdDev != NULL )
		FREE( miniStdDev );

	return error;
}

// k-means on the image LUT
// if you choose 4 classes you'll get an image with 4 grey level (0,1,2,3)
// p_src and p_dst can be the same pointer
INT32 kMeans( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_nbClasses )
{
	UINT32		j,qual;
	UINT32		min,max,dif,dist,temp;
	UINT32		*histo;
	UINT8		*belongTo,*index,*revIndex;
	UINT32		*center,*newCenter,*acc;
	BOOL8		changed;
	INT32		error,i;

	//init 
	histo = NULL;
	belongTo = NULL;
	index = NULL;
	revIndex = NULL;
	center = NULL;
	newCenter = NULL;
	acc = NULL;

	//allocations
	histo = (UINT32*) MALLOC( 256 * sizeof(UINT32) );
	if( histo == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	belongTo = (UINT8*) MALLOC( 256 * sizeof(UINT8) );
	if( belongTo == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	index = (UINT8*) MALLOC( p_nbClasses*sizeof(UINT8) );
	if( index == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	revIndex = (UINT8*) MALLOC( p_nbClasses*sizeof(UINT8) );
	if( revIndex == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	center = (UINT32*) MALLOC( p_nbClasses*sizeof(UINT32) );
	if( center == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	newCenter = (UINT32*) MALLOC( p_nbClasses*sizeof(UINT32) );
	if( newCenter == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	acc = (UINT32*) MALLOC( p_nbClasses*sizeof(UINT32) );
	if( acc == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// initialization of the histogram
	for( i=0; i<256; i++ )
		histo[i] = 0;

	// histogram calculation
	for( i=0 ;i<(INT32)(p_width*p_height); i++ )
		histo[p_src[i]]++;

	// initialization of the class centers
	for( i=0; i<256; i++ )
		if( histo[i] != 0 ) break;
	min = i;

	for( i=255; i>=0; i-- )
		if( histo[i] != 0 ) break;
	max = i;

	for( i=0; i<(INT32)p_nbClasses; i++ )
	{
		center[i] = min + (max-min)/(p_nbClasses*2) + i*(max-min)/p_nbClasses;
		newCenter[i] = 0;
		acc[i] = 0;
		index[i] = i;
	}

	// then we loop until stability
	do
	{
		for( i=0; i<256; i++ )
		{
			min = 256;
			for( j=0; j<p_nbClasses; j++ )
			{
				dist = zAbs( (INT32)center[j]-i );
				if( dist < min )
				{
					min = dist;
					belongTo[i] = j;
				}
			}
			newCenter[belongTo[i]] += i*histo[i];
			acc[belongTo[i]] += histo[i];
		}
		dif = 0;
		for( i=0; i<(INT32)p_nbClasses; i++ )
		{
			if( acc[i] == 0 ) acc[i] = 1;
			newCenter[i] /= acc[i];
			dif += zAbs( (INT32)newCenter[i]-(INT32)center[i] );
			center[i] = newCenter[i];
			newCenter[i] = 0;
			acc[i] = 0;
		}
	}
	while( dif > p_nbClasses-1 );

	// bubble sort... can be changed for a faster sort easily
	// but works quite well for small p_nbClasses
	changed = TRUE;
	while( changed )
	{
		changed = FALSE;
		for( i=0; i<(INT32)p_nbClasses-1; i++ )
		{
			if( center[i] > center[i+1] )
			{
				changed = TRUE;
				temp = center[i];
				center[i] = center[i+1];
				center[i+1] = temp;

				temp = index[i];
				index[i] = index[i+1];
				index[i+1] = temp;
			}
		}
	}
	
	// then fill the reversed tab
	for( i=0; i<(INT32)p_nbClasses; i++ )
		revIndex[index[i]] = i;
	
	// and write in the des image with the new classes
	for( i=0; i<(INT32)(p_width*p_height); i++ )
		p_dst[i] = revIndex[belongTo[p_src[i]]];

	qual = 1+(center[1]-center[0])*center[3]/256;

	// release memory
	FREE(histo);
	FREE(belongTo);
	FREE(index);
	FREE(revIndex);
	FREE(center);
	FREE(newCenter);
	FREE(acc);

	return qual; // qual index

	//error Handling
errorHandler:

	if( histo != NULL )
		FREE(histo);
	if( belongTo != NULL )
		FREE(belongTo);
	if( index != NULL )
		FREE(index);
	if( revIndex != NULL )
		FREE(revIndex);
	if( center != NULL )
		FREE(center);
	if( newCenter != NULL )
		FREE(newCenter);
	if( acc != NULL )
		FREE(acc);

	return error;
}

INT32 thresholdImage( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, UINT32 p_threshold )
{
	UINT32		i;


		for( i=0; i<p_width*p_height; i++ )
		{
			if( p_src[i] < p_threshold ) 
				p_dst[i] = 0;
			else
				p_dst[i] = 255;	
		}


	return ZEF_OK;
}

