/*
Company :		ZEFYR Technologies
File :			localParameters.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	11/06/2003

Description : local parameters determination

Features : fill template with orientation and quality infos	
    
Functions : 

	processLocalParameters : main function used to process local parameters
	setLocalParamSize : precalculate some info used to process local parameters (arrays size, shifts, etc.)
	fillLocalParams : processes local parameters (bloc orientation, bloc quality, etc)
	generateIndex : Genrate a local infos index based on the orientations around the main core

Function prototypes :

	INT32 processLocalParameters( BioEngineData *p_data );	
	INT32 setLocalParamSize( UINT8 *p_quality, UINT32 p_width, UINT32 p_height, Template *p_template );
	INT32 fillLocalParams( UINT8  *p_qual, UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template );
	INT32 generateIndex( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template );

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	zMaths.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>
#include <float.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "zMaths.h"

#include "localParameters.h"

INT32 processLocalParameters( BioEngineData *p_data )
{
	INT32	error;
#ifdef TIME_REPORT
	clock_t timeReport;
#endif


#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// set thge size of the local paralm arrays
	error = setLocalParamSize( p_data->m_quality, p_data->m_width/FUSION_BLOC_SIZE, p_data->m_height/FUSION_BLOC_SIZE, &(p_data->m_template) ); 
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "processLocalParameters - setLocalParamSize" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// fill template with local params
	error = fillLocalParams( p_data->m_quality, p_data->m_ori, p_data->m_width, p_data->m_height, &(p_data->m_template) );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "processLocalParameters - fillLocalParams" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	error = generateIndex( p_data->m_ori, p_data->m_width, p_data->m_height, &(p_data->m_template) );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "processLocalParameters - generateIndex" ,clock()-timeReport );
#endif

	return ZEF_OK;

	// error Handling
errorHandler:

	return error;
}

INT32 setLocalParamSize( UINT8 *p_quality, UINT32 p_width, UINT32 p_height, Template *p_template )
{
	INT32		i,j;
	BOOL8		startX,startY;
	Rect16		bound;
	INT32		xCenter,yCenter;
	INT32		error;
	INT32		localInfosBlocSize;
	INT32		localInfosBlocSizeParam,globalTemplateScaling;
	UINT32		globalImgScaling,maxNumberOfLocalInfos;

	// init
	LOCK_MUTEX( g_paramsMutex );
		localInfosBlocSizeParam = g_params.m_localInfosBlocSize;
		globalImgScaling = g_params.m_globalImgScaling;
		maxNumberOfLocalInfos = g_params.m_maxNumberOfLocalInfos;
		globalTemplateScaling = g_params.m_globalTemplateScaling;
	UNLOCK_MUTEX( g_paramsMutex );
	p_template->m_ori = NULL;
	p_template->m_qual = NULL;
	bound.m_left = SHRT_MAX;
	bound.m_right =	SHRT_MIN;
	bound.m_top = SHRT_MAX;
	bound.m_bottom = SHRT_MIN;

	startX = FALSE;
	startY = FALSE;

	// adjust local info block size to img scaling
	localInfosBlocSize = localInfosBlocSizeParam*globalImgScaling/256;

	// we are looking for the image boundaries
	for( j=0; j<(INT32)p_height; j++ )
	{
		startX = FALSE;
		for( i=0; i<(INT32)p_width; i++ )
		{
			if( p_quality[j*p_width+i] != 0 ) // beginning of a line (left limit)
			{
				if( (INT32)i<bound.m_left ) bound.m_left = i;
				startX = TRUE;
				break; // end of search for this line
			}
		}
		for( i=p_width-1; i>=0; i-- )
		{
			if( p_quality[j*p_width+i] != 0 ) // end of a line (right limit)
			{
				if( (INT32)i>bound.m_right ) bound.m_right = i;
				break; // end of search for this line
			}
		}

		// top of the image
		if( startX && !startY )
		{
			bound.m_top = j;
			startY = TRUE;
		}

		// bottom of the image
		/*if( !startX && startY )
		{
			bound.m_bottom = j;
			break; // end of search
		}*/
	}
	for( j=p_height-1; j>=0; j-- )
	{
		for( i=0; i<(INT32)p_width; i++ )
		{
			if( p_quality[j*p_width+i] != 0 ) // beginning of a line (bottom limit)
			{
				bound.m_bottom = j;
				j=0;
				break; // end of search
			}
		}
	}

	// now we handle some cases where we couldn't find the limits
	if( bound.m_left == SHRT_MAX ) bound.m_left = 0;
	if( bound.m_top == SHRT_MAX ) bound.m_top = 0;
	if( bound.m_right < bound.m_left ) bound.m_right = p_width;
	if( bound.m_bottom < bound.m_top ) bound.m_bottom = p_height;

	// now set boundaries to real coordinates (qual is a small image than the fingerprint)
	bound.m_left	*= FUSION_BLOC_SIZE;
	bound.m_right	*= FUSION_BLOC_SIZE;
	bound.m_top		*= FUSION_BLOC_SIZE;
	bound.m_bottom	*= FUSION_BLOC_SIZE;

	//set center of the tab
	xCenter = bound.m_left+(bound.m_right-bound.m_left)/2;
	yCenter = bound.m_top+(bound.m_bottom-bound.m_top)/2;

	// now set template localTab shift
	p_template->m_localShiftX = xCenter-p_template->m_cores[0].m_x;
	p_template->m_localShiftY = yCenter-p_template->m_cores[0].m_y;

	// then evaluate the number of blocs in X and Y direction
	if( localInfosBlocSize != 0 )
	{
		p_template->m_localW = (bound.m_right-bound.m_left)/localInfosBlocSize;
		p_template->m_localH = (bound.m_bottom-bound.m_top)/localInfosBlocSize;
	}
	else
		p_template->m_localW = p_template->m_localH = 0;
	

	// must be a multiple of two
	if( p_template->m_localW%2 != 0 ) p_template->m_localW++;
	if( p_template->m_localH%2 != 0 ) p_template->m_localH++;

	// handle the case where there are too much blocs
	if( p_template->m_localW*p_template->m_localH > (INT32)maxNumberOfLocalInfos )
	{
		p_template->m_localH = maxNumberOfLocalInfos/p_template->m_localW;
		// must still be a pultiple of two
		if( p_template->m_localH%2 != 0 ) p_template->m_localW--;
	}

	p_template->m_blocSize = localInfosBlocSizeParam*globalImgScaling*globalTemplateScaling/(256*256);

	// allocate localInfos memory
	p_template->m_ori = (UINT8*) MALLOC( p_template->m_localW*p_template->m_localH*sizeof(UINT8) );
	if( p_template->m_ori == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_template->m_qual = (UINT8*) MALLOC( p_template->m_localW*p_template->m_localH*sizeof(UINT8) );
	if( p_template->m_qual == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	return ZEF_OK;

	//error Handling
errorHandler:

	if( p_template->m_ori != NULL )
		FREE( p_template->m_ori );
	if( p_template->m_qual != NULL );
		FREE( p_template->m_qual);

	return error;
}

INT32 fillLocalParams( UINT8  *p_qual, UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template )
{
	INT32		i,j;
	UINT32		*tempQual;
	INT32		*tempSin;
	INT32		*tempCos;
	UINT32		*acc;
	INT32		xStart,yStart,val;
	UINT32		nbBlocs,width2,offset,offset1;
	INT32		error;
	INT32		localInfosBlocSize;
	INT32		localInfosBlocSizeParam,globalImgScaling;
	UINT32		qualityThreshold;

	// init
	LOCK_MUTEX( g_paramsMutex );
		localInfosBlocSizeParam = g_params.m_localInfosBlocSize;
		globalImgScaling = g_params.m_globalImgScaling;
		qualityThreshold = g_params.m_qualityThreshold;
	UNLOCK_MUTEX( g_paramsMutex );

	//init 
	tempQual = NULL;
	tempSin = NULL;
	tempCos = NULL;
	nbBlocs = p_template->m_localW*p_template->m_localH;
	width2 = p_width/FUSION_BLOC_SIZE;
	localInfosBlocSize = localInfosBlocSizeParam*globalImgScaling/256;

	// memory allocation
	tempQual = (UINT32*)MALLOC( nbBlocs*sizeof(UINT32) );
	if( tempQual == NULL )
	{	
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tempSin = (INT32*)MALLOC( nbBlocs*sizeof(INT32) );
	if( tempSin == NULL )
	{	
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	tempCos = (INT32*)MALLOC( nbBlocs*sizeof(INT32) );
	if( tempCos == NULL )
	{	
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	acc = (UINT32*)MALLOC( nbBlocs*sizeof(UINT32) );
	if( acc == NULL )
	{	
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// set start point in the src images
	xStart = p_template->m_cores[0].m_x + p_template->m_localShiftX-(localInfosBlocSize/2)*p_template->m_localW;
	yStart = p_template->m_cores[0].m_y + p_template->m_localShiftY-(localInfosBlocSize/2)*p_template->m_localH;


	// init arrays
	for( i=0; i<(INT32)nbBlocs; i++ )
	{
		tempQual[i] = 0;
		tempCos[i] = 0;
		tempSin[i] = 0;
		acc[i] = 0;
	}

	// loop into the images to fill the arrays
	for( j=0; j<p_template->m_localH*localInfosBlocSize; j++ )
	{
		offset1 = (j/localInfosBlocSize)*p_template->m_localW;
		for( i=0; i<p_template->m_localW*localInfosBlocSize; i++ )
		{
			if( (UINT32)(xStart+i)<p_width && (UINT32)(yStart+j) < p_height )
			{
				offset = offset1+(i/localInfosBlocSize);
				
				acc[offset]++;
				tempQual[offset] += p_qual[((yStart+j)/FUSION_BLOC_SIZE)*width2+((xStart+i)/FUSION_BLOC_SIZE)];
				tempSin[offset] += zSin127( p_ori[(yStart+j)*p_width+(xStart+i)] ) ;
				tempCos[offset] += zCos127( p_ori[(yStart+j)*p_width+(xStart+i)] ) ;
			}	
		}
	}

	// then write good values into template
	for( i=0; i<(INT32)nbBlocs; i++ )
	{
		if( acc[i] == 0 ) acc[i] = 1;
		if( (tempQual[i]/acc[i]) >= qualityThreshold )
			p_template->m_qual[i] = 16+(255-16)*((tempQual[i]/acc[i])-qualityThreshold)/(255-qualityThreshold);
		else
			p_template->m_qual[i] = 0;
		val = (zAtan(tempSin[i],tempCos[i]))/2;
		p_template->m_ori[i] = val;
	}

	FREE( tempQual );
	FREE( tempSin );
	FREE( tempCos );
	FREE( acc );
		
	return ZEF_OK;

	//error Handling
errorHandler:

	if( tempQual != NULL )
		FREE( tempQual );
	if( tempSin != NULL )
		FREE( tempSin );
	if( tempCos != NULL )
		FREE( tempCos );
	if( acc != NULL )
		FREE( acc );

	return error;

}

#if !defined(ZEF_FIXED_POINT)
INT32 generateIndex( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template )
{
	INT32	x,y,theta;
	INT32	diffs,distCore;
	INT32	ori,oldOri = -1;
	REAL32	coreOri;

	LOCK_MUTEX( g_paramsMutex );
		distCore = (g_params.m_distFindCoreAngle*g_params.m_globalImgScaling)>>8;
	UNLOCK_MUTEX( g_paramsMutex );

	// check if the core is not to close from the border
	if( p_template->m_cores[0].m_x >= distCore && p_template->m_cores[0].m_x < (INT32)p_width-distCore &&
		p_template->m_cores[0].m_y >= distCore && p_template->m_cores[0].m_y < (INT32)p_height-distCore )
	{
		coreOri = PI*p_template->m_oriCore/128; 

		// loop on ori to find core direction
		for( theta=0; theta<=16; theta++ )
		{
			x = p_template->m_cores[0].m_x+Real2Int32(distCore*zCos(coreOri+PI*theta/8));
			y = p_template->m_cores[0].m_y+Real2Int32(distCore*zSin(coreOri+PI*theta/8));

			ori = p_ori[y*p_width+x];

			if( oldOri != -1 )
			{
				diffs = ori-oldOri;
				if( diffs > 128 ) diffs  = 256 - diffs;
				if( diffs < -128 ) diffs  = 256 + diffs;
				
				p_template->m_sign[theta-1]=128+diffs;
			}
			oldOri = ori;

		}

	}
	else	
	{
		for( theta=0; theta<16; theta++ )
			p_template->m_sign[theta]=0;

		return ZEF_CORE_CLOSE_TO_BORDER;
	}

	return ZEF_OK;
	
}
#else
INT32 generateIndex( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, Template *p_template )
{
	INT32	x,y,theta;
	INT32	diffs;
	INT32	ori,oldOri = -1;
	INT32	coreOri,distCore;

	LOCK_MUTEX( g_paramsMutex );
		distCore = (g_params.m_distFindCoreAngle*g_params.m_globalImgScaling)>>8;
	UNLOCK_MUTEX( g_paramsMutex );

	// check if the core is not to close from the border
	if( p_template->m_cores[0].m_x >= distCore && p_template->m_cores[0].m_x < (INT32)p_width-distCore &&
		p_template->m_cores[0].m_y >= distCore && p_template->m_cores[0].m_y < (INT32)p_height-distCore )
	{
		// loop on ori to find core direction
		for( theta=0; theta<=16; theta++ )
		{
			// calcultate rotation
			coreOri = p_template->m_oriCore+256*theta/16;

			// apply rotation
			x = p_template->m_cores[0].m_x+((distCore*zCos127(coreOri))>>7);
			y = p_template->m_cores[0].m_y+((distCore*zSin127(coreOri))>>7);

			ori = p_ori[y*p_width+x];

			if( oldOri != -1 )
			{
				diffs = ori-oldOri;
				if( diffs > 128 ) diffs  = 256 - diffs;
				if( diffs < -128 ) diffs  = 256 + diffs;
				
				p_template->m_sign[theta-1]=128+diffs;
			}
			oldOri = ori;

		}

	}
	else	
	{
		for( theta=0; theta<16; theta++ )
			p_template->m_sign[theta]=0;

		return ZEF_CORE_CLOSE_TO_BORDER;
	}

	return ZEF_OK;
	
}
#endif

