/*
Company :		ZEFYR Technologies
File :			matchExport.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : SDK matching exported functions

Features : external interface functions		

Functions : 

	dpInitMatch		:	initialize matcher
	dpUnInitMatch	:	unititialize matcher
	setSecurity		:	set the global security threshold
	identifyTemplate :	compares a template to many others
	setDbMaxSearch	:	set maximum search depth during identification

Function prototypes : 

	INT32	dpInitMatch( void );
	INT32	dpUnInitMatch( void );
	INT32	setSecurity( UINT32 p_securityLevel );
	INT32	getSecurity( UINT32 *p_securityLevel );
	INT32	identifyTemplate( UINT8 **p_templatesArray, UINT32 *p_nbTemplates, UINT8 *p_templateBuffer, UINT32 *p_score );
	INT32	setDbMaxSearch( UINT32 p_maxSearch );

Necessary package : 

  	zMaths.h
	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	tools.h
	matcher.h

Comments : none

*/

#include <stdlib.h>

#include "zMaths.h"
#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "tools.h"
#include "tpsWarper.h"
#include "matcher.h"
#include "fxMatcher.h"
#include "matchRefine.h"
#include "quicksort.h"

#include "matchExport.h"

// global structures
extern BOOL8	g_initialized;

INT32 WINAPI dpInitMatcher( void )
{
	INT32	error;

	// create the mutex that will protect parms structure access
	CREATE_MUTEX( g_paramsMutex );
	// initialize ideniticaiton cancellation...
	CREATE_MUTEX( g_idCancelMutex );
	g_idCancelPointer = NULL;

	// initialize maths precalculated arrays
	error = zInitMaths();
	if( error != ZEF_OK )
		goto errorHandler;

	// mark initialisation flag to true
	g_initialized = TRUE;

	return ZEF_OK;

	//error Handling
errorHandler:

	// FREE memory 
	dpUnInitMatcher();

	return error;
}

INT32 WINAPI dpUnInitMatcher()
{
	INT32	error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// mark initialisation flag to false
	g_initialized = FALSE;

	return ZEF_OK;
	
errorHandler:

	return error;
}

INT32 WINAPI dpSetSecurity( UINT32 p_securityLevel )
{
	INT32	error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	LOCK_MUTEX( g_paramsMutex );
	if( p_securityLevel > g_params.m_maxScore )
	{
		UNLOCK_MUTEX( g_paramsMutex );
		error = ZEF_BAD_SECU_THRESHOLD;
		goto errorHandler;
	}
	else
	{
		g_params.m_defaultSecurity = p_securityLevel;
	}
	UNLOCK_MUTEX( g_paramsMutex );

	return ZEF_OK;

	// error handling
errorHandler:
	return error;
}

INT32 WINAPI dpGetSecurity( UINT32 *p_securityLevel )
{
	INT32	error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// test for NULL pointer in arguments
	if( p_securityLevel == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	LOCK_MUTEX( g_paramsMutex );
		*p_securityLevel = g_params.m_defaultSecurity;
	UNLOCK_MUTEX( g_paramsMutex );

	return ZEF_OK;

	// error handling
errorHandler:
	return error;
}

INT32 WINAPI dpSetDbMaxSearch( UINT32 p_maxSearch )
{
	INT32	error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	if( p_maxSearch >= 100 )
		return ZEF_BAD_MAX_SEARCH;
	else
	{
		LOCK_MUTEX( g_paramsMutex );
			g_params.m_maxDbSearch = p_maxSearch;
		UNLOCK_MUTEX( g_paramsMutex );
	}

	return ZEF_OK;

	// error handling
errorHandler:
	return error;
}

INT32 WINAPI dpStopIdentification( UINT8 *p_templateBuffer )
{
	INT32 error;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// test for NULL pointer in arguments
	if( p_templateBuffer == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// tell which identification must be aborted
	LOCK_MUTEX( g_idCancelMutex );
		g_idCancelPointer = p_templateBuffer;
	UNLOCK_MUTEX( g_idCancelMutex);

	return ZEF_OK;

errorHandler:
	return error;
}

INT32 WINAPI dpIdentifyTemplate( UINT8 **p_templatesArray, UINT32 p_nbTemplates, UINT8 *p_templateBuffer, 
								 UINT32 *p_score, UINT32 *p_templatePos )
{
	UINT32	*coreIndexes,*coreGlobDist,*noCoreIndexes,*noCoreGlobDist;
	UINT32	nbMaxItems;
	UINT32	i,index;
	INT32	error;
	UINT32	maxScore=0;
	INT32	indexOfMax=-1;
	UINT32	maxDbSearch,defaultSecurity;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxDbSearch = g_params.m_maxDbSearch;
		defaultSecurity = g_params.m_defaultSecurity;
	UNLOCK_MUTEX( g_paramsMutex );
	coreIndexes = NULL;
	noCoreIndexes = NULL;
	coreGlobDist = NULL;
	noCoreGlobDist = NULL;
	if( p_score != NULL )
		*p_score = 0;
	if( p_templatePos != NULL )
		*p_templatePos = p_nbTemplates;

	// test for NULL pointer in arguments
	// must be done after NULL pointers init
	if( p_templatesArray == NULL || p_templatePos == NULL || p_templateBuffer == NULL || p_score == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// extra test for template array
	for( i=0; i<(p_nbTemplates); i++ )
	{
		if( p_templatesArray[i] == NULL )
		{
			error = ZEF_NULL_POINTER;
			// set index to the faulty template position
			*p_templatePos = i;
			goto errorHandler;
		}
	}

	//check for zero nb of templates
	if( p_nbTemplates == 0 )
	{
		error = ZEF_BAD_SIZE;
		goto errorHandler;
	}

	// we allocate one more index for faulty template detection
	coreIndexes = MALLOC((p_nbTemplates+1)*sizeof(UINT32));
	if( coreIndexes == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	coreGlobDist = MALLOC((p_nbTemplates+1)*sizeof(UINT32));
	if( coreGlobDist == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	noCoreIndexes = MALLOC((p_nbTemplates+1)*sizeof(UINT32));
	if( noCoreIndexes == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	noCoreGlobDist = MALLOC((p_nbTemplates+1)*sizeof(UINT32));
	if( noCoreGlobDist == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	error = sortTemplates( p_templateBuffer, p_templatesArray, p_nbTemplates,
		coreIndexes, coreGlobDist, noCoreIndexes, noCoreGlobDist );
	if( error < ZEF_OK )
	{
		// set the index of the faulty template (if there is any)
		*p_templatePos = noCoreIndexes[p_nbTemplates];
		goto errorHandler;
	}

	if( coreIndexes[p_nbTemplates] < maxDbSearch*(p_nbTemplates)/100 )
		nbMaxItems = coreIndexes[p_nbTemplates];
	else
		nbMaxItems = maxDbSearch*(p_nbTemplates)/100;

	// first pass = fast
	for( i=0; i<nbMaxItems; i++ )
	{
		// test if identification must be aborted
		LOCK_MUTEX( g_idCancelMutex );
			if( p_templateBuffer == g_idCancelPointer )
			{
				g_idCancelPointer = NULL;
				error = ZEF_STOPPED;
				goto errorHandler;
			}
		UNLOCK_MUTEX( g_idCancelMutex);

		index = coreIndexes[i];
		// stop searching if global score is too high
		if( coreGlobDist[i] > 300 )  // <= this constant should be placed in a parameter in staticParameters.h & dynamicParameters.h
			break;
		error = fastVerifyTemplate( p_templateBuffer, p_templatesArray[index], p_score );
		if( error == ZEF_OK)
		{
			if( *p_score >= 2*defaultSecurity )
			{
				*p_templatePos = index;
				return ZEF_OK;
			}
			else
			{
				if( *p_score > maxScore )
				{
					maxScore = *p_score;
					indexOfMax = index;
				}
			}	
		}
		else if( error < ZEF_OK) // this test is useful.... error can be >0 !
		{
			// set the index of the faulty template
			*p_templatePos = index;
			goto errorHandler;
		}
	}

	
	if( indexOfMax != -1 )
	{
		error = dpVerifyTemplate( p_templateBuffer, p_templatesArray[indexOfMax], p_score );
		if( error < ZEF_OK )
		{
			// set the index of the faulty template
			*p_templatePos = (UINT32)indexOfMax;
			goto errorHandler;
		}
		if( *p_score >= 2*defaultSecurity )
		{
			*p_templatePos = indexOfMax;
			return ZEF_OK;
		}
	}

	// second pass = slower

	// sort nocore dists
	sortUINT32( noCoreGlobDist, noCoreIndexes, 0, p_nbTemplates-1 );

	nbMaxItems = maxDbSearch*(p_nbTemplates)/100;
	indexOfMax=-1;
	maxScore=0;
	for( i=0; i<nbMaxItems; i++ )
	{
		// test if identification must be aborted
		LOCK_MUTEX( g_idCancelMutex );
			if( p_templateBuffer == g_idCancelPointer )
			{
				g_idCancelPointer = NULL;
				error = ZEF_STOPPED;
				goto errorHandler;
			}
		UNLOCK_MUTEX( g_idCancelMutex);

		index = noCoreIndexes[i];
		error = dpVerifyTemplate( p_templateBuffer, p_templatesArray[index], p_score );
		if( error == ZEF_OK)
		{
			if( *p_score >= 2*defaultSecurity )
			{
				*p_templatePos = index;
				return ZEF_OK;
			}
			else
			{
				if( *p_score > maxScore )
				{
					maxScore = *p_score;
					indexOfMax = index;
				}
			}
			
		}
		else if( error < ZEF_OK)
		{
			// set the index of the faulty template
			*p_templatePos = index;
			goto errorHandler;
		}
	}

	if( indexOfMax != -1 )
	{
		*p_score = maxScore;
		*p_templatePos = indexOfMax;
		return ZEF_OK;
	}

	FREE( coreIndexes );
	FREE( coreGlobDist );
	FREE( noCoreIndexes );
	FREE( noCoreGlobDist );

	return ZEF_NO_MATCH;

	// error Handling
errorHandler:

	if( g_initialized )
	{
		if( coreIndexes != NULL )
			FREE( coreIndexes );
		if( coreGlobDist != NULL )
			FREE( coreGlobDist );
		if( noCoreIndexes != NULL )
			FREE( noCoreIndexes );
		if( noCoreGlobDist != NULL )
			FREE( noCoreGlobDist );
	}

	return error;
}

INT32	WINAPI dpVerifyAndRefineTemplate( UINT8 *p_templateBuffer1, UINT8 *p_templateBuffer2, UINT32 *p_score, BOOL8 *p_refined )
{
	INT32		error;
	Template	templ1,templ2,refinedTempl;
	INT32		nbMatches,i;
	INT32		**matches;
	INT32		matchArraySize;
	UINT32		maxNumberOfMinutiae,maxNumberOfCores,maxNumberOfDeltas;
	INT32		returnValue;
	UINT32		defaultSecurity;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}

	// init
	LOCK_MUTEX( g_paramsMutex );
		defaultSecurity = g_params.m_defaultSecurity;
	UNLOCK_MUTEX( g_paramsMutex );
	matches = NULL;

	// init template structures
	error = templateInit( &templ1 );	
	if( error < ZEF_OK )
		goto errorHandler;

	error = templateInit( &templ2 );
	if( error < ZEF_OK )
		goto errorHandler;

	// test for NULL pointer in arguments
	if( p_templateBuffer1 == NULL || p_templateBuffer2 == NULL || p_score == NULL || p_refined == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// unpack templates
	error = templateUnpacker( p_templateBuffer1, &templ1 );
	if( error < ZEF_OK )
		goto errorHandler;
	error = templateUnpacker( p_templateBuffer2, &templ2 );
	if( error < ZEF_OK )
		goto errorHandler;

	// init bis
	returnValue = ZEF_OK;
	matches = NULL;
	*p_refined = FALSE;
	maxNumberOfMinutiae = (templ1.m_nbMinutiae > templ2.m_nbMinutiae) ? templ1.m_nbMinutiae : templ2.m_nbMinutiae;
	maxNumberOfCores = (templ1.m_nbCores > templ2.m_nbCores) ? templ1.m_nbCores : templ2.m_nbCores;
	maxNumberOfDeltas = (templ1.m_nbDeltas > templ2.m_nbDeltas) ? templ1.m_nbDeltas : templ2.m_nbDeltas;

	// memory allocation
	matchArraySize = maxNumberOfCores+maxNumberOfDeltas+maxNumberOfMinutiae+1;
	matches = (INT32**)MALLOC( matchArraySize*sizeof(INT32*) );
	if( matches == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	for( i=0; i<matchArraySize; i++ )
	{
		matches[i] = (INT32*)MALLOC( 2*sizeof(INT32) );
		if( matches[i] == NULL )
		{
			error = ZEF_OUT_OF_MEMORY;
			goto errorHandler;
		}
	}
	
	// init score
	*p_score = 0;
	
	// match templates
	matchTemplate( &templ1, &templ2, matches, &nbMatches, p_score );

	// check if match is OK
	if( *p_score >= defaultSecurity )
	{
		// check if score is enough for refinement !
		if( *p_score >= 2*defaultSecurity )
		{
			// Init refined template
			error = templateInit( &refinedTempl );
			if( error < ZEF_OK )
				goto errorHandler;

			// create new template from templ1 + templ2
			if( templ1.m_blocSize == templ2.m_blocSize )
			{
				error = updateTemplate( &refinedTempl, &templ1, &templ2, matches, nbMatches );
				if( error < ZEF_OK )
					goto errorHandler;

				*p_refined = TRUE;

				// save updated template
				error = templatePacker( &refinedTempl, p_templateBuffer1 );
				if( error < ZEF_OK )
					goto errorHandler;

			}
			else
				*p_refined = FALSE;

		
			// FREE template structures
			error = templateUninit( &refinedTempl );
			if( error < ZEF_OK )
				goto errorHandler;

		}	
	}
	else
	{
		if( !RETURN_NON_MATCH_SCORE )
			*p_score = 0;

		returnValue = ZEF_NO_MATCH;
	}

	//FREE matches structure
	for( i=0; i<matchArraySize; i++ )
		FREE(matches[i]);
	FREE(matches);
	
	// FREE template structures
	error = templateUninit( &templ1 );
	if( error < ZEF_OK )
		goto errorHandler;
	error = templateUninit( &templ2 );
	if( error < ZEF_OK )
		goto errorHandler;
		
	return returnValue;

	// error Handling
errorHandler:

	if( g_initialized ) 
	{

		//FREE matches structure
		if( matches != NULL )
		{
			for( i=0; i<matchArraySize; i++ )
			{
				if( matches[i] != NULL )
					FREE(matches[i]);
			}
			FREE(matches);
		}	
		// FREE template structures
		templateUninit( &templ1 );
		templateUninit( &templ2 );

	}

	return error;
}

INT32	WINAPI dpVerifyTemplate( UINT8 *p_templateBuffer1, UINT8 *p_templateBuffer2, UINT32 *p_score )
{
	INT32		error;
	Template	templ1,templ2;
	INT32		nbMatches,i;
	INT32		**matches;
	INT32		matchArraySize;
	UINT32		maxNumberOfMinutiae,maxNumberOfCores,maxNumberOfDeltas;
	UINT32		defaultSecurity;

	//test bioEngine initialization
	if( !g_initialized ) 
	{
		error = ZEF_ENGINE_NOT_INITIALIZED;
		goto errorHandler;
	}
	
	// init
	LOCK_MUTEX( g_paramsMutex );
		defaultSecurity = g_params.m_defaultSecurity;
	UNLOCK_MUTEX( g_paramsMutex );
	matches = NULL;

	// init template structures
	error = templateInit( &templ1 );		
	if( error < ZEF_OK )
		goto errorHandler;	
	error = templateInit( &templ2 );		
	if( error < ZEF_OK )
		goto errorHandler;	

	// test for NULL pointer in arguments
	// must be done after matches initialized to NULL and templates initialized
	if( p_templateBuffer1 == NULL || p_templateBuffer2 == NULL || p_score == NULL )
	{
		error = ZEF_NULL_POINTER;
		goto errorHandler;
	}

	// unpack templates
	error = templateUnpacker( p_templateBuffer1, &templ1 );
	if( error < ZEF_OK )
		goto errorHandler;	
	error = templateUnpacker( p_templateBuffer2, &templ2 );
	if( error < ZEF_OK )
		goto errorHandler;
		
	// init score
	maxNumberOfMinutiae = (templ1.m_nbMinutiae > templ2.m_nbMinutiae) ? templ1.m_nbMinutiae : templ2.m_nbMinutiae;
	maxNumberOfCores = (templ1.m_nbCores > templ2.m_nbCores) ? templ1.m_nbCores : templ2.m_nbCores;
	maxNumberOfDeltas = (templ1.m_nbDeltas > templ2.m_nbDeltas) ? templ1.m_nbDeltas : templ2.m_nbDeltas;		

	// memory allocation
	matchArraySize = maxNumberOfCores+maxNumberOfDeltas+maxNumberOfMinutiae+1;
	matches = (INT32**)MALLOC( matchArraySize*sizeof(INT32*) );
	if( matches == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	for( i=0; i<matchArraySize; i++ )
	{
		matches[i] = (INT32*)MALLOC( 2*sizeof(INT32) );
		if( matches[i] == NULL )
		{
			error = ZEF_OUT_OF_MEMORY;
			goto errorHandler;
		}
	}
	
	// match templates
	matchTemplate( &templ1, &templ2, matches, &nbMatches, p_score );
	
	//FREE matches structure
	for( i=0; i<matchArraySize; i++ )
		FREE(matches[i]);
	FREE(matches);
	
	// FREE template structures
	error = templateUninit( &templ1 );
	if( error < ZEF_OK )
		goto errorHandler;	
	error = templateUninit( &templ2 );
	if( error < ZEF_OK )
		goto errorHandler;	
	
	// check if match is OK
	if( *p_score >= defaultSecurity )
		return ZEF_OK;
	else
	{
		if( !RETURN_NON_MATCH_SCORE )
			*p_score = 0;
		return ZEF_NO_MATCH;
	}
	
	// error handling
errorHandler:

	if( g_initialized ) 
	{

		//FREE matches structure
		if( matches != NULL )
		{
			for( i=0; i<matchArraySize; i++ )
			{
				if( matches[i] != NULL )
					FREE(matches[i]);
			}
			FREE(matches);
		}

		// FREE template structures
		templateUninit( &templ1 );
		templateUninit( &templ2 );

	}

	return error;
}
