/*
Company :		ZEFYR Technologies
File :			matchRefine.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	02/07/2003

Description : Template enhancement routines
Features : Makes a composite template from 2 templates
    
Functions : 

	updateTemplate : Build a composite template with two matching and aligned templates
	updateOri : build a composite bloc orentation map
	updateCores : keep the best cores for a composite template

Function prototypes : 

	INT32	updateTemplate( Template *refinedTemplate, Template *templ1, Template *templ2, INT32 **matches, INT32 nbMatches );
	INT32	updateOri( Template *p_refinedTemplate, Template *p_templ1, Template *p_templ2 );
	INT32	updateCores( Template *p_refinedTemplate, Template *p_templ1, Template *p_templ2, INT32 **p_matches );

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	zMaths.h
	tools.h
	tpsWarper.h
	quicksort.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>
#include <float.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "zMaths.h"
#include "tools.h"
#include "tpsWarper.h"
#include "quicksort.h"

#include "matchRefine.h"

INT32	updateTemplate( Template *p_refinedTemplate, Template *p_templ1, Template *p_templ2, INT32 **p_matches, INT32 p_nbMatches )
{
	INT32		i,j,x,y,ori,dOri;
	UINT32		nbMinutiae,maxNumberOfMinutiae,maxNumberOfCores,maxNumberOfDeltas;
	UINT32		*match1,*match2;
	INT32		xMin,xMax,yMin,yMax,error;
	Minutia		*minutiae;
	UINT32		maxNumberOfMinutiaeParam,maxNumberOfLocalInfos,maxNumberOfCoresParam,translationTolerance;

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfMinutiaeParam = g_params.m_maxNumberOfMinutiae;
		maxNumberOfLocalInfos = g_params.m_maxNumberOfLocalInfos;
		maxNumberOfCoresParam = g_params.m_maxNumberOfCores;
		translationTolerance = g_params.m_translationTolerance;
	UNLOCK_MUTEX( g_paramsMutex );
	nbMinutiae = 0;
	match1 = NULL;
	match2 = NULL;
	minutiae = NULL;
	maxNumberOfMinutiae = (p_templ1->m_nbMinutiae > p_templ2->m_nbMinutiae) ? p_templ1->m_nbMinutiae : p_templ2->m_nbMinutiae;
	maxNumberOfCores = (p_templ1->m_nbCores > p_templ2->m_nbCores) ? p_templ1->m_nbCores : p_templ2->m_nbCores;
	maxNumberOfDeltas = (p_templ1->m_nbDeltas > p_templ2->m_nbDeltas) ? p_templ1->m_nbDeltas : p_templ2->m_nbDeltas;

	// template memory allocations
	p_refinedTemplate->m_minutiae = (Minutia*)MALLOC( maxNumberOfMinutiaeParam*sizeof(Minutia));
	if( p_refinedTemplate->m_minutiae == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_refinedTemplate->m_ori = (UINT8*) MALLOC( maxNumberOfLocalInfos*sizeof(UINT8) );
	if( p_refinedTemplate->m_ori == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_refinedTemplate->m_qual = (UINT8*) MALLOC( maxNumberOfLocalInfos*sizeof(UINT8) );
	if( p_refinedTemplate->m_qual == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_refinedTemplate->m_cores = (Coord16*) MALLOC( maxNumberOfCoresParam*sizeof(Coord16) );
	if( p_refinedTemplate->m_cores == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_refinedTemplate->m_deltas = (Coord16*) MALLOC( maxNumberOfCoresParam*sizeof(Coord16) );
	if( p_refinedTemplate->m_deltas == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// allocate minutiae buffer
	minutiae = (Minutia*)MALLOC( (p_templ1->m_nbMinutiae+p_templ2->m_nbMinutiae)*sizeof(Minutia));
	if( minutiae == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}


	// fill template with matching minutiae
	for( i=maxNumberOfCores+maxNumberOfDeltas; i<p_nbMatches; i++ )
	{
		x = (p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_x+p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_x)/2;
		y = (p_templ1->m_minutiae[p_matches[i][0]].m_coord.m_y+p_templ2->m_minutiae[p_matches[i][1]].m_coord.m_y)/2;
		dOri = (p_templ2->m_minutiae[p_matches[i][1]].m_ori-p_templ1->m_minutiae[p_matches[i][0]].m_ori)*2;
		if( dOri > 256 ) dOri -= 512;
		if( dOri < -256 ) dOri += 512;
		ori = p_templ1->m_minutiae[p_matches[i][0]].m_ori + dOri;
		if( ori > 512 ) ori -= 512;
		if( ori < 0 ) ori += 512 ;
		minutiae[nbMinutiae].m_coord.m_x = x;
		minutiae[nbMinutiae].m_coord.m_y = y;
		minutiae[nbMinutiae].m_qual = 255;
		minutiae[nbMinutiae].m_ori = ori;
		nbMinutiae++;
	}

	// memory allocations for match tables
	match1 = (UINT32*)MALLOC(maxNumberOfMinutiae*sizeof(UINT32));
	if( match1 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	match2 = (UINT32*)MALLOC(maxNumberOfMinutiae*sizeof(UINT32));
	if( match2 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// init
	for( i=0; i<(INT32)maxNumberOfMinutiae; i++ )
	{
		match1[i] = match2[i] = 0;
	}
	
	// init overlapping area 
	xMin = (p_templ1->m_boundBox.m_left>p_templ2->m_boundBox.m_left)?p_templ1->m_boundBox.m_left:p_templ2->m_boundBox.m_left;
	yMin = (p_templ1->m_boundBox.m_top>p_templ2->m_boundBox.m_top)?p_templ1->m_boundBox.m_top:p_templ2->m_boundBox.m_top;
	xMax = (p_templ1->m_boundBox.m_right<p_templ2->m_boundBox.m_right)?p_templ1->m_boundBox.m_right:p_templ2->m_boundBox.m_right;
	yMax = (p_templ1->m_boundBox.m_bottom<p_templ2->m_boundBox.m_bottom)?p_templ1->m_boundBox.m_bottom:p_templ2->m_boundBox.m_bottom;
	
	// fill the match arrays
	for( i=maxNumberOfCores+maxNumberOfDeltas; i<p_nbMatches; i++ )
	{
		match1[p_matches[i][0]] = 1;
		match2[p_matches[i][1]] = 1;
	}

	// fill now with minutiae outside of overlaping area
	for( i=0; i<p_templ1->m_nbMinutiae; i++ )
	{
		if( !match1[i] )
		{
			x = p_templ1->m_minutiae[i].m_coord.m_x;
			y = p_templ1->m_minutiae[i].m_coord.m_y;

			minutiae[nbMinutiae].m_coord.m_x = x;
			minutiae[nbMinutiae].m_coord.m_y = y;
			minutiae[nbMinutiae].m_ori = p_templ1->m_minutiae[i].m_ori;
				
			// check if point is outside of overlapping area
			if( x<xMin || x>xMax || y<yMin || y>yMax )
				minutiae[nbMinutiae].m_qual = p_templ1->m_minutiae[i].m_qual;
			else
				minutiae[nbMinutiae].m_qual = p_templ1->m_minutiae[i].m_qual/2;

			nbMinutiae++;	
		}
	}
	for( i=0; i<p_templ2->m_nbMinutiae; i++ )
	{
		if( !match2[i] )
		{
			x = p_templ2->m_minutiae[i].m_coord.m_x;
			y = p_templ2->m_minutiae[i].m_coord.m_y;

			minutiae[nbMinutiae].m_coord.m_x = x;
			minutiae[nbMinutiae].m_coord.m_y = y;
			minutiae[nbMinutiae].m_ori = p_templ2->m_minutiae[i].m_ori;
					
			// check if point is outside of overlapping area
			if( x<xMin || x>xMax || y<yMin || y>yMax )		
				minutiae[nbMinutiae].m_qual = p_templ2->m_minutiae[i].m_qual;		
			else
				minutiae[nbMinutiae].m_qual = p_templ2->m_minutiae[i].m_qual/2;

			nbMinutiae++;	
		}
	}

	// now clean minutiae, remove minutiae that are to close !!!
	for( i=0; i<(INT32)nbMinutiae; i++ )
	{
		for( j=i+1; j<(INT32)nbMinutiae; j++ )
		{
			x = minutiae[i].m_coord.m_x-minutiae[j].m_coord.m_x;
			y = minutiae[i].m_coord.m_y-minutiae[j].m_coord.m_y;

			if( zSqrt(x*x+y*y) <= translationTolerance )
			{
				// set qual of minutiae with lowest qual to 0
				if( minutiae[i].m_qual < minutiae[j].m_qual )
					minutiae[i].m_qual = 0;
				else
					minutiae[j].m_qual = 0;
			}
		}
	}

	// copy minutiae to template
	p_refinedTemplate->m_nbMinutiae = 0;
	for( j=255; j>0; j-- )
	{
		for( i=0; i<(INT32)nbMinutiae; i++ )
		{
			if( minutiae[i].m_qual == j )
			{
				x = minutiae[i].m_coord.m_x;
				y = minutiae[i].m_coord.m_y;
				p_refinedTemplate->m_minutiae[p_refinedTemplate->m_nbMinutiae].m_coord.m_x = x;
				p_refinedTemplate->m_minutiae[p_refinedTemplate->m_nbMinutiae].m_coord.m_y = y;
				p_refinedTemplate->m_minutiae[p_refinedTemplate->m_nbMinutiae].m_ori = minutiae[i].m_ori;
				p_refinedTemplate->m_minutiae[p_refinedTemplate->m_nbMinutiae].m_qual = minutiae[i].m_qual;
				p_refinedTemplate->m_nbMinutiae++;

				// update bounding box
				if( x < p_refinedTemplate->m_boundBox.m_left ) p_refinedTemplate->m_boundBox.m_left = x;
				if( x > p_refinedTemplate->m_boundBox.m_right ) p_refinedTemplate->m_boundBox.m_right = x;
				if( y < p_refinedTemplate->m_boundBox.m_top ) p_refinedTemplate->m_boundBox.m_top = y;
				if( y > p_refinedTemplate->m_boundBox.m_bottom ) p_refinedTemplate->m_boundBox.m_bottom = y;

				if( p_refinedTemplate->m_nbMinutiae == maxNumberOfMinutiaeParam )
				{
					j=0;
					break;
				}
			}
		}
	}

	// update mean freq
	p_refinedTemplate->m_meanFreq = (p_templ1->m_meanFreq+p_templ2->m_meanFreq)/2;

	// update qual and ori maps
	error = updateOri( p_refinedTemplate, p_templ1, p_templ2 );
	if( error != ZEF_OK )
		goto errorHandler;
	error = updateCores( p_refinedTemplate, p_templ1, p_templ2, p_matches );
	if( error != ZEF_OK )
		goto errorHandler;
	
	FREE( minutiae );
	FREE( match1 );
	FREE( match2 );

	return ZEF_OK;

	// error Handling
errorHandler:

	if( minutiae != NULL )
		FREE( minutiae );
	if( match1 != NULL )
		FREE( match1 );
	if( match2 != NULL )
		FREE( match2 );

	// uninit composite template
	templateUninit(p_refinedTemplate);

	return error;

}

INT32 updateCores( Template *p_refinedTemplate, Template *p_templ1, Template *p_templ2, INT32 **p_matches )
{
	INT32		i,j,k,x,y,ori,dOri;
	INT32		decX=0,decY=0;
	BOOL8		*cores1,*deltas1,*cores2,*deltas2;
	BOOL8		coreOverflow,deltaOverflow;
	INT32		error;
	UINT32		maxNumberOfCores,maxNumberOfDeltas;
	UINT32		maxNumberOfCoresParam;

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfCoresParam = g_params.m_maxNumberOfCores;
	UNLOCK_MUTEX( g_paramsMutex );
	cores1 = NULL;
	deltas1 = NULL;
	cores2 = NULL;
	deltas2 = NULL;
	coreOverflow = FALSE;
	deltaOverflow = FALSE;
	maxNumberOfCores = (p_templ1->m_nbCores > p_templ2->m_nbCores) ? p_templ1->m_nbCores : p_templ2->m_nbCores;
	maxNumberOfDeltas = (p_templ1->m_nbDeltas > p_templ2->m_nbDeltas) ? p_templ1->m_nbDeltas : p_templ2->m_nbDeltas;

	// core and deltas used...
	cores1 = (BOOL8*)MALLOC(maxNumberOfCoresParam*sizeof(BOOL8));
	if( cores1 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	cores2 = (BOOL8*)MALLOC(maxNumberOfCoresParam*sizeof(BOOL8));
	if( cores2 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	deltas1 = (BOOL8*)MALLOC(maxNumberOfCoresParam*sizeof(BOOL8));
	if( deltas1 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	deltas2 = (BOOL8*)MALLOC(maxNumberOfCoresParam*sizeof(BOOL8));
	if( deltas2 == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	for( i=0; i<(INT32)maxNumberOfCoresParam; i++ )
	{
		cores1[i] = FALSE;
		cores2[i] = FALSE;
		deltas1[i] = FALSE;
		deltas2[i] = FALSE;
	}

	// if the main core of templ1 matches with main core of templ2
	// copy everything
	p_refinedTemplate->m_nbCores = 0;
	p_refinedTemplate->m_nbDeltas = 0;
	if( p_matches[0][0] == 0 && p_matches[0][1] == 0 )
	{
		// set used cores
		cores1[0] = TRUE;
		cores2[0] = TRUE;
		// update core datas
		x = (p_templ1->m_cores[p_matches[0][0]].m_x + p_templ2->m_cores[p_matches[0][1]].m_x)/2;
		y = (p_templ1->m_cores[p_matches[0][0]].m_y + p_templ2->m_cores[p_matches[0][1]].m_y)/2;
		p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_x = 0;
		p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_y = 0;
		p_refinedTemplate->m_nbCores++;
		if( p_refinedTemplate->m_nbCores == maxNumberOfCoresParam )
			coreOverflow = TRUE;
		dOri = (p_templ2->m_oriCore-p_templ1->m_oriCore);
		if( dOri > 127 ) dOri -= 255;
		if( dOri < -127 ) dOri += 255;
		ori = p_templ1->m_oriCore + dOri;
		if( ori > 255 ) ori -= 255;
		if( ori < 0 ) ori += 255;
		p_refinedTemplate->m_oriCore = ori;
		for( k=0; k<16; k++ )
			p_refinedTemplate->m_sign[k] = (p_templ1->m_sign[k] + p_templ2->m_sign[k] )/2;
	}
	else // we must find a main core !!!
	{
		if( p_templ1->m_nbCores > 0 )
		{
			// set used cores
			cores1[0] = TRUE;
			// update core datas
			p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_x = 0;
			p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_y = 0;
			ori = p_templ1->m_oriCore;
			p_refinedTemplate->m_oriCore = ori;
			p_refinedTemplate->m_nbCores++;	
			if( p_refinedTemplate->m_nbCores == maxNumberOfCoresParam )
				coreOverflow = TRUE;
			for( k=0; k<16; k++ )
				p_refinedTemplate->m_sign[k] = p_templ1->m_sign[k];
		}
		else if( p_templ2->m_nbCores > 0 )
		{
			// set used cores
			cores2[0] = TRUE;
			// update core datas
			decX = -p_templ2->m_cores[0].m_x;
			decY = -p_templ2->m_cores[0].m_y;
			p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_x = 0;
			p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_y = 0;
			ori = p_templ2->m_oriCore;
			p_refinedTemplate->m_oriCore = ori;
			p_refinedTemplate->m_nbCores++;	
			if( p_refinedTemplate->m_nbCores == maxNumberOfCoresParam )
				coreOverflow = TRUE;
			for( k=0; k<16; k++ )
				p_refinedTemplate->m_sign[k] = p_templ2->m_sign[k];

			// update minutiae positions
			for( i=0; i<p_refinedTemplate->m_nbMinutiae; i++ )
			{
				p_refinedTemplate->m_minutiae[i].m_coord.m_x += decX;
				p_refinedTemplate->m_minutiae[i].m_coord.m_y += decY;
			}
			p_refinedTemplate->m_localShiftX += decX;
			p_refinedTemplate->m_localShiftY += decY;
			p_refinedTemplate->m_boundBox.m_bottom += decY;
			p_refinedTemplate->m_boundBox.m_top += decY;
			p_refinedTemplate->m_boundBox.m_left += decX;
			p_refinedTemplate->m_boundBox.m_right += decX;

		}
		else // else we don't have any main core...
		{
			// just copy some data from templ1
			p_refinedTemplate->m_oriCore = 0;
			for( k=0; k<16; k++ )
				p_refinedTemplate->m_sign[k] = p_templ1->m_sign[k];
		}	
	}

	// add other matching cores
	for( i=0; i<p_templ1->m_nbCores && !coreOverflow; i++ )
	{
		for( j=0; j<(INT32)maxNumberOfCores; j++ )
		{
			if( !cores1[i] && p_matches[j][0] == i )
			{
				// set used cores
				cores1[i] = TRUE;
				cores2[p_matches[j][1]] = TRUE;
				// update core datas
				x = (p_templ1->m_cores[p_matches[j][0]].m_x + p_templ2->m_cores[p_matches[j][1]].m_x)/2;
				y = (p_templ1->m_cores[p_matches[j][0]].m_y + p_templ2->m_cores[p_matches[j][1]].m_y)/2;
				p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_x = x+decX;
				p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_y = y+decY;
				p_refinedTemplate->m_nbCores++;
				if( p_refinedTemplate->m_nbCores == maxNumberOfCoresParam )
					coreOverflow = TRUE;
				break;
			}
		}
	}
	// add matching deltas
	for( i=0; i<p_templ1->m_nbDeltas && !deltaOverflow; i++ )
	{
		for( j=maxNumberOfCores; j<(INT32)(maxNumberOfCores+maxNumberOfDeltas); j++ )
		{
			if( !deltas1[i] && p_matches[j][0] == i )
			{
				// set used cores
				deltas1[i] = TRUE;
				deltas2[p_matches[j][1]] = TRUE;
				// update core datas
				x = (p_templ1->m_deltas[p_matches[j][0]].m_x + p_templ2->m_deltas[p_matches[j][1]].m_x)/2;
				y = (p_templ1->m_deltas[p_matches[j][0]].m_y + p_templ2->m_deltas[p_matches[j][1]].m_y)/2;
				p_refinedTemplate->m_deltas[p_refinedTemplate->m_nbDeltas].m_x = x+decX;
				p_refinedTemplate->m_deltas[p_refinedTemplate->m_nbDeltas].m_y = y+decY;
				p_refinedTemplate->m_nbDeltas++;
				if( p_refinedTemplate->m_nbDeltas == maxNumberOfCoresParam )
					deltaOverflow = TRUE;
				break;
			}
		}
	}

	// process remaining cores and deltas
	for( i=0; i<p_templ1->m_nbCores && !coreOverflow; i++ )
	{
		if( !cores1[i] )
		{
			x = p_templ1->m_cores[i].m_x;
			y = p_templ1->m_cores[i].m_y;

			if( x < p_refinedTemplate->m_boundBox.m_left && x > p_refinedTemplate->m_boundBox.m_right &&
				y < p_refinedTemplate->m_boundBox.m_top && y > p_refinedTemplate->m_boundBox.m_bottom )
			{
				p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_x = x+decX;
				p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_y = y+decY;
				p_refinedTemplate->m_nbCores++;
				if( p_refinedTemplate->m_nbCores == maxNumberOfCoresParam )
					coreOverflow = TRUE;	
			}
		}
	}
	for( i=0; i<p_templ2->m_nbCores && !coreOverflow; i++ )
	{
		if( !cores2[i] )
		{
			x = p_templ2->m_cores[i].m_x;
			y = p_templ2->m_cores[i].m_y;

			if( x < p_refinedTemplate->m_boundBox.m_left && x > p_refinedTemplate->m_boundBox.m_right &&
				y < p_refinedTemplate->m_boundBox.m_top && y > p_refinedTemplate->m_boundBox.m_bottom )
			{
				p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_x = x+decX;
				p_refinedTemplate->m_cores[p_refinedTemplate->m_nbCores].m_y = y+decY;
				p_refinedTemplate->m_nbCores++;
				if( p_refinedTemplate->m_nbCores == maxNumberOfCoresParam )
					coreOverflow = TRUE;	
			}
		}
	}

	for( i=0; i<p_templ1->m_nbDeltas && !deltaOverflow; i++ )
	{
		if( !deltas1[i] )
		{
			x = p_templ1->m_deltas[i].m_x;
			y = p_templ1->m_deltas[i].m_y;

			if( x < p_refinedTemplate->m_boundBox.m_left && x > p_refinedTemplate->m_boundBox.m_right &&
				y < p_refinedTemplate->m_boundBox.m_top && y > p_refinedTemplate->m_boundBox.m_bottom )
			{
				p_refinedTemplate->m_deltas[p_refinedTemplate->m_nbDeltas].m_x = x+decX;
				p_refinedTemplate->m_deltas[p_refinedTemplate->m_nbDeltas].m_y = y+decY;
				p_refinedTemplate->m_nbDeltas++;
				if( p_refinedTemplate->m_nbDeltas == maxNumberOfCoresParam )
					deltaOverflow = TRUE;	
			}
		}
	}
	for( i=0; i<p_templ2->m_nbDeltas && !deltaOverflow; i++ )
	{
		if( !deltas2[i] )
		{
			x = p_templ2->m_deltas[i].m_x;
			y = p_templ2->m_deltas[i].m_y;

			if( x < p_refinedTemplate->m_boundBox.m_left && x > p_refinedTemplate->m_boundBox.m_right &&
				y < p_refinedTemplate->m_boundBox.m_top && y > p_refinedTemplate->m_boundBox.m_bottom )
			{
				p_refinedTemplate->m_deltas[p_refinedTemplate->m_nbDeltas].m_x = x+decX;
				p_refinedTemplate->m_deltas[p_refinedTemplate->m_nbDeltas].m_y = y+decY;
				p_refinedTemplate->m_nbDeltas++;
				if( p_refinedTemplate->m_nbDeltas == maxNumberOfCoresParam )
					deltaOverflow = TRUE;	
			}
		}
	}


	FREE( cores1 );
	FREE( cores2 );
	FREE( deltas1 );
	FREE( deltas2 );

	return ZEF_OK;

	// error Handling
errorHandler:

	if( cores1 != NULL )
		FREE( cores1 );
	if( cores2 != NULL )
		FREE( cores2 );
	if( deltas1 != NULL )
		FREE( deltas1 );
	if( deltas2 != NULL )
		FREE( deltas2 );

	return error;
}

INT32 updateOri( Template *p_refinedTemplate, Template *p_templ1, Template *p_templ2 )
{
	INT32		decX,decY,decJ,decI;
	INT32		i,j,y1,x1,y2,x2;
	INT32		qual1,qual2,ori1,ori2;
	BOOL8		bLength=TRUE;
	INT32		error;
	UINT32		maxNumberOfLocalInfos;

	// params init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfLocalInfos = g_params.m_maxNumberOfLocalInfos;
	UNLOCK_MUTEX( g_paramsMutex );

	// process displacement between the two maps
	decX = p_templ2->m_localShiftX-p_templ1->m_localShiftX;
	decY = p_templ2->m_localShiftY-p_templ1->m_localShiftY;
	
	decX = (decX/p_templ1->m_blocSize)+(decX%p_templ1->m_blocSize)/(p_templ1->m_blocSize/2);
	decY = (decY/p_templ1->m_blocSize)+(decY%p_templ1->m_blocSize)/(p_templ1->m_blocSize/2);
	
	decX -= ((INT32)p_templ2->m_localW-(INT32)p_templ1->m_localW)/2;
	decY -= ((INT32)p_templ2->m_localH-(INT32)p_templ1->m_localH)/2;

	// update some refined template data
	p_refinedTemplate->m_blocSize = p_templ1->m_blocSize;

	if( p_templ1->m_localW > p_templ2->m_localW+decX )
	{
		if( decX >= 0 )
			p_refinedTemplate->m_localW = p_templ1->m_localW;
		else
			p_refinedTemplate->m_localW = p_templ1->m_localW-decX;
	}
	else
		p_refinedTemplate->m_localW = p_templ2->m_localW+decX;


	if( p_templ1->m_localH > p_templ2->m_localH+decY )
	{
		if( decY >= 0 )
			p_refinedTemplate->m_localH = p_templ1->m_localH;
		else
			p_refinedTemplate->m_localH = p_templ1->m_localH-decY;
	}
	else
		p_refinedTemplate->m_localH = p_templ2->m_localH+decY;

	if( p_refinedTemplate->m_localH&1 )
		p_refinedTemplate->m_localH--;

	if( p_refinedTemplate->m_localW&1 )
		p_refinedTemplate->m_localW--;

	decJ = 0;
	decI = 0;
	while( p_refinedTemplate->m_localH*p_refinedTemplate->m_localW > (INT32)maxNumberOfLocalInfos )
	{
		if( bLength ) 
		{
			bLength = FALSE;
			p_refinedTemplate->m_localH -= 4;
			decJ+=2;

		}
		else
		{
			bLength = TRUE;
			p_refinedTemplate->m_localW -= 2;
			decI++;
		}
	}

	// init refined template data with template1 data

	p_refinedTemplate->m_localShiftX =  p_templ1->m_localShiftX;
	p_refinedTemplate->m_localShiftX -= (p_templ1->m_localW/2)*p_templ1->m_blocSize;
	if( decX < 0 )
		p_refinedTemplate->m_localShiftX += decX*p_templ1->m_blocSize;
	p_refinedTemplate->m_localShiftX += decI*p_templ1->m_blocSize;
	p_refinedTemplate->m_localShiftX += (p_refinedTemplate->m_localW/2)*p_templ1->m_blocSize;
		
	p_refinedTemplate->m_localShiftY =  p_templ1->m_localShiftY;
	p_refinedTemplate->m_localShiftY -= (p_templ1->m_localH/2)*p_templ1->m_blocSize;
	if( decY < 0 )
		p_refinedTemplate->m_localShiftY += decY*p_templ1->m_blocSize;
	p_refinedTemplate->m_localShiftY += decJ*p_templ1->m_blocSize;
	p_refinedTemplate->m_localShiftY += (p_refinedTemplate->m_localH/2)*p_templ1->m_blocSize;

	// allocate memory for qual and ori
	p_refinedTemplate->m_qual = (UINT8*)MALLOC(p_refinedTemplate->m_localH*p_refinedTemplate->m_localW*sizeof(UINT8));
	if( p_refinedTemplate->m_qual == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_refinedTemplate->m_ori = (UINT8*)MALLOC(p_refinedTemplate->m_localH*p_refinedTemplate->m_localW*sizeof(UINT8));
	if( p_refinedTemplate->m_ori == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	for( j=decJ; j<p_refinedTemplate->m_localH+decJ; j++ )
	{
		for( i=decI; i<p_refinedTemplate->m_localW+decI; i++ )
		{
			if( decX > 0 )
			{
				x2 = i-decX;
				x1 = i;
			}
			else
			{
				x1 = i+decX;
				x2 = i;
			}
			if( decY > 0 )
			{
				y2 = j-decY;
				y1 = j;
			}
			else
			{
				y1 = j+decY;
				y2 = j;
			}
			
			qual1=-1;qual2=-1;
			ori1=-1;ori2=-1;

			if( (UINT32)y1<p_templ1->m_localH && (UINT32)x1<p_templ1->m_localW )
			{
				qual1 = p_templ1->m_qual[y1*p_templ1->m_localW+x1];
				ori1 = p_templ1->m_ori[y1*p_templ1->m_localW+x1];
			}

			if( (UINT32)y2<p_templ2->m_localH && (UINT32)x2<p_templ2->m_localW )
			{
				qual2 = p_templ2->m_qual[y2*p_templ2->m_localW+x2];
				ori2 = p_templ2->m_ori[y2*p_templ2->m_localW+x2];
			}

			if( qual1 != -1 || qual2 != -1 )
			{
				if( (j-decJ)*(INT32)p_refinedTemplate->m_localW+(i-decI)< 0 || (j-decJ)*(INT32)p_refinedTemplate->m_localW+(i-decI) > p_refinedTemplate->m_localH*p_refinedTemplate->m_localW )
					i = 0;
				if( qual1 > qual2 )
				{
					p_refinedTemplate->m_ori[(j-decJ)*p_refinedTemplate->m_localW+(i-decI)] = ori1;
					p_refinedTemplate->m_qual[(j-decJ)*p_refinedTemplate->m_localW+(i-decI)] = 16*qual1;
				}
				else
				{
					p_refinedTemplate->m_ori[(j-decJ)*p_refinedTemplate->m_localW+(i-decI)] = ori2;
					p_refinedTemplate->m_qual[(j-decJ)*p_refinedTemplate->m_localW+(i-decI)] = 16*qual2;
				}

			}
			else // no data in templ1 or templ2 
			{
					p_refinedTemplate->m_ori[(j-decJ)*p_refinedTemplate->m_localW+(i-decI)] = 0;
					p_refinedTemplate->m_qual[(j-decJ)*p_refinedTemplate->m_localW+(i-decI)] = 0;
			}
		}
	}
	
	return ZEF_OK;

	// error handling
errorHandler:

	// FREE composite template
	templateUninit( p_refinedTemplate );

	return error;
}

