/*
Company :		ZEFYR Technologies
File :			minutiae.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	13/06/2003

Description : Minutiae pre-processing, finding and post-processing

Features : differents fingerprint specific enhancements
    
Functions :

	extractMinutiae : main function used for minutiae extraction
	detectMinutiae : find minutiae in a skeletonize binary image
	followLine : internal function used to find minutiae neighboors
	writeMinutiae : write minutiae into a template
	cleanMinutiae : remove fake minutiae
	checkMinutia : internal function used when cleaning minutiae, check if a minutiae have to be erased
	reconnectSkeleton : reconnect some minutiae that are in front of each other and with the same orientation
	detectMinutiaeOri : determines each minutia orientation

Function prototypes : 

	INT32 extractMinutiae( BioEngineData *p_data );
	INT32 detectMinutiae( UINT8 *p_img, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 *p_nbMinutiae, UINT32 p_searchSize );
	UINT32 followLine( UINT8 *p_img, UINT32 p_width, UINT32 p_height, UINT32 p_x, UINT32 p_y, UINT32 *p_retX, UINT32 *p_retY, UINT32 p_searchSize, UINT32 p_oldDir );
	INT32 writeMinutiae( UINT8 *p_ori,UINT8 *p_qual, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 p_nbMinutiae, Template *p_template );
	INT32 cleanMinutiae( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 *p_nbMinutiae, UINT32 p_size );
	BOOL8 checkMinutia( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 p_index, UINT32 p_size, UINT32 *oldMinutiae, UINT32 *oldCount );
	INT32 reconnectSkeleton( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae, UINT32 *p_nbMinutiae, UINT32 p_reconSize );
	INT32 detectMinutiaeOri( UINT8 *p_img, UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae,
						UINT32 p_nbMinutiae, UINT32 p_searchSize )

Necessary package : 

	customTypes.h
	staticParameters.h
	dynamicParameters.h
	errorCodes.h
	skeleton.h
	zMaths.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>

#include "customTypes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "errorCodes.h"
#include "skeleton.h"
#include "zMaths.h"
#include "BMPfiles.h"

#include "minutiae.h"

// arrays for line following
INT8 g_dirX[8][8] = {
	{  0,  1, -1,  1, -1,  1, -1,  0 },
	{  1,  1,  0,  1, -1,  0, -1, -1 },
	{ -1,  0, -1,  1, -1,  1,  0,  1 }, 
	{  1,  1,  1,  0,  0, -1, -1, -1 },
	{ -1, -1, -1,  0,  0,  1,  1,  1 },
	{  1,  0,  1, -1,  1, -1,  0, -1 },
	{ -1, -1,  0, -1,  1,  0,  1,  1 },
	{  0, -1,  1, -1,  1, -1,  1,  0 },	
};
 
INT8 g_dirY[8][8] = {
	{ -1, -1, -1,  0,  0,  1,  1,  1 },
	{ -1,  0, -1,  1, -1,  1,  0,  1 },
	{ -1, -1,  0, -1,  1,  0,  1,  1 },
	{  0,  1, -1,  1, -1,  1, -1,  0 },
	{  0, -1,  1, -1,  1, -1,  1,  0 },
	{  1,  1,  0,  1, -1,  0, -1, -1 },
	{  1,  0,  1, -1,  1, -1,  0, -1 },
	{  1,  1,  1,  0,  0, -1, -1, -1 },
};

INT32 extractMinutiae( BioEngineData *p_data )
{
	INT32				error;
	MinutiaNeighborhood	*minutiae;
	UINT32				nbMinutiae;
#ifdef TIME_REPORT
	clock_t timeReport;
#endif
	UINT32 maxNumberOfDetectedMinutiae,minutiaeNeiborhoodSize,branchMinSize,maxReconnectSize;

	//for debug mode
	#ifdef DP_DEBUG
		UINT32	xCore,yCore;
		xCore = p_data->m_template.m_cores[0].m_x;
		yCore = p_data->m_template.m_cores[0].m_y;
	#endif


	// init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfDetectedMinutiae = g_params.m_maxNumberOfDetectedMinutiae;
		minutiaeNeiborhoodSize = g_params.m_minutiaeNeiborhoodSize;
		branchMinSize = g_params.m_branchMinSize;
		maxReconnectSize = g_params.m_maxReconnectSize;
	UNLOCK_MUTEX( g_paramsMutex );
	minutiae = NULL;

	// memory allocation
	minutiae = (MinutiaNeighborhood*) MALLOC( maxNumberOfDetectedMinutiae*sizeof(MinutiaNeighborhood));
	if( minutiae == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// first : skeletonization of the fingerprint image
	error = skeletonization( p_data->m_workImg, p_data->m_workImg, p_data->m_width, p_data->m_height );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "extractMinutiae - skeletonization" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		saveBmp( "Extract-01skeleton.bmp", p_data->m_workImg, p_data->m_width, p_data->m_height );
	#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// search minutiae on the skeleton
	error = detectMinutiae( p_data->m_workImg, p_data->m_quality, p_data->m_width, p_data->m_height, minutiae, &nbMinutiae, minutiaeNeiborhoodSize );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "extractMinutiae - detectMinutiae" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// get minutiae orientation
	error = detectMinutiaeOri( p_data->m_workImg, p_data->m_ori, p_data->m_width, p_data->m_height, minutiae, nbMinutiae, 8);
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "extractMinutiae - detectMinutiaeOri" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// clean minutiae
	error = cleanMinutiae( p_data->m_ori, p_data->m_width, p_data->m_height, minutiae, &nbMinutiae, branchMinSize );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "extractMinutiae - cleanMinutiae" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// reconnect skeleton
	error = reconnectSkeleton( p_data->m_ori, p_data->m_width, p_data->m_height, minutiae, &nbMinutiae, maxReconnectSize );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "extractMinutiae - reconnectSkeleton" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// write minutiae into template
	error = writeMinutiae( p_data->m_quality, p_data->m_width, p_data->m_height, minutiae, nbMinutiae, &(p_data->m_template) );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "extractMinutiae - writeMinutiae" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		displayMinutiae( p_data->m_workImg, p_data->m_workImg, p_data->m_quality, p_data->m_width, p_data->m_height, &(p_data->m_template), xCore, yCore);
		saveBmp("Extract-02minutiae.bmp",p_data->m_workImg, p_data->m_width, p_data->m_height );
	#endif

	// FREE minutiae
	FREE(minutiae);

	return ZEF_OK;

	// error handling
errorHandler:

	// FREE minutiae
	if( minutiae != NULL )
		FREE(minutiae);

	return error;
}

INT32 detectMinutiae( UINT8 *p_img, UINT8 *p_qual, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae,
					UINT32 *p_nbMinutiae, UINT32 p_searchSize )
{
	UINT32	x,y,a,i,j,u,v;
	UINT32	miniWidth,decX,decY;
	UINT32	offset,offset2;
	UINT32	nbNeighbors, nbTransitions, nbMinutiae, minutiaeCpt;
	UINT32	neighbX,neighbY;
	BOOL8	minutiaeOverflow;
	UINT32	maxNumberOfDetectedMinutiae;

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfDetectedMinutiae = g_params.m_maxNumberOfDetectedMinutiae;
	UNLOCK_MUTEX( g_paramsMutex );
	miniWidth = p_width/FUSION_BLOC_SIZE;
	decX = IMAGE_BORDER_LIMIT/FUSION_BLOC_SIZE;
	decY = decX *  miniWidth;
	minutiaeOverflow = FALSE;
	nbMinutiae = 0;

	// search for ridge ending and ridge bifurcations
	for( y=FFT_BLOC_SIZE/4; y<p_height-FFT_BLOC_SIZE/4; y++ )
	{
		for( x=FFT_BLOC_SIZE/4; x<p_width-FFT_BLOC_SIZE/4; x++ )
		{
			offset = y*p_width+x;
			offset2 = (y/FUSION_BLOC_SIZE)*miniWidth+(x/FUSION_BLOC_SIZE);
			// if point is part of the skeleton and quality is good enough
			if( p_img[offset] == 0 && p_qual[offset2] )
			{
				// analyse minutiae neighborhood
				analyzeNeighborhood( &(p_img[offset]), p_width, &nbNeighbors, &nbTransitions );

				// nbNeighbors == 1 : ridge ending
				// nbNeighbors == 3 && nbTransitions == 3 : ridge bifurcation
				if( (nbNeighbors == 1) || (nbNeighbors == 3 && nbTransitions == 3 ) )
				{
					// check if minutia is not to close from image borders
					if( p_qual[offset2+decX] && p_qual[offset2-decX] && p_qual[offset2+decY] && p_qual[offset2-decY] &&
						p_qual[offset2+decX+decY] && p_qual[offset2-decX-decY] && p_qual[offset2+decX-decY] && p_qual[offset2-decX+decY] )
					{
						// if we find to much minutiae, go to next step...
						if( nbMinutiae == maxNumberOfDetectedMinutiae )
						{
							minutiaeOverflow = TRUE;
							x=p_width;
							y=p_height;
							break;
						}
						
						// write minutiae information in the array
						p_minutiae[nbMinutiae].m_coord.m_x = x;
						p_minutiae[nbMinutiae].m_coord.m_y = y;
						p_minutiae[nbMinutiae].m_typeMultiple = (nbNeighbors == 3)? TRUE:FALSE;
						p_minutiae[nbMinutiae].m_nbMinutiae = 1;
						p_minutiae[nbMinutiae].m_maxNbMinutiae = 1;
						p_minutiae[nbMinutiae].m_neighbors[0] = -1;
						nbMinutiae++;
					}
				}
			}
		}
	}

	// then find relatioship between minutiae
	for( i=0; i<nbMinutiae; i++ )
	{
		// inits
		x = p_minutiae[i].m_coord.m_x;
		y = p_minutiae[i].m_coord.m_y;

		minutiaeCpt = 1;

		// find minutia neighbors
		// mark point a treated
		p_img[y*p_width+x] = 1;
		for( a=0; a<8; a++ )
		{
			u = g_dirX[0][a];
			v = g_dirY[0][a];

			if( p_img[(y+v)*p_width+(x+u)] == 0 )
			{
				nbNeighbors = followLine( p_img, p_width, p_height,x+u, y+v, &neighbX, &neighbY, p_searchSize, a);
				if( nbNeighbors )
				{
					// find neighbor index
					for( j=0; j<nbMinutiae; j++ )
					{
						if( p_minutiae[j].m_coord.m_x == (INT32)neighbX && p_minutiae[j].m_coord.m_y == (INT32)neighbY )
						{	
							p_minutiae[i].m_neighbors[minutiaeCpt++] = j;
							p_minutiae[i].m_nbMinutiae = minutiaeCpt;
							p_minutiae[i].m_maxNbMinutiae= minutiaeCpt;
							break;
						}
					}
				}
			}
		}
		// put point to previous value
		p_img[y*p_width+x] = 0;
	}

	*p_nbMinutiae = nbMinutiae;
	
	if( !minutiaeOverflow )
		return ZEF_OK;
	else
		return ZEF_TOO_MANY_MINUTIAE;
}



// warning : this function doesn't return an error code
// the return value is the number of neighboring minutiae
UINT32 followLine( UINT8 *p_img, UINT32 p_width, UINT32 p_height, UINT32 p_x, UINT32 p_y, 
				  UINT32 *p_retX, UINT32 *p_retY, UINT32 p_searchSize, UINT32 p_oldDir )
{
	UINT32	nbNeighbors, nbTransitions;
	UINT32	u,v,i;
	UINT32	val;

	// check image boundaries
	if( p_x == 0 || p_x == p_width-1 || p_y == 0 || p_y == p_height-1 )
		return 0;

	// set point coordinates
	*p_retX = p_x;
	*p_retY = p_y;

	// check if this point is a minutia
	analyzeNeighborhood( &(p_img[p_y*p_width+p_x]), p_width, &nbNeighbors, &nbTransitions );
	if( nbNeighbors == 1 || ( nbNeighbors == 3 && nbTransitions == 3 ) )
	{
		return nbNeighbors;
	}

	// check end of recursion condition
	if( p_searchSize == 0 )
		return 0;

	// else, recurse to find nearest minutia point
	// mark point a treated
	p_img[p_y*p_width+p_x] = 1;
	for( i=0; i<8; i++ )
	{
		u = g_dirX[p_oldDir][i];
		v = g_dirY[p_oldDir][i];

		if( p_img[(p_y+v)*p_width+(p_x+u)] == 0 )
		{
			val = followLine( p_img, p_width, p_height,p_x+u, p_y+v, p_retX, p_retY, p_searchSize-1, i);
			// put point to previous value
			p_img[p_y*p_width+p_x] = 0;
			return val;
		}
	}
	// put point to previous value
	p_img[p_y*p_width+p_x] = 0;

	return 0; // should never happen
}

INT32 detectMinutiaeOri( UINT8 *p_img, UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae,
					UINT32 p_nbMinutiae, UINT32 p_searchSize )
{
	UINT32	x,y,a,i,j,u,v;
	UINT32	cpt;
	UINT32	neighbX,neighbY;
	INT16	ori[8],tempOri,minOri,minIndex;

	// for all minutiae in our minutiae array
	for( i=0; i<p_nbMinutiae; i++ )
	{
		// inits
		x = p_minutiae[i].m_coord.m_x;
		y = p_minutiae[i].m_coord.m_y;

		cpt = 0;

		// find minutia neighbors
		// mark point a treated
		p_img[y*p_width+x] = 1;
		for( a=0; a<8; a++ )
		{
			u = g_dirX[0][a];
			v = g_dirY[0][a];

			if( p_img[(y+v)*p_width+(x+u)] == 0 )
			{
				followLine( p_img, p_width, p_height,x+u, y+v, &neighbX, &neighbY, p_searchSize, a);
				
				// process ori between minutiae and point returned
				ori[cpt] = zAtan(neighbY-y,neighbX-x);
				cpt++;
			}
		}

		minIndex = -1;
		minOri = 512;

		if( cpt == 1 )
		{
			minOri = ori[0];
		}
		else if( cpt == 3 )
		{
			// search for the min angle between two lines
			for( j=0; j<3; j++ )
			{
				tempOri=zAbs(ori[j]-ori[(j+1)%3]);
				if( tempOri > 256 ) tempOri = 512-tempOri; 
				if( tempOri < minOri )
				{
					minOri = tempOri;
					minIndex = j;
				}
			}

			// once the min ori is found : the minutiae orientation is in the middle of this angle
			if( ori[minIndex] < ori[(minIndex+1)%3] )
				minOri = ori[minIndex]+minOri/2;
			else
				minOri = ori[(minIndex+1)%3]+minOri/2;
		}
		/*else // debug
			printf("error in minutiae ori detection... \n");*/

		tempOri = p_ori[x+y*p_width];
		tempOri = zAbs(tempOri-minOri);
		if( tempOri > 256 ) tempOri = 512-tempOri; 
		if( tempOri > 128 )
			p_minutiae[i].m_ori =  p_ori[x+y*p_width]+256;
		else
			p_minutiae[i].m_ori =  p_ori[x+y*p_width];
	
		// put point to previous value
		p_img[y*p_width+x] = 0;
	}

	return ZEF_OK;
}
 

INT32 writeMinutiae( UINT8 *p_qual, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae,
					UINT32 p_nbMinutiae, Template *p_template )
{
	UINT32	i;
	UINT32	acc,miniWidth;
	INT32	xC,yC,x,y,q,xScaled,yScaled;
	BOOL8	minutiaeOverflow;
	INT32	error;
	UINT32	maxNumberOfMinutiae;
	INT32	globalTemplateScaling;

	// init
	LOCK_MUTEX( g_paramsMutex );
		maxNumberOfMinutiae = g_params.m_maxNumberOfMinutiae;
		globalTemplateScaling = g_params.m_globalTemplateScaling;
	UNLOCK_MUTEX( g_paramsMutex );
	minutiaeOverflow = FALSE;
	acc=0;
	miniWidth = p_width / FUSION_BLOC_SIZE;

	xC = p_template->m_cores[0].m_x;
	yC = p_template->m_cores[0].m_y;

	p_template->m_boundBox.m_left = SHRT_MAX;
	p_template->m_boundBox.m_right = SHRT_MIN;
	p_template->m_boundBox.m_top = SHRT_MAX;
	p_template->m_boundBox.m_bottom = SHRT_MIN;

	// minutiae memory allocation
	p_template->m_minutiae = (Minutia*)MALLOC(maxNumberOfMinutiae*sizeof(Minutia) );
	if( p_template->m_minutiae == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// process minutiae quality
	for( i=0; i<p_nbMinutiae; i++ )
	{
		x = p_minutiae[i].m_coord.m_x;
		y = p_minutiae[i].m_coord.m_y;
	
		if( p_minutiae[i].m_nbMinutiae > 0 )
			p_minutiae[i].m_qual = p_qual[(y/FUSION_BLOC_SIZE)*miniWidth+(x/FUSION_BLOC_SIZE)];
		else
			p_minutiae[i].m_qual = 0;
	}

	// browse minutiae sorted by decreasing quality
	for( q=255; q>0; q-- )
	{
		for( i=0; i<p_nbMinutiae; i++ )
		{
			x = p_minutiae[i].m_coord.m_x;
			y = p_minutiae[i].m_coord.m_y;
	
			// if minutia is good, write it to template
			if( p_minutiae[i].m_qual == q )
			{
				// check if max nb of minutiae is found
				if( acc == maxNumberOfMinutiae )
				{
					minutiaeOverflow = TRUE;
					q=0;
					i=p_nbMinutiae;
					break;
				}

				// scale minutiae coordinates
				xScaled = (x-xC)*globalTemplateScaling/256;
				yScaled = (y-yC)*globalTemplateScaling/256;

				// write minutiae coordinates
				p_template->m_minutiae[acc].m_coord.m_x = xScaled;
				p_template->m_minutiae[acc].m_coord.m_y = yScaled;

				// update bounding box
				if( xScaled < p_template->m_boundBox.m_left ) p_template->m_boundBox.m_left = xScaled;
				if( xScaled > p_template->m_boundBox.m_right ) p_template->m_boundBox.m_right = xScaled;
				if( yScaled < p_template->m_boundBox.m_top ) p_template->m_boundBox.m_top = yScaled;
				if( yScaled > p_template->m_boundBox.m_bottom ) p_template->m_boundBox.m_bottom = yScaled;

				// write ori
				p_template->m_minutiae[acc].m_ori = p_minutiae[i].m_ori/2;

				// write qual
				p_template->m_minutiae[acc].m_qual = p_qual[(y/FUSION_BLOC_SIZE)*miniWidth+(x/FUSION_BLOC_SIZE)];

				acc++;			
			}
		}
	}

	// update cores and deltas
	for( i=0; i<p_template->m_nbCores; i++ )
	{
		p_template->m_cores[i].m_x = globalTemplateScaling*(p_template->m_cores[i].m_x-xC)/256;
		p_template->m_cores[i].m_y = globalTemplateScaling*(p_template->m_cores[i].m_y-yC)/256;
	}
	for( i=0; i<p_template->m_nbDeltas; i++ )
	{
		p_template->m_deltas[i].m_x = globalTemplateScaling*(p_template->m_deltas[i].m_x-xC)/256;
		p_template->m_deltas[i].m_y = globalTemplateScaling*(p_template->m_deltas[i].m_y-yC)/256;
	}

	p_template->m_nbMinutiae = acc;

	if( !minutiaeOverflow )
		return ZEF_OK;
	else
		return ZEF_TOO_MANY_MINUTIAE;

	// error handling
errorHandler:

	if( p_template->m_minutiae != NULL )
	{
		FREE( p_template->m_minutiae );
		p_template->m_minutiae = NULL;
	}

	return error;
}

INT32 cleanMinutiae( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae,
					 UINT32 *p_nbMinutiae, UINT32 p_size )
{
	UINT32	i;
	UINT32	oldMinutiae[MAX_NB_OLD_MINUTIAE];
	UINT32	oldCount;

	// check each minutia
	for( i=0; i<*p_nbMinutiae; i++ )
	{
		oldCount = 0;
		checkMinutia( p_ori, p_width, p_height, p_minutiae, i, p_size, oldMinutiae, &oldCount );
	}

	return ZEF_OK;
}

BOOL8 checkMinutia( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae,
				    UINT32 p_index, UINT32 p_size, UINT32 *p_oldMinutiae, UINT32 *p_oldCount )
{
	UINT32	a,acc;
	INT32	i,x,y,u,v;
	INT32	index;
	INT32	ori0,ori1,ori2,angle,angle2;
	UINT32	maxSupprAngle;

	// init 
	LOCK_MUTEX( g_paramsMutex );
		maxSupprAngle = g_params.m_maxSupprAngle;
	UNLOCK_MUTEX( g_paramsMutex );
	acc=0;

	//condition to avoid infinite loops
	if( (*p_oldCount) == MAX_NB_OLD_MINUTIAE )
		return TRUE;

	// mark that we have treated this minutia
	p_oldMinutiae[(*p_oldCount)] = p_index;
	(*p_oldCount)++;

	// if there is more than one minutia in the neighborhood
	// check if we can remove it...
	if( p_minutiae[p_index].m_nbMinutiae > 1 )
	{
		x = p_minutiae[p_index].m_coord.m_x;
		y = p_minutiae[p_index].m_coord.m_y;

		// loop on all the neighbors
		for( i=1; i<p_minutiae[p_index].m_maxNbMinutiae; i++ )
		{
			// check if this minutiae has not already been treated (to avoid infinite loops)
			index = p_minutiae[p_index].m_neighbors[i];
			for( a=0; a<(*p_oldCount); a++ )
			{
				if( index == (INT32)p_oldMinutiae[a] )
				{
					index = -1;
					acc++;
					break;
				}
			}
			// if the minutia has not already been treated
			if( index != -1 )
			{
				// check it !!!
				if( !checkMinutia( p_ori, p_width, p_height, p_minutiae, index, p_size, p_oldMinutiae, p_oldCount) );
				{
					// It can be suppressed, we have to do some verification on it's parent minutia
					
					// if minutiae is a bifurcation
					if( p_minutiae[index].m_typeMultiple == TRUE )
					{
						u = p_minutiae[index].m_coord.m_x;
						v = p_minutiae[index].m_coord.m_y;

						// we check the diference of angles between the two minutiae

						ori0 = zAtan(v-y, u-x );
						ori1 = p_ori[y*p_width+x];
						ori2 = p_ori[v*p_width+u];
						angle = ori2-ori0;
						angle2 = ori1-ori0;
						if( angle <= -128 ) angle += 256;
						if( angle >= 128 ) angle -= 256;
						if( angle2 <= -128 ) angle2 += 256;
						if( angle2 >= 128 ) angle2 -= 256;

						angle = zAbs(angle);
						angle2 = zAbs(angle);

						// if the difference of orientations between the two minutiae is too high
						// we remove both of the minutiae
						if( angle  >= (INT32)maxSupprAngle && angle2 >= (INT32)maxSupprAngle )
						{
							p_minutiae[index].m_nbMinutiae = 0;
							p_minutiae[p_index].m_neighbors[i] = -1;
							p_minutiae[p_index].m_nbMinutiae = 0;
						}
						else
						{
							(*p_oldCount)--;
							return FALSE; // we don't remove anything
						}
					}
					else
					{
						// ridge ending
						p_minutiae[index].m_nbMinutiae = 0;
						p_minutiae[p_index].m_neighbors[i] = -1;
						p_minutiae[p_index].m_nbMinutiae = 0;
					}
				}
			}
		}
		if( p_minutiae[p_index].m_nbMinutiae-acc <= 1 )
		{
			(*p_oldCount)--;
			return FALSE;
		}
	}
	else
	{
		(*p_oldCount)--;
		return FALSE;
	}

	(*p_oldCount)--;
	return TRUE; // return TRUE if we keep the minutiae
}

INT32 reconnectSkeleton( UINT8 *p_ori, UINT32 p_width, UINT32 p_height, MinutiaNeighborhood *p_minutiae,
						UINT32 *p_nbMinutiae, UINT32 p_reconSize )
{
	INT32	u,v,x,y,threshold;
	UINT32	i,j,jMin;
	INT32	ori0,ori1,ori2,angle,angle2;
	INT32	minAngle,minAngle2;
	UINT32	minReconnectionAngle;

	// init 
	LOCK_MUTEX( g_paramsMutex );
		minReconnectionAngle = g_params.m_minReconnectionAngle;
	UNLOCK_MUTEX( g_paramsMutex );
	threshold = p_reconSize*p_reconSize;

	// loop on all minutiae
	for(i=0; i<*p_nbMinutiae; i++ )
	{	
		// for ridge endings
		if( p_minutiae[i].m_typeMultiple == FALSE )
		{
			x = p_minutiae[i].m_coord.m_x;
			y = p_minutiae[i].m_coord.m_y;
			minAngle = INT_MAX;
			minAngle2 = INT_MAX;
			// we are looking for another minutia, close to the first one
			for( j=0; j<*p_nbMinutiae; j++ )
			{
				// for ridge endings
				if( p_minutiae[j].m_typeMultiple == FALSE )
				{
					u = p_minutiae[j].m_coord.m_x;
					v = p_minutiae[j].m_coord.m_y;

					// if the minutiae are close enough
					if( (j != i) && ( (u-x)*(u-x) + (v-y)*(v-y) ) <= threshold )
					{
						ori0 =zAtan(v-y, u-x );
						ori1 = p_ori[y*p_width+x];
						ori2 = p_ori[v*p_width+u];
						
						angle = ori2-ori0;
						angle2 = ori1-ori0;
						if( angle <= -128 ) angle += 256;
						if( angle >= 128 ) angle -= 256;
						if( angle2 <= -128 ) angle2 += 256;
						if( angle2 >= 128 ) angle2 -= 256;

						angle = zAbs(angle);
						angle2 = zAbs(angle);

						// if the ori difference is not too high, we bridge them
						if( angle < minAngle && angle2 <= minAngle2 )
						{
							minAngle = angle;
							minAngle2 = angle2;
							jMin = j;
						}
					}
				}
			}
			// if we have to bridge, we bridge
			if( minAngle <= (INT32)minReconnectionAngle && minAngle2 <= (INT32)minReconnectionAngle )
			{
				u = p_minutiae[jMin].m_coord.m_x;
				v = p_minutiae[jMin].m_coord.m_y;
				p_minutiae[i].m_nbMinutiae = 0;
				p_minutiae[jMin].m_nbMinutiae = 0;
			}
		}
	}

	return ZEF_OK;
}

