/*
Company :		ZEFYR Technologies
File :			preProcessings.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	09/05/2003

Description : fingerprint pre-Processings

Features : differents fingerprint specific pre-processings		
    
Functions : 

	preProcess : main function applying the preprocessings
	firstMask : quickly evaluate a first mask to crop the image and only process real fingrprint data (less background)
	binarizeMask : binarize the mask calculated with firstMask
	cleanMask : keep only the biggest 8-connected region of a mask
	calculateCropSize : calculate the size of the image once pre-processed
	cropFromMask : crop the image to reduce its size and adapt it the the processings taht will follow

Function prototypes : 

	INT32 preProcess( BioEngineData *p_data );
	INT32 firstMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );
	INT32 binarizeMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );
	INT32 cleanMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );
	INT32 calculateCropSize( UINT8 *p_src, UINT8 *p_mask, UINT32 p_width, UINT32 p_height, CropStruct *p_crop );
	INT32 cropFromMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, CropStruct *p_crop );

Necessary package : 

	customTypes.h
	errorCodes.h
	staticParameters.h
	dynamicParameters.h
	greyLevelTrans.h
	filters.h
	geometry.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "greyLevelTrans.h"
#include "filters.h"
#include "geometry.h"
#include "BMPfiles.h"

#include "preProcessings.h"

// processes all the pre-enhancement steps on the image
// that's to say :
// resizing, and normalizing
// the new treated image is allocated and put in m_workImg in p_data
// it wil have to be deallocated later
INT32 preProcess( BioEngineData *p_data )
{
	INT32		error;
	CropStruct	crop;
	UINT8		*temp;
#ifdef TIME_REPORT
	clock_t timeReport;
#endif
	UINT32		normalizationMean,normalizationStdDev;

	// init 
	LOCK_MUTEX( g_paramsMutex );
		normalizationMean = g_params.m_normalizationMean;
		normalizationStdDev = g_params.m_normalizationStdDev;
	UNLOCK_MUTEX( g_paramsMutex );
	temp = NULL;

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// resize image to SEGM_BLOC_SIZE multiples
	error = checkBlocMultiple( p_data->m_image , &(p_data->m_width), &(p_data->m_height) );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "preProcess - checkBlocMultiple" ,clock()-timeReport );
#endif

	temp = (UINT8*)MALLOC( (p_data->m_width/SEGM_BLOC_SIZE)*(p_data->m_height/SEGM_BLOC_SIZE)*sizeof(UINT8) );
	if( temp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// generate temporary mask
	error = firstMask( p_data->m_image, temp, p_data->m_width, p_data->m_height );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "preProcess - firstMask" ,clock()-timeReport );
#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// normalize image locally
	error = normalizeLocally( p_data->m_image, p_data->m_image, p_data->m_width, p_data->m_height, normalizationMean, normalizationStdDev );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "preProcess - normalizeLocally" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		saveBmp( "Preprocess-04NormaliseLocally.bmp", p_data->m_image, p_data->m_width, p_data->m_height );
	#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// binarize mask
	error = binarizeMask( temp, temp, p_data->m_width/SEGM_BLOC_SIZE, p_data->m_height/SEGM_BLOC_SIZE );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "preProcess - binarizeMask" ,clock()-timeReport );
#endif

	#ifdef DP_DEBUG
		saveBmp( "Preprocess-07BinaryMask.bmp", temp, p_data->m_width/SEGM_BLOC_SIZE, p_data->m_height/SEGM_BLOC_SIZE );
	#endif

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// calculate size of the new image
	error = calculateCropSize( temp, p_data->m_width, p_data->m_height, &crop );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "preProcess - calculateCropSize" ,clock()-timeReport );
#endif

	// FREE temporary mask buffer
	FREE( temp );
	temp = NULL;
	
		// allocate memory for workImg
	p_data->m_workImg = (UINT8*)MALLOC( crop.m_newWidth*crop.m_newHeight*sizeof(UINT8) );
	if( p_data->m_workImg  == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

#ifdef TIME_REPORT
	timeReport = clock();
#endif
	// resize image 
	error = cropFromMask( p_data->m_image, p_data->m_workImg, p_data->m_width, p_data->m_height, &crop );
	if( error < ZEF_OK )
		goto errorHandler;
#ifdef TIME_REPORT
	printf("Function: %s (time=%d)\n", "preProcess - cropFromMask" ,clock()-timeReport );
#endif

	// set the new width and height
	p_data->m_width = crop.m_newWidth;
	p_data->m_height = crop.m_newHeight;

	#ifdef DP_DEBUG
		saveBmp( "Preprocess-08result.bmp", p_data->m_workImg, crop.m_newWidth, crop.m_newHeight );
	#endif

	return ZEF_OK;

	// error handling
errorHandler:

	if( temp != NULL )
		FREE( temp );
	if( p_data->m_workImg != NULL )
		FREE( p_data->m_workImg );

	return error;
}


// evaluate roughly a first mask for the image
// processing is based on differences of std deviation in the image
// p_src is left unchanged, p_dst contains a mask, SEGM_BLOC_SIZE times smaller (with 255 = image and 0 = background )
// p_dst must be allocated before calling firstMask, its size is  (p_width/SEGM_BLOC_SIZE)*(p_height/SEGM_BLOC_SIZE)
// p_width and p_height MUST BE SEGM_BLOC_SIZE multiples
INT32 firstMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height )
{
	INT32	i,j,error;	
	UINT32	miniWidth, miniHeight;
	UINT32	max,val;
	UINT8	*temp;
	UINT32	*mean,*stdDev;
	UINT32	segmMedianFilterSize,normalizationMean,normalizationStdDev;

	// init the max
	LOCK_MUTEX( g_paramsMutex );
		segmMedianFilterSize = g_params.m_segmMedianFilterSize;
		normalizationMean = g_params.m_normalizationMean;
		normalizationStdDev = g_params.m_normalizationStdDev;
	UNLOCK_MUTEX( g_paramsMutex );
	max = 0;

	// set miniWidth and height
	miniWidth = p_width / SEGM_BLOC_SIZE;
	if( p_width%SEGM_BLOC_SIZE) 
	{
		error = ZEF_NOT_BLOC_MULTIPLE;
		goto errorHandler;
	}
	miniHeight = p_height / SEGM_BLOC_SIZE;
	if( p_height%SEGM_BLOC_SIZE) 
	{
		error = ZEF_NOT_BLOC_MULTIPLE;
		goto errorHandler;
	}

	// allocations
	temp = (UINT8*) MALLOC( p_width*p_height*sizeof( UINT8 ) );
	if( temp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	mean = (UINT32*) MALLOC( miniWidth*miniHeight*sizeof( UINT32 ) );
	if( mean == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	stdDev = (UINT32*) MALLOC( miniWidth*miniHeight*sizeof( UINT32 ) );
	if( stdDev == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// global normalization
	error = normalize( p_src, temp, p_width, p_height, normalizationMean, normalizationStdDev );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp( "Preprocess-00normalize.bmp", temp, p_width, p_height );
	#endif

	// calculation stdDev of the input image
	error = localMeanStdDev( temp, p_width, p_height, mean, stdDev );
	if( error < ZEF_OK )
		goto errorHandler;

	#ifdef DP_DEBUG
		saveBmp32( "Preprocess-01localMean.bmp", mean, miniWidth, miniHeight );
		saveBmp32( "Preprocess-02localStdDev.bmp", stdDev, miniWidth, miniHeight );
	#endif

	FREE(temp);
	
	// we convert stdDev to a UINT8 image
	for( j=2; j<(INT32)miniHeight-2; j++ )
	{
		for( i=2; i<(INT32)miniWidth-2; i++ )
		{
 			if( stdDev[j*miniWidth+i] > max ) 
				max = stdDev[j*miniWidth+i];
		}
	}

	if( max == 0 ) max = 1; 
	max = (1L<<16)/max; // to avoid division in loops

	for( i=0; i<(INT32)(miniWidth*miniHeight); i++ )
	{
		val = (stdDev[i]*max)>>8;  // p_dst[i] = 255*stdDev[i]/max;
		if( val > 255 ) val = 255;
		p_dst[i] = val;
	}

	// apply a recursive median filter to smooth everything
	error = medianFilter( p_dst, p_dst, miniWidth, miniHeight, segmMedianFilterSize );	 
	if( error < ZEF_OK )
		goto errorHandler;
	#ifdef DP_DEBUG
		saveBmp( "Preprocess-03localStdDevFiltered.bmp", p_dst, miniWidth, miniHeight );
	#endif

	// and give memory back
	FREE(mean);
	FREE(stdDev);

	return ZEF_OK;

	// error handling
errorHandler:

	if( mean != NULL )
		FREE(mean);
	if( stdDev != NULL )
		FREE(stdDev);

	return error;
}

INT32 binarizeMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height )
{
	INT32	error;
	UINT32	qual;
	UINT32	i;
	UINT32	nbClassesSegmentation,nbClassesRejectSegmentation;

	// params init
	LOCK_MUTEX( g_paramsMutex );
		nbClassesSegmentation = g_params.m_nbClassesSegmentation;
		nbClassesRejectSegmentation = g_params.m_nbClassesRejectSegmentation;
	UNLOCK_MUTEX( g_paramsMutex );

	// Do a kMean classification on pixels to keep only NB_CLASSES_SEGMENTATION classes
	error = kMeans( p_src, p_dst, p_width, p_height, nbClassesSegmentation );
	if( error < ZEF_OK )
		goto errorHandler;

	qual = error; //quality of segmentation returned by kMeans
	
	#ifdef DP_DEBUG
		saveBmp( "Preprocess-05kMeans.bmp", p_dst, p_width, p_height );
	#endif

	// and finaly reject the first NB_CLASSES_REJECT_SEGMENTATION classes
	for( i=0; i<p_width*p_height; i++ )
		if( p_dst[i] < nbClassesRejectSegmentation )
			p_dst[i] = 0;
		else
			p_dst[i] = 255;

	error = cleanMask( p_dst, p_dst, p_width, p_height );
	if( error < ZEF_OK )
		goto errorHandler;

	return qual; // quality of segmentation

	// error handling
errorHandler:

	return error;
}

// multiply image by its mask... 
INT32 multImage( UINT8 *p_img, UINT8 *p_mask, UINT32 p_width, UINT32 p_height, UINT32 p_qual )
{
	UINT32	x,y;
	INT32	val,val2;
	UINT32	miniWidth;
	INT32	error;

	// set miniWidth and height
	miniWidth = p_width / SEGM_BLOC_SIZE;
	if( p_width%SEGM_BLOC_SIZE) 
	{
		error = ZEF_NOT_BLOC_MULTIPLE;
		goto errorHandler;
	}
	if( p_height%SEGM_BLOC_SIZE) 
		{
		error = ZEF_NOT_BLOC_MULTIPLE;
		goto errorHandler;
	}

	if( p_qual > 50 ) p_qual = 50;
	for( y=0; y<p_height; y++ )
	{
		for( x=0; x<p_width; x++ )
		{
			if( !p_mask[y/SEGM_BLOC_SIZE*miniWidth+x/SEGM_BLOC_SIZE] )
			{
				val = (p_img[y*p_width+x]-127);
				val2 = val*(INT32)p_qual/50;
				p_img[y*p_width+x] = 127 + val - val2;
			}
		}
	}

	return ZEF_OK;

	// error handling
errorHandler:

	return error;

}

// if p_src is a mask with values 0 and 255
// keep the biggest 8-connected region and put it in p_dst
// p_src and p_dst can be the same
INT32 cleanMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height )
{
	UINT32		i,x,y,xx,yy;
	UINT32		val,oldVal,acc;
	UINT32		max,regionMax;
	UINT32		*regions,*accumulator;
	UINT32		rWidth,rHeight;
	UINT32		offset,offset2;
	INT32		error;

	// fast lookup table for 8-connected pixel search
	INT8		dirX[8] = {  0,  1, -1,  1, -1,  1, -1,  0 };
	INT8		dirY[8] = { -1, -1, -1,  0,  0,  1,  1,  1 };

	// initialisations
	acc = 1;
	max = 0;
	rWidth = p_width+2;
	rHeight = p_height+2;

	// allocations and init
	regions = (UINT32*) MALLOC( rWidth*rHeight*sizeof(UINT32) );
	if( regions == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	for( i=0; i<rWidth*rHeight; i++ )
		regions[i] = 0;

	// main loop : we are looking for all the regions defined by 
	// a region = a set of 8-connected pixels
	offset=0;
	offset2=rWidth+1;
	for( y=1; y<p_height+1; y++ )
	{
		for( x=1; x<p_width+1; x++ )
		{
			// for each pixel with a non zero-mask (a pixels that really belongs to the fingerprint image)
			if( p_src[offset] != 0 )
			{
				// we are looking for a 8-connected pixel in the region map
				for( i=0; i<8; i++ )
				{
					val = regions[(y+dirY[i])*rWidth+(x+dirX[i])];

					// if we find one
					if( val )
					{
						// we give the value of the current region to the current point
						if( regions[offset2] == 0 )
							regions[offset2] = val; // case where the point is not associated with another region
						else
						{
							// if the point is associated to another region we've got to change that !
							if( regions[offset2] != val ) // if it is associated to another region
							{
								oldVal = regions[offset2];	// save the old value
								// go through all the pixels already treated
								for( yy=1; yy<=y; yy++ )
								{
									for( xx=1; xx<p_width+1; xx++ )
									{
										// and replace the old region values with the new ones
										if( regions[yy*rWidth+xx] == oldVal )
										{
											regions[yy*rWidth+xx] = val;
										}
									}
								}

							}
						}
					}
				}
				// if there's no 8-connected pixel we create a new region
				if( regions[offset2] == 0 )
				{
					regions[offset2] = acc++;
				}
			}
			offset2++;
			offset++;
		}
		offset2 += 2;
	}

	// now we are looking for the biggest region
	accumulator = (UINT32*) MALLOC( (acc+1)*sizeof(UINT32) );
	if( accumulator == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	for( i=0; i<acc+1; i++ )
		accumulator[i] = 0;

	offset2=rWidth+1;
	for( y=1; y<p_height+1; y++ )
	{
		for( x=1; x<p_width+1; x++ )
		{
			// for each pixel belonging to a region
			if( regions[offset2] != 0 )
			{
				// we add it to its accumulator
				accumulator[regions[offset2]]++;	
			}
			offset2++;
		}
		offset2+=2;
	}

	// the we find the max
	for( i=0; i<acc+1; i++ )
	{
		if( accumulator[i] > max )	
		{
			max = accumulator[i];
			regionMax = i;
		}
	}

	// and we only keep this region in the destination image
	offset=0;
	offset2=rWidth+1;
	for( y=1; y<p_height+1; y++ )
	{
		for( x=1; x<p_width+1; x++ )
		{
			if( regions[offset2] == regionMax )
				p_dst[offset] = 255;
			else
				p_dst[offset] = 0;

			offset++;
			offset2++;
		}
		offset2+=2;
	}

	#ifdef DP_DEBUG
		saveBmp32( "Preprocess-06regions.bmp",regions, rWidth, rHeight );
	#endif

	FREE( accumulator );
	FREE( regions );

	return ZEF_OK;

	// error handling
errorHandler:

	if( accumulator != NULL )
		FREE( accumulator );
	if( regions != NULL )
		FREE( regions );

	return error;

}

INT32 calculateCropSize( UINT8 *p_mask, UINT32 p_width, UINT32 p_height, CropStruct *p_crop )
{
	INT32		mWidth,mHeight;
	INT32		i,j;
	BOOL8		startX,startY;
	UINT32		max_img_width,max_img_height,fftBlocOverlap;

	// initializations
	LOCK_MUTEX( g_paramsMutex );
		max_img_width = g_params.m_max_img_width;
		max_img_height = g_params.m_max_img_height;
		fftBlocOverlap = g_params.m_fftBlocOverlap;
	UNLOCK_MUTEX( g_paramsMutex );
	p_crop->m_xMin = p_crop->m_yMin = INT_MAX;
	p_crop->m_xMax = p_crop->m_yMax = 0;
	startY = FALSE;

	// calculate mask size
	mWidth = (INT32)p_width/SEGM_BLOC_SIZE;
	mHeight = (INT32)p_height/SEGM_BLOC_SIZE;
	if( p_width%SEGM_BLOC_SIZE ) mWidth++;
	if( p_height%SEGM_BLOC_SIZE ) mHeight++;


	// we are looking for the new image boundaries
	for( j=0; j<mHeight; j++ )
	{
		startX = FALSE;
		for( i=0; i<mWidth; i++ )
		{
			if( p_mask[j*mWidth+i] != 0 ) // beginning of a line (left limit)
			{
				if( i<(INT32)p_crop->m_xMin ) p_crop->m_xMin = i;
				startX = TRUE;
				break; // end of search for this line
			}
		}
		for( i=mWidth-1; i>=0; i-- )
		{
			if( p_mask[j*mWidth+i] != 0 ) // end of a line (right limit)
			{
				if( i>(INT32)p_crop->m_xMax ) p_crop->m_xMax = i;
				break; // end of search for this line
			}
		}

		// top of the image
		if( startX && !startY )
		{
			p_crop->m_yMin = j;
			startY = TRUE;
		}

		// bottom of the image
		if( !startX && startY )
		{
			p_crop->m_yMax = j;
			break; // end of search
		}
	}

	// now we handle some cases where we couldn't find the limits
	if( p_crop->m_xMin == INT_MAX ) p_crop->m_xMin = 0;
	if( p_crop->m_yMin == INT_MAX ) p_crop->m_yMin = 0;
	if( p_crop->m_xMax < p_crop->m_xMin ) p_crop->m_xMax = mWidth;
	if( p_crop->m_yMax < p_crop->m_yMin ) p_crop->m_yMax = mHeight;

	// set the new width and height (in SEGM_BLOC_SIZE blocs)
	p_crop->m_newWidth = p_crop->m_xMax-p_crop->m_xMin;
	p_crop->m_newHeight = p_crop->m_yMax-p_crop->m_yMin;

	// some additionnal tests for very little images... (less than one FFT bloc)
	if( p_crop->m_newWidth <= FFT_BLOC_SIZE/SEGM_BLOC_SIZE )
	{
		p_crop->m_xMin = 0;
		p_crop->m_xMax = p_crop->m_newWidth = mWidth;
	}
	if( p_crop->m_newHeight <= FFT_BLOC_SIZE/SEGM_BLOC_SIZE )
	{
		p_crop->m_yMin = 0;
		p_crop->m_yMax = p_crop->m_newHeight = mHeight;
	}

	if( p_crop->m_newWidth > max_img_width/SEGM_BLOC_SIZE )
		p_crop->m_newWidth = max_img_width/SEGM_BLOC_SIZE;
	
	if( p_crop->m_newHeight > max_img_height/SEGM_BLOC_SIZE )
		p_crop->m_newHeight = max_img_height/SEGM_BLOC_SIZE;

	// we add some space for FFT overlap on the edges
	p_crop->m_newWidth += FFT_BLOC_SIZE/(2*SEGM_BLOC_SIZE);
	p_crop->m_newHeight += FFT_BLOC_SIZE/(2*SEGM_BLOC_SIZE);


	// here we adjust width and height so that we have an integer number
	// of FFT_BLOC_SIZE blocs with g_params.m_fftBlocOverlap overlap.
	if( (p_crop->m_newWidth-(fftBlocOverlap)/SEGM_BLOC_SIZE)%((FFT_BLOC_SIZE-fftBlocOverlap)/SEGM_BLOC_SIZE) )
		p_crop->m_newWidth += ((FFT_BLOC_SIZE-fftBlocOverlap)/SEGM_BLOC_SIZE) - (p_crop->m_newWidth-(fftBlocOverlap)/SEGM_BLOC_SIZE)%((FFT_BLOC_SIZE-fftBlocOverlap)/SEGM_BLOC_SIZE);
	if( (p_crop->m_newHeight-(fftBlocOverlap)/SEGM_BLOC_SIZE)%((FFT_BLOC_SIZE-fftBlocOverlap)/SEGM_BLOC_SIZE) )
		p_crop->m_newHeight += ((FFT_BLOC_SIZE-fftBlocOverlap)/SEGM_BLOC_SIZE) - (p_crop->m_newHeight-(fftBlocOverlap)/SEGM_BLOC_SIZE)%((FFT_BLOC_SIZE-fftBlocOverlap)/SEGM_BLOC_SIZE);

	p_crop->m_newHeight *= SEGM_BLOC_SIZE;
	p_crop->m_newWidth *= SEGM_BLOC_SIZE;
	p_crop->m_xMin *= SEGM_BLOC_SIZE;
	p_crop->m_yMin *= SEGM_BLOC_SIZE;
	p_crop->m_xMax *= SEGM_BLOC_SIZE;
	p_crop->m_yMax *= SEGM_BLOC_SIZE;

	return ZEF_OK;
}

INT32 cropFromMask( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height, CropStruct *p_crop )
{
	UINT32	i,j,offset;
	INT32	x,y;
	INT32	shiftX,shiftY;

	// init 
	offset = 0;
	shiftX = ((INT32)p_crop->m_newWidth-(INT32)(p_crop->m_xMax-p_crop->m_xMin))/2;
	shiftY = ((INT32)p_crop->m_newHeight-(INT32)(p_crop->m_yMax-p_crop->m_yMin))/2;

	// loop thru the image to copy it !
	for( j=0; j<p_crop->m_newHeight; j++ )
	{
		y = p_crop->m_yMin+j-shiftY;
		for( i=0; i<p_crop->m_newWidth; i++ )
		{
			x=p_crop->m_xMin+i-shiftX;
			if( (UINT32)x<p_width && (UINT32)y<p_height )
				p_dst[offset++] = p_src[y*p_width+x];
			else
				p_dst[offset++] = 127;
		}
	}

	return ZEF_OK;
}

