/*
Company :		ZEFYR Technologies
File :			skeleton.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	13/06/2003

Description : Skeletonization algorithm

Features : fast skeletonization		
    
Functions : 

	analyzeNeighborhood : internal function used to give information about a pixel (Black to White transition in neighboorhoud, etc.)
	skeletonization : main skeletonisation function, should be quite fast

Function prototypes :

	UINT32 analyzeNeighborhood( UINT8 *p_src, UINT32 p_width, UINT32 *p_nbNeighbors, UINT32 *p_nbTransitions );
	INT32 skeletonization( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height );

Necessary package : 

	customTypes.h
	staticParameters.h
	dynamicParameters.h
	errorCodes.h

Comments : none

*/

#include <stdlib.h>
#include <string.h>

#include "customTypes.h"
#include "staticParameters.h"
#include "dynamicParameters.h"
#include "errorCodes.h"

#include "skeleton.h"

INT32 skeletonization( UINT8 *p_src, UINT8 *p_dst, UINT32 p_width, UINT32 p_height )
{
	UINT32	x,y,nbTransitions,nbNeighbors,code,acc;
	UINT32	y1;
	UINT32	g;
	BOOL8	loop;
	UINT8	*tmp;
	INT32	error;

	// init
	tmp = NULL;
	loop = TRUE;
	g=1;
	acc=0;

	// allocations
	tmp = (UINT8*) MALLOC( p_width*p_height*sizeof(UINT8) );
	if( tmp == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	// init bis
	memcpy(tmp,p_src,p_width*p_height*sizeof(UINT8));
	memcpy(p_dst,p_src,p_width*p_height*sizeof(UINT8));

	// loop while we can change pixels
	while( loop && acc < MAX_LOOPS )
	{
		acc++;
		loop = FALSE;
		g=1-g;

		// loop thru the image
		for( y=1; y<p_height-1; y++ )
		{
			y1 = y*p_width;
			for( x=1; x<p_width-1; x++ )
			{
				if( p_dst[y1+x] == 0 ) // if the pixel is ON
				{
					code = analyzeNeighborhood( &(p_dst[y1+x]), p_width, &nbNeighbors, &nbTransitions );
					if( nbNeighbors>1 && nbNeighbors<7 && E(code) && (nbTransitions==1 ||  C(code) ) )
					{
						tmp[y*p_width+x] = 255;
						loop = TRUE;
					}
				}
			}
		}

		memcpy(p_dst,tmp,p_width*p_height*sizeof(UINT8));

	}

	// some more cleanup
	for( y=1; y<p_height-1; y++ )
	{
		y1 = y*p_width;
		for( x=1; x<p_width-1; x++ )
		{
			if( p_dst[y1+x] == 0 ) // if the pixel is ON
			{
				code = analyzeNeighborhood( &(p_dst[y1+x]), p_width, &nbNeighbors, &nbTransitions );
				if( nbNeighbors>1 && nbNeighbors<7 && nbTransitions==3  && (!E(code) || !F(code)) )
				{
					p_dst[y*p_width+x] = 255;
				}
			}
		}
	}

	// some one last more cleanup
	for( y=1; y<p_height-1; y++ )
	{
		y1 = y*p_width;
		for( x=1; x<p_width-1; x++ )
		{
			if( p_dst[y1+x] == 0 ) // if the pixel is ON
			{
				code = analyzeNeighborhood( &(p_dst[y1+x]), p_width, &nbNeighbors, &nbTransitions );
				if( nbNeighbors>1 && nbNeighbors<7 && nbTransitions==2 &&
					( ((code&5) == 5) || ((code&20) == 20) || ((code&65) == 65) || ((code&80) == 80 ) ) )
				{
					p_dst[y*p_width+x] = 255;
				}
			}
		}
	}

	FREE( tmp );

	return ZEF_OK;

	//error handling
errorHandler:

	if( tmp != NULL )
		FREE(tmp);

	return error;
}

UINT32 analyzeNeighborhood( UINT8 *p_src, UINT32 p_width, UINT32 *p_nbNeighbors, UINT32 *p_nbTransitions )
{
	UINT32	code;
	UINT32	oldPix;
	UINT32	nbTransitions,nbNeighbors;
	UINT8	*src1,*src2;

	// init
	code = 0;
	nbTransitions = 0;
	nbNeighbors = 0;
	src1 = p_src-p_width;
	src2 = p_src+p_width;

	//construct code and nbNeighbors + nbTransitions
	if( (src1[0]&128) == 0 ) // is equivalent to "if( src1[0] < 127 )" but is faster 
	{
		code |= 1;
		nbNeighbors++;
		oldPix=0;
	}
	else
		oldPix = 1;

	if( (src1[1]&128) == 0 )
	{
		code |= 2;
		nbNeighbors++;
		nbTransitions += oldPix;
		oldPix=0;
	}
	else
		oldPix = 1;

	if( (p_src[1]&128) == 0 )
	{
		code |= 4;
		nbNeighbors++;
		nbTransitions += oldPix;
		oldPix=0;
	}
	else
		oldPix = 1;

	if( (src2[1]&128) == 0 )
	{
		code |= 8;
		nbNeighbors++;
		nbTransitions += oldPix;
		oldPix=0;
	}
	else
		oldPix = 1;

	if( (src2[0]&128) == 0 )
	{
		code |= 16;
		nbNeighbors++;
		nbTransitions += oldPix;
		oldPix=0;
	}
	else
		oldPix = 1;

	if( (src2[-1]&128) == 0 )
	{
		code |= 32;
		nbNeighbors++;
		nbTransitions += oldPix;
		oldPix=0;
	}
	else
		oldPix = 1;

	if( (p_src[-1]&128) == 0 )
	{
		code |= 64;
		nbNeighbors++;
		nbTransitions += oldPix;
		oldPix=0;
	}
	else
		oldPix = 1;

	if( (src1[-1]&128) == 0 )
	{
		code |= 128;
		nbNeighbors++;
		nbTransitions += oldPix;
		oldPix=0;
	}
	else
		oldPix = 1;

	
	if( (src1[0]&128) == 0 )
		nbTransitions += oldPix;


	*p_nbNeighbors = nbNeighbors;
	*p_nbTransitions = nbTransitions;

	return code;
}

