/*
Company :		ZEFYR Technologies
File :			tpsWarper.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	04/07/2003

Description : calculates a non rigid deformation between two sets of points using thin plate splines

Features : estimates local deformation of fingerprints
    
Functions : 

	initTpsWarper : initialize a new estimation with new data
	initMatrixA : build tyhe A matrix for TPS system resolution
	unInitTpsWarper : FREE memory used during TPS warping
	warp : calculate the new position of a poirn according to the deformation calculated with initTpsWarper
	getDiffs : get the "energy" needed for the deformation

Function prototypes : 

	INT32	initTpsWarper( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, INT32 *p_dstX, INT32 *p_dstY, REAL32 p_alpha, TpsData *p_data);
	INT32	initMatrixA( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, REAL32 p_alpha, TpsData *p_data );
	INT32	unInitTpsWarper( TpsData *p_data );
	INT32	warp( INT32 *p_x, INT32 *p_y, TpsData *p_data );
	UINT32  getDiffs( TpsData *p_data );

Necessary package : 

	customtypes.h
	errorCodes.h
	linAlgebra.h
	zMaths.h

Comments : none

*/

#include <stdlib.h>
#include <limits.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "linAlgebra.h"
#include "zMaths.h"

#include "tpsWarper.h"

#if !defined(ZEF_FIXED_POINT) && defined(ZEF_CORRECT_DEFORMATION)

INT32 initTpsWarper( INT32 p_nbPts, INT32 *p_srcX, INT32 *p_srcY, INT32 *p_dstX, INT32 *p_dstY, REAL32 p_alpha, TpsData *p_data)
{
	INT32	error;
	INT32	i;
	REAL32	even;

	// init
	p_data->m_index = NULL;
	p_data->m_matrixA = NULL;
	p_data->m_srcX = NULL;
	p_data->m_srcY = NULL;
	p_data->m_vectorBx = NULL;
	p_data->m_vectorBy = NULL;

	// initialise differences indicator
	p_data->m_error = 0;
	p_data->m_cptPts = 0;

	// then the number of control points
	p_data->m_nbPts = p_nbPts;

	// index
	p_data->m_index = (INT32*)MALLOC((p_nbPts+3)*sizeof(INT32));
	if( p_data->m_index == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	// and source points
	p_data->m_srcX = (REAL32*)MALLOC((p_nbPts)*sizeof(REAL32));
	if( p_data->m_srcX == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	p_data->m_srcY = (REAL32*)MALLOC((p_nbPts)*sizeof(REAL32));
	if( p_data->m_srcY == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_srcX[i] = p_srcX[i]/512.0f; // we divide by 512 to be consistant with fxTpsWarper.c
		p_data->m_srcY[i] = p_srcY[i]/512.0f; // we divide by 512 to be consistant with fxTpsWarper.c
	}

	// initialisation of matrix A
	error = initMatrixA( p_nbPts, p_data->m_srcX, p_data->m_srcY, p_alpha, p_data );
	if( error < ZEF_OK )
		goto errorHandler;

	// then we fill the right side with X  source coords
	p_data->m_vectorBx = (REAL32*)MALLOC((p_nbPts+3)*sizeof(REAL32));
	if( p_data->m_vectorBx == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}

	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_vectorBx[i] = p_dstX[i]/512.0f; // we divide by 512 tp be consistant with fxTpsWarper.c
	}
	for( i=p_nbPts; i<p_nbPts+3; i++ )
	{
		p_data->m_vectorBx[i] = 0;
	}
	

	p_data->m_vectorBy = (REAL32*)MALLOC((p_nbPts+3)*sizeof(REAL32));
	if( p_data->m_vectorBy == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	
	// solve the system
	error = luDecompose(p_data->m_matrixA, p_data->m_nbPts+3, p_data->m_index, &even);
	if( error < ZEF_OK )
		goto errorHandler;

	error = luBackSubstitute(p_data->m_matrixA, p_data->m_nbPts+3, p_data->m_index, p_data->m_vectorBx );
	if( error < ZEF_OK )
		goto errorHandler;
	
	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_vectorBy[i] = p_dstY[i]/512.0f; // we divide by 512 tp be consistant with fxTpsWarper.c
	}
	for( i=p_nbPts; i<p_nbPts+3; i++ )
	{
		p_data->m_vectorBy[i] = 0;
	}

	// and solve a second time
	error = luBackSubstitute(p_data->m_matrixA, p_data->m_nbPts+3, p_data->m_index, p_data->m_vectorBy );
	if( error < ZEF_OK )
		goto errorHandler;
	
	return ZEF_OK;


	// error handling
errorHandler:

	unInitTpsWarper(p_data);
	
	return error;

}
INT32 initMatrixA( INT32 p_nbPts, REAL32 *p_srcX, REAL32 *p_srcY, REAL32 p_alpha, TpsData *p_data )
{
	INT32		i,j;
	REAL32		r;
	REAL32		subLog;
	INT32		error;

	// init
	subLog = zLog(32768);

	// allocation of matrix A
	p_data->m_matrixA = (REAL32**)MALLOC( (p_nbPts+3)*sizeof(REAL32*) );
	if( p_data->m_matrixA == NULL )
	{
		error = ZEF_OUT_OF_MEMORY;
		goto errorHandler;
	}
	for( i=0; i<p_nbPts+3; i++ )
	{
		p_data->m_matrixA[i] = (REAL32*)MALLOC( (p_nbPts+3)*sizeof(REAL32) );
		if( p_data->m_matrixA[i] == NULL )
		{
			error = ZEF_OUT_OF_MEMORY;
			goto errorHandler;
		}
	}

	// We fill the K part of the matrix
	for( i=0; i<p_nbPts; i++ )
	{
		for( j=0; j<=i; j++ )
		{
			r = (p_srcX[i]-p_srcX[j])*(p_srcX[i]-p_srcX[j]) + (p_srcY[i]-p_srcY[j])*(p_srcY[i]-p_srcY[j]);
			if( r == 0 ) 
			{
				p_data->m_matrixA[i][j] = 0;
				p_data->m_matrixA[j][i] = 0;
			}
			else
			{
				// here we should calculate r*ln(sqrt(r)) = r*ln(r)/2
				// but to avoid division in this loop (for speed improvement on machines where division is not cabled
				// we do not devide by two (which doesn't change very muche the final result !)
				p_data->m_matrixA[i][j] = r*(zLog(zReal2Int32(r*32768))-subLog); // equivalent to r*ln(r)
				p_data->m_matrixA[j][i] = p_data->m_matrixA[i][j];
			}
		}
	}


	// then we add I*alpha in order to have K+I*alpha
	for( i=0; i<p_nbPts; i++ )
	{
		p_data->m_matrixA[i][i] += p_alpha; // Ok p_alpha has already been converted to float
	}

	// and add P in the top right of the matrix
	for( i=0; i< p_nbPts; i++ )
	{
		p_data->m_matrixA[i][p_nbPts] = 1.0f;
		p_data->m_matrixA[i][p_nbPts+1] = (REAL32)p_srcX[i];
		p_data->m_matrixA[i][p_nbPts+2] = (REAL32)p_srcY[i];
	}

	// Add P transposed in the bottom left of the matrix
	for( j=0; j< p_nbPts; j++ )
	{
		p_data->m_matrixA[p_nbPts][j] = 1.0f;
		p_data->m_matrixA[p_nbPts+1][j] = (REAL32)p_srcX[j];
		p_data->m_matrixA[p_nbPts+2][j] = (REAL32)p_srcY[j];
	}

	// and zeros in the bottom right of the matrix
	for( i=0; i<3; i++)
	{
		for( j=0; j<3; j++)
		{
			p_data->m_matrixA[p_nbPts+i][p_nbPts+j] = 0;	
		}
	}

	return ZEF_OK;

	// error handling
errorHandler:

	unInitTpsWarper(p_data);

	return error;
}
INT32 unInitTpsWarper( TpsData *p_data )
{
	INT32	i;

	for( i=0; i<p_data->m_nbPts+3; i++ )
	{
		if( p_data->m_matrixA[i] != NULL )
			FREE( p_data->m_matrixA[i] );
	}

	if( p_data->m_matrixA != NULL )
		FREE(p_data->m_matrixA);
	if( p_data->m_vectorBx != NULL )
		FREE(p_data->m_vectorBx);
	if( p_data->m_vectorBy != NULL )
		FREE(p_data->m_vectorBy);
	if( p_data->m_srcX != NULL )
		FREE(p_data->m_srcX);
	if( p_data->m_srcY != NULL )
		FREE(p_data->m_srcY);
	if( p_data->m_index != NULL );
		FREE(p_data->m_index);

	p_data->m_nbPts = 0;
	p_data->m_error = 0;
	p_data->m_cptPts = 0;

	return ZEF_OK;
}
INT32 warp( INT32 *p_x, INT32 *p_y, TpsData *p_data )
{
	INT32	i;
	REAL32	r,x,y;
	REAL32	tmpX,tmpY;
	REAL32	Ur,subLog;

	// init
	x = (*p_x)/512.0f;
	y = (*p_y)/512.0f;
	subLog = zLog(32768);

	tmpX = p_data->m_vectorBx[p_data->m_nbPts] + p_data->m_vectorBx[p_data->m_nbPts+1]*(x) + p_data->m_vectorBx[p_data->m_nbPts+2]*(y);
	tmpY = p_data->m_vectorBy[p_data->m_nbPts] + p_data->m_vectorBy[p_data->m_nbPts+1]*(x) + p_data->m_vectorBy[p_data->m_nbPts+2]*(y);

	// process new coordinates
	for( i=0; i<p_data->m_nbPts; i++ )
	{
		r = (p_data->m_srcX[i]-(x))*(p_data->m_srcX[i]-(x))+(p_data->m_srcY[i]-(y))*(p_data->m_srcY[i]-(y));
		if( r != 0 )
		{
			// here we should calculate r*ln(sqrt(r)) = r*ln(r)/2
			// but to avoid division in this loop (for speed improvement on machines where division is not cabled
			// we do not devide by two (which doesn't change very muche the final result !)
			Ur = r*(zLog(zReal2Int32(r*32768))-subLog); // equivalent to r*ln(r)
		}
		else
			Ur = 0;

		tmpX += p_data->m_vectorBx[i] * Ur;
		tmpY += p_data->m_vectorBy[i] * Ur;
	}

	// update deformation indicator
	p_data->m_error += zReal2Int32(((x)-tmpX)*((x)-tmpX) + ((y)-tmpY)*((y)-tmpY));
	(p_data->m_cptPts)++;

	*p_x = zReal2Int32(tmpX*512); 
	*p_y = zReal2Int32(tmpY*512);


	return ZEF_OK;
}
UINT32  getDiffs( TpsData *p_data )
{
	if( p_data->m_cptPts != 0 )
		return p_data->m_error/p_data->m_cptPts;
	else
		return 0;
}


#endif // ZEF_FIXED_POINT

