/*
Company :		ZEFYR Technologies
File :			zMaths.c
Author :		Matthieu Perreira Da Silva 
Fisrt release :	06/05/2003

Description : custom maths functions

Features : provides functions that are faster or more compatible with the architecture used		
    
Functions : 

	zSqrt : processes the square root of an integer number
	zfSqrt : processes the square root of a real number
	zSin : Sinus
	zCos : Cosinus
	zAbs : Absolute value
	zfAbs : Real Absolute Value
	zSinSpectrum : fast Sinus for spectrum processing (table based)
	zCosSpectrum : fast Cosinus for spectrum processing (table based)
	zReal2Int32 : fast real to int conversions for x86
	zAtan : fast Atan
	zAtanSpec : fast atan for spectrum processing (table based)
	zSqrtSpec : fast sqrt for spectrum processing (table based)
	zLog : logarithm
	zCos127 : fast integar cosinus
	zSin127 : fast integer sinus


Function prototypes : 

	INT16	zAtan(INT32 y, INT32 x);
	REAL32	zLog( UINT32 p_nb );
	INT8	zCos127( INT32 p_angle );
	INT8	zSin127( INT32 p_angle );
	INT32	Real2Int32(REAL32 fFloat)

	All other mathematical functions are macros

Necessary package : 

	math.h
	errorCodes.h
	customTypes.h

Comments : none

*/

#include <stdlib.h>

#include "customTypes.h"
#include "errorCodes.h"
#include "staticParameters.h"
#include "zMaths.h"

#if !defined(USE_FX_PRECOMP) || !defined(ZEF_FIXED_POINT)
INT32	g_precSinSpectrum[SPECTRUM_RESOLUTION];
INT32	g_precCosSpectrum[SPECTRUM_RESOLUTION];
INT8    g_precSin[256];
INT8	g_precCos[256];
UINT8	g_sqrtSpec[FFT_BLOC_SIZE][FFT_BLOC_SIZE/2];
UINT8	g_atanSpec[FFT_BLOC_SIZE][FFT_BLOC_SIZE/2];
UINT8	g_atanTable[257];
#endif
#ifndef ZEF_FIXED_POINT
REAL32	g_invExp[512];
REAL32	g_log2Table[1025];
#else if !defined(USE_FX_PRECOMP)
FIXED16	g_invExp[512];
FIXED32	g_log2Table[1025];
#endif

#if !defined(USE_FX_PRECOMP) || !defined(ZEF_FIXED_POINT)
// initialize precomputed arrays
INT32	zInitMaths( void )
{
	INT32	i,j, theta;

	// initialize g_precCos32Dec & g_preSin32Dec arrays
	for( i=0; i<SPECTRUM_RESOLUTION/2; i++ )
	{
		g_precCosSpectrum[i] = zReal2Int32((REAL32)cos(i*PI/SPECTRUM_RESOLUTION)*65536.0f+0.5f);
		g_precSinSpectrum[i] = zReal2Int32((REAL32)sin(i*PI/SPECTRUM_RESOLUTION)*65536.0f+0.5f);
	}
	for( i=SPECTRUM_RESOLUTION/2; i<SPECTRUM_RESOLUTION; i++ )
	{
		g_precCosSpectrum[i] = zReal2Int32((REAL32)cos(-PI+i*PI/SPECTRUM_RESOLUTION)*65536.0f+0.5f);
		g_precSinSpectrum[i] = zReal2Int32((REAL32)sin(-PI+i*PI/SPECTRUM_RESOLUTION)*65536.0f+0.5f);
	}

	// init precomp Cos and Sin tables
	for( i=0; i<256; i++ )
	{
		INT32 val;
		val = zReal2Int32(127.0f*(REAL32)cos(2.0f*PI*i/256.0f)+0.5f);
		if( val > 127 ) val = 127;
		if( val < -128 ) val = -128;
		g_precCos[i] = val;
		val = zReal2Int32(127.0f*(REAL32)sin(2.0f*PI*i/256.0f)+0.5f);
		if( val > 127 ) val = 127;
		if( val < -128 ) val = -128;
		g_precSin[i] = val;
	}

	// init precomp atan and sqrt for spectrum
	for( j=0; j<FFT_BLOC_SIZE; j++ )
	{
		for( i=0; i<FFT_BLOC_SIZE/2; i++ )
		{
			g_sqrtSpec[j][i] = (UINT8)zSqrt(i*i+(j-32)*(j-32));
			theta = zReal2Int32((REAL32)atan2(j-32,i)*SPECTRUM_RESOLUTION/PI+0.5f);
			if( theta < 0 ) theta += SPECTRUM_RESOLUTION;
			g_atanSpec[j][i] = theta;
		}
	}

	// log table
	for( i=0; i<=1024; i++ )
	{
		#ifndef ZEF_FIXED_POINT
			g_log2Table[i] = (REAL32)(log(1+i/1024.0)/log(2));
		#else
			g_log2Table[i] = FTOFIX32(log(1+i/1024.0)/log(2),11);
		#endif
	}
	//atan table
	for( i=0; i<257; i++ )
		g_atanTable[i] = zReal2Int32(256.0f*(REAL32)atan(i/256.0)/PI+0.5f);

	// exp(-x) for 0 <= x <= 8
	for( i=0; i<512; i++ )
	{
		#ifndef ZEF_FIXED_POINT
			g_invExp[i] = (REAL32)exp(TOFLT(-i,6));
		#else
			g_invExp[i] = FTOFIX16(exp(TOFLT(-i,6)),Q1_14);	
		#endif
	}

	return ZEF_OK;
}
#else
INT32	zInitMaths( void )
{
	return ZEF_OK;
}
#endif //USE_FX_PRECOMP

#ifndef ZEF_FIXED_POINT
	REAL32 zLog( UINT32 p_nb )
	{
		UINT32	g,j,x_minus_a;
		REAL32	gf;

		#if defined(_M_IX86) && defined(WIN32) 
			// faster code
			_asm {
				bsr eax,p_nb
				mov g,eax
			}
			p_nb<<=(31-g);
		#else
			// portable code
			g = 31;
			while(g && !(p_nb&0x80000000) )   // loop until msb of x occupies bit 31
			{
				p_nb <<= 1;
				g--;
			}
		#endif //(_M_IX86) 

		// If g==0, then we're done.
		if(!g) return(0);
		// We have the integer portion of the log(x). Now get the fractional part.
		p_nb <<= 1;                 // "normalize" x, get rid of the MSB.
		j = p_nb >> 22;             // Get the array index
		x_minus_a = p_nb & 0x3fffff;   // The lower bits of x are all that's left 
										//   after subtracting "a".
		gf = g_log2Table[j] + (x_minus_a * (g_log2Table[j+1] - g_log2Table[j]))*LOGDIV;

		return (g+gf)*INVLOG2E;
	}
#else
	FIXED32 zLog( UINT32 p_nb )
	{
		UINT32	g,j,x_minus_a;
		FIXED32	gf;

		#ifdef _M_IX86 
			// faster code
			_asm {
				bsr eax,p_nb
				mov g,eax
			}
			p_nb<<=(31-g);
		#else
			// portable code
			g = 31;
			while(g && !(p_nb&0x80000000) )   // loop until msb of x occupies bit 31
			{
				p_nb <<= 1;
				g--;
			}
		#endif //(_M_IX86) 

		// If g==0, then we're done.
		if(!g) return(0);
		// We have the integer portion of the log(x). Now get the fractional part.
		p_nb <<= 1;                 // "normalize" x, get rid of the MSB.
		j = p_nb >> 22;             // Get the array index
		x_minus_a = (p_nb>>11) & 0xffff;   // The lower bits of x are all that's left 
										//   after subtracting "a". (11 bits precision)
		gf = FADD(g_log2Table[j],FMUL(x_minus_a,FSUB(g_log2Table[j+1],g_log2Table[j]),11));
		g = ITOFIX32(g,11);

		return FMUL(FADD(g,gf),INVLOG2E,11);
	}
#endif

// slow on a PIII, but could be usefull on a dsp ?
#ifndef _M_IX86
UINT16 zSqrt( UINT32 r) 
{
  unsigned int temp, g=0;

  if (r >= 0x40000000) 
  {
    g = 0x8000; 
    r -= 0x40000000;
  }

#define INNER_MBGSQRT(s) temp = (g << (s)) + (1L << ((s) * 2 - 2));if (r >= temp) {g += 1L << ((s)-1); r -= temp; }

  INNER_MBGSQRT (15)
  INNER_MBGSQRT (14)
  INNER_MBGSQRT (13)
  INNER_MBGSQRT (12)
  INNER_MBGSQRT (11)
  INNER_MBGSQRT (10)
  INNER_MBGSQRT ( 9)
  INNER_MBGSQRT ( 8)
  INNER_MBGSQRT ( 7)
  INNER_MBGSQRT ( 6)
  INNER_MBGSQRT ( 5)
  INNER_MBGSQRT ( 4)
  INNER_MBGSQRT ( 3)
  INNER_MBGSQRT ( 2)

#undef INNER_MBGSQRT

  temp = g+g+1;
  if (r >= temp) g++;

  return g;
}
#endif



INT16 zAtan(INT32 y, INT32 x)
{
	INT32 xabs, yabs;
	INT32 f, g;

		
	if (x > 0)
		xabs = x;
	else
	{
		if( x == 0 ) 
		{
			if( y == 0 ) 
				return 0;
			else
			{
				if( y >= 0 )
					return  PI2_256;
				else
					return  -PI2_256;
			}
		}
		xabs = -x;
	
	}

	if (y > 0)
		yabs = y;
	else
	{
		if( y == 0 )
		{
			if( x >= 0 )
				return 0;
			else
				return PI_256;
		}
		yabs = -y;
	}
	if (yabs <= xabs)
	{
		f = (yabs * 256) / xabs;
		g = g_atanTable[f];
	}
	else
	{
		f = (xabs * 256) / yabs;
		g = PI2_256 - g_atanTable[f];
	}
	
	if (x >= 0)
	{
		if (y >= 0)  // first quadrant 
			return(g);
		else         // fourth quadrant
			return(-g);
	}
	else
	{
		if (y >= 0)  // second quadrant 
			return( PI_256 - g );
		else         // third quadrant 
			return(g - PI_256 );
	}
}
